// scripts/core/analyze_selection.js

import { Player, system, world } from "@minecraft/server";
import { getPlayerSelection } from "../state/index.js";
import { MAX_COPY_VOLUME } from "../constants/index.js";
import { getMinCorner, getMaxCorner, calculateVolume } from "../utils/location.js";
import { showMessage } from "../ui/message_display.js";
import {
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Counts the occurrences of each block type within the selection.
 * @param {Player} player The player initiating the action.
 */
export async function countBlocks(player) {
    const selection = getPlayerSelection(player);
    const { posA, posB, selectionState } = selection;
    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return;
    }
    const min = getMinCorner(posA, posB);
    const max = getMaxCorner(posA, posB);
    const volume = calculateVolume(min, max);

    if (volume <= 0) {
        sendErrorMessage(player, "Selection empty or invalid.");
        return;
    }
    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}) to count blocks.`);
        return;
    }

    sendInfoMessage(player, `Analyzing ${volume} blocks...`);
    const blockCounts = new Map();
    let blocksChecked = 0;
    let airBlocks = 0;
    const dimension = player.dimension;
    const checkInterval = Math.max(5000, Math.floor(volume / 5));

    try {
        for (let x = min.x; x <= max.x; x++) {
            for (let y = min.y; y <= max.y; y++) {
                for (let z = min.z; z <= max.z; z++) {
                    blocksChecked++;
                    if (blocksChecked % checkInterval === 0) {
                        await system.run(() => { });
                    }
                    try {
                        const block = dimension.getBlock({ x, y, z });
                        if (block) {
                            if (block.isAir) {
                                airBlocks++;
                            } else {
                                const typeId = block.typeId;
                                blockCounts.set(typeId, (blockCounts.get(typeId) ?? 0) + 1);
                            }
                        }
                    } catch (e) { /* Ignore single block errors */ }
                }
            }
        }

        let totalBlocks = blocksChecked - airBlocks;
        let resultBody = `§f--- Block Counts (${totalBlocks} blocks) ---\n`;
        if (airBlocks > 0) {
            resultBody += `§7minecraft:air: ${airBlocks}\n`;
        }

        if (blockCounts.size === 0 && airBlocks === 0) {
            resultBody = "§cCould not find or access any blocks in the selection (possibly unloaded?).";
        } else if (blockCounts.size === 0 && airBlocks > 0) {
            resultBody = `§fSelection contains only Air (${airBlocks} blocks).`;
        } else {
            const sortedCounts = Array.from(blockCounts.entries()).sort(([, countA], [, countB]) => countB - countA);
            const maxLines = 20;
            for (let i = 0; i < Math.min(sortedCounts.length, maxLines); i++) {
                const [id, count] = sortedCounts[i];
                resultBody += `§b${id.replace('minecraft:', '')}§f: ${count}\n`;
            }
            if (sortedCounts.length > maxLines) {
                resultBody += `§7... and ${sortedCounts.length - maxLines} more types.\n`;
            }
        }
        // showMessage uses MessageFormData, which always shows. This is fine for detailed results.
        await showMessage(player, "Block Count Results", resultBody);
        // No separate success message needed here as showMessage provides the feedback.

    } catch (error) {
        console.error(`[Count Blocks] Error during operation: ${error}\n${error.stack}`);
        sendErrorMessage(player, "An error occurred while counting blocks.");
    }
}

/**
 * Analyzes the player's selection to find unique non-air block types.
 * Respects MAX_COPY_VOLUME limit.
 * @param {Player} player The player whose selection to analyze.
 * @returns {Promise<string[] | null>} A promise that resolves with an array of unique block type IDs, or null on failure.
 */
export async function analyzeSelectionForPalette(player) {
    const selection = getPlayerSelection(player);
    const { posA, posB, selectionState } = selection;
    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return null;
    }
    const min = getMinCorner(posA, posB);
    const max = getMaxCorner(posA, posB);
    const volume = calculateVolume(min, max);

    if (volume <= 0) {
        sendErrorMessage(player, "Selection empty or invalid.");
        return null;
    }
    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}) to analyze for palette.`);
        return null;
    }

    sendInfoMessage(player, `Analyzing selection for palette...`);
    const uniqueBlockIds = new Set();
    let blocksChecked = 0;
    const dimension = player.dimension;
    const checkInterval = Math.max(10000, Math.floor(volume / 4));

    try {
        for (let x = min.x; x <= max.x; x++) {
            for (let y = min.y; y <= max.y; y++) {
                for (let z = min.z; z <= max.z; z++) {
                    blocksChecked++;
                    if (blocksChecked % checkInterval === 0) {
                        await system.run(() => { });
                    }
                    try {
                        const block = dimension.getBlock({ x, y, z });
                        if (block && !block.isAir) {
                            uniqueBlockIds.add(block.typeId);
                        }
                    } catch (e) { /* Ignore single block errors */ }
                }
            }
        }

        if (uniqueBlockIds.size === 0) {
            sendInfoMessage(player, "No non-air blocks found in selection to create palette from.");
            return null; // Still return null as no palette was created
        }

        const sortedBlockIds = Array.from(uniqueBlockIds).sort();
        // This function is usually called by UI that then uses the result.
        // A direct success message here might be redundant if the UI shows the palette.
        // However, if called directly and not by UI, a message might be good.
        // For now, let's assume the caller (UI) will provide feedback.
        // sendSuccessMessage(player, `Found ${sortedBlockIds.length} unique block types for palette.`);
        return sortedBlockIds;

    } catch (error) {
        console.error(`[Analyze Palette] Error during operation: ${error}\n${error.stack}`);
        sendErrorMessage(player, "An error occurred while analyzing selection for palette.");
        return null;
    }
}
