// scripts/core/brush.js

import { Player } from "@minecraft/server";
import {
    getPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName
} from "../state/index.js";
import {
    DEFAULT_BRUSH_MATERIAL,
    MAX_VIEW_DISTANCE,
    MAX_BRUSH_VOLUME,
    UNDO_TOOL_ID
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { parseWeightedBlockInput, selectWeightedBlock } from "../utils/parsing.js";
import { validateBlockIds, getSingleBlockPermutation } from "../utils/validation.js";
import { calculateVolume } from "../utils/location.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Applies the brush tool to the world based on the player's current settings.
 * Handles different shapes, sizes, hollow option, softness, and material sources.
 * Material source can be 'manual', 'active', or a specific saved palette name.
 * @param {Player} player The player using the brush tool.
 */
export async function applyBrush(player) {
    const selectionData = getPlayerSelection(player);
    const {
        brushShape,
        brushSize,
        brushHollow,
        brushSoftnessAmount,
        brushMaterialSource,
        brushManualMaterial,
        activePaletteData
    } = selectionData;

    // 1. Determine Center Location
    const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
    if (!viewBlockHit?.block) {
        sendErrorMessage(player, "No block in view to apply brush.");
        return;
    }
    const center = viewBlockHit.block.location;
    const dimension = player.dimension;

    // 2. Determine Block Permutation(s) and Source Description
    let blockPermutation = undefined;
    let blockSourceDesc = ""; // For user feedback
    let useWeighted = false;
    let weightedBlocks = [];

    const source = brushMaterialSource ?? 'manual';

    try {
        if (source === 'manual') {
            blockSourceDesc = "manual input";
            if (!brushManualMaterial || brushManualMaterial.trim() === '') {
                sendInfoMessage(player, `Manual material empty. Using default: ${DEFAULT_BRUSH_MATERIAL}`, true);
                blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            } else {
                const parsedResult = parseWeightedBlockInput(brushManualMaterial);
                if (parsedResult.errors.length > 0 || parsedResult.blocks.length === 0) {
                    sendErrorMessage(player, `Could not parse manual material. Using default: ${DEFAULT_BRUSH_MATERIAL}`);
                    blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                } else {
                    const { validPermutations, invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id));
                    if (invalidIds.length > 0) {
                        sendErrorMessage(player, `Invalid block ID(s) in manual material: ${invalidIds.join(', ')}. Using default.`);
                        blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                        blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                    } else {
                        weightedBlocks = parsedResult.blocks;
                        useWeighted = weightedBlocks.length > 1 || weightedBlocks[0]?.weight !== 1;
                        if (!useWeighted) {
                            blockPermutation = validPermutations[0];
                            blockSourceDesc = `manual (${weightedBlocks[0].id.replace('minecraft:', '')})`;
                        } else {
                            blockSourceDesc = `manual (${weightedBlocks.length} types)`;
                        }
                    }
                }
            }
        } else if (source === 'active') {
            blockSourceDesc = "Active Palette";
            if (!activePaletteData || activePaletteData.length === 0) {
                sendInfoMessage(player, `Active Palette empty. Using default: ${DEFAULT_BRUSH_MATERIAL}`, true);
                blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                useWeighted = false;
                weightedBlocks = [];
            } else {
                const { validPermutations, invalidIds } = validateBlockIds(activePaletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Active Palette contains invalid IDs: ${invalidIds.join(', ')}. Using default.`);
                    blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                    useWeighted = false;
                    weightedBlocks = [];
                } else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Active Palette has no valid blocks. Using default.`);
                    blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                    useWeighted = false;
                    weightedBlocks = [];
                } else {
                    weightedBlocks = activePaletteData;
                    useWeighted = weightedBlocks.length > 1 || weightedBlocks[0]?.weight !== 1;
                    if (!useWeighted) {
                        blockPermutation = validPermutations[0];
                    }
                }
            }
        } else { // Source is a specific saved palette name
            const palettes = getPlayerPalettes(player);
            const paletteNameToUse = source;
            const paletteData = palettes[paletteNameToUse];
            blockSourceDesc = `saved palette ('${paletteNameToUse}')`;
            useWeighted = false;
            weightedBlocks = [];
            blockPermutation = undefined;

            if (!paletteData || paletteData.length === 0) {
                sendErrorMessage(player, `Saved palette '${paletteNameToUse}' is empty or invalid. Using default.`);
                blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            } else {
                const { validPermutations, invalidIds } = validateBlockIds(paletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' contains invalid IDs: ${invalidIds.join(', ')}. Using default.`);
                    blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' has no valid blocks. Using default.`);
                    blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                }
                else {
                    weightedBlocks = paletteData;
                    useWeighted = weightedBlocks.length > 1 || weightedBlocks[0]?.weight !== 1;
                    if (!useWeighted) { blockPermutation = validPermutations[0]; }
                }
            }
            if (!useWeighted && !blockPermutation) { // Fallback if palette somehow failed to resolve to single or weighted
                blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            }
        }

        if (!useWeighted && !blockPermutation) { // Ultimate fallback
            if (!DEFAULT_BRUSH_MATERIAL) throw new Error("CRITICAL: Could not resolve any block permutation and default material is not set.");
            blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
            if (!blockPermutation) throw new Error(`CRITICAL: Default brush material '${DEFAULT_BRUSH_MATERIAL}' is invalid.`);
            console.warn(`[ApplyBrush] Material resolution failed unexpectedly, falling back to default: ${DEFAULT_BRUSH_MATERIAL}`);
            blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            useWeighted = false;
        }
    } catch (error) {
        console.error(`[ApplyBrush] Error determining material: ${error}\n${error.stack}`);
        sendErrorMessage(player, `Error setting brush material: ${error.message}. Using default.`);
        blockPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
        blockSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
        useWeighted = false;
        weightedBlocks = [];
        if (!blockPermutation) {
            sendErrorMessage(player, `CRITICAL: Default brush material '${DEFAULT_BRUSH_MATERIAL}' is invalid. Cannot apply brush.`);
            return;
        }
    }

    // 3. Calculate Target Blocks
    const targetLocations = [];
    const radius = Math.max(0, Math.floor(brushSize));
    const radiusSq = radius * radius;
    const innerRadius = Math.max(0, radius - 1);
    const innerRadiusSq = innerRadius * innerRadius;
    const coneHeight = radius;
    const torusMinorRadius = 2;
    const torusMinorRadiusSq = torusMinorRadius * torusMinorRadius;
    const torusMajorRadius = Math.max(0, radius - torusMinorRadius);
    const softnessFactor = (brushSoftnessAmount ?? 0) / 100.0;

    let yMinOffset = -radius; let yMaxOffset = radius;
    switch (brushShape) {
        case 'cone_down': yMinOffset = -coneHeight; yMaxOffset = 0; break;
        case 'pyramid': case 'cone_up': yMinOffset = 0; yMaxOffset = radius; break;
        case 'dome': yMinOffset = 0; yMaxOffset = radius; break;
        case 'bowl': yMinOffset = -radius; yMaxOffset = 0; break;
        case 'cylinder': yMinOffset = -(radius > 0 ? radius - 1 : 0); yMaxOffset = radius; if (radius === 0) { yMinOffset = 0; yMaxOffset = 0; } break;
        case 'disk': case 'plate': yMinOffset = 0; yMaxOffset = 0; break;
        case 'torus': yMinOffset = -torusMinorRadius; yMaxOffset = torusMinorRadius; break;
    }
    const startY = center.y + yMinOffset;
    const endY = center.y + yMaxOffset;
    let minBound = { x: center.x - radius, y: Math.min(startY, endY), z: center.z - radius };
    let maxBound = { x: center.x + radius, y: Math.max(startY, endY), z: center.z + radius };

    let estimatedVolume = 0;
    const softDesc = softnessFactor > 0 ? ` (Soft ${brushSoftnessAmount}%)` : '';
    const hollowDesc = softnessFactor <= 0 && brushHollow ? ' (Hollow)' : '';

    try {
        for (let x = minBound.x; x <= maxBound.x; x++) {
            for (let y = minBound.y; y <= maxBound.y; y++) {
                for (let z = minBound.z; z <= maxBound.z; z++) {
                    let include = false;
                    const dx = x - center.x; const dy = y - center.y; const dz = z - center.z;
                    const distSqXZ = dx * dx + dz * dz; const distSqXYZ = distSqXZ + dy * dy;
                    const manhattanDist = Math.abs(dx) + Math.abs(dy) + Math.abs(dz);
                    let isInsideHard = false;
                    // Shape logic (unchanged, complex)
                    switch (brushShape) {
                        case 'sphere': { let minCornerDistSq = Infinity; for (let cX = x; cX <= x + 1; cX++)for (let cY = y; cY <= y + 1; cY++)for (let cZ = z; cZ <= z + 1; cZ++) { const cdx = cX - (center.x + 0.5), cdy = cY - (center.y + 0.5), cdz = cZ - (center.z + 0.5); minCornerDistSq = Math.min(minCornerDistSq, cdx * cdx + cdy * cdy + cdz * cdz); } if (minCornerDistSq <= radiusSq) { if (!brushHollow || radius === 0 || distSqXYZ > innerRadiusSq) isInsideHard = true; } break; }
                        case 'cube': if (Math.abs(dx) <= radius && Math.abs(dy) <= radius && Math.abs(dz) <= radius) { if (!brushHollow || radius === 0 || Math.abs(dx) === radius || Math.abs(dy) === radius || Math.abs(dz) === radius) isInsideHard = true; } break;
                        case 'cylinder': { if (y >= startY && y <= endY) { let minCornerDistSqXZ = Infinity; for (let cX = x; cX <= x + 1; cX++)for (let cZ = z; cZ <= z + 1; cZ++) { const cdx = cX - (center.x + 0.5), cdz = cZ - (center.z + 0.5); minCornerDistSqXZ = Math.min(minCornerDistSqXZ, cdx * cdx + cdz * cdz); } if (minCornerDistSqXZ <= radiusSq) { let blockIsHollowedOut = false; if (brushHollow && radius > 0 && distSqXZ <= innerRadiusSq && y > startY && y < endY) blockIsHollowedOut = true; if (!blockIsHollowedOut) isInsideHard = true; } } break; }
                        case 'pyramid': { const apexY = center.y + coneHeight; const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = radius * (coneHeight - relY) / coneHeight; if (Math.abs(dx) <= maxDist && Math.abs(dz) <= maxDist) { if (!brushHollow || radius === 0 || relY === 0 || Math.abs(dx) >= maxDist - 1 || Math.abs(dz) >= maxDist - 1) { isInsideHard = true; } } } break; }
                        case 'cone_up': { const apexY = center.y + coneHeight; const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = radius * (coneHeight - relY) / coneHeight; const maxDistSq = maxDist * maxDist; if (distSqXZ <= maxDistSq) { if (!brushHollow || radius === 0 || relY === 0 || distSqXZ >= Math.pow(Math.max(0, maxDist - 1), 2)) { isInsideHard = true; } } } break; }
                        case 'cone_down': { const apexY = center.y - coneHeight; const relY = y - center.y; if (relY <= 0 && relY >= -coneHeight) { const maxDist = radius * (coneHeight + relY) / coneHeight; const maxDistSq = maxDist * maxDist; if (distSqXZ <= maxDistSq) { if (!brushHollow || radius === 0 || relY === 0 || distSqXZ >= Math.pow(Math.max(0, maxDist - 1), 2)) { isInsideHard = true; } } } break; }
                        case 'octahedron': if (manhattanDist <= radius) { if (!brushHollow || radius === 0 || manhattanDist === radius) isInsideHard = true; } break;
                        case 'dome': if (y >= center.y && distSqXYZ <= radiusSq) { if (!brushHollow || radius === 0 || distSqXYZ > innerRadiusSq || y === center.y) isInsideHard = true; } break;
                        case 'bowl': if (y <= center.y && distSqXYZ <= radiusSq) { if (!brushHollow || radius === 0 || distSqXYZ > innerRadiusSq || y === center.y) isInsideHard = true; } break;
                        case 'disk': if (y === center.y && distSqXZ <= radiusSq) { if (!brushHollow || radius === 0 || distSqXZ > innerRadiusSq) isInsideHard = true; } break;
                        case 'plate': if (y === center.y && Math.abs(dx) <= radius && Math.abs(dz) <= radius) { if (!brushHollow || radius === 0 || Math.abs(dx) === radius || Math.abs(dz) === radius) isInsideHard = true; } break;
                        case 'torus': { const hDist = Math.sqrt(distSqXZ); const dSqTube = Math.pow(hDist - torusMajorRadius, 2) + dy * dy; if (dSqTube <= torusMinorRadiusSq) { const innerTubeSq = Math.pow(Math.max(0, torusMinorRadius - 1), 2); if (!brushHollow || torusMinorRadius === 0 || dSqTube > innerTubeSq) isInsideHard = true; } break; }
                    }
                    // Softness logic (unchanged, complex)
                    if (softnessFactor <= 0) { include = isInsideHard; }
                    else {
                        let normalizedDistance = 1.1;
                        switch (brushShape) {
                            case 'sphere': normalizedDistance = (radius > 0) ? Math.sqrt(distSqXYZ) / radius : 1.1; break;
                            case 'cube': normalizedDistance = (radius > 0) ? Math.max(Math.abs(dx), Math.abs(dy), Math.abs(dz)) / radius : 1.1; break;
                            case 'cylinder': normalizedDistance = (radius > 0 && y >= startY && y <= endY) ? Math.sqrt(distSqXZ) / radius : 1.1; break;
                            case 'pyramid': { const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = (coneHeight > 0) ? radius * (coneHeight - relY) / coneHeight : 0; normalizedDistance = (maxDist > 0) ? Math.max(Math.abs(dx), Math.abs(dz)) / maxDist : 1.1; } else { normalizedDistance = 1.1; } break; }
                            case 'cone_up': { const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = (coneHeight > 0) ? radius * (coneHeight - relY) / coneHeight : 0; normalizedDistance = (maxDist > 0) ? Math.sqrt(distSqXZ) / maxDist : 1.1; } else { normalizedDistance = 1.1; } break; }
                            case 'cone_down': { const relY = y - center.y; if (relY <= 0 && relY >= -coneHeight) { const maxDist = (coneHeight > 0) ? radius * (coneHeight + relY) / coneHeight : 0; normalizedDistance = (maxDist > 0) ? Math.sqrt(distSqXZ) / maxDist : 1.1; } else { normalizedDistance = 1.1; } break; }
                            case 'octahedron': normalizedDistance = (radius > 0) ? manhattanDist / radius : 1.1; break;
                            case 'dome': normalizedDistance = (radius > 0 && y >= center.y) ? Math.sqrt(distSqXYZ) / radius : 1.1; break;
                            case 'bowl': normalizedDistance = (radius > 0 && y <= center.y) ? Math.sqrt(distSqXYZ) / radius : 1.1; break;
                            case 'disk': normalizedDistance = (radius > 0 && y === center.y) ? Math.sqrt(distSqXZ) / radius : 1.1; break;
                            case 'plate': normalizedDistance = (radius > 0 && y === center.y) ? Math.max(Math.abs(dx), Math.abs(dz)) / radius : 1.1; break;
                            case 'torus': { const hDist = Math.sqrt(distSqXZ); normalizedDistance = (torusMinorRadius > 0) ? Math.sqrt(Math.pow(hDist - torusMajorRadius, 2) + dy * dy) / torusMinorRadius : 1.1; break; }
                        }
                        normalizedDistance = Math.max(0.0, Math.min(1.0, normalizedDistance)); const t = 1.0 - normalizedDistance; let p_falloff = 0.0;
                        if (brushShape === 'cone_up' || brushShape === 'cone_down' || brushShape === 'pyramid') { p_falloff = t * t * t; } else { p_falloff = t * t * (3.0 - 2.0 * t); }
                        const hardEdgeProbability = isInsideHard ? 1.0 : 0.0; const finalProbability = hardEdgeProbability * (1.0 - softnessFactor) + p_falloff * softnessFactor;
                        if (finalProbability >= 0.0001 && Math.random() < finalProbability) { include = true; }
                    }
                    if (include) { targetLocations.push({ x, y, z }); }
                }
            }
        }
        estimatedVolume = targetLocations.length;
    } catch (e) {
        console.error(`[Brush Calculation] Error: ${e}\n${e.stack}`);
        sendErrorMessage(player, "Error calculating brush shape.");
        return;
    }

    if (targetLocations.length === 0) {
        sendInfoMessage(player, "Brush calculation resulted in zero blocks to place.");
        return;
    }

    if (estimatedVolume > MAX_BRUSH_VOLUME) {
        sendErrorMessage(player, `Brush area too large (${estimatedVolume} > ${MAX_BRUSH_VOLUME}). Reduce brush size or adjust shape.`);
        return;
    }

    // 5. Save Undo State
    const beforeStateBlocks = getBlocksInArea(dimension, minBound, maxBound); // Use original bounds for undo capture
    let actualMin = targetLocations.length > 0 ? { ...targetLocations[0] } : { ...center };
    let actualMax = targetLocations.length > 0 ? { ...targetLocations[0] } : { ...center };
    if (targetLocations.length > 0) {
        for (const loc of targetLocations) {
            actualMin.x = Math.min(actualMin.x, loc.x); actualMin.y = Math.min(actualMin.y, loc.y); actualMin.z = Math.min(actualMin.z, loc.z);
            actualMax.x = Math.max(actualMax.x, loc.x); actualMax.y = Math.max(actualMax.y, loc.y); actualMax.z = Math.max(actualMax.z, loc.z);
        }
    }
    const undoState = { affectedArea: { min: actualMin, max: actualMax }, blocks: beforeStateBlocks, timestamp: Date.now() };
    if (beforeStateBlocks.length === 0 && calculateVolume(minBound, maxBound) > 0 && estimatedVolume > 0) {
        sendInfoMessage(player, "Warning: Failed to capture undo state (area possibly unloaded?). Undo might be incomplete.", true);
    }

    // 6. Perform Brush Placement
    let blocksPlaced = 0;
    sendInfoMessage(player, `Applying ${brushShape} brush with ${blockSourceDesc}`);
    const permutationCache = new Map();

    try {
        for (const loc of targetLocations) {
            try {
                const block = dimension.getBlock(loc);
                if (block) {
                    let permToSet = blockPermutation;
                    if (useWeighted) {
                        const selectedId = selectWeightedBlock(weightedBlocks);
                        if (selectedId) {
                            if (permutationCache.has(selectedId)) {
                                permToSet = permutationCache.get(selectedId);
                            } else {
                                permToSet = getSingleBlockPermutation(selectedId);
                                if (permToSet) {
                                    permutationCache.set(selectedId, permToSet);
                                } else {
                                    // console.warn(`[ApplyBrush] Failed to resolve weighted block ID: ${selectedId}`); 
                                    continue;
                                }
                            }
                        } else {
                            // console.warn(`[ApplyBrush] selectWeightedBlock failed`); 
                            continue;
                        }
                    }
                    if (permToSet) {
                        block.setPermutation(permToSet);
                        blocksPlaced++;
                    }
                }
            } catch (e) { /* Ignore single block errors */ }
        }

        // 7. Push Undo and Report Result
        if (blocksPlaced > 0) {
            pushUndoState(player, undoState);
        } else {
        }

    } catch (error) {
        console.error(`[WorldEdit Brush Apply] Error during operation: ${error}\n${error.stack}`);
        sendErrorMessage(player, `An error occurred during brush application. Undo might be needed.`);
        if (beforeStateBlocks.length > 0) {
            pushUndoState(player, undoState);
            sendInfoMessage(player, "Undo state saved despite error.", true);
        }
    }
}
