// scripts/core/clipboard.js

import {
    Player,
    BlockPermutation,
    world,
    system,
    StructureSaveMode,
    StructureRotation,
    StructureMirrorAxis,
} from "@minecraft/server";
import {
    getPlayerSelection,
    getPlayerClipboard,
    setPlayerClipboard
} from "../state/index.js";
import {
    MAX_COPY_VOLUME,
    STRUCTURE_NAMESPACE,
    UNDO_TOOL_ID,
    PASTE_TOOL_ID,
    TICKING_AREA_THRESHOLD_VOLUME,
    TICKING_AREA_LOAD_DELAY_TICKS
} from "../constants/index.js";
import {
    getMinCorner,
    getMaxCorner,
    calculateVolume,
    getPlayerDirection
} from "../utils/location.js";
import {
    getBlocksInArea,
    pushUndoState
} from "../utils/history_helpers.js";
import {
    addTemporaryTickingArea,
    removeTemporaryTickingArea
} from "../utils/tickingarea_manager.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

export function copySelection(player) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;
    const playerStateCopyEntitiesFlag = selectionData.copyIncludesEntities;
    const includeEntitiesFinal = playerStateCopyEntitiesFlag === true;

    const { posA, posB, selectionState } = selectionData;

    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return null;
    }
    const min = getMinCorner(posA, posB);
    const max = getMaxCorner(posA, posB);
    const volume = calculateVolume(min, max);

    if (volume <= 0) {
        sendErrorMessage(player, "Selection empty or invalid.");
        return null;
    }
    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}).`);
        return null;
    }
    const structureId = `${STRUCTURE_NAMESPACE}:clipboard_${player.id}`;
    const dimension = player.dimension;
    // player.sendMessage(`§eCopying ${volume} blocks${includeEntitiesFinal ? ' and entities' : ''} to structure '${structureId}'...`); // REMOVED intermediate message

    try {
        try {
            world.structureManager.delete(structureId);
        } catch (e) {
            if (e.message && !e.message.toLowerCase().includes("structure not found")) {
                // Silently ignore non-critical delete errors
            }
        }
        const createOptions = {
            includeEntities: includeEntitiesFinal,
            includeBlocks: true,
            saveMode: StructureSaveMode.Memory
        };
        const structure = world.structureManager.createFromWorld(structureId, dimension, min, max, createOptions);


        if (!structure || !structure.isValid) {
            throw new Error("Structure creation failed or returned an invalid structure object.");
        }
        const clipboardData = {
            structureId: structure.id,
            size: { ...structure.size },
            includesEntities: includeEntitiesFinal,
            timestamp: Date.now(),
            origin: { ...min }
        };
        setPlayerClipboard(player, clipboardData);
        if (showFeedback) {
            // player.sendMessage(`§aCopied ${volume} blocks${includeEntitiesFinal ? ' and entities' : ''} to memory clipboard (${structure.id}). Size: ${structure.size.x}x${structure.size.y}x${structure.size.z}`);
            sendSuccessMessage(player, "Selection copied."); // Simplified message
        }
        return clipboardData;
    } catch (error) {
        sendErrorMessage(player, `Error copying selection: ${error.message}`);
        try { world.structureManager.delete(structureId); } catch { /* ignore cleanup error */ }
        setPlayerClipboard(player, undefined);
        return null;
    }
}

export async function cutSelection(player) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;

    // player.sendMessage("§eAttempting to cut selection..."); // REMOVED

    const clipboardData = copySelection(player);

    if (clipboardData) {
        const currentSelectionData = getPlayerSelection(player);
        const { posA, posB, selectionState } = currentSelectionData;
        if (!posA || !posB || selectionState !== 2) {
            sendErrorMessage(player, "Cannot remove blocks: Selection is invalid (internal error after copy).");
            return;
        }
        const originalMin = getMinCorner(posA, posB);
        const originalMax = getMaxCorner(posA, posB);
        const volumeToClear = calculateVolume(originalMin, originalMax);
        const dimension = player.dimension;
        let tempTickingAreaName = null;
        let tickingAreaBoundsForUndo = null;
        try {
            if (volumeToClear > TICKING_AREA_THRESHOLD_VOLUME) {
                tempTickingAreaName = await addTemporaryTickingArea(dimension, originalMin, originalMax, "cut_remove");
                if (tempTickingAreaName) {
                    // player.sendMessage("§eEnsuring cut area is loaded for removal..."); // REMOVED
                    tickingAreaBoundsForUndo = { min: { ...originalMin }, max: { ...originalMax } };
                    await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
                } else {
                    sendInfoMessage(player, "Warning: Could not create temporary ticking area for cut removal. Operation might be slow or unstable.", true);
                }
            }
            // player.sendMessage("§eSaving undo state for cut area..."); // REMOVED
            const beforeStateBlocks = getBlocksInArea(dimension, originalMin, originalMax);
            if (beforeStateBlocks.length === 0 && volumeToClear > 0) {
                sendInfoMessage(player, "Warning: Failed to capture undo state (area possibly unloaded?). Undo might be incomplete.", true);
            }
            const undoState = {
                affectedArea: { min: originalMin, max: originalMax },
                blocks: beforeStateBlocks,
                timestamp: Date.now()
            };
            if (tickingAreaBoundsForUndo) {
                undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
                undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
            }
            let blocksRemoved = 0;
            let airPermutation;
            try { airPermutation = BlockPermutation.resolve("minecraft:air"); }
            catch (e) {
                sendErrorMessage(player, "Internal error: Could not find air block for removal.");
                return;
            }
            // player.sendMessage(`§eRemoving ${volumeToClear} blocks...`); // REMOVED
            for (let x = originalMin.x; x <= originalMax.x; x++) {
                for (let y = originalMin.y; y <= originalMax.y; y++) {
                    for (let z = originalMin.z; z <= originalMax.z; z++) {
                        try {
                            const block = dimension.getBlock({ x, y, z });
                            if (block && !block.isAir) {
                                block.setPermutation(airPermutation);
                                blocksRemoved++;
                            }
                        } catch (e) { /* Ignore single block errors */ }
                    }
                }
            }
            if (blocksRemoved > 0 || (beforeStateBlocks.length === 0 && volumeToClear > 0)) {
                pushUndoState(player, undoState);
                if (showFeedback) {
                    // player.sendMessage(`§aCut ${clipboardData.size.x}x${clipboardData.size.y}x${clipboardData.size.z} selection to memory clipboard (${blocksRemoved} blocks removed). Use Paste tool (${PASTE_TOOL_ID.replace('bn:', '')}) to place.`);
                    sendSuccessMessage(player, `Selection cut. Use Paste tool to place.`);
                }
            } else {
                if (showFeedback) {
                    // player.sendMessage(`§eCut complete. No blocks needed removing (area already empty).`);
                    sendInfoMessage(player, "Cut complete. No blocks needed removing.");
                }
            }
        } catch (error) {
            sendErrorMessage(player, `An error occurred while removing blocks during cut. Undo might be needed.`);
        } finally {
            if (tempTickingAreaName) {
                await removeTemporaryTickingArea(dimension, tempTickingAreaName);
            }
        }
    }
}

export async function pasteClipboardRotated(player, targetBlockLocation) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;
    const clipboardData = getPlayerClipboard(player);
    const pasteOptionsFromState = selectionData.pasteOptions ?? {
        pasteOrigin: false,
        placeEntities: true,
        integrity: 1.0,
        seed: "",
        rotation: "None (0°)",
        mirror: "None"
    };

    if (!clipboardData?.structureId || !clipboardData?.size) {
        sendErrorMessage(player, "Memory clipboard is empty or invalid. Use Copy/Cut first.");
        return;
    }

    const { structureId, size: structureSize, includesEntities: structureActuallyIncludesEntities, origin: copyOrigin } = clipboardData;
    const dimension = player.dimension;

    let finalRotation = StructureRotation.None;
    switch (pasteOptionsFromState.rotation) {
        case "90°": finalRotation = StructureRotation.Rotate90; break;
        case "180°": finalRotation = StructureRotation.Rotate180; break;
        case "270°": finalRotation = StructureRotation.Rotate270; break;
    }

    let placeSize = { ...structureSize };
    if (finalRotation === StructureRotation.Rotate90 || finalRotation === StructureRotation.Rotate270) {
        placeSize = { x: structureSize.z, y: structureSize.y, z: structureSize.x };
    }

    let placementLocation = { x: 0, y: 0, z: 0 };
    if (pasteOptionsFromState.pasteOrigin && copyOrigin) {
        placementLocation = { ...copyOrigin };
        // player.sendMessage("§ePasting at original copy location..."); // REMOVED
    } else {
        placementLocation.y = targetBlockLocation.y;
        const playerDirection = getPlayerDirection(player.getViewDirection());
        switch (playerDirection) {
            case "North": placementLocation.x = targetBlockLocation.x - Math.floor(placeSize.x / 2); placementLocation.z = targetBlockLocation.z - placeSize.z + 1; break;
            case "South": placementLocation.x = targetBlockLocation.x - Math.floor(placeSize.x / 2); placementLocation.z = targetBlockLocation.z + 1; break;
            case "West": placementLocation.x = targetBlockLocation.x - placeSize.x + 1; placementLocation.z = targetBlockLocation.z - Math.floor(placeSize.z / 2); break;
            case "East": placementLocation.x = targetBlockLocation.x + 1; placementLocation.z = targetBlockLocation.z - Math.floor(placeSize.z / 2); break;
            default: placementLocation.x = targetBlockLocation.x - Math.floor(placeSize.x / 2); placementLocation.z = targetBlockLocation.z - placeSize.z + 1; break;
        }
    }

    const placementMin = { ...placementLocation };
    const placementMax = {
        x: placementLocation.x + placeSize.x - 1,
        y: placementLocation.y + placeSize.y - 1,
        z: placementLocation.z + placeSize.z - 1,
    };
    const affectedAreaForPaste = { min: placementMin, max: placementMax };
    const affectedVolume = calculateVolume(affectedAreaForPaste.min, affectedAreaForPaste.max);
    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;

    try {
        if (affectedVolume > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, affectedAreaForPaste.min, affectedAreaForPaste.max, "paste");
            if (tempTickingAreaName) {
                // player.sendMessage("§eEnsuring paste area is loaded..."); // REMOVED
                tickingAreaBoundsForUndo = { min: { ...affectedAreaForPaste.min }, max: { ...affectedAreaForPaste.max } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area. Paste operation might be slow or unstable.", true);
            }
        }
        // player.sendMessage("§eSaving undo state for paste area..."); // REMOVED
        const beforeStateBlocks = getBlocksInArea(dimension, affectedAreaForPaste.min, affectedAreaForPaste.max);
        if (beforeStateBlocks.length === 0 && affectedVolume > 0) {
            sendInfoMessage(player, "Warning: Failed to capture undo state (paste area possibly unloaded?). Undo might be incomplete.", true);
        }
        const undoState = {
            affectedArea: affectedAreaForPaste,
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }

        const includeEntitiesForPlacement = structureActuallyIncludesEntities && pasteOptionsFromState.placeEntities === true;

        const placeOptionsObj = {
            rotation: finalRotation,
            mirror: StructureMirrorAxis.None,
            includeEntities: includeEntitiesForPlacement,
            includeBlocks: true,
            integrity: pasteOptionsFromState.integrity,
            seed: (pasteOptionsFromState.integrity < 1.0 && pasteOptionsFromState.seed) ? pasteOptionsFromState.seed : undefined,
        };

        switch (pasteOptionsFromState.mirror) {
            case "X": placeOptionsObj.mirror = StructureMirrorAxis.X; break;
            case "Z": placeOptionsObj.mirror = StructureMirrorAxis.Z; break;
            case "XZ": placeOptionsObj.mirror = StructureMirrorAxis.XZ; break;
        }

        // const rotationDesc = pasteOptionsFromState.rotation === "None (0°)" ? "0°" : pasteOptionsFromState.rotation;
        // player.sendMessage(`§ePlacing structure '${structureId}' (Rotation: ${rotationDesc}, Mirror: ${pasteOptionsFromState.mirror}, Integrity: ${pasteOptionsFromState.integrity * 100}%, Entities: ${includeEntitiesForPlacement ? 'ON' : 'OFF'})...`); // REMOVED

        world.structureManager.place(structureId, dimension, placementLocation, placeOptionsObj);

        if (beforeStateBlocks.length > 0 || affectedVolume > 0) {
            pushUndoState(player, undoState);
        }
        if (showFeedback) {
            sendSuccessMessage(player, `Paste complete. Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
        }
    } catch (error) {
        let errorMsg = error.message || "An unknown error occurred.";
        if (error.name === "InvalidStructureError") {
            errorMsg = `Structure '${structureId}' is invalid or not found in memory. Try copying again.`;
            setPlayerClipboard(player, undefined);
        } else if (error.message?.includes("outside the world bounds")) {
            errorMsg = "Placement location is partially outside the world bounds.";
        } else if (error.message?.includes("Structure placement failed")) {
            errorMsg = "Structure placement failed (internal engine error).";
        }
        sendErrorMessage(player, `Error pasting structure: ${errorMsg}`);
    } finally {
        if (tempTickingAreaName) {
            await removeTemporaryTickingArea(dimension, tempTickingAreaName);
        }
    }
}
