// scripts/core/edit_selection.js

import {
    Player,
    BlockPermutation,
    world,
    StructureSaveMode,
    StructureRotation,
    StructureMirrorAxis,
    system,
    Dimension
} from "@minecraft/server";
import { getPlayerSelection } from "../state/index.js";
import {
    MAX_COPY_VOLUME,
    MAX_STACK_AFFECTED_VOLUME,
    STRUCTURE_NAMESPACE,
    UNDO_TOOL_ID,
    BLOCKS_PER_CHUNK,
    TICKING_AREA_THRESHOLD_VOLUME,
    TICKING_AREA_LOAD_DELAY_TICKS
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { getMinCorner, getMaxCorner, calculateVolume } from "../utils/location.js";
import {
    addTemporaryTickingArea,
    removeTemporaryTickingArea
} from "../utils/tickingarea_manager.js";
import {
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Asynchronous generator to clear an area in chunks.
 * Yields execution to the system periodically to prevent watchdog timeouts.
 * @param {Dimension} dimension
 * @param {{x:number, y:number, z:number}} min
 * @param {{x:number, y:number, z:number}} max
 * @param {BlockPermutation} airPermutation
 * @yields {number} Number of blocks cleared in the processed segment.
 */
async function* processClearChunk(dimension, min, max, airPermutation) {
    let blocksClearedInCurrentSegment = 0;
    let totalProcessedInSegment = 0;

    for (let x = min.x; x <= max.x; x++) {
        for (let y = min.y; y <= max.y; y++) {
            for (let z = min.z; z <= max.z; z++) {
                try {
                    const block = dimension.getBlock({ x, y, z });
                    if (block && !block.isAir) {
                        block.setPermutation(airPermutation);
                        blocksClearedInCurrentSegment++;
                    }
                } catch (e) { /* Ignore single block errors, e.g., block became invalid */ }

                totalProcessedInSegment++;
                if (totalProcessedInSegment >= BLOCKS_PER_CHUNK) {
                    await system.run(() => { });
                    if (blocksClearedInCurrentSegment > 0) {
                        yield blocksClearedInCurrentSegment;
                    }
                    blocksClearedInCurrentSegment = 0;
                    totalProcessedInSegment = 0;
                }
            }
        }
    }

    if (blocksClearedInCurrentSegment > 0) {
        yield blocksClearedInCurrentSegment;
    }
}

/**
 * Edits the blocks within the current selection "in-place" by applying
 * rotation, mirroring, and offsets using a temporary structure.
 *
 * @param {Player} player The player performing the action.
 * @param {string} rotationString String representation from dropdown ("None (0°)", "90°", "180°", "270°").
 * @param {string} mirrorString String representation from dropdown ("None", "X", "Z", "XZ").
 * @param {number} offsetX Relative offset in X.
 * @param {number} offsetY Relative offset in Y.
 * @param {number} offsetZ Relative offset in Z.
 * @param {boolean} includeEntities Whether to include entities in the copy/paste process.
 */
export async function editSelectionInPlace(player, rotationString, mirrorString, offsetX, offsetY, offsetZ, includeEntities) {
    const selection = getPlayerSelection(player);
    const { posA, posB, selectionState } = selection;
    const dimension = player.dimension;

    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return;
    }
    const originalMin = getMinCorner(posA, posB);
    const originalMax = getMaxCorner(posA, posB);
    const originalVolume = calculateVolume(originalMin, originalMax);
    if (originalVolume <= 0) {
        sendErrorMessage(player, "Selection empty or invalid.");
        return;
    }
    if (originalVolume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${originalVolume} > ${MAX_COPY_VOLUME}) to edit.`);
        return;
    }

    let rotation = StructureRotation.None;
    if (rotationString === "90°") rotation = StructureRotation.Rotate90;
    else if (rotationString === "180°") rotation = StructureRotation.Rotate180;
    else if (rotationString === "270°") rotation = StructureRotation.Rotate270;

    let mirror = StructureMirrorAxis.None;
    if (mirrorString === "X") mirror = StructureMirrorAxis.X;
    else if (mirrorString === "Z") mirror = StructureMirrorAxis.Z;
    else if (mirrorString === "XZ") mirror = StructureMirrorAxis.XZ;

    const tempStructureId = `${STRUCTURE_NAMESPACE}:tempedit_${player.id}`;
    sendInfoMessage(player, "Preparing to edit selection...");

    const placementLocation = { x: originalMin.x + offsetX, y: originalMin.y + offsetY, z: originalMin.z + offsetZ };
    let tempSizeForCalc = { x: originalMax.x - originalMin.x + 1, y: originalMax.y - originalMin.y + 1, z: originalMax.z - originalMin.z + 1 };
    if (rotation === StructureRotation.Rotate90 || rotation === StructureRotation.Rotate270) {
        tempSizeForCalc = { x: tempSizeForCalc.z, y: tempSizeForCalc.y, z: tempSizeForCalc.x };
    }
    const placementMin = { ...placementLocation };
    const placementMax = { x: placementLocation.x + tempSizeForCalc.x - 1, y: placementLocation.y + tempSizeForCalc.y - 1, z: placementLocation.z + tempSizeForCalc.z - 1 };

    const finalMinUndo = { x: Math.min(originalMin.x, placementMin.x), y: Math.min(originalMin.y, placementMin.y), z: Math.min(originalMin.z, placementMin.z) };
    const finalMaxUndo = { x: Math.max(originalMax.x, placementMax.x), y: Math.max(originalMax.y, placementMax.y), z: Math.max(originalMax.z, placementMax.z) };
    const affectedVolumeForUndo = calculateVolume(finalMinUndo, finalMaxUndo);

    if (affectedVolumeForUndo > MAX_STACK_AFFECTED_VOLUME) {
        sendErrorMessage(player, `Operation affects too large an area (${affectedVolumeForUndo} > ${MAX_STACK_AFFECTED_VOLUME}). Reduce selection or offset.`);
        return;
    }

    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;

    try {
        if (affectedVolumeForUndo > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, finalMinUndo, finalMaxUndo, "edit_selection");
            if (tempTickingAreaName) {
                sendInfoMessage(player, "Ensuring operation area is loaded...");
                tickingAreaBoundsForUndo = { min: { ...finalMinUndo }, max: { ...finalMaxUndo } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area. Operation might be slow or unstable.", true);
            }
        }

        sendInfoMessage(player, `Saving selection to temporary structure...`);
        try { world.structureManager.delete(tempStructureId); } catch { /* Ignore */ }
        const createOptions = { includesEntities: includeEntities, includesBlocks: true, saveMode: StructureSaveMode.Memory };
        const tempStructure = world.structureManager.createFromWorld(tempStructureId, dimension, originalMin, originalMax, createOptions);
        if (!tempStructure || !tempStructure.isValid) { throw new Error("Failed to save selection to temporary structure."); }

        sendInfoMessage(player, "Saving undo state...");
        const beforeStateBlocks = getBlocksInArea(dimension, finalMinUndo, finalMaxUndo);
        if (beforeStateBlocks.length === 0 && affectedVolumeForUndo > 0) {
            sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
        }
        const undoState = {
            affectedArea: { min: finalMinUndo, max: finalMaxUndo },
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }

        sendInfoMessage(player, "Clearing original selection area...");
        let airPermutation;
        try { airPermutation = BlockPermutation.resolve("minecraft:air"); }
        catch (e) { throw new Error("Could not resolve air block for clearing."); }

        let totalBlocksCleared = 0;
        const clearIterator = processClearChunk(dimension, originalMin, originalMax, airPermutation);
        for await (const blocksClearedInThisChunk of clearIterator) {
            totalBlocksCleared += blocksClearedInThisChunk;
        }

        sendInfoMessage(player, `Original area cleared. Placing transformed selection...`);

        if (tempTickingAreaName) {
            await system.runTimeout(() => { }, Math.floor(TICKING_AREA_LOAD_DELAY_TICKS / 2));
        }

        const placeOptions = { rotation: rotation, mirror: mirror, includesEntities: includeEntities, includesBlocks: true, ignoreStructureBlocks: false };
        world.structureManager.place(tempStructureId, dimension, placementLocation, placeOptions);

        pushUndoState(player, undoState);
        sendSuccessMessage(player, `Selection edited. Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);

    } catch (error) {
        console.error(`[Edit Selection] Error: ${error}\n${error.stack}`);
        sendErrorMessage(player, `Error editing selection: ${error.message}`);
    } finally {
        try {
            world.structureManager.delete(tempStructureId);
        } catch (deleteError) {
            if (deleteError.message && !deleteError.message.toLowerCase().includes("structure not found")) {
                // console.warn(`[Edit Selection] Non-critical error deleting temp structure ${tempStructureId}: ${deleteError}`);
            }
        }
        if (tempTickingAreaName) {
            await removeTemporaryTickingArea(dimension, tempTickingAreaName);
        }
    }
}
