// scripts/core/fill.js

import { Player, BlockPermutation, system, world } from "@minecraft/server";
// ModalFormData is used by the UI layer, not directly here.
// import { ModalFormData } from "@minecraft/server-ui"; 
import {
    getPlayerSelection,
    // getPlayerPalettes, // Not directly used, UI resolves material
    // getActiveSavedPaletteName // Not directly used
} from "../state/index.js";
import {
    UNDO_TOOL_ID,
    MAX_COPY_VOLUME,
    BLOCKS_PER_CHUNK,
    TICKING_AREA_THRESHOLD_VOLUME,
    TICKING_AREA_LOAD_DELAY_TICKS, // Added for ticking area delay
    // DEFAULT_BRUSH_MATERIAL // Not directly used, UI resolves material
} from "../constants/index.js";
import {
    getMinCorner,
    getMaxCorner,
    calculateVolume
} from "../utils/location.js";
// import { parseWeightedBlockInput, selectWeightedBlock } from "../utils/parsing.js"; // UI handles parsing
import { selectWeightedBlock } from "../utils/parsing.js"; // Only selectWeightedBlock is needed here
import { validateBlockIds, getSingleBlockPermutation } from "../utils/validation.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { addTemporaryTickingArea, removeTemporaryTickingArea } from "../utils/tickingarea_manager.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

// Helper to determine if a block should be filled based on type and position
function shouldFillBlock(fillType, x, y, z, min, max) {
    switch (fillType) {
        case "Solid":
            return true;
        case "Hollow":
            return (x === min.x || x === max.x || y === min.y || y === max.y || z === min.z || z === max.z);
        case "Frame":
            const onX = (x === min.x || x === max.x);
            const onY = (y === min.y || y === max.y);
            const onZ = (z === min.z || z === max.z);
            return ((onX ? 1 : 0) + (onY ? 1 : 0) + (onZ ? 1 : 0)) >= 2;
        case "Walls":
            return (x === min.x || x === max.x || z === min.z || z === max.z);
        default:
            return false;
    }
}

/**
 * Asynchronous generator to process fill operations in chunks.
 * @yields {number} Number of blocks filled in the current chunk.
 */
async function* processFillChunk(dimension, min, max, fillType, blocksToUseWeighted, permutationCache) {
    let blocksFilledInChunk = 0;
    let currentProcessed = 0;
    const blockPermutationCache = new Map(permutationCache);

    for (let x = min.x; x <= max.x; x++) {
        for (let y = min.y; y <= max.y; y++) {
            for (let z = min.z; z <= max.z; z++) {
                if (shouldFillBlock(fillType, x, y, z, min, max)) {
                    try {
                        const blockToPlace = dimension.getBlock({ x, y, z });
                        if (blockToPlace) {
                            const selectedBlockId = selectWeightedBlock(blocksToUseWeighted);
                            let permutationToSet;

                            if (selectedBlockId) {
                                if (blockPermutationCache.has(selectedBlockId)) {
                                    permutationToSet = blockPermutationCache.get(selectedBlockId);
                                } else {
                                    const perm = getSingleBlockPermutation(selectedBlockId);
                                    if (perm) {
                                        blockPermutationCache.set(selectedBlockId, perm);
                                        permutationToSet = perm;
                                    } else {
                                        // console.warn(`[Fill Chunk] Could not resolve permutation for ${selectedBlockId}. Skipping.`);
                                        continue;
                                    }
                                }
                            } else {
                                // console.warn(`[Fill Chunk] selectWeightedBlock returned undefined. Skipping.`);
                                continue;
                            }

                            if (permutationToSet) {
                                if (!blockToPlace.permutation.matches(permutationToSet.type.id, permutationToSet.getAllStates())) {
                                    blockToPlace.setPermutation(permutationToSet);
                                    blocksFilledInChunk++;
                                }
                            }
                        }
                    } catch (e) { /* Ignore single block errors */ }
                }

                currentProcessed++;
                if (currentProcessed >= BLOCKS_PER_CHUNK) {
                    await system.run(() => { }); // Yield execution
                    if (blocksFilledInChunk > 0) yield blocksFilledInChunk;
                    blocksFilledInChunk = 0;
                    currentProcessed = 0;
                }
            }
        }
    }

    if (blocksFilledInChunk > 0) {
        yield blocksFilledInChunk;
    }
}


/** * Fills the player's selection based on the specified fill type.
 * The UI (`showFillOptionsMenu` and its helpers) is responsible for parsing and validating
 * the material source and providing `blocksToUseWeighted` and `sourceDesc`.
 * @param {Player} player The player initiating the action.
 * @param {string} fillType The type of fill ("Solid", "Hollow", "Frame", "Walls").
 * @param {Array<{id: string, weight: number}>} blocksToUseWeighted The resolved list of weighted blocks.
 * @param {string} sourceDesc A description of the material source for messages.
 */
export async function fillSelection(player, fillType, blocksToUseWeighted, sourceDesc) {
    const selection = getPlayerSelection(player);
    const { posA, posB, selectionState } = selection;

    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return;
    }
    const minSelection = getMinCorner(posA, posB);
    const maxSelection = getMaxCorner(posA, posB);
    const volume = calculateVolume(minSelection, maxSelection);

    if (volume <= 0) {
        sendErrorMessage(player, "Selected volume is zero or invalid."); return;
    }
    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}).`); return;
    }

    if (!blocksToUseWeighted || blocksToUseWeighted.length === 0) {
        sendErrorMessage(player, "No blocks provided for filling. Operation cancelled.");
        return;
    }
    // Validation of blocksToUseWeighted is assumed to be done by the UI calling this function.
    // However, a quick re-check for permutations can be good.
    const permutationCache = new Map();
    for (const blockInfo of blocksToUseWeighted) {
        if (!permutationCache.has(blockInfo.id)) {
            const perm = getSingleBlockPermutation(blockInfo.id);
            if (perm) {
                permutationCache.set(blockInfo.id, perm);
            } else {
                sendErrorMessage(player, `Error: Could not resolve fill block ${blockInfo.id}. Aborting.`);
                return;
            }
        }
    }

    const dimension = player.dimension;
    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;

    try {
        if (volume > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, minSelection, maxSelection, "fill");
            if (tempTickingAreaName) {
                sendInfoMessage(player, "Ensuring fill area is loaded...");
                tickingAreaBoundsForUndo = { min: { ...minSelection }, max: { ...maxSelection } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS); // Wait for ticking area
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area. Fill might be slow.", true);
            }
        }

        sendInfoMessage(player, "Saving undo state...");
        const beforeStateBlocks = getBlocksInArea(dimension, minSelection, maxSelection);
        if (beforeStateBlocks.length === 0 && volume > 0) {
            sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
        }
        const undoState = {
            affectedArea: { min: minSelection, max: maxSelection },
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }

        sendInfoMessage(player, `Filling selection (${fillType}, ${volume} blocks)...`);
        let totalBlocksFilled = 0;

        const chunkIterator = processFillChunk(dimension, minSelection, maxSelection, fillType, blocksToUseWeighted, permutationCache);

        const runNextChunk = async () => {
            if (!player.isValid()) {
                // console.log(`Player ${player.name ?? player.id} disconnected during chunked fill.`);
                if (tempTickingAreaName) {
                    system.run(async () => {
                        await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                        // console.log("[FillSelection] Cleaned up ticking area after player disconnect.");
                    });
                }
                return;
            }

            const iteratorResult = await chunkIterator.next();
            if (iteratorResult.value !== undefined) totalBlocksFilled += iteratorResult.value;

            if (!iteratorResult.done) {
                system.run(runNextChunk);
            } else {
                // Operation complete
                if (totalBlocksFilled > 0 || (beforeStateBlocks.length === 0 && volume > 0)) {
                    pushUndoState(player, undoState);
                }
                const undoMessage = (totalBlocksFilled > 0 || (beforeStateBlocks.length === 0 && volume > 0)) ? ` Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.` : '';
                sendSuccessMessage(player, `${fillType} fill complete. ${totalBlocksFilled} blocks changed.${undoMessage}`);

                if (tempTickingAreaName) {
                    await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                }
            }
        };
        system.run(runNextChunk);

    } catch (error) {
        console.error(`[WorldEdit Fill Operation] Error: ${error}\n${error.stack}`);
        sendErrorMessage(player, `An error occurred during the ${fillType} fill. Undo might be needed.`);
        if (tempTickingAreaName) {
            system.run(async () => {
                await removeTemporaryTickingArea(player.dimension, tempTickingAreaName);
                // console.log("[FillSelection] Cleaned up ticking area after setup error.");
            });
        }
    }
}
