// scripts/core/foliage_brush.js
import { Player, system, world, BlockPermutation, Dimension } from "@minecraft/server";
import {
    getPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName
} from "../state/index.js";
import {
    UNDO_TOOL_ID,
    MAX_VIEW_DISTANCE,
    // BLOCKS_PER_CHUNK, // Not directly used for yielding in this file's loop
    FOLIAGE_PRESETS,
    VALID_FOLIAGE_GROUND_TYPES,
    DEFAULT_FOLIAGE_RADIUS,
    DEFAULT_FOLIAGE_DENSITY,
    DEFAULT_FOLIAGE_SOURCE,
    DEFAULT_FOLIAGE_PRESET,
    DEFAULT_FOLIAGE_CUSTOM,
    DEFAULT_BRUSH_MATERIAL // For fallback
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { parseWeightedBlockInput, selectWeightedBlock, validateBlockIds, getSingleBlockPermutation } from "../utils/index.js";
import { calculateVolume } from "../utils/location.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Applies the foliage brush to scatter plants around the target location,
 * respecting placement rules defined in constants.
 * Plant source can be 'preset', 'custom', 'active', or a specific saved palette name.
 * @param {Player} player
 */
export async function applyFoliageBrush(player) {
    const selectionData = getPlayerSelection(player);
    const {
        foliageBrushRadius: radius,
        foliageBrushDensity: densityPercent,
        foliageBrushPlantSource: source,
        foliageBrushPresetType: presetType,
        foliageBrushCustomInput: customInput,
        activePaletteData
    } = selectionData;

    // 1. Determine Center Location 
    const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
    if (!viewBlockHit?.block) {
        sendErrorMessage(player, "No block in view to apply foliage brush.");
        return;
    }
    const center = viewBlockHit.block.location;
    const dimension = player.dimension;

    // 2. Determine Plant List & Resolve Permutations 
    let plantInfoList = [];
    let sourceDesc = "";
    let useDefaultFallback = false;

    const effectiveSource = source ?? DEFAULT_FOLIAGE_SOURCE;

    try {
        if (effectiveSource === 'preset') {
            sourceDesc = `preset '${presetType}'`;
            plantInfoList = FOLIAGE_PRESETS[presetType] ?? [];
            if (plantInfoList.length === 0) {
                sendInfoMessage(player, `Warning: Foliage preset "${presetType}" is empty or not found. Using default stone.`, true);
                useDefaultFallback = true;
            }
        } else if (effectiveSource === 'manual') {
            sourceDesc = `manual input`;
            if (!customInput || customInput.trim() === '') {
                useDefaultFallback = true;
                sendInfoMessage(player, `Manual input is empty. Using default stone.`, true);
            }
            else {
                const parsedResult = parseWeightedBlockInput(customInput);
                if (parsedResult.errors.length > 0) {
                    useDefaultFallback = true;
                    sendErrorMessage(player, `Manual input errors:\n${parsedResult.errors.join('\n')}\nUsing default stone.`);
                }
                else if (parsedResult.blocks.length === 0) {
                    useDefaultFallback = true;
                    sendErrorMessage(player, `No valid blocks in manual input. Using default stone.`);
                }
                else {
                    const { invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id));
                    if (invalidIds.length > 0) {
                        useDefaultFallback = true;
                        sendErrorMessage(player, `Invalid IDs in manual input: ${invalidIds.join(', ')}. Using default stone.`);
                    }
                    else {
                        plantInfoList = parsedResult.blocks.map(b => ({ id: b.id, weight: b.weight, state: b.state }));
                        sourceDesc = `manual input (${plantInfoList.length} types)`;
                    }
                }
            }
        } else if (effectiveSource === 'active') {
            sourceDesc = "Active Palette";
            if (!activePaletteData || activePaletteData.length === 0) {
                sendInfoMessage(player, `Active Palette empty. Using default: ${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}`, true);
                useDefaultFallback = true;
            } else {
                const { validPermutations, invalidIds } = validateBlockIds(activePaletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Active Palette contains invalid IDs: ${invalidIds.join(', ')}. Using default...`);
                    useDefaultFallback = true;
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Active Palette has no valid blocks. Using default...`);
                    useDefaultFallback = true;
                }
                else {
                    plantInfoList = activePaletteData.map(b => ({ id: b.id, weight: b.weight, state: b.state }));
                    sourceDesc = `Active Palette (${plantInfoList.length} types)`;
                }
            }
        } else { // Saved palette name
            const palettes = getPlayerPalettes(player);
            const paletteNameToUse = effectiveSource;
            const paletteData = palettes[paletteNameToUse];
            sourceDesc = `saved palette ('${paletteNameToUse}')`;
            if (!paletteData || paletteData.length === 0) {
                sendErrorMessage(player, `Saved palette '${paletteNameToUse}' is empty or invalid. Using default...`);
                useDefaultFallback = true;
            }
            else {
                const { validPermutations, invalidIds } = validateBlockIds(paletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' contains invalid IDs. Using default...`);
                    useDefaultFallback = true;
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' has no valid blocks. Using default...`);
                    useDefaultFallback = true;
                }
                else {
                    plantInfoList = paletteData.map(b => ({ id: b.id, weight: b.weight, state: b.state }));
                    sourceDesc = `saved palette '${paletteNameToUse}' (${plantInfoList.length} types)`;
                }
            }
        }

        if (useDefaultFallback) {
            plantInfoList = [{ id: DEFAULT_BRUSH_MATERIAL, weight: 1 }];
            sourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
        }

    } catch (error) {
        console.error(`[FoliageBrush] Error determining plant list: ${error}\n${error.stack}`);
        sendErrorMessage(player, "Error processing plant source. Using default stone.");
        plantInfoList = [{ id: DEFAULT_BRUSH_MATERIAL, weight: 1 }];
        sourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
    }

    if (plantInfoList.length === 0) {
        sendErrorMessage(player, `No plants selected to place from source: ${sourceDesc}.`);
        return;
    }

    const weightedChoices = [];
    let totalWeight = 0;
    for (const plantInfo of plantInfoList) {
        try {
            const permutation = BlockPermutation.resolve(plantInfo.id, plantInfo.state);
            const weight = plantInfo.weight > 0 ? plantInfo.weight : 1;
            weightedChoices.push({
                permutation: permutation,
                weight: weight,
                validGround: plantInfo.validGround,
                needsWater: plantInfo.needsWater ?? false,
                id: plantInfo.id
            });
            totalWeight += weight;
        } catch (e) {
            // console.warn(`[Foliage Brush] Failed to resolve permutation for ${plantInfo.id}: ${e}`);
        }
    }

    if (weightedChoices.length === 0 || totalWeight <= 0) {
        sendErrorMessage(player, "Could not resolve any valid plant permutations to place.");
        return;
    }

    // 3. Calculate Area and Save Undo State
    const yRangeForUndo = 8;
    const minUndo = { x: Math.floor(center.x - radius), y: Math.floor(center.y - yRangeForUndo), z: Math.floor(center.z - radius) };
    const maxUndo = { x: Math.ceil(center.x + radius), y: Math.ceil(center.y + yRangeForUndo), z: Math.ceil(center.z + radius) };
    // sendInfoMessage(player, "Saving undo state..."); // REMOVED
    const beforeStateBlocks = getBlocksInArea(dimension, minUndo, maxUndo);
    const undoState = { affectedArea: { min: minUndo, max: maxUndo }, blocks: beforeStateBlocks, timestamp: Date.now() };
    if (beforeStateBlocks.length === 0 && calculateVolume(minUndo, maxUndo) > 0) {
        sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
    }

    // 4. Perform Scattering
    // sendInfoMessage(player, `Applying foliage using ${sourceDesc}...`); // REMOVED
    let blocksPlaced = 0;
    const density = Math.max(0, Math.min(100, densityPercent)) / 100.0;
    const numAttempts = Math.max(1, Math.ceil(Math.PI * radius * radius * density * 1.2));
    const effectiveMaxAttempts = Math.min(5000, numAttempts * 5);
    const sandTypes = new Set(["minecraft:sand", "minecraft:red_sand", "minecraft:suspicious_sand", "minecraft:suspicious_gravel"]);
    const waterType = "minecraft:water";
    const cardinalDirections = ["north", "south", "east", "west"];

    for (let i = 0; i < effectiveMaxAttempts && blocksPlaced < numAttempts * 2; i++) {
        const angle = Math.random() * 2 * Math.PI;
        const dist = Math.sqrt(Math.random()) * radius;
        const targetX = Math.round(center.x + dist * Math.cos(angle));
        const targetZ = Math.round(center.z + dist * Math.sin(angle));

        try {
            const topBlock = dimension.getTopmostBlock({ x: targetX, z: targetZ }, center.y + yRangeForUndo);
            if (!topBlock) continue;

            const groundLoc = topBlock.location;
            let plantLoc = { x: groundLoc.x, y: groundLoc.y + 1, z: groundLoc.z };
            const blockAbove = dimension.getBlock(plantLoc);
            if (!blockAbove?.isAir) continue;

            let randomWeight = Math.random() * totalWeight;
            let selectedChoice = weightedChoices[weightedChoices.length - 1];
            for (const choice of weightedChoices) { if (randomWeight < choice.weight) { selectedChoice = choice; break; } randomWeight -= choice.weight; }

            let canPlace = false;
            const groundBlockId = topBlock.typeId;
            const requiredGround = selectedChoice.validGround ?? VALID_FOLIAGE_GROUND_TYPES;

            if (selectedChoice.needsWater) {
                if (groundBlockId === waterType) {
                    canPlace = true;
                    plantLoc.y = groundLoc.y;
                }
            } else if (requiredGround.has(groundBlockId)) {
                const plantId = selectedChoice.id;
                if ((plantId === "minecraft:cactus" || plantId === "minecraft:dead_bush") && !sandTypes.has(groundBlockId)) {
                    canPlace = false;
                } else {
                    canPlace = true;
                }
            }

            if (canPlace) {
                const blockToSet = dimension.getBlock(plantLoc);
                if (blockToSet && blockToSet.isAir) {
                    let finalPermutation = selectedChoice.permutation;
                    if (selectedChoice.id === "minecraft:pink_petals") {
                        const randomGrowth = Math.floor(Math.random() * 4);
                        const randomDirectionString = cardinalDirections[Math.floor(Math.random() * cardinalDirections.length)];
                        try {
                            finalPermutation = BlockPermutation.resolve(selectedChoice.id, {
                                "amount": randomGrowth,
                                "minecraft:cardinal_direction": randomDirectionString
                            });
                        } catch (petalError) {
                            // console.warn(`[Foliage Brush] Failed pink_petals state resolve: ${petalError}`); 
                        }
                    }

                    blockToSet.setPermutation(finalPermutation);
                    blocksPlaced++;

                    if (finalPermutation.hasTag("upper_block_bit") && !finalPermutation.getState("upper_block_bit")) {
                        try {
                            const upperLoc = { x: plantLoc.x, y: plantLoc.y + 1, z: plantLoc.z };
                            const upperBlock = dimension.getBlock(upperLoc);
                            if (upperBlock?.isAir) {
                                const upperPermutation = BlockPermutation.resolve(selectedChoice.id, { upper_block_bit: true });
                                upperBlock.setPermutation(upperPermutation);
                            }
                        } catch (upperError) { /* ignore error placing upper block */ }
                    }
                }
            }
        } catch (e) {
            if (!(e?.message?.includes("Failed to find a topmost block"))) {
                // console.warn(`[Foliage Brush] Error during scatter attempt at ${targetX},${targetZ}: ${e}`);
            }
        }
    }

    // 5. Push Undo & Report
    if (blocksPlaced > 0) {
        pushUndoState(player, undoState);
        sendSuccessMessage(player, `Foliage applied (${blocksPlaced} blocks placed). Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
    } else {
        sendInfoMessage(player, "Foliage brush applied, but no suitable locations found or no blocks placed.");
    }
}
