// scripts/core/height_brush.js

import {
    Player,
    BlockPermutation,
    system,
    world,
    Dimension
} from "@minecraft/server";
import { getPlayerSelection } from "../state/index.js";
import {
    MAX_VIEW_DISTANCE,
    // MAX_BRUSH_VOLUME, // Not directly used for limiting this brush type
    UNDO_TOOL_ID,
    MAX_HEIGHT_BRUSH_AMOUNT
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { calculateVolume } from "../utils/location.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";


/**
 * Finds the highest non-air block location at a given x, z coordinate within a dimension.
 * Scans downwards from a starting Y level to a minimum Y level.
 * @param {Dimension} dimension The dimension to scan within.
 * @param {number} x The X coordinate of the column.
 * @param {number} z The Z coordinate of the column.
 * @param {number} startY The maximum Y level to start scanning from (inclusive).
 * @param {number} minY The minimum Y level to scan down to (inclusive).
 * @returns { {location: {x:number, y:number, z:number}, permutation: BlockPermutation} | null }
 */
function findSurfaceBlockY(dimension, x, z, startY, minY) {
    const worldMinY = dimension.heightRange.min;
    const worldMaxY = dimension.heightRange.max;
    const effectiveStartY = Math.min(worldMaxY, Math.floor(startY));
    const effectiveMinY = Math.max(worldMinY, Math.floor(minY));

    for (let y = effectiveStartY; y >= effectiveMinY; y--) {
        try {
            const block = dimension.getBlock({ x, y, z });
            if (block && !block.isAir) {
                return { location: { x, y, z }, permutation: block.permutation };
            }
        } catch (e) { /* Ignore errors */ }
    }
    return null;
}


/**
 * Applies the height brush, raising or lowering terrain based on player settings.
 * @param {Player} player The player using the height brush tool.
 */
export async function applyHeightBrush(player) {
    const selectionData = getPlayerSelection(player);
    const {
        brushShape,
        brushSize,
        heightBrushAmount,
        heightBrushMode
    } = selectionData;

    const isLowering = (heightBrushMode === "Lower");
    const amount = Math.max(1, Math.min(heightBrushAmount ?? 1, MAX_HEIGHT_BRUSH_AMOUNT));

    const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
    if (!viewBlockHit?.block) {
        sendErrorMessage(player, "No block in view to apply height brush.");
        return;
    }
    const center = viewBlockHit.block.location;
    const dimension = player.dimension;
    const worldHeightRange = dimension.heightRange;

    const targetLocationsXZ = new Set();
    const radius = Math.max(0, Math.floor(brushSize));
    const radiusSq = radius * radius;
    let minBoundsXZ = { x: center.x, z: center.z };
    let maxBoundsXZ = { x: center.x, z: center.z };

    // sendInfoMessage(player, `Calculating ${brushShape} height brush footprint (Size: ${radius})...`); // REMOVED

    for (let x = center.x - radius; x <= center.x + radius; x++) {
        for (let z = center.z - radius; z <= center.z + radius; z++) {
            let includeXZ = false;
            const dx = x - center.x;
            const dz = z - center.z;
            const distSqXZ = dx * dx + dz * dz;

            switch (brushShape) {
                case 'sphere': case 'cylinder': case 'dome': case 'bowl':
                case 'cone_up': case 'cone_down': case 'disk': case 'torus':
                    if (distSqXZ <= radiusSq) includeXZ = true;
                    break;
                case 'cube': case 'pyramid': case 'plate':
                    if (Math.abs(dx) <= radius && Math.abs(dz) <= radius) includeXZ = true;
                    break;
                case 'octahedron':
                    if (Math.abs(dx) + Math.abs(dz) <= radius) includeXZ = true;
                    break;
                default:
                    if (distSqXZ <= radiusSq) includeXZ = true;
                    break;
            }

            if (includeXZ) {
                const xzKey = `${x},${z}`;
                if (!targetLocationsXZ.has(xzKey)) {
                    targetLocationsXZ.add(xzKey);
                    minBoundsXZ.x = Math.min(minBoundsXZ.x, x);
                    minBoundsXZ.z = Math.min(minBoundsXZ.z, z);
                    maxBoundsXZ.x = Math.max(maxBoundsXZ.x, x);
                    maxBoundsXZ.z = Math.max(maxBoundsXZ.z, z);
                }
            }
        }
    }

    if (targetLocationsXZ.size === 0) {
        sendInfoMessage(player, "Brush area calculation resulted in zero columns.");
        return;
    }

    let currentMinSurfaceY = worldHeightRange.max;
    let currentMaxSurfaceY = worldHeightRange.min;
    const scanStartY = Math.min(worldHeightRange.max, Math.floor(center.y) + radius + (isLowering ? 5 : amount + 5));
    const scanMinY = worldHeightRange.min;

    for (const xzKey of targetLocationsXZ) {
        const [xStr, zStr] = xzKey.split(',');
        const surface = findSurfaceBlockY(dimension, parseInt(xStr), parseInt(zStr), scanStartY, scanMinY);
        if (surface) {
            currentMinSurfaceY = Math.min(currentMinSurfaceY, surface.location.y);
            currentMaxSurfaceY = Math.max(currentMaxSurfaceY, surface.location.y);
        }
    }
    if (currentMinSurfaceY > currentMaxSurfaceY) { // Check if any surface was found
        sendInfoMessage(player, "Could not find terrain surface within brush area to modify.");
        return;
    }
    const undoMinY = Math.max(worldHeightRange.min, currentMinSurfaceY - (isLowering ? amount : 0));
    const undoMaxY = Math.min(worldHeightRange.max, currentMaxSurfaceY + (isLowering ? 0 : amount));
    const undoMin = { x: minBoundsXZ.x, y: undoMinY, z: minBoundsXZ.z };
    const undoMax = { x: maxBoundsXZ.x, y: undoMaxY, z: maxBoundsXZ.z };
    const undoVolume = calculateVolume(undoMin, undoMax);

    // sendInfoMessage(player, "Saving undo state..."); // REMOVED
    const beforeStateBlocks = getBlocksInArea(dimension, undoMin, undoMax);
    if (beforeStateBlocks.length === 0 && undoVolume > 0) {
        sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
    }
    const undoState = { affectedArea: { min: undoMin, max: undoMax }, blocks: beforeStateBlocks, timestamp: Date.now() };

    const modeDisplay = isLowering ? "Lowering" : "Raising";
    // sendInfoMessage(player, `Applying Height Brush (${modeDisplay} by ${amount})...`); // REMOVED
    let blocksChanged = 0;
    let airPermutation = undefined;
    if (isLowering) {
        try { airPermutation = BlockPermutation.resolve("minecraft:air"); }
        catch (e) {
            sendErrorMessage(player, "Error resolving air block. Aborting.");
            return;
        }
    }

    try {
        for (const xzKey of targetLocationsXZ) {
            const [xStr, zStr] = xzKey.split(',');
            const x = parseInt(xStr);
            const z = parseInt(zStr);

            const surfaceInfo = findSurfaceBlockY(dimension, x, z, undoMaxY + 1, undoMinY);
            if (!surfaceInfo) continue;

            const surfaceY = surfaceInfo.location.y;
            const surfacePermutation = surfaceInfo.permutation;

            if (isLowering) {
                const targetBottomY = Math.max(worldHeightRange.min, surfaceY - amount + 1);
                for (let y = surfaceY; y >= targetBottomY; y--) {
                    try {
                        const block = dimension.getBlock({ x, y, z });
                        if (block && !block.isAir) {
                            block.setPermutation(airPermutation);
                            blocksChanged++;
                        }
                    } catch (e) { /* ignore single block errors */ }
                }
            } else {
                if (!surfacePermutation || surfacePermutation.matches("minecraft:air")) {
                    // console.warn(`Height Brush: Skipping raise at ${x},${z} due to air/invalid surface block.`);
                    continue;
                }
                const targetTopY = Math.min(worldHeightRange.max, surfaceY + amount);
                for (let y = surfaceY + 1; y <= targetTopY; y++) {
                    try {
                        const block = dimension.getBlock({ x, y, z });
                        if (block && block.isAir) {
                            block.setPermutation(surfacePermutation);
                            blocksChanged++;
                        } else if (!block) { /* Block outside loaded/sim area */ }
                    } catch (e) { /* ignore single block errors */ }
                }
            }
        }

        if (blocksChanged > 0) {
            pushUndoState(player, undoState);
            sendSuccessMessage(player, `Height Brush applied (${modeDisplay}). Changed ${blocksChanged} blocks. Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
        } else {
            sendInfoMessage(player, "Height Brush applied, but no blocks were changed.");
        }

    } catch (error) {
        console.error(`[Height Brush] Error during operation: ${error}\n${error.stack}`);
        sendErrorMessage(player, "An error occurred during the height brush operation. Undo might be needed.");
        if (beforeStateBlocks.length > 0) {
            pushUndoState(player, undoState);
            sendInfoMessage(player, "Undo state saved despite error.", true);
        }
    }
}
