// scripts/core/history.js
import { Player, system } from "@minecraft/server";
import {
    getPlayerUndoStack,
    setPlayerUndoStack,
    getPlayerRedoStack,
    setPlayerRedoStack,
    getGlobalSettings
} from "../state/index.js";
import {
    getBlocksInArea,
    restoreBlocksFromState,
} from "../utils/history_helpers.js";
import { calculateVolume } from "../utils/location.js";
import {
    addTemporaryTickingArea,
    removeTemporaryTickingArea
} from "../utils/tickingarea_manager.js";
import {
    TICKING_AREA_LOAD_DELAY_TICKS
} from "../constants/index.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";


/** Undoes the last action for the player */
export async function undoLastAction(player) {
    const undoStack = getPlayerUndoStack(player);
    const globalSettings = getGlobalSettings();
    const maxHistory = globalSettings.maxUndoHistory;

    if (maxHistory <= 0 && undoStack.length > 0) {
        sendInfoMessage(player, "Undo history is disabled in global settings.", true);
        return;
    }
    if (maxHistory <= 0 || undoStack.length === 0) {
        sendInfoMessage(player, "Nothing to undo.", true);
        return;
    }

    const stateToRestore = undoStack.pop();
    if (!stateToRestore?.affectedArea?.min || !stateToRestore?.affectedArea?.max || !stateToRestore?.blocks) {
        sendErrorMessage(player, "Invalid undo state found. Cannot undo.");
        setPlayerUndoStack(player, undoStack); // Push it back if it was popped but invalid for some reason
        return;
    }
    setPlayerUndoStack(player, undoStack);

    let tempTickingAreaNameUndo = null;
    const dimension = player.dimension;

    try {
        if (stateToRestore.tickingAreaMin && stateToRestore.tickingAreaMax) {
            tempTickingAreaNameUndo = await addTemporaryTickingArea(
                dimension,
                stateToRestore.tickingAreaMin,
                stateToRestore.tickingAreaMax,
                "undo"
            );
            if (tempTickingAreaNameUndo) {
                // sendInfoMessage(player, "Ensuring undo area is loaded..."); // REMOVED
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area for undo. Operation might be unstable.", true);
            }
        }

        let redoState = undefined;
        try {
            player.dimension.getBlock(stateToRestore.affectedArea.min);
            const currentStateBlocks = getBlocksInArea(player.dimension, stateToRestore.affectedArea.min, stateToRestore.affectedArea.max);
            if (currentStateBlocks.length === 0 && calculateVolume(stateToRestore.affectedArea.min, stateToRestore.affectedArea.max) > 0 && stateToRestore.blocks.length > 0) {
                sendInfoMessage(player, "Warning: Failed to capture current state for redo. Redo might be incomplete.", true);
            }
            if (currentStateBlocks.length > 0 || stateToRestore.blocks.length === 0) { // Only save redo if current state is not empty or original state was empty
                redoState = {
                    affectedArea: { ...stateToRestore.affectedArea },
                    blocks: currentStateBlocks,
                    timestamp: Date.now(),
                    tickingAreaMin: stateToRestore.tickingAreaMin,
                    tickingAreaMax: stateToRestore.tickingAreaMax
                };
            }
        } catch (e) {
            sendErrorMessage(player, "Failed to save redo state (area likely unloaded). Cannot undo safely.");
            undoStack.push(stateToRestore);
            setPlayerUndoStack(player, undoStack);
            throw e;
        }

        // sendInfoMessage(player, "Undoing last action..."); // REMOVED
        restoreBlocksFromState(player, stateToRestore);

        if (redoState) {
            const redoStack = getPlayerRedoStack(player);
            redoStack.push(redoState);
            while (redoStack.length > maxHistory && maxHistory > 0) {
                redoStack.shift();
            }
            setPlayerRedoStack(player, redoStack);
        } else {
            if (calculateVolume(stateToRestore.affectedArea.min, stateToRestore.affectedArea.max) > 0 && stateToRestore.blocks.length > 0) {
                sendInfoMessage(player, "Warning: Redo state not saved (current area appears empty?).", true);
            }
        }
        sendSuccessMessage(player, "Undo successful.");

    } catch (error) {
        // console.error(`[UndoLastAction] Error during undo process: ${error}`); // Log for server console
        // Player already received a more specific error message if it was redo capture failure
    } finally {
        if (tempTickingAreaNameUndo) {
            await removeTemporaryTickingArea(dimension, tempTickingAreaNameUndo);
        }
    }
}

/** Redoes the last undone action for the player */
export async function redoLastAction(player) {
    const redoStack = getPlayerRedoStack(player);
    const globalSettings = getGlobalSettings();
    const maxHistory = globalSettings.maxUndoHistory;

    if (maxHistory <= 0 && redoStack.length > 0) {
        sendInfoMessage(player, "Undo/Redo history is disabled in global settings.", true);
        return;
    }
    if (maxHistory <= 0 || redoStack.length === 0) {
        sendInfoMessage(player, "Nothing to redo.", true);
        return;
    }

    const stateToRestore = redoStack.pop();
    if (!stateToRestore?.affectedArea?.min || !stateToRestore?.affectedArea?.max || !stateToRestore?.blocks) {
        sendErrorMessage(player, "Invalid redo state found. Cannot redo.");
        setPlayerRedoStack(player, redoStack); // Push back if invalid
        return;
    }
    setPlayerRedoStack(player, redoStack);

    let tempTickingAreaNameRedo = null;
    const dimension = player.dimension;

    try {
        if (stateToRestore.tickingAreaMin && stateToRestore.tickingAreaMax) {
            tempTickingAreaNameRedo = await addTemporaryTickingArea(
                dimension,
                stateToRestore.tickingAreaMin,
                stateToRestore.tickingAreaMax,
                "redo"
            );
            if (tempTickingAreaNameRedo) {
                // sendInfoMessage(player, "Ensuring redo area is loaded..."); // REMOVED
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area for redo. Operation might be unstable.", true);
            }
        }

        let undoState = undefined;
        try {
            player.dimension.getBlock(stateToRestore.affectedArea.min);
            const currentStateBlocks = getBlocksInArea(player.dimension, stateToRestore.affectedArea.min, stateToRestore.affectedArea.max);
            if (currentStateBlocks.length === 0 && calculateVolume(stateToRestore.affectedArea.min, stateToRestore.affectedArea.max) > 0 && stateToRestore.blocks.length > 0) {
                sendInfoMessage(player, "Warning: Failed to capture current state for undo. Further undos might be incomplete.", true);
            }
            if (currentStateBlocks.length > 0 || stateToRestore.blocks.length === 0) {
                undoState = {
                    affectedArea: { ...stateToRestore.affectedArea },
                    blocks: currentStateBlocks,
                    timestamp: Date.now(),
                    tickingAreaMin: stateToRestore.tickingAreaMin,
                    tickingAreaMax: stateToRestore.tickingAreaMax
                };
            }
        } catch (e) {
            sendErrorMessage(player, "Failed to save undo state (area likely unloaded). Cannot redo safely.");
            redoStack.push(stateToRestore);
            setPlayerRedoStack(player, redoStack);
            throw e;
        }

        // sendInfoMessage(player, "Redoing last action..."); // REMOVED
        restoreBlocksFromState(player, stateToRestore);

        if (undoState) {
            const undoStack = getPlayerUndoStack(player);
            undoStack.push(undoState);
            while (undoStack.length > maxHistory && maxHistory > 0) {
                undoStack.shift();
            }
            setPlayerUndoStack(player, undoStack);
        } else {
            if (calculateVolume(stateToRestore.affectedArea.min, stateToRestore.affectedArea.max) > 0 && stateToRestore.blocks.length > 0) {
                sendInfoMessage(player, "Warning: Undo state not saved (current area appears empty?).", true);
            }
        }
        sendSuccessMessage(player, "Redo successful.");

    } catch (error) {
        // console.error(`[RedoLastAction] Error during redo process: ${error}`); // Log for server console
    } finally {
        if (tempTickingAreaNameRedo) {
            await removeTemporaryTickingArea(dimension, tempTickingAreaNameRedo);
        }
    }
}
