// scripts/core/line.js

import { Player, BlockPermutation, system, world } from "@minecraft/server";
import {
    getPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName
} from "../state/index.js";
import {
    MAX_LINE_LENGTH,
    UNDO_TOOL_ID,
    DEFAULT_BRUSH_MATERIAL // For fallback
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { getMinCorner, getMaxCorner, calculateVolume } from "../utils/location.js";
import {
    parseWeightedBlockInput,
    selectWeightedBlock,
    validateBlockIds,
    getSingleBlockPermutation
} from "../utils/index.js";
import { getLinePoints3D } from "../utils/algorithms.js";
import {
    sendPlayerMessage, // Though not directly used here now, good to keep for potential direct use
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Draws a 1-block thick line between two points using specified material.
 * Material source can be 'manual', 'active', or a specific saved palette name.
 * @param {Player} player The player initiating the action.
 * @param {{x:number, y:number, z:number}} posA Start point.
 * @param {{x:number, y:number, z:number}} posB End point.
 */
export async function drawLine(player, posA, posB) {
    if (!posA || !posB) {
        sendErrorMessage(player, "Line points not properly defined.");
        return;
    }

    const selectionData = getPlayerSelection(player); // Get player-specific settings
    const showFeedback = selectionData.showActionFeedbackMessages ?? false; // Get the flag for conditional feedback

    const {
        lineMaterialSource,
        lineManualMaterial,
        activePaletteData
    } = selectionData;

    let linePermutation = undefined;
    let useWeightedLine = false;
    let weightedLineBlocks = [];
    const source = lineMaterialSource ?? 'manual';

    try {
        if (source === 'manual') {
            if (!lineManualMaterial || lineManualMaterial.trim() === '') {
                sendInfoMessage(player, `Manual Line Material empty. Using default: ${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}`, true);
                linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
            } else {
                const parsedResult = parseWeightedBlockInput(lineManualMaterial);
                if (parsedResult.errors.length > 0 || parsedResult.blocks.length === 0) {
                    sendErrorMessage(player, `Could not parse Manual Line Material. Using default: ${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}`);
                    if (parsedResult.errors.length > 0) console.warn(`Line material parse errors: ${parsedResult.errors.join('; ')}`);
                    linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                } else {
                    const { validPermutations, invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id));
                    if (invalidIds.length > 0) {
                        sendErrorMessage(player, `Invalid block ID(s) in Manual Line Material: ${invalidIds.join(', ')}. Using default.`);
                        linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    } else {
                        weightedLineBlocks = parsedResult.blocks;
                        useWeightedLine = weightedLineBlocks.length > 1 || weightedLineBlocks[0]?.weight !== 1;
                        if (!useWeightedLine) {
                            linePermutation = validPermutations[0];
                        }
                    }
                }
            }
        } else if (source === 'active') {
            if (!activePaletteData || activePaletteData.length === 0) {
                sendInfoMessage(player, `Active Palette empty for line. Using default: ${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}`, true);
                linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                useWeightedLine = false; weightedLineBlocks = [];
            } else {
                const { validPermutations, invalidIds } = validateBlockIds(activePaletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Active Palette contains invalid IDs. Using default...`);
                    linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    useWeightedLine = false; weightedLineBlocks = [];
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Active Palette has no valid blocks. Using default...`);
                    linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    useWeightedLine = false; weightedLineBlocks = [];
                }
                else {
                    weightedLineBlocks = activePaletteData;
                    useWeightedLine = weightedLineBlocks.length > 1 || weightedLineBlocks[0]?.weight !== 1;
                    if (!useWeightedLine) { linePermutation = validPermutations[0]; }
                }
            }
        } else { // Saved palette
            const palettes = getPlayerPalettes(player);
            const paletteNameToUse = source;
            const paletteData = palettes[paletteNameToUse];
            useWeightedLine = false; weightedLineBlocks = []; linePermutation = undefined;

            if (!paletteData || paletteData.length === 0) {
                sendErrorMessage(player, `Saved line palette '${paletteNameToUse}' is empty/invalid. Using default...`);
                linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
            }
            else {
                const { validPermutations, invalidIds } = validateBlockIds(paletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' contains invalid IDs. Using default...`);
                    linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' has no valid blocks. Using default...`);
                    linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                }
                else {
                    weightedLineBlocks = paletteData;
                    useWeightedLine = weightedLineBlocks.length > 1 || weightedLineBlocks[0]?.weight !== 1;
                    if (!useWeightedLine) { linePermutation = validPermutations[0]; }
                }
            }
            if (!useWeightedLine && !linePermutation) { // Fallback for saved palette
                linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
            }
        }

        if (!useWeightedLine && !linePermutation) { // Ultimate fallback
            if (!DEFAULT_BRUSH_MATERIAL) throw new Error("CRITICAL: Could not resolve any line permutation and default material is not set.");
            linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
            if (!linePermutation) throw new Error(`CRITICAL: Default brush material '${DEFAULT_BRUSH_MATERIAL}' is invalid.`);
            useWeightedLine = false;
        }
    } catch (error) {
        console.error(`[DrawLine] Error determining material: ${error}\n${error.stack}`);
        sendErrorMessage(player, `Error setting line material: ${error.message}. Using default.`);
        linePermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
        useWeightedLine = false;
        weightedLineBlocks = [];
        if (!linePermutation) {
            sendErrorMessage(player, `CRITICAL: Default brush material '${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}' is invalid. Cannot draw line.`);
            return;
        }
    }

    const linePoints = getLinePoints3D(posA, posB);
    const lineLength = linePoints.length;

    if (lineLength === 0) {
        if (showFeedback) { // Explicit check before sending optional feedback
            sendInfoMessage(player, "Line calculation resulted in zero points.");
        }
        return;
    }

    if (lineLength > MAX_LINE_LENGTH) {
        sendErrorMessage(player, `Line too long (${lineLength} > ${MAX_LINE_LENGTH}).`);
        return;
    }

    let min = { ...linePoints[0] }; let max = { ...linePoints[0] };
    for (const point of linePoints) {
        min.x = Math.min(min.x, point.x); max.x = Math.max(max.x, point.x);
        min.y = Math.min(min.y, point.y); max.y = Math.max(max.y, point.y);
        min.z = Math.min(min.z, point.z); max.z = Math.max(max.z, point.z);
    }
    const dimension = player.dimension;
    const beforeStateBlocks = getBlocksInArea(dimension, min, max);
    const undoState = { affectedArea: { min, max }, blocks: beforeStateBlocks, timestamp: Date.now() };
    if (beforeStateBlocks.length === 0 && calculateVolume(min, max) > 0 && linePoints.length > 0) {
        // This is a warning about a potential issue, so it should always be shown.
        sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
    }

    let blocksPlaced = 0;
    const permutationCache = new Map();

    try {
        for (const point of linePoints) {
            try {
                let permToSet = linePermutation;
                if (useWeightedLine) {
                    const selectedId = selectWeightedBlock(weightedLineBlocks);
                    if (!selectedId) continue;

                    permToSet = permutationCache.get(selectedId);
                    if (!permToSet) {
                        permToSet = getSingleBlockPermutation(selectedId);
                        if (permToSet) permutationCache.set(selectedId, permToSet);
                        else {
                            // console.warn(`[LineTool] Failed to resolve block ID: ${selectedId}`); 
                            continue;
                        }
                    }
                }

                if (permToSet) {
                    const block = dimension.getBlock(point);
                    if (block) {
                        block.setPermutation(permToSet);
                        blocksPlaced++;
                    }
                }
            } catch (e) {
                // console.warn(`[LineTool] Error placing single block: ${e?.message}`); 
            }
        }

        if (blocksPlaced > 0) {
            pushUndoState(player, undoState);
            if (showFeedback) { // Explicit check for feedback message
                sendSuccessMessage(player, `Line drawn (${blocksPlaced} blocks). Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
            }
        } else {
            if (showFeedback) { // Explicit check for feedback message
                sendInfoMessage(player, "Line drawn, but no blocks were placed.");
            }
        }

    } catch (error) {
        console.error(`[LineTool] Error during operation: ${error}\n${error.stack}`);
        sendErrorMessage(player, "An error occurred while drawing the line. Undo might be needed.");
        if (beforeStateBlocks.length > 0) { // If an error occurred but we have an undo state
            pushUndoState(player, undoState);
            // This is important information, so force show it.
            sendInfoMessage(player, "Undo state saved despite error during line drawing.", true);
        }
    }
}
