// scripts/core/paint_brush.js

import { Player, BlockPermutation, world } from "@minecraft/server";
import {
    getPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName
} from "../state/index.js";
import {
    DEFAULT_BRUSH_MATERIAL,
    MAX_VIEW_DISTANCE,
    MAX_BRUSH_VOLUME,
    UNDO_TOOL_ID
} from "../constants/index.js";
import {
    getBlocksInArea,
    pushUndoState
} from "../utils/history_helpers.js";
import {
    parseWeightedBlockInput,
    selectWeightedBlock,
    validateBlockIds,
    getSingleBlockPermutation
} from "../utils/index.js";
import {
    calculateVolume
} from "../utils/location.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Applies the paint brush tool to the world, replacing existing non-air blocks within the specified shape.
 * Handles different shapes, sizes, masking, softness, and material sources. Ignores hollow setting.
 * Material source can be 'manual', 'active', or a specific saved palette name.
 * @param {Player} player The player using the paint brush tool.
 */
export async function applyPaintBrush(player) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;

    const {
        brushShape,
        brushSize,
        brushSoftnessAmount,
        brushMaterialSource,
        brushManualMaterial,
        activePaletteData,
        paintBrushMaskEnabled,
        paintBrushMaskInput
    } = selectionData;

    // 1. Determine Center Location & Dimension
    const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
    if (!viewBlockHit?.block) {
        sendErrorMessage(player, "No block in view to apply paint brush.");
        return;
    }
    const center = viewBlockHit.block.location;
    const dimension = player.dimension;

    // 2. Determine Paint Material Permutation(s) and Source Description
    let paintPermutation = undefined;
    let paintSourceDesc = ""; // For user feedback, if needed for success/error messages
    let useWeightedPaint = false;
    let weightedPaintBlocks = [];
    const source = brushMaterialSource ?? 'manual';

    try {
        if (source === 'manual') {
            paintSourceDesc = "manual input";
            if (!brushManualMaterial || brushManualMaterial.trim() === '') {
                sendInfoMessage(player, `Manual Paint Blocks empty. Using default: ${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}`, true);
                paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            } else {
                const parsedResult = parseWeightedBlockInput(brushManualMaterial);
                if (parsedResult.errors.length > 0 || parsedResult.blocks.length === 0) {
                    sendErrorMessage(player, `Could not parse Manual Paint Blocks. Using default: ${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}`);
                    if (parsedResult.errors.length > 0) console.warn(`Paint material parse errors: ${parsedResult.errors.join('; ')}`);
                    paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                } else {
                    const { validPermutations, invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id));
                    if (invalidIds.length > 0) {
                        sendErrorMessage(player, `Invalid block ID(s) in Manual Paint Blocks: ${invalidIds.join(', ')}. Using default.`);
                        paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                        paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                    } else {
                        weightedPaintBlocks = parsedResult.blocks;
                        useWeightedPaint = weightedPaintBlocks.length > 1 || weightedPaintBlocks[0]?.weight !== 1;
                        if (!useWeightedPaint) {
                            paintPermutation = validPermutations[0];
                            paintSourceDesc = `manual (${weightedPaintBlocks[0].id.replace('minecraft:', '')})`;
                        } else {
                            paintSourceDesc = `manual (${weightedPaintBlocks.length} types)`;
                        }
                    }
                }
            }
        } else if (source === 'active') {
            paintSourceDesc = "Active Palette";
            if (!activePaletteData || activePaletteData.length === 0) {
                sendInfoMessage(player, `Active Palette (Paint Material) empty. Using default: ${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}`, true);
                paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                useWeightedPaint = false; weightedPaintBlocks = [];
            } else {
                const { validPermutations, invalidIds } = validateBlockIds(activePaletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Active Palette contains invalid IDs. Using default...`);
                    paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                    useWeightedPaint = false; weightedPaintBlocks = [];
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Active Palette has no valid blocks. Using default...`);
                    paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                    useWeightedPaint = false; weightedPaintBlocks = [];
                }
                else {
                    weightedPaintBlocks = activePaletteData;
                    useWeightedPaint = weightedPaintBlocks.length > 1 || weightedPaintBlocks[0]?.weight !== 1;
                    if (!useWeightedPaint) { paintPermutation = validPermutations[0]; }
                }
            }
        } else { // Saved palette
            const palettes = getPlayerPalettes(player);
            const paletteNameToUse = source;
            const paletteData = palettes[paletteNameToUse];
            paintSourceDesc = `saved palette ('${paletteNameToUse}')`;
            useWeightedPaint = false; weightedPaintBlocks = []; paintPermutation = undefined;

            if (!paletteData || paletteData.length === 0) {
                sendErrorMessage(player, `Saved paint palette '${paletteNameToUse}' is empty/invalid. Using default...`);
                paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            }
            else {
                const { validPermutations, invalidIds } = validateBlockIds(paletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' contains invalid IDs. Using default...`);
                    paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' has no valid blocks. Using default...`);
                    paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                    paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
                }
                else {
                    weightedPaintBlocks = paletteData;
                    useWeightedPaint = weightedPaintBlocks.length > 1 || weightedPaintBlocks[0]?.weight !== 1;
                    if (!useWeightedPaint) { paintPermutation = validPermutations[0]; }
                }
            }
            if (!useWeightedPaint && !paintPermutation) { // Fallback for saved palette
                paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
                paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            }
        }

        if (!useWeightedPaint && !paintPermutation) { // Ultimate fallback
            if (!DEFAULT_BRUSH_MATERIAL) throw new Error("CRITICAL: Could not resolve any paint permutation and default material is not set.");
            paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
            if (!paintPermutation) throw new Error(`CRITICAL: Default brush material '${DEFAULT_BRUSH_MATERIAL}' is invalid.`);
            // console.warn(`[ApplyPaintBrush] Material resolution failed, falling back to default: ${DEFAULT_BRUSH_MATERIAL}`);
            paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
            useWeightedPaint = false;
        }
    } catch (error) {
        console.error(`[ApplyPaintBrush] Error determining material: ${error}\n${error.stack}`);
        sendErrorMessage(player, `Error setting paint material: ${error.message}. Using default.`);
        paintPermutation = getSingleBlockPermutation(DEFAULT_BRUSH_MATERIAL);
        paintSourceDesc = `default (${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')})`;
        useWeightedPaint = false; weightedPaintBlocks = [];
        if (!paintPermutation) {
            sendErrorMessage(player, `CRITICAL: Default brush material '${DEFAULT_BRUSH_MATERIAL.replace('minecraft:', '')}' is invalid. Cannot apply paint brush.`);
            return;
        }
    }

    // 2b. Resolve Mask Block IDs
    let maskTypeIds = new Set(); let useMasking = false; let maskDesc = "";
    if (paintBrushMaskEnabled) {
        if (!paintBrushMaskInput || paintBrushMaskInput.trim() === "") {
            sendInfoMessage(player, "Paint Mask enabled but input is empty. Masking skipped.", true);
        }
        else {
            const parsedMask = parseWeightedBlockInput(paintBrushMaskInput); // Weights are ignored for mask
            if (parsedMask.blocks.length === 0) {
                sendInfoMessage(player, "No valid blocks in Paint Mask input. Masking skipped.", true);
            }
            else {
                const { validPermutations, invalidIds } = validateBlockIds(parsedMask.blocks.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Invalid Mask IDs: ${invalidIds.join(', ')}. Masking skipped.`);
                }
                else {
                    validPermutations.forEach(perm => maskTypeIds.add(perm.type.id));
                    if (maskTypeIds.size > 0) {
                        useMasking = true;
                        maskDesc = ` masking ${maskTypeIds.size} type${maskTypeIds.size === 1 ? '' : 's'}`;
                    }
                }
            }
        }
    }

    // 3. Calculate Bounding Box
    const radius = Math.max(0, Math.floor(brushSize));
    const radiusSq = radius * radius; // Not used for all shapes but common
    const coneHeight = radius;
    const torusMinorRadius = 2;
    const torusMajorRadius = Math.max(0, radius - torusMinorRadius);
    const torusMinorRadiusSq = torusMinorRadius * torusMinorRadius;
    let yMinOffset = -radius; let yMaxOffset = radius;
    switch (brushShape) {
        case 'cone_down': yMinOffset = -coneHeight; yMaxOffset = 0; break;
        case 'pyramid': case 'cone_up': yMinOffset = 0; yMaxOffset = radius; break;
        case 'dome': yMinOffset = 0; yMaxOffset = radius; break;
        case 'bowl': yMinOffset = -radius; yMaxOffset = 0; break;
        case 'cylinder': yMinOffset = -(radius > 0 ? radius - 1 : 0); yMaxOffset = radius; if (radius === 0) { yMinOffset = 0; yMaxOffset = 0; } break;
        case 'disk': case 'plate': yMinOffset = 0; yMaxOffset = 0; break;
        case 'torus': yMinOffset = -torusMinorRadius; yMaxOffset = torusMinorRadius; break;
    }
    const startY = center.y + yMinOffset; const endY = center.y + yMaxOffset;
    let minBound = { x: center.x - radius, y: Math.min(startY, endY), z: center.z - radius };
    let maxBound = { x: center.x + radius, y: Math.max(startY, endY), z: center.z + radius };
    const estimatedBoundVolume = calculateVolume(minBound, maxBound);

    // 4. Check Bounding Box Volume Limit
    if (estimatedBoundVolume > MAX_BRUSH_VOLUME) {
        sendErrorMessage(player, `Brush area too large (${estimatedBoundVolume} > ${MAX_BRUSH_VOLUME}). Reduce size.`);
        return;
    }

    // 5. Save Undo State
    // sendInfoMessage(player, "Saving undo state..."); // REMOVED
    const beforeStateBlocks = getBlocksInArea(dimension, minBound, maxBound);
    const undoState = { affectedArea: { min: minBound, max: maxBound }, blocks: beforeStateBlocks, timestamp: Date.now() };
    if (beforeStateBlocks.length === 0 && estimatedBoundVolume > 0) {
        sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
    }

    // 6. Perform Painting Loop
    let blocksPainted = 0;
    const softnessFactor = (brushSoftnessAmount ?? 0) / 100.0;
    const softDesc = softnessFactor > 0 ? ` (Soft ${brushSoftnessAmount}%)` : '';
    // sendInfoMessage(player, `Painting eligible blocks within ${brushShape}${softDesc} area using ${paintSourceDesc}${maskDesc}...`); // REMOVED
    const permutationCache = new Map();

    try {
        for (let x = minBound.x; x <= maxBound.x; x++) {
            for (let y = minBound.y; y <= maxBound.y; y++) {
                for (let z = minBound.z; z <= maxBound.z; z++) {
                    let include = false;
                    const dx = x - center.x; const dy = y - center.y; const dz = z - center.z;
                    const distSqXZ = dx * dx + dz * dz; const distSqXYZ = distSqXZ + dy * dy; const manhattanDist = Math.abs(dx) + Math.abs(dy) + Math.abs(dz);
                    let isInsideHardSolid = false;
                    // Shape logic (unchanged)
                    switch (brushShape) {
                        case 'sphere': { let minCornerDistSq = Infinity; for (let cX = x; cX <= x + 1; cX++)for (let cY = y; cY <= y + 1; cY++)for (let cZ = z; cZ <= z + 1; cZ++) { const cdx = cX - (center.x + 0.5), cdy = cY - (center.y + 0.5), cdz = cZ - (center.z + 0.5); minCornerDistSq = Math.min(minCornerDistSq, cdx * cdx + cdy * cdy + cdz * cdz); } if (minCornerDistSq <= radiusSq) isInsideHardSolid = true; break; }
                        case 'cube': if (Math.abs(dx) <= radius && Math.abs(dy) <= radius && Math.abs(dz) <= radius) isInsideHardSolid = true; break;
                        case 'cylinder': { if (y >= startY && y <= endY) { let minCornerDistSqXZ = Infinity; for (let cX = x; cX <= x + 1; cX++)for (let cZ = z; cZ <= z + 1; cZ++) { const cdx = cX - (center.x + 0.5), cdz = cZ - (center.z + 0.5); minCornerDistSqXZ = Math.min(minCornerDistSqXZ, cdx * cdx + cdz * cdz); } if (minCornerDistSqXZ <= radiusSq) isInsideHardSolid = true; } break; }
                        case 'pyramid': { const apexY = center.y + coneHeight; const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = radius * (coneHeight - relY) / coneHeight; if (Math.abs(dx) <= maxDist && Math.abs(dz) <= maxDist) isInsideHardSolid = true; } break; }
                        case 'cone_up': { const apexY = center.y + coneHeight; const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = radius * (coneHeight - relY) / coneHeight; const maxDistSq = maxDist * maxDist; if (distSqXZ <= maxDistSq) isInsideHardSolid = true; } break; }
                        case 'cone_down': { const apexY = center.y - coneHeight; const relY = y - center.y; if (relY <= 0 && relY >= -coneHeight) { const maxDist = radius * (coneHeight + relY) / coneHeight; const maxDistSq = maxDist * maxDist; if (distSqXZ <= maxDistSq) isInsideHardSolid = true; } break; }
                        case 'octahedron': if (manhattanDist <= radius) isInsideHardSolid = true; break;
                        case 'dome': if (y >= center.y && distSqXYZ <= radiusSq) isInsideHardSolid = true; break;
                        case 'bowl': if (y <= center.y && distSqXYZ <= radiusSq) isInsideHardSolid = true; break;
                        case 'disk': if (y === center.y && distSqXZ <= radiusSq) isInsideHardSolid = true; break;
                        case 'plate': if (y === center.y && Math.abs(dx) <= radius && Math.abs(dz) <= radius) isInsideHardSolid = true; break;
                        case 'torus': { const hDist = Math.sqrt(distSqXZ); const dSqTube = Math.pow(hDist - torusMajorRadius, 2) + dy * dy; if (dSqTube <= torusMinorRadiusSq) isInsideHardSolid = true; break; }
                    }
                    // Softness logic (unchanged)
                    if (softnessFactor <= 0) { include = isInsideHardSolid; }
                    else {
                        let normalizedDistance = 1.1;
                        switch (brushShape) {
                            case 'sphere': normalizedDistance = (radius > 0) ? Math.sqrt(distSqXYZ) / radius : 1.1; break;
                            case 'cube': normalizedDistance = (radius > 0) ? Math.max(Math.abs(dx), Math.abs(dy), Math.abs(dz)) / radius : 1.1; break;
                            case 'cylinder': normalizedDistance = (radius > 0 && y >= startY && y <= endY) ? Math.sqrt(distSqXZ) / radius : 1.1; break;
                            case 'pyramid': { const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = (coneHeight > 0) ? radius * (coneHeight - relY) / coneHeight : 0; normalizedDistance = (maxDist > 0) ? Math.max(Math.abs(dx), Math.abs(dz)) / maxDist : 1.1; } else { normalizedDistance = 1.1; } break; }
                            case 'cone_up': { const relY = y - center.y; if (relY >= 0 && relY <= coneHeight) { const maxDist = (coneHeight > 0) ? radius * (coneHeight - relY) / coneHeight : 0; normalizedDistance = (maxDist > 0) ? Math.sqrt(distSqXZ) / maxDist : 1.1; } else { normalizedDistance = 1.1; } break; }
                            case 'cone_down': { const relY = y - center.y; if (relY <= 0 && relY >= -coneHeight) { const maxDist = (coneHeight > 0) ? radius * (coneHeight + relY) / coneHeight : 0; normalizedDistance = (maxDist > 0) ? Math.sqrt(distSqXZ) / maxDist : 1.1; } else { normalizedDistance = 1.1; } break; }
                            case 'octahedron': normalizedDistance = (radius > 0) ? manhattanDist / radius : 1.1; break;
                            case 'dome': normalizedDistance = (radius > 0 && y >= center.y) ? Math.sqrt(distSqXYZ) / radius : 1.1; break;
                            case 'bowl': normalizedDistance = (radius > 0 && y <= center.y) ? Math.sqrt(distSqXYZ) / radius : 1.1; break;
                            case 'disk': normalizedDistance = (radius > 0 && y === center.y) ? Math.sqrt(distSqXZ) / radius : 1.1; break;
                            case 'plate': normalizedDistance = (radius > 0 && y === center.y) ? Math.max(Math.abs(dx), Math.abs(dz)) / radius : 1.1; break;
                            case 'torus': { const hDist = Math.sqrt(distSqXZ); normalizedDistance = (torusMinorRadius > 0) ? Math.sqrt(Math.pow(hDist - torusMajorRadius, 2) + dy * dy) / torusMinorRadius : 1.1; break; }
                        }
                        normalizedDistance = Math.max(0.0, Math.min(1.0, normalizedDistance)); const t = 1.0 - normalizedDistance; let p_falloff = 0.0;
                        if (brushShape === 'cone_up' || brushShape === 'cone_down' || brushShape === 'pyramid') { p_falloff = t * t * t; } else { p_falloff = t * t * (3.0 - 2.0 * t); }
                        const hardEdgeProbability = isInsideHardSolid ? 1.0 : 0.0; const finalProbability = hardEdgeProbability * (1.0 - softnessFactor) + p_falloff * softnessFactor;
                        if (finalProbability >= 0.0001 && Math.random() < finalProbability) { include = true; }
                    }

                    if (include) {
                        try {
                            const existingBlock = dimension.getBlock({ x, y, z });
                            if (!existingBlock || existingBlock.isAir) continue; // Paint brush only paints non-air blocks
                            if (useMasking && !maskTypeIds.has(existingBlock.typeId)) continue; // Skip if mask enabled and block not in mask

                            let permToSet = paintPermutation;
                            if (useWeightedPaint) {
                                const selectedId = selectWeightedBlock(weightedPaintBlocks);
                                if (!selectedId) continue;
                                if (permutationCache.has(selectedId)) { permToSet = permutationCache.get(selectedId); }
                                else {
                                    permToSet = getSingleBlockPermutation(selectedId);
                                    if (permToSet) { permutationCache.set(selectedId, permToSet); }
                                    else {
                                        // console.warn(`[PaintBrush] Failed resolve: ${selectedId}`); 
                                        continue;
                                    }
                                }
                            }

                            if (permToSet) {
                                existingBlock.setPermutation(permToSet);
                                blocksPainted++;
                            }
                        } catch (e) { /* Ignore single block paint errors */ }
                    }
                }
            }
        }

        // 7. Push Undo State and Report Result
        if (blocksPainted > 0) {
            pushUndoState(player, undoState);
            // Simplified success message
            sendSuccessMessage(player, `Paint Brush applied. Painted ${blocksPainted} blocks. Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
        } else {
            if (showFeedback) { // Only show this if feedback is enabled
                sendInfoMessage(player, `Paint Brush applied, but no blocks met the criteria.`);
            }
        }

    } catch (error) {
        console.error(`[WorldEdit PaintBrush Apply] Error during operation: ${error}\n${error.stack}`);
        sendErrorMessage(player, `An error occurred during painting. Undo might be needed.`);
        if (beforeStateBlocks.length > 0) {
            pushUndoState(player, undoState);
            sendInfoMessage(player, "Undo state saved despite error.", true);
        }
    }
}
