// scripts/core/path.js

import { Player, BlockPermutation, system, world } from "@minecraft/server";
import {
    getPlayerSelection,
    setPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName
} from "../state/index.js";
import {
    MAX_PATH_LENGTH,
    UNDO_TOOL_ID,
    DEFAULT_PATH_MATERIAL,
    DEFAULT_PATH_CURVE_TYPE,
    DEFAULT_PATH_SMOOTHNESS,
    DEFAULT_PATH_WIDTH
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { getMinCorner, getMaxCorner, calculateVolume } from "../utils/location.js";
import {
    parseWeightedBlockInput,
    selectWeightedBlock,
    validateBlockIds,
    getSingleBlockPermutation
} from "../utils/index.js";
import {
    getLinePoints3D,
    getCatmullRomSplinePoints,
    getBezierCurvePoints,
    getCircularArcPoints
} from "../utils/algorithms.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";


/**
 * Normalizes a 2D vector (XZ plane). Returns {x: 0, z: 0} if magnitude is zero.
 * @param {{x: number, z: number}} vec The vector to normalize.
 * @returns {{x: number, z: number}} The normalized vector.
 */
function normalizeXZ(vec) {
    const mag = Math.sqrt(vec.x * vec.x + vec.z * vec.z);
    if (mag === 0) return { x: 0, z: 0 };
    return { x: vec.x / mag, z: vec.z / mag };
}


/**
 * Draws a path connecting the points set by the player.
 * Supports 'Linear', 'Smooth', 'Bezier', 'Circular Arc' curve types, path width, and closed paths.
 * Material source can be 'manual', 'active', or a specific saved palette name.
 * @param {Player} player The player initiating the action.
 */
export async function drawPath(player) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;

    const {
        pathPoints: originalPathPoints,
        pathMaterialSource,
        pathManualMaterial,
        pathCurveType,
        pathSmoothness,
        pathWidth,
        pathClosed,
        activePaletteData
    } = selectionData;

    // 1. Validate Path Points
    const isClosed = pathClosed ?? false;
    const curve = pathCurveType ?? DEFAULT_PATH_CURVE_TYPE;
    let minPoints = 2;
    if (isClosed) minPoints = 3;
    if (curve === 'Bezier') minPoints = isClosed ? 3 : 3; // Bezier needs at least 3 for open, 3 for closed (becomes a loop of segments)
    if (curve === 'Circular Arc') minPoints = isClosed ? 3 : 3; // 3 points define an arc

    if (!originalPathPoints || originalPathPoints.length < minPoints) {
        sendErrorMessage(player, `Need at least ${minPoints} points to draw a${isClosed ? ' closed' : ''} ${curve} path.`);
        return;
    }

    let pathPoints = [...originalPathPoints];
    const width = pathWidth ?? DEFAULT_PATH_WIDTH;
    const radius = Math.max(0, Math.floor(width / 2));

    // 2. Determine Material
    let pathPermutation = undefined;
    let useWeightedPath = false;
    let weightedPathBlocks = [];
    // let materialDesc = ""; // REMOVED - No longer needed for intermediate messages
    const source = pathMaterialSource ?? 'manual';

    try {
        if (source === 'manual') {
            const materialInput = pathManualMaterial ?? DEFAULT_PATH_MATERIAL;
            if (!materialInput || materialInput.trim() === '') {
                sendInfoMessage(player, `Manual Path Material empty. Using default: ${DEFAULT_PATH_MATERIAL.replace('minecraft:', '')}`, true);
                pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
            } else {
                const parsedResult = parseWeightedBlockInput(materialInput);
                if (parsedResult.errors.length > 0 || parsedResult.blocks.length === 0) {
                    sendErrorMessage(player, `Could not parse Manual Path Material. Using default: ${DEFAULT_PATH_MATERIAL.replace('minecraft:', '')}`);
                    if (parsedResult.errors.length > 0) console.warn(`Path material parse errors: ${parsedResult.errors.join('; ')}`);
                    pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
                } else {
                    const { validPermutations, invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id));
                    if (invalidIds.length > 0) {
                        sendErrorMessage(player, `Invalid block ID(s) in Manual Path Material: ${invalidIds.join(', ')}. Using default.`);
                        pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
                    } else {
                        weightedPathBlocks = parsedResult.blocks;
                        useWeightedPath = weightedPathBlocks.length > 1 || weightedPathBlocks[0]?.weight !== 1;
                        if (!useWeightedPath) {
                            pathPermutation = validPermutations[0];
                        }
                    }
                }
            }
        } else if (source === 'active') {
            if (!activePaletteData || activePaletteData.length === 0) {
                sendInfoMessage(player, `Active Palette empty for path. Using default: ${DEFAULT_PATH_MATERIAL.replace('minecraft:', '')}`, true);
                pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
                useWeightedPath = false; weightedPathBlocks = [];
            } else {
                const { validPermutations, invalidIds } = validateBlockIds(activePaletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Active Palette contains invalid IDs. Using default...`);
                    pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
                    useWeightedPath = false; weightedPathBlocks = [];
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Active Palette has no valid blocks. Using default...`);
                    pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
                    useWeightedPath = false; weightedPathBlocks = [];
                }
                else {
                    weightedPathBlocks = activePaletteData;
                    useWeightedPath = weightedPathBlocks.length > 1 || weightedPathBlocks[0]?.weight !== 1;
                    if (!useWeightedPath) { pathPermutation = validPermutations[0]; }
                }
            }
        } else { // Saved Palette
            const palettes = getPlayerPalettes(player);
            const paletteNameToUse = source;
            const paletteData = palettes[paletteNameToUse];
            useWeightedPath = false; weightedPathBlocks = []; pathPermutation = undefined;
            if (!paletteData || paletteData.length === 0) {
                sendErrorMessage(player, `Saved path palette '${paletteNameToUse}' is empty/invalid. Using default...`);
                pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
            }
            else {
                const { validPermutations, invalidIds } = validateBlockIds(paletteData.map(b => b.id));
                if (invalidIds.length > 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' contains invalid IDs. Using default...`);
                    pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
                }
                else if (validPermutations.length === 0) {
                    sendErrorMessage(player, `Palette '${paletteNameToUse}' has no valid blocks. Using default...`);
                    pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
                }
                else {
                    weightedPathBlocks = paletteData;
                    useWeightedPath = weightedPathBlocks.length > 1 || weightedPathBlocks[0]?.weight !== 1;
                    if (!useWeightedPath) { pathPermutation = validPermutations[0]; }
                }
            }
            if (!useWeightedPath && !pathPermutation) { // Fallback for saved palette
                pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
            }
        }
        if (!useWeightedPath && !pathPermutation) { // Ultimate fallback
            if (!DEFAULT_PATH_MATERIAL) throw new Error("CRITICAL: Could not resolve any path permutation and default material is not set.");
            pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
            if (!pathPermutation) throw new Error(`CRITICAL: Default path material '${DEFAULT_PATH_MATERIAL}' is invalid.`);
            useWeightedPath = false;
        }
    } catch (error) {
        console.error(`[DrawPath] Error determining material: ${error}\n${error.stack}`);
        sendErrorMessage(player, `Error setting path material: ${error.message}. Using default.`);
        pathPermutation = getSingleBlockPermutation(DEFAULT_PATH_MATERIAL);
        useWeightedPath = false; weightedPathBlocks = [];
        if (!pathPermutation) {
            sendErrorMessage(player, `CRITICAL: Default path material '${DEFAULT_PATH_MATERIAL.replace('minecraft:', '')}' is invalid. Cannot draw path.`);
            return;
        }
    }

    // 3. Calculate ALL Centerline Points First
    // sendInfoMessage(player, "Calculating path centerline..."); // REMOVED
    let centerlinePoints = [];
    const smoothness = pathSmoothness ?? DEFAULT_PATH_SMOOTHNESS;
    const pointCount = pathPoints.length;

    const addCurveSegmentPoints = (segmentPoints) => {
        if (segmentPoints.length > 0) {
            if (centerlinePoints.length === 0) {
                centerlinePoints.push({ x: Math.floor(segmentPoints[0].x), y: Math.floor(segmentPoints[0].y), z: Math.floor(segmentPoints[0].z) });
            }
            for (let j = 1; j < segmentPoints.length; j++) {
                const lineSegment = getLinePoints3D(segmentPoints[j - 1], segmentPoints[j]);
                centerlinePoints.push(...lineSegment.slice(1));
            }
        }
    };
    const numSegments = isClosed ? pointCount : pointCount - 1;

    if (numSegments > 0) {
        switch (curve) {
            case 'Linear':
                for (let i = 0; i < numSegments; i++) {
                    const p1 = pathPoints[i]; const p2 = pathPoints[(i + 1) % pointCount];
                    const segment = getLinePoints3D(p1, p2);
                    centerlinePoints.push(...(i === 0 ? segment : segment.slice(1)));
                }
                break;
            case 'Smooth':
                for (let i = 0; i < numSegments; i++) {
                    const p1 = pathPoints[i]; const p2 = pathPoints[(i + 1) % pointCount];
                    const p0 = pathPoints[(i - 1 + pointCount) % pointCount]; const p3 = pathPoints[(i + 2) % pointCount];
                    const splineSegmentPoints = getCatmullRomSplinePoints(p0, p1, p2, p3, smoothness);
                    addCurveSegmentPoints(splineSegmentPoints);
                }
                break;
            case 'Bezier':
                for (let i = 0; i < numSegments; i++) {
                    const p_i = pathPoints[i]; const p_ip1 = pathPoints[(i + 1) % pointCount];
                    const p_im1 = pathPoints[(i - 1 + pointCount) % pointCount]; const p_ip2 = pathPoints[(i + 2) % pointCount];
                    const control1 = { x: p_i.x + (p_ip1.x - p_im1.x) / 6, y: p_i.y + (p_ip1.y - p_im1.y) / 6, z: p_i.z + (p_ip1.z - p_im1.z) / 6 };
                    const control2 = { x: p_ip1.x - (p_ip2.x - p_i.x) / 6, y: p_ip1.y - (p_ip2.y - p_i.y) / 6, z: p_ip1.z - (p_ip2.z - p_i.z) / 6 };
                    const bezierSegmentPoints = getBezierCurvePoints(p_i, control1, control2, p_ip1, smoothness);
                    addCurveSegmentPoints(bezierSegmentPoints);
                }
                break;
            case 'Circular Arc':
                if (pointCount < 3) { // Arc needs at least 3 points for a segment.
                    sendErrorMessage(player, "Circular Arc path needs at least 3 points per segment.");
                    return; // Exit if not enough points for any arc segment.
                }
                // For closed circular arc, ensure last segment connects back
                const arcNumSegments = isClosed ? pointCount : Math.max(0, pointCount - 2);
                for (let i = 0; i < arcNumSegments; i++) {
                    const p1 = pathPoints[i];
                    const p2 = pathPoints[(i + 1) % pointCount];
                    const p3 = pathPoints[(i + 2) % pointCount];
                    const arcSegmentPoints = getCircularArcPoints(p1, p2, p3, smoothness);
                    addCurveSegmentPoints(arcSegmentPoints);
                }
                break;
            default: // Fallback to Linear
                sendInfoMessage(player, `Unknown path curve type: ${curve}. Using Linear.`, true);
                for (let i = 0; i < numSegments; i++) {
                    const p1 = pathPoints[i]; const p2 = pathPoints[(i + 1) % pointCount];
                    const segment = getLinePoints3D(p1, p2);
                    centerlinePoints.push(...(i === 0 ? segment : segment.slice(1)));
                }
                break;
        }
    }

    if (centerlinePoints.length === 0) {
        if (showFeedback) { sendInfoMessage(player, "Path centerline calculation resulted in zero points."); }
        return;
    }
    const uniqueCenterlinePoints = centerlinePoints.filter((p, i, arr) => {
        if (i === 0) return true; const prev = arr[i - 1];
        return p.x !== prev.x || p.y !== prev.y || p.z !== prev.z;
    });
    centerlinePoints = uniqueCenterlinePoints;
    if (centerlinePoints.length === 0) {
        if (showFeedback) { sendInfoMessage(player, "Path centerline calculation resulted in zero unique points."); }
        return;
    }

    // 4. Calculate Final Block Locations (Including Width)
    // sendInfoMessage(player, "Calculating path blocks including width..."); // REMOVED
    const finalBlockLocations = new Set();
    if (width <= 1) {
        centerlinePoints.forEach(p => finalBlockLocations.add(`${p.x},${p.y},${p.z}`));
    } else {
        for (let i = 0; i < centerlinePoints.length; i++) {
            const p = centerlinePoints[i]; let tangent = { x: 0, z: 0 }; const len = centerlinePoints.length;
            if (len === 1) { tangent = { x: 1, z: 0 }; }
            else if (i === 0) {
                const prevIndex = isClosed ? len - 1 : 0; const nextIndex = 1;
                tangent = { x: centerlinePoints[nextIndex].x - centerlinePoints[prevIndex].x, z: centerlinePoints[nextIndex].z - centerlinePoints[prevIndex].z };
                if (!isClosed) tangent = { x: centerlinePoints[1].x - p.x, z: centerlinePoints[1].z - p.z };
            } else if (i === len - 1) {
                const prevIndex = i - 1; const nextIndex = isClosed ? 0 : i;
                tangent = { x: centerlinePoints[nextIndex].x - centerlinePoints[prevIndex].x, z: centerlinePoints[nextIndex].z - centerlinePoints[prevIndex].z };
                if (!isClosed) tangent = { x: p.x - centerlinePoints[i - 1].x, z: p.z - centerlinePoints[i - 1].z };
            } else {
                tangent = { x: centerlinePoints[i + 1].x - centerlinePoints[i - 1].x, z: centerlinePoints[i + 1].z - centerlinePoints[i - 1].z };
            }
            const tangentXZ = normalizeXZ(tangent); const perpXZ = { x: -tangentXZ.z, z: tangentXZ.x };
            finalBlockLocations.add(`${p.x},${p.y},${p.z}`);
            for (let d = 1; d <= radius; d++) {
                const posOffsetX = p.x + perpXZ.x * d; const posOffsetZ = p.z + perpXZ.z * d;
                finalBlockLocations.add(`${Math.round(posOffsetX)},${p.y},${Math.round(posOffsetZ)}`);
                const negOffsetX = p.x - perpXZ.x * d; const negOffsetZ = p.z - perpXZ.z * d;
                finalBlockLocations.add(`${Math.round(negOffsetX)},${p.y},${Math.round(negOffsetZ)}`);
            }
        }
    }
    const pathLength = finalBlockLocations.size;
    if (pathLength === 0) {
        if (showFeedback) { sendInfoMessage(player, "Path calculation resulted in zero blocks after applying width."); }
        return;
    }

    // 5. Check Length Limit
    if (pathLength > MAX_PATH_LENGTH) {
        sendErrorMessage(player, `Path too long (${pathLength} > ${MAX_PATH_LENGTH}). Reduce points, smoothness, or width.`);
        return;
    }

    // 6. Save Undo State
    let min = { x: Infinity, y: Infinity, z: Infinity }; let max = { x: -Infinity, y: -Infinity, z: -Infinity };
    finalBlockLocations.forEach(locString => {
        const [x, y, z] = locString.split(',').map(Number);
        min.x = Math.min(min.x, x); max.x = Math.max(max.x, x);
        min.y = Math.min(min.y, y); max.y = Math.max(max.y, y);
        min.z = Math.min(min.z, z); max.z = Math.max(max.z, z);
    });
    if (min.x === Infinity) { sendErrorMessage(player, "Error calculating path bounds."); return; }

    const dimension = player.dimension;
    // sendInfoMessage(player, "Saving undo state..."); // REMOVED
    const beforeStateBlocks = getBlocksInArea(dimension, min, max);
    const undoState = { affectedArea: { min, max }, blocks: beforeStateBlocks, timestamp: Date.now() };
    if (beforeStateBlocks.length === 0 && calculateVolume(min, max) > 0 && pathLength > 0) {
        sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true); // Force show
    }

    // 7. Place Blocks
    // sendInfoMessage(player, `Drawing path (${pathLength} blocks) using ${materialDesc}...`); // REMOVED
    let blocksPlaced = 0;
    const permutationCache = new Map();

    try {
        for (const locString of finalBlockLocations) {
            const [x, y, z] = locString.split(',').map(Number);
            try {
                let permToSet = pathPermutation;
                if (useWeightedPath) {
                    const selectedId = selectWeightedBlock(weightedPathBlocks);
                    if (!selectedId) continue;
                    permToSet = permutationCache.get(selectedId);
                    if (!permToSet) {
                        permToSet = getSingleBlockPermutation(selectedId);
                        if (permToSet) permutationCache.set(selectedId, permToSet);
                        else {
                            // console.warn(`[PathTool] Failed to resolve block ID: ${selectedId}`); 
                            continue;
                        }
                    }
                }
                if (permToSet) {
                    const block = dimension.getBlock({ x, y, z });
                    if (block) {
                        block.setPermutation(permToSet);
                        blocksPlaced++;
                    }
                }
            } catch (e) {
                // console.warn(`[PathTool] Error placing single block: ${e?.message}`); 
            }
            if (blocksPlaced > 0 && blocksPlaced % 5000 === 0) await system.run(() => { }); // Yield for very long paths
        }

        // 8. Push Undo State & Clear Path Points
        if (blocksPlaced > 0) {
            pushUndoState(player, undoState);
            if (showFeedback) { // Explicit check for feedback
                sendSuccessMessage(player, `Path drawn (${blocksPlaced} blocks). Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
            }
        } else {
            if (showFeedback) { // Explicit check for feedback
                sendInfoMessage(player, "Path drawn, but no blocks were placed.");
            }
        }
        selectionData.pathPoints = []; // Clear points after drawing
        setPlayerSelection(player, selectionData);

    } catch (error) {
        console.error(`[PathTool] Error during operation: ${error}\n${error.stack}`);
        sendErrorMessage(player, "An error occurred while drawing the path. Undo might be needed.");
        if (beforeStateBlocks.length > 0) {
            pushUndoState(player, undoState);
            sendInfoMessage(player, "Undo state saved despite error.", true); // Force show
        }
    }
}
