// scripts/core/presets.js

import { Player, BlockPermutation, system, world } from "@minecraft/server"; // Added world for consistency
import { ModalFormData } from "@minecraft/server-ui";
import { getPlayerSelection } from "../state/index.js";
import {
    MAX_COPY_VOLUME,
    UNDO_TOOL_ID,
    TICKING_AREA_THRESHOLD_VOLUME,
    TICKING_AREA_LOAD_DELAY_TICKS
} from "../constants/index.js";
import { getMinCorner, getMaxCorner, calculateVolume } from "../utils/location.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { getSingleBlockPermutation } from "../utils/validation.js";
import {
    addTemporaryTickingArea,
    removeTemporaryTickingArea
} from "../utils/tickingarea_manager.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/** Generic function to remove blocks matching a set of IDs within the selection */
export async function presetRemoveBlocks(player, blockIdSet, blockTypeName) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;
    const { posA, posB, selectionState } = selectionData;

    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return;
    }
    const selectionMin = getMinCorner(posA, posB);
    const selectionMax = getMaxCorner(posA, posB);
    const volume = calculateVolume(selectionMin, selectionMax);

    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}). Cannot perform preset action.`);
        return;
    }
    if (volume <= 0) {
        sendErrorMessage(player, "Selection volume is zero or invalid.");
        return;
    }

    const dimension = player.dimension;
    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;

    try {
        if (volume > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, selectionMin, selectionMax, `preset_remove_${blockTypeName.toLowerCase()}`);
            if (tempTickingAreaName) {
                // sendInfoMessage(player, `Ensuring area for '${blockTypeName}' removal is loaded...`); // REMOVED
                tickingAreaBoundsForUndo = { min: { ...selectionMin }, max: { ...selectionMax } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, `Warning: Could not create temporary ticking area for '${blockTypeName}' removal. Operation might be slow or unstable.`, true);
            }
        }

        // sendInfoMessage(player, `Preparing to remove ${blockTypeName} blocks... Saving undo state.`); // REMOVED
        const beforeStateBlocks = getBlocksInArea(dimension, selectionMin, selectionMax);
        const undoState = {
            affectedArea: { min: selectionMin, max: selectionMax },
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }
        if (beforeStateBlocks.length === 0 && volume > 0) { // Check if undo state might be incomplete
            sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
        }


        let blocksChanged = 0;
        let airPermutation;
        try {
            airPermutation = BlockPermutation.resolve("minecraft:air");
        } catch (e) {
            console.error(`[WorldEdit Preset Remove ${blockTypeName}] Could not resolve air permutation!`);
            sendErrorMessage(player, "Internal error: Could not find air block.");
            return;
        }

        // sendInfoMessage(player, `Scanning ${volume} blocks to remove ${blockTypeName}...`); // REMOVED
        for (let x = selectionMin.x; x <= selectionMax.x; x++) {
            for (let y = selectionMin.y; y <= selectionMax.y; y++) {
                for (let z = selectionMin.z; z <= selectionMax.z; z++) {
                    try {
                        const block = dimension.getBlock({ x, y, z });
                        if (block && blockIdSet.has(block.typeId)) {
                            block.setPermutation(airPermutation);
                            blocksChanged++;
                        }
                    } catch (e) { /* Ignore single block errors */ }
                }
            }
        }

        if (blocksChanged > 0) {
            pushUndoState(player, undoState);
            if (showFeedback) {
                sendSuccessMessage(player, `Removed ${blocksChanged} ${blockTypeName} blocks. Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
            }
        } else {
            if (showFeedback) {
                sendInfoMessage(player, `No ${blockTypeName} blocks found in the selection.`);
            }
        }
    } catch (error) {
        console.error(`[WorldEdit Preset Remove ${blockTypeName}] Error during operation: ${error}`);
        sendErrorMessage(player, `An error occurred while removing ${blockTypeName} blocks. Undo might be needed if partially completed.`);
    } finally {
        if (tempTickingAreaName) {
            await removeTemporaryTickingArea(dimension, tempTickingAreaName);
        }
    }
}


/** Preset action to flatten terrain to a specified Y-level */
export async function presetFlattenTerrain(player) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;
    const { posA, posB, selectionState } = selectionData;

    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return;
    }
    const selectionMin = getMinCorner(posA, posB);
    const selectionMax = getMaxCorner(posA, posB);
    const volume = calculateVolume(selectionMin, selectionMax);

    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}). Cannot flatten.`);
        return;
    }
    if (volume <= 0) {
        sendErrorMessage(player, "Selection volume is zero or invalid.");
        return;
    }

    // UI for flatten parameters is handled in presets_menu.js
    // This function now expects targetYLevel and fillBlockId to be passed if we were to separate UI fully.
    // For now, keeping the UI form call here as it's specific to this preset.
    const form = new ModalFormData()
        .title("Flatten Terrain")
        .textField("Target Y Level", `Flatten area to this Y level (between ${selectionMin.y} and ${selectionMax.y})`, String(Math.floor((selectionMin.y + selectionMax.y) / 2)))
        .textField("Fill Block ID below Target Y", "Block to use for filling (e.g., dirt, stone)", "minecraft:dirt");

    let targetYLevel, fillBlockId, fillPermutation, airPermutation;
    const dimension = player.dimension;
    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            if (showFeedback) sendInfoMessage(player, "Flatten operation cancelled.");
            return;
        }
        targetYLevel = parseInt(response.formValues[0], 10);
        fillBlockId = response.formValues[1].trim();

        if (isNaN(targetYLevel)) {
            sendErrorMessage(player, "Invalid Target Y Level entered. Must be a number."); return;
        }
        if (!fillBlockId) {
            sendErrorMessage(player, "Fill Block ID cannot be empty."); return;
        }
        if (!fillBlockId.includes(':')) {
            fillBlockId = `minecraft:${fillBlockId}`;
        }
        try {
            fillPermutation = getSingleBlockPermutation(fillBlockId);
            if (!fillPermutation) {
                sendErrorMessage(player, `Invalid Fill Block ID: ${fillBlockId}.`); return;
            }
            airPermutation = BlockPermutation.resolve("minecraft:air");
        } catch (e) {
            sendErrorMessage(player, "Error resolving block types for flatten."); return;
        }


        if (volume > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, selectionMin, selectionMax, "flatten");
            if (tempTickingAreaName) {
                // sendInfoMessage(player, "Ensuring flatten area is loaded..."); // REMOVED
                tickingAreaBoundsForUndo = { min: { ...selectionMin }, max: { ...selectionMax } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area for flatten. Operation might be slow or unstable.", true);
            }
        }

        // sendInfoMessage(player, `Preparing to flatten to Y=${targetYLevel} with ${fillBlockId.replace('minecraft:', '')}... Saving undo state.`); // REMOVED
        const beforeStateBlocks = getBlocksInArea(dimension, selectionMin, selectionMax);
        const undoState = {
            affectedArea: { min: selectionMin, max: selectionMax },
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }
        if (beforeStateBlocks.length === 0 && volume > 0) { // Check if undo state might be incomplete
            sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true);
        }

        let blocksChanged = 0;
        // sendInfoMessage(player, `Scanning ${volume} blocks to flatten...`); // REMOVED
        for (let x = selectionMin.x; x <= selectionMax.x; x++) {
            for (let y = selectionMin.y; y <= selectionMax.y; y++) {
                for (let z = selectionMin.z; z <= selectionMax.z; z++) {
                    try {
                        const block = dimension.getBlock({ x, y, z });
                        if (!block) continue;

                        if (y > targetYLevel) {
                            if (!block.permutation.matches("minecraft:air")) {
                                block.setPermutation(airPermutation);
                                blocksChanged++;
                            }
                        } else { // y <= targetYLevel
                            if (!block.permutation.matches(fillPermutation.type.id, fillPermutation.getAllStates())) {
                                block.setPermutation(fillPermutation);
                                blocksChanged++;
                            }
                        }
                    } catch (e) { /* Ignore single block errors */ }
                }
            }
        }

        if (blocksChanged > 0) {
            pushUndoState(player, undoState);
            if (showFeedback) {
                sendSuccessMessage(player, `Flatten complete to Y=${targetYLevel}. Changed ${blocksChanged} blocks. Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
            }
        } else {
            if (showFeedback) {
                sendInfoMessage(player, `No blocks needed changing for flatten operation.`);
            }
        }
    } catch (error) {
        console.error(`[WorldEdit Flatten] Error during operation: ${error}`);
        sendErrorMessage(player, `An error occurred during flatten. Undo might be needed if partially completed.`);
    } finally {
        if (tempTickingAreaName) {
            await removeTemporaryTickingArea(dimension, tempTickingAreaName);
        }
    }
}
