// scripts/core/replace_blocks.js

import { Player, BlockPermutation, system, world } from "@minecraft/server"; // Added world for consistency
import { getPlayerSelection } from "../state/index.js";
import {
    MAX_COPY_VOLUME,
    UNDO_TOOL_ID,
    BLOCKS_PER_CHUNK,
    TICKING_AREA_THRESHOLD_VOLUME,
    TICKING_AREA_LOAD_DELAY_TICKS // Added for ticking area
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { getMinCorner, getMaxCorner, calculateVolume } from "../utils/location.js";
import { parseWeightedBlockInput, selectWeightedBlock } from "../utils/parsing.js";
import { validateBlockIds, getSingleBlockPermutation } from "../utils/validation.js";
import {
    addTemporaryTickingArea,
    removeTemporaryTickingArea
} from "../utils/tickingarea_manager.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Asynchronous generator to process replace operations in chunks.
 * @yields {number} Number of blocks replaced in the current chunk.
 */
async function* processReplaceChunk(dimension, min, max, targetTypeIdSet, replacementWeightedBlocks, replacementPermutationCache, useWeightedReplace) {
    let blocksReplacedInChunk = 0;
    let currentProcessed = 0;
    const blockPermutationCache = new Map(replacementPermutationCache);

    for (let x = min.x; x <= max.x; x++) {
        for (let y = min.y; y <= max.y; y++) {
            for (let z = min.z; z <= max.z; z++) {
                try {
                    const block = dimension.getBlock({ x, y, z });
                    if (block) {
                        if (targetTypeIdSet.has(block.typeId)) {
                            let permToSet;
                            const selectedId = selectWeightedBlock(replacementWeightedBlocks);

                            if (selectedId) {
                                if (blockPermutationCache.has(selectedId)) {
                                    permToSet = blockPermutationCache.get(selectedId);
                                } else {
                                    const resolvedPerm = getSingleBlockPermutation(selectedId);
                                    if (resolvedPerm) {
                                        blockPermutationCache.set(selectedId, resolvedPerm);
                                        permToSet = resolvedPerm;
                                    } else {
                                        // console.warn(`[Replace Chunk] Could not resolve permutation for ${selectedId}. Skipping.`);
                                        continue;
                                    }
                                }
                            } else {
                                // console.warn(`[Replace Chunk] selectWeightedBlock returned undefined. Skipping.`);
                                continue;
                            }

                            if (permToSet) {
                                if (!block.permutation.matches(permToSet.type.id, permToSet.getAllStates())) {
                                    block.setPermutation(permToSet);
                                    blocksReplacedInChunk++;
                                }
                            }
                        }
                    }
                } catch (e) {
                    // console.warn(`[Replace Chunk] Error processing block at ${x},${y},${z}: ${e}`);
                }

                currentProcessed++;
                if (currentProcessed >= BLOCKS_PER_CHUNK) {
                    await system.run(() => { }); // Yield
                    if (blocksReplacedInChunk > 0) yield blocksReplacedInChunk;
                    blocksReplacedInChunk = 0;
                    currentProcessed = 0;
                }
            }
        }
    }

    if (blocksReplacedInChunk > 0) {
        yield blocksReplacedInChunk;
    }
}

/**
 * Replaces specified block types within the selection with new block(s).
 * @param {Player} player The player initiating the action.
 * @param {string} targetBlocksString Comma-separated string of block IDs to replace (can include "air").
 * @param {string} replacementString Single block ID or weighted palette string to replace with.
 */
export async function replaceBlocks(player, targetBlocksString, replacementString) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;
    const { posA, posB, selectionState } = selectionData;

    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed."); return;
    }
    const selectionMin = getMinCorner(posA, posB);
    const selectionMax = getMaxCorner(posA, posB);
    const volume = calculateVolume(selectionMin, selectionMax);
    if (volume <= 0) { sendErrorMessage(player, "Selection empty or invalid."); return; }
    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}) to perform replace operation.`); return;
    }

    // sendInfoMessage(player, "Parsing block inputs..."); // REMOVED
    const parsedTargets = parseWeightedBlockInput(targetBlocksString);
    if (parsedTargets.blocks.length === 0 && parsedTargets.errors.length > 0) {
        sendErrorMessage(player, `Target block input errors: ${parsedTargets.errors.join(', ')}`); return;
    }
    if (parsedTargets.blocks.length === 0) {
        sendErrorMessage(player, "No valid block IDs found in 'Blocks to Replace' input."); return;
    }

    const targetValidation = validateBlockIds(parsedTargets.blocks.map(b => b.id));
    if (targetValidation.invalidIds.length > 0) {
        sendErrorMessage(player, `Invalid target block IDs: ${targetValidation.invalidIds.join(', ')}`); return;
    }
    const targetTypeIdSet = new Set(targetValidation.validPermutations.map(p => p.type.id));
    if (targetTypeIdSet.size === 0) {
        sendErrorMessage(player, "Failed to resolve any valid target block types."); return;
    }

    const parsedReplacements = parseWeightedBlockInput(replacementString);
    if (parsedReplacements.errors.length > 0) {
        sendErrorMessage(player, `Replacement input errors:\n${parsedReplacements.errors.join('\n')}`);
    }
    if (parsedReplacements.blocks.length === 0) {
        sendErrorMessage(player, "No valid blocks found in 'Replace With' input."); return;
    }

    const replacementValidation = validateBlockIds(parsedReplacements.blocks.map(b => b.id));
    if (replacementValidation.invalidIds.length > 0) {
        sendErrorMessage(player, `Invalid replacement block IDs: ${replacementValidation.invalidIds.join(', ')}`); return;
    }
    if (replacementValidation.validPermutations.length === 0) {
        sendErrorMessage(player, "No valid replacement blocks could be resolved."); return;
    }

    const replacementWeightedBlocks = parsedReplacements.blocks;
    // const useWeightedReplace = replacementWeightedBlocks.length > 1 || (replacementWeightedBlocks[0] && replacementWeightedBlocks[0].weight !== 1); // Not directly used after cache creation

    const replacementPermutationCache = new Map();
    for (const perm of replacementValidation.validPermutations) {
        replacementPermutationCache.set(perm.type.id, perm);
    }
    for (const blockInfo of replacementWeightedBlocks) {
        if (!replacementPermutationCache.has(blockInfo.id)) {
            const perm = getSingleBlockPermutation(blockInfo.id);
            if (perm) {
                replacementPermutationCache.set(blockInfo.id, perm);
            } else {
                sendErrorMessage(player, `Error: Could not resolve replacement block ${blockInfo.id}. Aborting.`);
                return;
            }
        }
    }

    const dimension = player.dimension;
    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;

    try {
        if (volume > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, selectionMin, selectionMax, "replace");
            if (tempTickingAreaName) {
                // sendInfoMessage(player, "Ensuring replace area is loaded..."); // REMOVED
                tickingAreaBoundsForUndo = { min: { ...selectionMin }, max: { ...selectionMax } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area. Replace operation might be slow or unstable.", true); // Force show
            }
        }

        // sendInfoMessage(player, "Saving undo state..."); // REMOVED
        const beforeStateBlocks = getBlocksInArea(dimension, selectionMin, selectionMax);
        if (beforeStateBlocks.length === 0 && volume > 0) {
            sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true); // Force show
        }
        const undoState = {
            affectedArea: { min: selectionMin, max: selectionMax },
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }

        // sendInfoMessage(player, `Replacing blocks within ${volume} block area... (Processing in chunks)`); // REMOVED
        let totalBlocksReplaced = 0;

        const chunkIterator = processReplaceChunk(dimension, selectionMin, selectionMax, targetTypeIdSet, replacementWeightedBlocks, replacementPermutationCache, (replacementWeightedBlocks.length > 1));

        const runNextChunk = async () => {
            if (!player.isValid()) {
                // console.log(`Player ${player.name ?? player.id} disconnected during chunked replace.`);
                if (tempTickingAreaName) {
                    system.run(async () => {
                        await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                        // console.log("[ReplaceBlocks] Cleaned up ticking area after player disconnect.");
                    });
                }
                return;
            }

            const iteratorResult = await chunkIterator.next();
            if (iteratorResult.value !== undefined) {
                totalBlocksReplaced += iteratorResult.value;
            }

            if (!iteratorResult.done) {
                system.run(runNextChunk);
            } else {
                // Operation complete
                if (totalBlocksReplaced > 0 || (beforeStateBlocks.length === 0 && volume > 0)) { // Push undo if blocks changed or if area was empty (to restore emptiness)
                    pushUndoState(player, undoState);
                }

                if (showFeedback) {
                    const undoMessage = (totalBlocksReplaced > 0 || (beforeStateBlocks.length === 0 && volume > 0)) ? ` Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.` : '';
                    sendSuccessMessage(player, `Replace complete. Replaced ${totalBlocksReplaced} blocks.${undoMessage}`);
                }

                if (tempTickingAreaName) {
                    await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                }
            }
        };
        system.run(runNextChunk);

    } catch (error) {
        console.error(`[Replace Blocks Setup] Error: ${error}\n${error.stack}`);
        sendErrorMessage(player, "An error occurred setting up replace operation. Undo might be needed.");
        if (tempTickingAreaName) {
            system.run(async () => {
                await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                // console.log("[ReplaceBlocks] Cleaned up ticking area after setup error.");
            });
        }
    }
}
