// scripts/core/stack_repeat.js

import {
    Player,
    system,
    Dimension,
    world,
    StructureSaveMode,
} from "@minecraft/server";
import { getPlayerSelection } from "../state/index.js";
import {
    MAX_COPY_VOLUME,
    MAX_STACK_AFFECTED_VOLUME,
    UNDO_TOOL_ID,
    STRUCTURE_NAMESPACE,
    TICKING_AREA_THRESHOLD_VOLUME,
    TICKING_AREA_LOAD_DELAY_TICKS
} from "../constants/index.js";
import { getBlocksInArea, pushUndoState } from "../utils/history_helpers.js";
import { getMinCorner, getMaxCorner, calculateVolume } from "../utils/location.js";
import {
    addTemporaryTickingArea,
    removeTemporaryTickingArea
} from "../utils/tickingarea_manager.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";


/**
 * Stacks or repeats the current selection multiple times using the Structure API.
 * Always includes air blocks as per Structure API behavior.
 * @param {Player} player The player performing the action.
 * @param {number} repeatCount The number of *additional* copies to make.
 * @param {{x:number, y:number, z:number}} directionVec Unit vector indicating stack direction.
 * @param {number} offsetX Additional offset along X between copies.
 * @param {number} offsetY Additional offset along Y between copies.
 * @param {number} offsetZ Additional offset along Z between copies.
 */
export async function stackRepeatSelection(player, repeatCount, directionVec, offsetX, offsetY, offsetZ) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;
    const { posA, posB, selectionState } = selectionData;
    const dimension = player.dimension;

    // 1. Initial Validation
    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return;
    }
    if (repeatCount <= 0) {
        sendErrorMessage(player, "Repeat count must be at least 1.");
        return;
    }

    const originalMin = getMinCorner(posA, posB);
    const originalMax = getMaxCorner(posA, posB);
    const sourceSize = { x: originalMax.x - originalMin.x + 1, y: originalMax.y - originalMin.y + 1, z: originalMax.z - originalMin.z + 1 };
    const sourceVolume = sourceSize.x * sourceSize.y * sourceSize.z;

    if (sourceVolume <= 0) {
        sendErrorMessage(player, "Selection empty or invalid.");
        return;
    }
    if (sourceVolume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Original selection too large (${sourceVolume} > ${MAX_COPY_VOLUME}) to stack.`);
        return;
    }

    const totalDisplacement = {
        x: directionVec.x * sourceSize.x + offsetX,
        y: directionVec.y * sourceSize.y + offsetY,
        z: directionVec.z * sourceSize.z + offsetZ
    };

    const lastCopyMin = { x: originalMin.x + repeatCount * totalDisplacement.x, y: originalMin.y + repeatCount * totalDisplacement.y, z: originalMin.z + repeatCount * totalDisplacement.z };
    const lastCopyMax = { x: originalMax.x + repeatCount * totalDisplacement.x, y: originalMax.y + repeatCount * totalDisplacement.y, z: originalMax.z + repeatCount * totalDisplacement.z };
    const finalMinUndo = { x: Math.min(originalMin.x, lastCopyMin.x), y: Math.min(originalMin.y, lastCopyMin.y), z: Math.min(originalMin.z, lastCopyMin.z) };
    const finalMaxUndo = { x: Math.max(originalMax.x, lastCopyMax.x), y: Math.max(originalMax.y, lastCopyMax.y), z: Math.max(originalMax.z, lastCopyMax.z) };
    const affectedVolumeForUndo = calculateVolume(finalMinUndo, finalMaxUndo);

    if (affectedVolumeForUndo > MAX_STACK_AFFECTED_VOLUME) {
        sendErrorMessage(player, `Operation affects too large an area (${affectedVolumeForUndo} > ${MAX_STACK_AFFECTED_VOLUME}). Reduce count or selection size.`);
        return;
    }

    const tempStructureId = `${STRUCTURE_NAMESPACE}:tempstack_${player.id}`;
    let undoState = null;
    let structureSaved = false;
    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;
    // sendInfoMessage(player, "Stacking selection..."); // REMOVED

    try {
        if (affectedVolumeForUndo > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, finalMinUndo, finalMaxUndo, "stack_repeat");
            if (tempTickingAreaName) {
                // sendInfoMessage(player, "Ensuring operation area is loaded..."); // REMOVED
                tickingAreaBoundsForUndo = { min: { ...finalMinUndo }, max: { ...finalMaxUndo } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area. Operation might be slow or unstable.", true); // Force show
            }
        }

        // sendInfoMessage(player, `Saving selection to temporary structure...`); // REMOVED
        try { world.structureManager.delete(tempStructureId); } catch { /* ignore */ }
        const createOptions = { includesEntities: false, includesBlocks: true, saveMode: StructureSaveMode.Memory }; // Stack/Repeat usually doesn't include entities by default
        const tempStructure = world.structureManager.createFromWorld(tempStructureId, dimension, originalMin, originalMax, createOptions);
        if (!tempStructure || !tempStructure.isValid) { throw new Error("Failed to save selection to temporary structure."); }
        structureSaved = true;

        // sendInfoMessage(player, "Saving undo state for affected area..."); // REMOVED
        const beforeStateBlocks = getBlocksInArea(dimension, finalMinUndo, finalMaxUndo);
        if (beforeStateBlocks.length === 0 && affectedVolumeForUndo > 0) {
            sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true); // Force show
        }
        undoState = {
            affectedArea: { min: finalMinUndo, max: finalMaxUndo },
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }

        // sendInfoMessage(player, `Placing ${repeatCount} copies...`); // REMOVED
        let copiesPlaced = 0;
        const placeOptions = { ignoreStructureBlocks: false, includesEntities: false }; // Match createOptions for entities

        for (let i = 1; i <= repeatCount; i++) {
            const copyOrigin = {
                x: originalMin.x + i * totalDisplacement.x,
                y: originalMin.y + i * totalDisplacement.y,
                z: originalMin.z + i * totalDisplacement.z,
            };
            try {
                world.structureManager.place(tempStructureId, dimension, copyOrigin, placeOptions);
                copiesPlaced++;
            } catch (placeError) {
                console.error(`[Stack/Repeat] Error placing copy ${i}: ${placeError}`);
                sendErrorMessage(player, `Error placing copy ${i}. Operation partially completed.`);
                break;
            }
            if (i % 5 === 0 && i < repeatCount) await system.run(() => { }); // Yield for very large counts
        }

        if (copiesPlaced > 0) {
            if (undoState && (undoState.blocks.length > 0 || affectedVolumeForUndo === 0)) {
                pushUndoState(player, undoState);
            }
            if (showFeedback) {
                sendSuccessMessage(player, `Stack/Repeat complete. Placed ${copiesPlaced} copies. Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.`);
            }
        } else {
            if (showFeedback) {
                sendInfoMessage(player, "Stack/Repeat finished, but no copies were placed successfully.");
            }
            // Still push undo if area was empty, to restore emptiness if an error occurred before any placement
            if (undoState && (undoState.blocks.length === 0 && affectedVolumeForUndo > 0) && copiesPlaced === 0) {
                if (undoState.blocks.length > 0 || affectedVolumeForUndo === 0) { // Check if it's a valid undo state to push
                    pushUndoState(player, undoState);
                }
            }
        }

    } catch (error) {
        console.error(`[Stack/Repeat API Method] Error: ${error}\n${error.stack}`);
        sendErrorMessage(player, `An error occurred during stack/repeat: ${error.message}`);
        if (undoState && undoState.blocks && (undoState.blocks.length > 0 || affectedVolumeForUndo === 0)) {
            pushUndoState(player, undoState);
            sendInfoMessage(player, "Undo state saved despite error.", true); // Force show
        }
    } finally {
        if (structureSaved) {
            try { world.structureManager.delete(tempStructureId); } catch { /* ignore cleanup error */ }
        }
        if (tempTickingAreaName) {
            await removeTemporaryTickingArea(dimension, tempTickingAreaName);
        }
    }
}
