// scripts/core/terrain_generator.js

import {
    Player,
    BlockPermutation,
    Dimension,
    system,
    world
} from "@minecraft/server";
import {
    getPlayerSelection,
    // getPlayerPalettes // Not directly used, UI resolves material
} from "../state/index.js";
import {
    MAX_COPY_VOLUME,
    UNDO_TOOL_ID,
    BLOCKS_PER_CHUNK,
    TICKING_AREA_THRESHOLD_VOLUME,
    TICKING_AREA_LOAD_DELAY_TICKS, // Added for ticking area
    // DEFAULT_TERRAIN_BASE_MATERIAL, // Not directly used, UI resolves material
    // DEFAULT_TERRAIN_SURFACE_MATERIAL // Not directly used
} from "../constants/index.js";
import {
    getBlocksInArea,
    pushUndoState
} from "../utils/history_helpers.js";
import {
    getMinCorner,
    getMaxCorner,
    calculateVolume
} from "../utils/location.js";
import {
    getSingleBlockPermutation,
    validateBlockIds
} from "../utils/validation.js";
import {
    SimplexNoise
} from "../libs/index.js";
// import { parseWeightedBlockInput, selectWeightedBlock } from "../utils/parsing.js"; // UI handles parsing
import { selectWeightedBlock } from "../utils/parsing.js"; // Only selectWeightedBlock is needed
import {
    addTemporaryTickingArea,
    removeTemporaryTickingArea
} from "../utils/tickingarea_manager.js";
import { // Import new messaging functions
    sendPlayerMessage,
    sendSuccessMessage,
    sendInfoMessage,
    sendErrorMessage
} from "../utils/messaging.js";

/**
 * Asynchronous generator to process terrain generation in chunks.
 * @yields {number} Number of blocks modified in the current chunk.
 */
async function* processTerrainChunk(
    dimension, min, max, noiseGenerator,
    baseMaterialWeightedBlocks, surfaceMaterialWeightedBlocks, airPermutation,
    surfaceDepth, noiseType, frequency, amplitude,
    octaves, lacunarity, persistence,
    warpEnabled, warpAmplitude, warpFrequency,
    jitter
) {
    let blocksModifiedInChunk = 0;
    let currentProcessed = 0;
    const selectionHeight = max.y - min.y + 1;

    const primaryNoiseFunc = (noiseType === 'FBM')
        ? (nx, ny) => noiseGenerator.fbm2D(nx, ny, octaves, lacunarity, persistence)
        : (nx, ny) => noiseGenerator.noise2D(nx, ny);

    const basePermutationCache = new Map();
    const surfacePermutationCache = new Map();

    for (let x = min.x; x <= max.x; x++) {
        for (let z = min.z; z <= max.z; z++) {
            const noiseInputX = x * frequency;
            const noiseInputZ = z * frequency;
            let noiseValue;

            if (noiseType === 'Worley') {
                const worleyDist = noiseGenerator.worleyF1_2D(noiseInputX, noiseInputZ, jitter);
                const maxExpectedDist = Math.sqrt(0.5 * 0.5 + 0.5 * 0.5) * (1 + jitter); // Approx max distance in a unit cell
                const normalizedDist = Math.min(1.0, worleyDist / maxExpectedDist);
                noiseValue = (1.0 - normalizedDist) * 2.0 - 1.0; // Map to [-1, 1]
            } else {
                if (warpEnabled) {
                    noiseValue = noiseGenerator.domainWarp2D(
                        noiseInputX, noiseInputZ,
                        warpAmplitude, warpFrequency,
                        primaryNoiseFunc
                    );
                } else {
                    noiseValue = primaryNoiseFunc(noiseInputX, noiseInputZ);
                }
            }

            const baseHeight = min.y + selectionHeight / 2; // Center of selection vertically
            const heightOffset = noiseValue * amplitude;
            let targetSurfaceY = Math.round(baseHeight + heightOffset);
            targetSurfaceY = Math.max(min.y, Math.min(max.y, targetSurfaceY)); // Clamp within selection bounds

            for (let y = min.y; y <= max.y; y++) {
                let permutationToSet = airPermutation;
                let selectedMaterialId;

                if (y < targetSurfaceY - surfaceDepth) { // Base material
                    selectedMaterialId = selectWeightedBlock(baseMaterialWeightedBlocks);
                    if (selectedMaterialId) {
                        if (basePermutationCache.has(selectedMaterialId)) {
                            permutationToSet = basePermutationCache.get(selectedMaterialId);
                        } else {
                            const perm = getSingleBlockPermutation(selectedMaterialId);
                            if (perm) {
                                basePermutationCache.set(selectedMaterialId, perm);
                                permutationToSet = perm;
                            } else {
                                // console.warn(`[TerrainGen Chunk] Failed to resolve base material: ${selectedMaterialId}`);
                                permutationToSet = airPermutation;
                            }
                        }
                    } else {
                        permutationToSet = airPermutation;
                    }
                } else if (y <= targetSurfaceY) { // Surface material
                    selectedMaterialId = selectWeightedBlock(surfaceMaterialWeightedBlocks);
                    if (selectedMaterialId) {
                        if (surfacePermutationCache.has(selectedMaterialId)) {
                            permutationToSet = surfacePermutationCache.get(selectedMaterialId);
                        } else {
                            const perm = getSingleBlockPermutation(selectedMaterialId);
                            if (perm) {
                                surfacePermutationCache.set(selectedMaterialId, perm);
                                permutationToSet = perm;
                            } else {
                                // console.warn(`[TerrainGen Chunk] Failed to resolve surface material: ${selectedMaterialId}`);
                                permutationToSet = airPermutation;
                            }
                        }
                    } else {
                        permutationToSet = airPermutation;
                    }
                }
                // Else (y > targetSurfaceY), it remains air (permutationToSet is already airPermutation)

                try {
                    const block = dimension.getBlock({ x, y, z });
                    if (block) {
                        if (!block.permutation.matches(permutationToSet.type.id, permutationToSet.getAllStates())) {
                            block.setPermutation(permutationToSet);
                            blocksModifiedInChunk++;
                        }
                    }
                } catch (e) { /* Ignore single block errors */ }

                currentProcessed++;
                if (currentProcessed >= BLOCKS_PER_CHUNK) {
                    await system.run(() => { }); // Yield
                    if (blocksModifiedInChunk > 0) yield blocksModifiedInChunk;
                    blocksModifiedInChunk = 0;
                    currentProcessed = 0;
                }
            }
        }
    }

    if (blocksModifiedInChunk > 0) {
        yield blocksModifiedInChunk;
    }
}


/**
 * Generates terrain within the player's selection using noise.
 * @param {Player} player The player initiating the action.
 * @param {object} parameters Terrain generation parameters from UI.
 */
export async function generateTerrain(player, parameters) {
    const selectionData = getPlayerSelection(player);
    const showFeedback = selectionData.showActionFeedbackMessages ?? false;
    const { posA, posB, selectionState } = selectionData;

    if (!posA || !posB || selectionState !== 2) {
        sendErrorMessage(player, "Selection incomplete or not confirmed.");
        return;
    }
    const selectionMin = getMinCorner(posA, posB);
    const selectionMax = getMaxCorner(posA, posB);
    const volume = calculateVolume(selectionMin, selectionMax);

    if (volume <= 0) {
        sendErrorMessage(player, "Selection empty or invalid.");
        return;
    }
    if (volume > MAX_COPY_VOLUME) {
        sendErrorMessage(player, `Selection too large (${volume} > ${MAX_COPY_VOLUME}) to generate terrain.`);
        return;
    }

    const baseMaterialWeightedBlocks = parameters.baseMaterialParam;
    const surfaceMaterialWeightedBlocks = parameters.surfaceMaterialParam;

    if (!baseMaterialWeightedBlocks || baseMaterialWeightedBlocks.length === 0) {
        sendErrorMessage(player, "Base material not defined or invalid for terrain generation.");
        return;
    }
    if (!surfaceMaterialWeightedBlocks || surfaceMaterialWeightedBlocks.length === 0) {
        sendErrorMessage(player, "Surface material not defined or invalid for terrain generation.");
        return;
    }
    // Validation of material IDs is assumed to be done by the UI calling this function.

    const surfaceDepth = Math.max(1, Math.floor(parameters.surfaceDepth ?? 3));
    const noiseType = parameters.noiseType || 'Simplex';
    const frequency = Math.max(0.001, parameters.frequency ?? 0.05);
    const amplitude = Math.max(1, parameters.amplitude ?? 10);
    const seed = parameters.seed; // Seed is resolved by UI
    const octaves = Math.max(1, Math.floor(parameters.octaves ?? 4));
    const lacunarity = Math.max(1.1, parameters.lacunarity ?? 2.0);
    const persistence = Math.max(0.1, Math.min(1.0, parameters.persistence ?? 0.5));
    const warpEnabled = parameters.warpEnabled ?? false;
    const warpAmplitude = Math.max(0, parameters.warpAmplitude ?? 5.0);
    const warpFrequency = Math.max(0.001, parameters.warpFrequency ?? 0.1);
    const jitter = Math.max(0, Math.min(1.0, parameters.jitter ?? 0.8));

    // The detailed parameters message is now part of the UI, so we don't need to send it here.
    // if (showFeedback) {
    //     let message = `Generating terrain (${noiseType}${warpEnabled && noiseType !== 'Worley' ? ' + Warp' : ''}): Using selected materials (depth ${surfaceDepth}), Freq ${frequency.toFixed(3)}, Amp ${amplitude}, Seed ${seed}`;
    //     if (noiseType === 'FBM') message += `, Oct ${octaves}, Lac ${lacunarity.toFixed(2)}, Per ${persistence.toFixed(2)}`;
    //     else if (noiseType === 'Worley') message += `, Jitter ${jitter.toFixed(2)}`;
    //     if (warpEnabled && noiseType !== 'Worley') message += `, WarpAmp ${warpAmplitude.toFixed(1)}, WarpFreq ${warpFrequency.toFixed(3)}`;
    //     sendInfoMessage(player, message);
    // }

    const airPermutation = getSingleBlockPermutation("minecraft:air");
    if (!airPermutation) {
        sendErrorMessage(player, "Error: Could not resolve air block type.");
        return;
    }

    const noiseGenerator = new SimplexNoise(seed);
    const dimension = player.dimension;
    let tempTickingAreaName = null;
    let tickingAreaBoundsForUndo = null;

    try {
        if (volume > TICKING_AREA_THRESHOLD_VOLUME) {
            tempTickingAreaName = await addTemporaryTickingArea(dimension, selectionMin, selectionMax, "terrain_gen");
            if (tempTickingAreaName) {
                // sendInfoMessage(player, "Ensuring terrain generation area is loaded..."); // REMOVED
                tickingAreaBoundsForUndo = { min: { ...selectionMin }, max: { ...selectionMax } };
                await system.runTimeout(() => { }, TICKING_AREA_LOAD_DELAY_TICKS);
            } else {
                sendInfoMessage(player, "Warning: Could not create temporary ticking area. Terrain generation might be slow.", true); // Force show
            }
        }

        // sendInfoMessage(player, "Saving undo state..."); // REMOVED
        const beforeStateBlocks = getBlocksInArea(dimension, selectionMin, selectionMax);
        if (beforeStateBlocks.length === 0 && volume > 0) {
            sendInfoMessage(player, "Warning: Failed to capture undo state. Undo might be incomplete.", true); // Force show
        }
        const undoState = {
            affectedArea: { min: selectionMin, max: selectionMax },
            blocks: beforeStateBlocks,
            timestamp: Date.now()
        };
        if (tickingAreaBoundsForUndo) {
            undoState.tickingAreaMin = tickingAreaBoundsForUndo.min;
            undoState.tickingAreaMax = tickingAreaBoundsForUndo.max;
        }

        // sendInfoMessage(player, `Generating terrain within ${volume} block area... (Processing in chunks)`); // REMOVED
        let totalBlocksModified = 0;

        const chunkIterator = processTerrainChunk(
            dimension, selectionMin, selectionMax, noiseGenerator,
            baseMaterialWeightedBlocks, surfaceMaterialWeightedBlocks, airPermutation,
            surfaceDepth, noiseType, frequency, amplitude,
            octaves, lacunarity, persistence,
            warpEnabled, warpAmplitude, warpFrequency,
            jitter
        );

        const runNextChunk = async () => {
            if (!player.isValid()) {
                // console.log(`Player ${player.name ?? player.id} disconnected during terrain generation.`);
                if (tempTickingAreaName) {
                    system.run(async () => {
                        await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                        // console.log("[TerrainGen] Cleaned up ticking area after player disconnect.");
                    });
                }
                return;
            }

            const iteratorResult = await chunkIterator.next();
            if (iteratorResult.value !== undefined) {
                totalBlocksModified += iteratorResult.value;
            }

            if (!iteratorResult.done) {
                system.run(runNextChunk);
            } else {
                // Operation complete
                if (totalBlocksModified > 0 || (beforeStateBlocks.length === 0 && volume > 0)) {
                    pushUndoState(player, undoState);
                }

                if (showFeedback) {
                    const undoMessage = (totalBlocksModified > 0 || (beforeStateBlocks.length === 0 && volume > 0)) ? ` Use '${UNDO_TOOL_ID.replace('bn:', '')}' to undo.` : '';
                    sendSuccessMessage(player, `Terrain generation complete. Modified ${totalBlocksModified} blocks.${undoMessage}`);
                }

                if (tempTickingAreaName) {
                    await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                }
            }
        };
        system.run(runNextChunk);

    } catch (error) {
        console.error(`[Terrain Generation Setup] Error: ${error}\n${error.stack}`);
        sendErrorMessage(player, "An error occurred setting up terrain generation. Undo might be needed if partially complete.");
        if (tempTickingAreaName) {
            system.run(async () => {
                await removeTemporaryTickingArea(dimension, tempTickingAreaName);
                // console.log("[TerrainGen] Cleaned up ticking area after setup error.");
            });
        }
    }
}
