// scripts/events/interval.js

import { world, system, Player, EquipmentSlot, StructureRotation, GameMode } from "@minecraft/server";
import {
    // Constants needed by the interval
    OUTLINE_INTERVAL,
    PASTE_TOOL_ID,
    BRUSH_TOOL_ID,
    PAINT_BRUSH_TOOL_ID,
    HEIGHT_BRUSH_TOOL_ID,
    LINE_TOOL_ID,
    FOLIAGE_BRUSH_TOOL_ID,
    PATH_TOOL_ID,
    SELECTION_TOOL_ID,
    COPY_TOOL_ID,
    CUT_TOOL_ID,
    UNDO_TOOL_ID,
    REDO_TOOL_ID,
    EDIT_TOOL_ID,
    PALETTE_PICKER_TOOL_ID,
    MASK_PICKER_TOOL_ID,
    SETTINGS_TOOL_ID,
    MAX_VIEW_DISTANCE,
    PREVIEW_PARTICLE_DENSITY,
    MIN_PREVIEW_PARTICLES_PER_EDGE,
    DEFAULT_PATH_CURVE_TYPE,
    DEFAULT_PATH_SMOOTHNESS
} from "../constants/index.js";
import {
    // State functions needed
    getPlayerSelection,
    getPlayerClipboard,
    getActiveSavedPaletteName, // Renamed getter
    getGlobalSettings
} from "../state/index.js";
import {
    // Utility functions needed
    getPlayerDirection,
    getMinCorner,
    getMaxCorner
} from "../utils/location.js";
import { determineLookTargetPos } from "../utils/raycasting.js";
import {
    // Particle drawing functions
    drawBoxOutline,
    drawShapeOutline,
    silentSpawnParticle,
    drawParticleLine
} from "../utils/particle.js";
// Import spline algorithm
import { getCatmullRomSplinePoints } from "../utils/algorithms.js";

// Set of all WorldEdit tool IDs for efficient checking
const WORLDEDIT_TOOL_IDS = new Set([
    SELECTION_TOOL_ID,
    BRUSH_TOOL_ID,
    PAINT_BRUSH_TOOL_ID,
    HEIGHT_BRUSH_TOOL_ID,
    LINE_TOOL_ID,
    FOLIAGE_BRUSH_TOOL_ID,
    PATH_TOOL_ID,
    COPY_TOOL_ID,
    CUT_TOOL_ID,
    PASTE_TOOL_ID,
    UNDO_TOOL_ID,
    REDO_TOOL_ID,
    EDIT_TOOL_ID,
    PALETTE_PICKER_TOOL_ID,
    MASK_PICKER_TOOL_ID,
    SETTINGS_TOOL_ID
]);


// --- Helper Function for Circle Outline ---
function drawCircleOutline(player, center, radius) {
    const globalSettings = getGlobalSettings();
    const outlineParticleId = globalSettings.outlineParticleId;
    const dimension = player.dimension;
    radius = Math.max(0, radius); // Ensure non-negative radius
    const density = PREVIEW_PARTICLE_DENSITY;
    const minParticles = MIN_PREVIEW_PARTICLES_PER_EDGE;
    // Calculate number of steps for the circle outline
    const steps = Math.max(minParticles * 2, Math.ceil(2 * Math.PI * radius * density));
    const groundY = Math.floor(center.y); // Particles are drawn at the Y level of the center block's base

    // Handle radius 0 case (single particle at center)
    if (steps <= 0 && radius === 0) {
        const pointBounds = { x: Math.floor(center.x), y: groundY, z: Math.floor(center.z) };
        silentSpawnParticle(dimension, outlineParticleId, pointBounds, pointBounds, pointBounds);
        return;
    }
    if (steps <= 0) return; // Should not happen if radius > 0 due to Math.max with minParticles

    // Define bounding box for particle offsetting (used by silentSpawnParticle)
    const boundsMin = { x: center.x - radius, y: groundY, z: center.z - radius };
    const boundsMax = { x: center.x + radius, y: groundY, z: center.z + radius };

    const particleLocations = new Set(); // Use a Set to store unique particle locations
    // Calculate particle positions around the circle
    for (let i = 0; i < steps; i++) {
        const angle = (i / steps) * 2 * Math.PI;
        const x = center.x + radius * Math.cos(angle);
        const z = center.z + radius * Math.sin(angle);
        // Round to nearest integer block coordinates for particle placement
        particleLocations.add(`${Math.round(x)},${groundY},${Math.round(z)}`);
    }
    // Spawn particles at the calculated locations
    particleLocations.forEach(locString => {
        const [x, y, z] = locString.split(',').map(Number);
        silentSpawnParticle(dimension, outlineParticleId, { x, y, z }, boundsMin, boundsMax);
    });
}


/**
 * Sets up the periodic interval timer for UI updates (previews, action bar).
 */
export function setupIntervalListener() {
    system.runInterval(() => {
        try {
            const globalSettings = getGlobalSettings();
            const showActionBarSetting = globalSettings.showActionBarInfo;
            const outlineParticleId = globalSettings.outlineParticleId; // Get from global settings

            for (const player of world.getAllPlayers()) {
                if (!player || !player.isValid()) continue;

                let actionBarMsg = "";
                let isHoldingTool = false;
                let drewToolPreview = false; // Flag to track if a tool-specific preview was drawn

                const selectionData = getPlayerSelection(player);
                const { posA, posB, selectionState } = selectionData; // Common selection properties
                let heldItemId = undefined;

                // Get the item ID the player is holding in their main hand
                try {
                    const equippable = player.getComponent("minecraft:equippable");
                    if (equippable) {
                        const mainHandItem = equippable.getEquipment(EquipmentSlot.Mainhand);
                        heldItemId = mainHandItem?.typeId;
                        isHoldingTool = WORLDEDIT_TOOL_IDS.has(heldItemId);
                    }
                } catch (e) { /* Ignore errors getting held item */ }

                // If the player is holding a WorldEdit tool
                if (isHoldingTool) {
                    // Add player facing direction to action bar if enabled
                    if (showActionBarSetting) {
                        try {
                            const viewDir = player.getViewDirection();
                            const direction = getPlayerDirection(viewDir);
                            actionBarMsg = `§eF:${direction}`;
                        } catch (e) { /* Ignore errors getting direction */ }
                    }

                    // --- Tool Specific Logic (Previews & Action Bar Messages) ---
                    if (heldItemId === PASTE_TOOL_ID) {
                        const clipboard = getPlayerClipboard(player);
                        if (clipboard?.size && clipboard.size.x > 0 && clipboard.size.y > 0 && clipboard.size.z > 0) {
                            const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
                            if (viewBlockHit?.block) {
                                const targetBlockLocation = viewBlockHit.block.location;
                                const structureSize = clipboard.size;
                                const pasteOptions = selectionData.pasteOptions ?? { rotation: "None (0°)" }; // Default if not set
                                let finalRotation = StructureRotation.None;
                                switch (pasteOptions.rotation) {
                                    case "90°": finalRotation = StructureRotation.Rotate90; break;
                                    case "180°": finalRotation = StructureRotation.Rotate180; break;
                                    case "270°": finalRotation = StructureRotation.Rotate270; break;
                                }
                                let previewSize = { ...structureSize };
                                if (finalRotation === StructureRotation.Rotate90 || finalRotation === StructureRotation.Rotate270) {
                                    previewSize = { x: structureSize.z, y: structureSize.y, z: structureSize.x }; // Swap X and Z for rotation
                                }
                                let previewMin = { x: 0, y: targetBlockLocation.y, z: 0 };
                                const playerDirection = getPlayerDirection(player.getViewDirection());
                                // Calculate preview location based on player facing and structure size
                                switch (playerDirection) {
                                    case "North":
                                        previewMin.x = targetBlockLocation.x - Math.floor(previewSize.x / 2);
                                        previewMin.z = targetBlockLocation.z - previewSize.z + 1;
                                        break;
                                    case "South":
                                        previewMin.x = targetBlockLocation.x - Math.floor(previewSize.x / 2);
                                        previewMin.z = targetBlockLocation.z + 1;
                                        break;
                                    case "West":
                                        previewMin.x = targetBlockLocation.x - previewSize.x + 1;
                                        previewMin.z = targetBlockLocation.z - Math.floor(previewSize.z / 2);
                                        break;
                                    case "East":
                                        previewMin.x = targetBlockLocation.x + 1;
                                        previewMin.z = targetBlockLocation.z - Math.floor(previewSize.z / 2);
                                        break;
                                    default: // Fallback, should ideally not happen
                                        previewMin.x = targetBlockLocation.x - Math.floor(previewSize.x / 2);
                                        previewMin.z = targetBlockLocation.z - previewSize.z + 1;
                                        break;
                                }
                                const previewMax = {
                                    x: previewMin.x + previewSize.x - 1,
                                    y: previewMin.y + previewSize.y - 1,
                                    z: previewMin.z + previewSize.z - 1
                                };
                                if (previewSize.x > 0 && previewSize.y > 0 && previewSize.z > 0) {
                                    drawBoxOutline(player, previewMin, previewMax);
                                    drewToolPreview = true;
                                }
                            }
                        }
                        // Update action bar based on clipboard status
                        if (drewToolPreview) {
                            actionBarMsg += " §d| Paste Preview";
                        } else if (clipboard?.structureId) {
                            actionBarMsg += " §d| Clip Ready";
                        } else {
                            actionBarMsg += " §7| Clip Empty";
                        }
                    } else if (heldItemId === BRUSH_TOOL_ID) {
                        const { brushShape, brushSize, brushHollow, brushMaterialSource } = selectionData;
                        const sourceDesc = brushMaterialSource === 'active' ? 'ActPal' : (brushMaterialSource === 'manual' ? 'Manual' : 'Saved');
                        actionBarMsg += ` §3| Brush: ${brushShape}${brushHollow ? '(H)' : ''} sz${brushSize} ${sourceDesc}`;
                        const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
                        if (viewBlockHit?.block) {
                            drawShapeOutline(player, viewBlockHit.block.location, brushShape, brushSize);
                            drewToolPreview = true;
                        }
                    } else if (heldItemId === PAINT_BRUSH_TOOL_ID) {
                        const { brushShape, brushSize, brushMaterialSource, paintBrushMaskEnabled } = selectionData;
                        const sourceDesc = brushMaterialSource === 'active' ? 'ActPal' : (brushMaterialSource === 'manual' ? 'Manual' : 'Saved');
                        const maskDesc = paintBrushMaskEnabled ? ` | Mask: ON` : ` | Mask: OFF`;
                        actionBarMsg += ` §3| Paint: ${brushShape} sz${brushSize} ${sourceDesc}${maskDesc}`;
                        const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
                        if (viewBlockHit?.block) {
                            drawShapeOutline(player, viewBlockHit.block.location, brushShape, brushSize);
                            drewToolPreview = true;
                        }
                    } else if (heldItemId === HEIGHT_BRUSH_TOOL_ID) {
                        const { brushShape, brushSize, heightBrushAmount, heightBrushMode } = selectionData;
                        const mode = heightBrushMode ?? "Raise"; // Default to Raise if not set
                        actionBarMsg += ` §3| Height: ${brushShape} sz${brushSize} Amt:${heightBrushAmount} Mode:${mode}`;
                        const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
                        if (viewBlockHit?.block) {
                            drawShapeOutline(player, viewBlockHit.block.location, brushShape, brushSize);
                            drewToolPreview = true;
                        }
                    } else if (heldItemId === LINE_TOOL_ID) {
                        const { lineState, linePosA, lineMaterialSource } = selectionData;
                        const sourceDesc = lineMaterialSource === 'active' ? 'ActPal' : (lineMaterialSource === 'manual' ? 'Manual' : 'Saved');
                        const matInfo = ` | Mat: ${sourceDesc}`;
                        if (lineState === 1 && linePosA) { // If PosA is set for the line
                            const currentTarget = determineLookTargetPos(player, 9); // Line tool PosB is within 9 blocks
                            if (currentTarget) {
                                drawParticleLine(player, linePosA, currentTarget); // Draw preview line
                                drewToolPreview = true;
                                actionBarMsg += ` §aLine: PosA->Look (${currentTarget.x},${currentTarget.y},${currentTarget.z})${matInfo}`;
                            } else {
                                actionBarMsg += ` §eLine: PosA Set | Target > 9 blocks${matInfo}`;
                            }
                        } else { // Ready to set PosA
                            actionBarMsg += ` §eLine: Set Pos A${matInfo}`;
                            const potentialTarget = determineLookTargetPos(player, MAX_VIEW_DISTANCE);
                            if (potentialTarget) { // Show particle at potential PosA
                                silentSpawnParticle(player.dimension, outlineParticleId, potentialTarget, potentialTarget, potentialTarget);
                                drewToolPreview = true;
                            }
                        }
                    } else if (heldItemId === FOLIAGE_BRUSH_TOOL_ID) {
                        const { foliageBrushRadius: radius, foliageBrushPlantSource: source, foliageBrushPresetType: preset, foliageBrushDensity: density } = selectionData;
                        const sourceDesc = source === 'preset' ? `Preset(${preset})` : (source === 'active' ? 'ActPal' : (source === 'custom' ? 'Custom' : 'Saved'));
                        actionBarMsg += ` §2| Foliage: R${radius} Dens${density}% ${sourceDesc}`;
                        const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE });
                        if (viewBlockHit?.block) {
                            drawCircleOutline(player, viewBlockHit.block.location, radius); // Draw circular preview
                            drewToolPreview = true;
                        }
                    } else if (heldItemId === PATH_TOOL_ID) {
                        const points = selectionData.pathPoints ?? [];
                        const pointCount = points.length;
                        const curveType = selectionData.pathCurveType ?? DEFAULT_PATH_CURVE_TYPE;
                        const smoothness = selectionData.pathSmoothness ?? DEFAULT_PATH_SMOOTHNESS; // Used for Smooth curve preview
                        actionBarMsg += ` §b| Path (${curveType}): ${pointCount} points.`;

                        // Draw particles at each set path point
                        if (pointCount > 0) {
                            points.forEach(p => {
                                const pBounds = { x: Math.floor(p.x), y: Math.floor(p.y), z: Math.floor(p.z) };
                                silentSpawnParticle(player.dimension, outlineParticleId, p, pBounds, pBounds);
                            });
                            drewToolPreview = true;
                        }
                        // Draw lines between points for preview (linear or smooth)
                        if (pointCount >= 2) {
                            if (curveType === 'Smooth') { // Catmull-Rom preview
                                for (let i = 0; i < pointCount - 1; i++) {
                                    const p1 = points[i];
                                    const p2 = points[i + 1];
                                    // Get control points (p0 and p3) for the segment
                                    const p0 = (i === 0) ? p1 : points[i - 1]; // Use p1 if it's the first point
                                    const p3 = (i === pointCount - 2) ? p2 : points[i + 2]; // Use p2 if it's the last segment
                                    const splineSegmentPoints = getCatmullRomSplinePoints(p0, p1, p2, p3, smoothness);
                                    for (let j = 1; j < splineSegmentPoints.length; j++) {
                                        drawParticleLine(player, splineSegmentPoints[j - 1], splineSegmentPoints[j]);
                                    }
                                }
                            } else { // Linear, Bezier, Circular Arc (show linear preview for simplicity for now)
                                for (let i = 0; i < pointCount - 1; i++) {
                                    drawParticleLine(player, points[i], points[i + 1]);
                                }
                            }
                        }
                    }
                    // Fallback Action Bar Messages for other tools if no specific preview was drawn
                    else if (!drewToolPreview) {
                        if (heldItemId === SELECTION_TOOL_ID) {
                            if (selectionState === 1 && posA) actionBarMsg += " §a| PosA Set";
                            else if (selectionState === 2 && posA && posB) actionBarMsg += " §b| Sel Set";
                            else actionBarMsg += " §7| Sel: Set Pos A";
                        } else if (heldItemId === COPY_TOOL_ID) actionBarMsg += " §6| Copy Tool Ready";
                        else if (heldItemId === CUT_TOOL_ID) actionBarMsg += " §6| Cut Tool Ready";
                        else if (heldItemId === UNDO_TOOL_ID) actionBarMsg += " §6| Undo Tool Ready";
                        else if (heldItemId === REDO_TOOL_ID) actionBarMsg += " §6| Redo Tool Ready";
                        else if (heldItemId === EDIT_TOOL_ID) actionBarMsg += " §6| Edit Tool Ready";
                        else if (heldItemId === PALETTE_PICKER_TOOL_ID) actionBarMsg += " §6| Palette Picker";
                        else if (heldItemId === MASK_PICKER_TOOL_ID) actionBarMsg += " §6| Mask Picker";
                        else if (heldItemId === SETTINGS_TOOL_ID) actionBarMsg += " §6| Addon Settings Tool";
                        else { // Generic message if holding a WE tool without a specific preview
                            const { activePaletteData, activePaletteIsDirty } = selectionData;
                            const activeSavedPalette = getActiveSavedPaletteName(player);
                            if (activePaletteData?.length > 0) {
                                actionBarMsg += ` §6| ActPal (${activePaletteData.length}${activePaletteIsDirty ? '*' : ''})`;
                            } else if (activeSavedPalette) {
                                actionBarMsg += ` §e| LoadSav:${activeSavedPalette}`;
                            } else {
                                actionBarMsg += " §7| WorldEdit Tool";
                            }
                        }
                    }
                } // End of isHoldingTool block

                // Draw Selection Preview if no specific tool preview was drawn
                if (!drewToolPreview) {
                    if (selectionState === 1 && posA) { // PosA set, previewing to current look target
                        const tempPosB = determineLookTargetPos(player);
                        if (tempPosB) {
                            const min = getMinCorner(posA, tempPosB);
                            const max = getMaxCorner(posA, tempPosB);
                            if (min.x !== max.x || min.y !== max.y || min.z !== max.z) { // Check if it's more than a single point
                                drawBoxOutline(player, min, max);
                            } else { // Single point, just show particle at PosA
                                silentSpawnParticle(player.dimension, outlineParticleId, posA, posA, posA);
                            }
                        } else { // No valid target, just show particle at PosA
                            silentSpawnParticle(player.dimension, outlineParticleId, posA, posA, posA);
                        }
                    } else if (selectionState === 2 && posA && posB) { // Selection complete, draw outline
                        const min = getMinCorner(posA, posB);
                        const max = getMaxCorner(posA, posB);
                        if (min.x !== max.x || min.y !== max.y || min.z !== max.z) {
                            drawBoxOutline(player, min, max);
                        } else { // Selection is a single block/point
                            silentSpawnParticle(player.dimension, outlineParticleId, min, min, max);
                        }
                    }
                } // End Selection Preview Logic

                // Set Final Action Bar Message
                if (showActionBarSetting && isHoldingTool && actionBarMsg) {
                    try {
                        player.onScreenDisplay.setActionBar(actionBarMsg);
                    } catch (e) { /* Ignore errors setting action bar */ }
                } else if (showActionBarSetting && !isHoldingTool) {
                    // Optionally clear action bar if not holding a WorldEdit tool
                    // player.onScreenDisplay.setActionBar(" "); 
                }

            } // End player loop
        } catch (intervalError) {
            console.error(`[WorldEdit Interval] Critical Error: ${intervalError}\n${intervalError.stack}`);
        }
    }, OUTLINE_INTERVAL);
}
