// scripts/events/item_use.js

import { world, system, Player, EquipmentSlot } from "@minecraft/server";
import {
    // Tool IDs handled by this listener (normal use)
    SELECTION_TOOL_ID,
    BRUSH_TOOL_ID,
    PAINT_BRUSH_TOOL_ID,
    HEIGHT_BRUSH_TOOL_ID,
    LINE_TOOL_ID,
    FOLIAGE_BRUSH_TOOL_ID,
    PATH_TOOL_ID,
    COPY_TOOL_ID,
    CUT_TOOL_ID,
    PASTE_TOOL_ID,
    UNDO_TOOL_ID,
    REDO_TOOL_ID,
    EDIT_TOOL_ID,
    SETTINGS_TOOL_ID,
    // Settings
    MAX_VIEW_DISTANCE,
    MAX_SELECTION_DISTANCE
} from "../constants/index.js";
import {
    // State functions
    getPlayerSelection,
    setPlayerSelection
} from "../state/index.js";
import {
    // Utility functions
    determineLookTargetPos,
    calculateVolume
} from "../utils/index.js";
import {
    // Core logic functions
    copySelection,
    cutSelection,
    pasteClipboardRotated,
    undoLastAction,
    redoLastAction,
    applyBrush,
    applyPaintBrush,
    applyHeightBrush,
    drawLine,
    applyFoliageBrush,
    // drawPath // drawPath is called from UI, not directly here
} from "../core/index.js";
import {
    showEditMenu,
    showMainSettingsHubForm // <<< UPDATED Import: Use the new main settings hub
} from "../ui/index.js";

/**
 * Sets up the listener for regular (non-sneak) item use actions.
 */
export function setupItemUseListener() {
    world.afterEvents.itemUse.subscribe(eventData => {
        if (!(eventData.source instanceof Player)) return;

        const player = eventData.source;
        const itemStack = eventData.itemStack;
        const itemId = itemStack?.typeId;

        // --- Global Settings Tool ---
        if (itemId === SETTINGS_TOOL_ID) {
            system.run(async () => {
                try {
                    if (player.isValid()) await showMainSettingsHubForm(player); // <<< UPDATED Call
                }
                catch (e) {
                    console.error(`[ItemUse-SettingsMenu] Error: ${e}\n${e.stack}`);
                    player?.sendMessage("§cError opening settings menu.");
                }
            });
            return;
        }

        // --- Selection Tool ---
        if (itemId === SELECTION_TOOL_ID && !player.isSneaking) {
            system.run(() => {
                try {
                    if (!player.isValid()) return;
                    const currentSelection = getPlayerSelection(player);
                    const currentState = currentSelection.selectionState ?? 0;
                    if (currentState === 0 || currentState === 2) {
                        const targetPosA = determineLookTargetPos(player, MAX_SELECTION_DISTANCE);
                        if (targetPosA) {
                            try {
                                const targetBlock = player.dimension.getBlock(targetPosA);
                                if (targetBlock && !targetBlock.isAir) {
                                    currentSelection.posA = targetPosA;
                                    currentSelection.posB = undefined;
                                    currentSelection.selectionState = 1;
                                    setPlayerSelection(player, currentSelection);
                                    player.sendMessage(`§aPos A Set: §f${targetPosA.x}, ${targetPosA.y}, ${targetPosA.z}`);
                                    player.sendMessage(`§eLook where you want Pos B and use the wand again.`);
                                } else { player.sendMessage("§cPlease look at a solid block to set position A."); }
                            } catch (e) { console.warn(`Error getting target block for Pos A: ${e}`); player.sendMessage("§cError getting target block. Is the area loaded?"); }
                        } else { player.sendMessage("§cCould not determine target block within range for Pos A."); }
                    } else if (currentState === 1) {
                        const targetPosB = determineLookTargetPos(player, MAX_SELECTION_DISTANCE);
                        if (targetPosB) {
                            currentSelection.posB = targetPosB;
                            currentSelection.selectionState = 2;
                            setPlayerSelection(player, currentSelection);
                            player.sendMessage(`§bPos B Set: §f${targetPosB.x}, ${targetPosB.y}, ${targetPosB.z}`);
                            if (currentSelection.posA) { const volume = calculateVolume(currentSelection.posA, currentSelection.posB); player.sendMessage(`§eVolume: §f${volume}`); }
                            player.sendMessage(`§eSelection complete. Use wand again to start new selection, or sneak-use for options.`);
                        } else { player.sendMessage("§cCould not determine target block for Pos B within range."); }
                    }
                } catch (e) { console.error(`[ItemUse-Selection] Error: ${e}\n${e.stack}`); player?.sendMessage("§cAn error occurred while setting selection position."); }
            });
            return;
        }

        // --- Regular Brush Tool ---
        if (itemId === BRUSH_TOOL_ID && !player.isSneaking) {
            system.run(async () => { try { if (player.isValid()) await applyBrush(player); } catch (e) { console.error(`[ItemUse-ApplyBrush] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError applying brush."); } });
            return;
        }
        // --- Paint Brush Tool ---
        if (itemId === PAINT_BRUSH_TOOL_ID && !player.isSneaking) {
            system.run(async () => { try { if (player.isValid()) await applyPaintBrush(player); } catch (e) { console.error(`[ItemUse-ApplyPaintBrush] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError applying paint brush."); } });
            return;
        }
        // --- Height Brush Tool ---
        // Note: Height brush might handle sneak internally for mode switching,
        // but its primary action (apply) is triggered on normal use.
        if (itemId === HEIGHT_BRUSH_TOOL_ID) {
            system.run(async () => { try { if (player.isValid()) await applyHeightBrush(player); } catch (e) { console.error(`[ItemUse-ApplyHeightBrush] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError applying height brush."); } });
            return;
        }
        // --- Line Tool ---
        if (itemId === LINE_TOOL_ID && !player.isSneaking) {
            system.run(async () => {
                try {
                    if (!player.isValid()) return;
                    const selectionData = getPlayerSelection(player);
                    const currentState = selectionData.lineState ?? 0;
                    if (currentState === 0) { // Set Pos A
                        const targetPos = determineLookTargetPos(player, MAX_VIEW_DISTANCE);
                        if (!targetPos) { player.sendMessage("§cCould not determine target point for Pos A."); return; }
                        selectionData.linePosA = { ...targetPos };
                        selectionData.lineState = 1;
                        setPlayerSelection(player, selectionData);
                        player.sendMessage(`§aLine Pos A Set: §f${targetPos.x}, ${targetPos.y}, ${targetPos.z}`);
                        player.sendMessage(`§eUse tool again at target point (within 9 blocks) for Pos B to draw line.`);
                    } else if (currentState === 1) { // Set Pos B and Draw
                        const posA = selectionData.linePosA;
                        const targetPosB = determineLookTargetPos(player, 9); // Line tool places PosB within 9 blocks
                        if (!targetPosB) { player.sendMessage("§cCould not determine target point within 9 blocks for Pos B."); return; }
                        const posB = { ...targetPosB };
                        player.sendMessage(`§bLine Pos B Set: §f${posB.x}, ${posB.y}, ${posB.z}`);
                        // Reset line state before drawing
                        selectionData.linePosA = undefined;
                        selectionData.lineState = 0;
                        setPlayerSelection(player, selectionData);
                        if (posA) { await drawLine(player, posA, posB); }
                        else { player.sendMessage("§cError: Line Pos A was lost before drawing."); }
                    }
                } catch (e) { console.error(`[ItemUse-LineTool] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError processing line tool action."); }
            });
            return;
        }
        // --- Foliage Brush Tool ---
        if (itemId === FOLIAGE_BRUSH_TOOL_ID && !player.isSneaking) {
            system.run(async () => { try { if (player.isValid()) await applyFoliageBrush(player); } catch (e) { console.error(`[ItemUse-ApplyFoliageBrush] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError applying foliage brush."); } });
            return;
        }

        // --- Path Tool (Add Point) ---
        if (itemId === PATH_TOOL_ID && !player.isSneaking) {
            system.run(() => {
                try {
                    if (!player.isValid()) return;
                    const targetPos = determineLookTargetPos(player, MAX_VIEW_DISTANCE);
                    if (!targetPos) {
                        player.sendMessage("§cCould not determine target point.");
                        return;
                    }
                    const selectionData = getPlayerSelection(player);
                    selectionData.pathPoints = selectionData.pathPoints ?? [];
                    selectionData.pathPoints.push({ ...targetPos });
                    setPlayerSelection(player, selectionData);
                    const pointNum = selectionData.pathPoints.length;
                    player.sendMessage(`§aPath Point ${pointNum} added at §f${targetPos.x}, ${targetPos.y}, ${targetPos.z}§a.`);
                    player.sendMessage(`§eClick again for next point, or Sneak-Use for options.`);
                } catch (e) {
                    console.error(`[ItemUse-PathTool] Error: ${e}\n${e.stack}`);
                    player?.sendMessage("§cError adding path point.");
                }
            });
            return;
        }


        // --- Other Tools (Normal Use) ---
        if (itemId === COPY_TOOL_ID) { system.run(() => { try { if (player.isValid()) copySelection(player); } catch (e) { console.error(`[ItemUse-Copy] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError during copy."); } }); }
        else if (itemId === CUT_TOOL_ID) { system.run(async () => { try { if (player.isValid()) await cutSelection(player); } catch (e) { console.error(`[ItemUse-Cut] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError during cut."); } }); }
        else if (itemId === PASTE_TOOL_ID && !player.isSneaking) { const viewBlockHit = player.getBlockFromViewDirection({ maxDistance: MAX_VIEW_DISTANCE }); if (viewBlockHit?.block) { const targetLoc = viewBlockHit.block.location; system.run(async () => { try { if (player.isValid()) await pasteClipboardRotated(player, targetLoc); } catch (e) { console.error(`[ItemUse-Paste] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError during paste."); } }); } else { player.sendMessage("§cNo block in view to paste relative to."); } }
        else if (itemId === UNDO_TOOL_ID) { system.run(async () => { try { if (player.isValid()) await undoLastAction(player); } catch (e) { console.error(`[ItemUse-Undo] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError during undo."); } }); } // Made async
        else if (itemId === REDO_TOOL_ID) { system.run(async () => { try { if (player.isValid()) await redoLastAction(player); } catch (e) { console.error(`[ItemUse-Redo] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError during redo."); } }); } // Made async
        else if (itemId === EDIT_TOOL_ID) { system.run(async () => { try { if (player.isValid()) await showEditMenu(player); } catch (e) { console.error(`[ItemUse-EditMenu] Error: ${e}\n${e.stack}`); player?.sendMessage("§cError opening edit menu."); } }); }

    }); // End of subscribe callback
} // End of setup function
