// scripts/events/item_use_before.js

// Ensure your manifest.json targets a @minecraft/server API version
// where BlockFace is available (e.g., "1.7.0" or higher).
import { world, system, Player, BlockPermutation, GameMode, EquipmentSlot } from "@minecraft/server";
import {
    // Tool IDs that have sneak + use actions to open menus
    SELECTION_TOOL_ID,
    PASTE_TOOL_ID,
    BRUSH_TOOL_ID,       // Regular Brush
    PAINT_BRUSH_TOOL_ID,   // Paint Brush
    COPY_TOOL_ID,          // Copy Tool
    HEIGHT_BRUSH_TOOL_ID,  // Height Brush
    LINE_TOOL_ID,          // Line Tool
    FOLIAGE_BRUSH_TOOL_ID,
    PATH_TOOL_ID,
    // Extended Reach constant
    EXTENDED_REACH_DISTANCE,
    // Set of all WorldEdit tool IDs
    WORLDEDIT_TOOL_IDS
} from "../constants/index.js";
import {
    // UI forms triggered by sneak + use
    showSelectionOptionsMenu,
    showPasteOptionsForm,
    showRegularBrushSettingsForm, // For regular brush
    showPaintBrushSettingsForm,   // For paint brush
    showCopyOptionsForm,          // For copy tool
    showHeightBrushSettingsForm,  // For height brush
    showLineToolSettingsForm,     // For line tool
    showFoliageBrushSettingsForm,
    showPathToolOptionsMenu
} from "../ui/index.js";
import { getPlayerSelection, setPlayerSelection } from "../state/index.js";

/**
 * Sets up the listener for the 'before item use' event.
 * This is used to intercept sneak + use actions for specific tools,
 * cancel the default behavior, and open corresponding settings menus.
 * Also handles the "Extended Reach" feature.
 */
export function setupItemUseBeforeListener() {
    // Subscribe to the event that fires just BEFORE an item is used
    world.beforeEvents.itemUse.subscribe(eventData => {
        const { source, itemStack } = eventData;

        // --- Initial Checks ---
        // 1. Only handle actions performed by players.
        if (!(source instanceof Player)) {
            return;
        }
        const player = source; // Cast source to Player type

        // --- Extended Reach Feature Logic ---
        const selectionData = getPlayerSelection(player);
        if (selectionData.extendedReachEnabled && !player.isSneaking && itemStack) {
            const heldItemId = itemStack.typeId;
            system.run(() => {
                // Check if it's a WorldEdit tool; if so, let normal WE logic handle it.
                // This prevents Extended Reach from interfering with WorldEdit tool actions.
                if (WORLDEDIT_TOOL_IDS.has(heldItemId)) {
                    // Do nothing here, let other event handlers or default behavior for WE tools proceed.
                } else {
                    // Attempt to resolve as a block. If it fails, it's not a placeable block.
                    let blockPermutationToPlace;
                    try {
                        blockPermutationToPlace = BlockPermutation.resolve(heldItemId);
                    } catch (e) {
                        // Not a placeable block item, so Extended Reach doesn't apply.
                        // Let the event proceed normally for other item uses (e.g., food, weapons).
                        return;
                    }

                    // If we reached here, it's a placeable block and Extended Reach is active.
                    // Perform raycasting and placement.
                    const blockHit = player.getBlockFromViewDirection({ maxDistance: EXTENDED_REACH_DISTANCE });

                    if (blockHit?.block) {
                        const targetBlock = blockHit.block;
                        const hitFace = blockHit.face; // The face of the targetBlock that was hit
                        let placementLocation = { ...targetBlock.location };

                        // Adjust placementLocation based on the hit face
                        // This determines where the new block will be placed relative to the target block
                        switch (hitFace) {
                            case "Up": placementLocation.y++; break;
                            case "Down": placementLocation.y--; break;
                            case "North": placementLocation.z--; break; // Negative Z
                            case "South": placementLocation.z++; break; // Positive Z
                            case "East": placementLocation.x++; break;  // Positive X
                            case "West": placementLocation.x--; break;  // Negative X
                            default:
                                // This case should ideally not be reached if blockHit.face is valid
                                player.sendMessage("§cInvalid block face hit for Extended Reach.");
                                return;
                        }

                        try {
                            const dimension = player.dimension;
                            dimension.setBlockPermutation(placementLocation, blockPermutationToPlace);

                            // Consume item if not in creative mode
                            if (player.getGameMode() !== GameMode.creative) {
                                const equippable = player.getComponent("minecraft:equippable");
                                if (equippable) {
                                    const mainHandItem = equippable.getEquipment(EquipmentSlot.Mainhand);
                                    // Double check the item in hand before consuming
                                    if (mainHandItem && mainHandItem.typeId === heldItemId) {
                                        if (mainHandItem.amount === 1) {
                                            // If it's the last item, remove it
                                            equippable.setEquipment(EquipmentSlot.Mainhand, undefined);
                                        } else {
                                            // Otherwise, decrement the amount
                                            mainHandItem.amount--;
                                            equippable.setEquipment(EquipmentSlot.Mainhand, mainHandItem);
                                        }
                                    }
                                }
                            }
                            // player.sendMessage("§aBlock placed (Extended Reach)."); // Optional feedback
                            eventData.cancel = true; // Cancel the original item use event as we've handled it
                            return; // Extended Reach action taken, exit handler.

                        } catch (placementError) {
                            // Log error if placement fails for some reason
                            console.warn(`[Extended Reach] Error placing block at ${JSON.stringify(placementLocation)}: ${placementError}`);
                            player.sendMessage("§cCould not place block with Extended Reach at that location.");
                            // Decide whether to cancel the event even if our placement fails.
                            // If we don't cancel, vanilla short-range placement might still occur.
                            // For now, let's not cancel so vanilla behavior can take over on failure.
                            // eventData.cancel = true;
                            // return;
                        }
                    }
                    // If no block was hit by raycast (e.g., player looking at air too far away),
                    // let the event proceed normally (e.g., throwing an egg, eating food).
                }
            })

        }


        // --- Sneak + Use Tool Menu Logic ---
        // This part of the code only runs if the Extended Reach logic above did not `return`.
        // It specifically checks if the player IS sneaking to open tool option menus.
        if (!player.isSneaking) {
            return; // Ignore non-sneaking players for menu opening if Extended Reach conditions weren't met
        }

        const itemId = itemStack?.typeId; // Get the ID of the item being used

        // Determine which UI function to call based on the item ID for sneak-use actions
        let uiFunctionToShow = null;
        let shouldCancelSneakUse = false; // Flag to determine if event should be cancelled for menu

        if (itemId === SELECTION_TOOL_ID) { uiFunctionToShow = showSelectionOptionsMenu; shouldCancelSneakUse = true; }
        else if (itemId === PASTE_TOOL_ID) { uiFunctionToShow = showPasteOptionsForm; shouldCancelSneakUse = true; }
        else if (itemId === BRUSH_TOOL_ID) { uiFunctionToShow = showRegularBrushSettingsForm; shouldCancelSneakUse = true; }
        else if (itemId === PAINT_BRUSH_TOOL_ID) { uiFunctionToShow = showPaintBrushSettingsForm; shouldCancelSneakUse = true; }
        else if (itemId === COPY_TOOL_ID) { uiFunctionToShow = showCopyOptionsForm; shouldCancelSneakUse = true; }
        else if (itemId === HEIGHT_BRUSH_TOOL_ID) { uiFunctionToShow = showHeightBrushSettingsForm; shouldCancelSneakUse = true; }
        else if (itemId === LINE_TOOL_ID) { uiFunctionToShow = showLineToolSettingsForm; shouldCancelSneakUse = true; }
        else if (itemId === FOLIAGE_BRUSH_TOOL_ID) { uiFunctionToShow = showFoliageBrushSettingsForm; shouldCancelSneakUse = true; }
        else if (itemId === PATH_TOOL_ID) { uiFunctionToShow = showPathToolOptionsMenu; shouldCancelSneakUse = true; }

        // If a relevant tool was sneak-used, cancel the event and show the UI
        if (shouldCancelSneakUse && uiFunctionToShow) {
            eventData.cancel = true; // *** CRITICAL: Prevent default item use action ***

            // Use system.run to reliably execute async UI calls after the event handler finishes
            system.run(async () => {
                // Re-check player validity inside the async callback, as they might disconnect
                if (player.isValid()) {
                    try {
                        await uiFunctionToShow(player); // Call the determined UI function
                    } catch (uiError) {
                        // Log errors specifically from showing the UI forms
                        console.error(`[ItemUseBefore-UI] Error showing form for ${itemId}: ${uiError}\n${uiError.stack}`);
                        // Avoid sending message if player is no longer valid
                        if (player.isValid()) {
                            player.sendMessage("§cError opening tool options menu.");
                        }
                    }
                }
            });
        }
        // If no matching tool ID for sneak-use, do nothing (event continues normally, if not cancelled by Extended Reach).

    }); // End subscribe
} // End setup function
