// scripts/events/item_use_on_before.js

import { world, system, Player } from "@minecraft/server";
import {
    PALETTE_PICKER_TOOL_ID,
    MASK_PICKER_TOOL_ID
} from "../constants/index.js";
import {
    getPlayerSelection,
    setPlayerSelection
} from "../state/index.js";
import {
    parseWeightedBlockInput
} from "../utils/index.js";

/**
 * Sets up the listener for the 'before item use on block' event.
 * Handles actions for tools that need to interact with a specific block
 * *before* any default behavior occurs (like Palette Picker and Mask Picker).
 */
export function setupItemUseOnBeforeListener() {
    world.beforeEvents.itemUseOn.subscribe(eventData => {
        const { source, itemStack, block } = eventData;

        if (!(source instanceof Player) || !itemStack) {
            return;
        }

        const player = source;
        const itemId = itemStack.typeId;

        const isPalettePicker = (itemId === PALETTE_PICKER_TOOL_ID);
        const isMaskPicker = (itemId === MASK_PICKER_TOOL_ID);

        if (isPalettePicker || isMaskPicker) {
            eventData.cancel = true; // Cancel the default item use on block action

            let targetBlockId;
            try {
                // It's crucial to get the block from the event data directly
                // as player.getBlockFromViewDirection() might not be what we want here.
                if (!block || !block.isValid()) { // Check if the block from eventData is valid
                    player.sendMessage("§cInvalid target block.");
                    return;
                }
                targetBlockId = block.typeId; // Get typeId from the event's block
                if (!targetBlockId || targetBlockId === "minecraft:air") {
                    player.sendMessage("§cCannot pick air.");
                    return;
                }
            } catch (e) {
                // Log the error with more context if possible
                console.warn(`[Picker Tool] Failed to get typeId from event block at ${block?.location?.x},${block?.location?.y},${block?.location?.z}: ${e}`);
                // Use system.run for UI calls after an event handler
                system.run(() => player.isValid() && player.sendMessage("§cCould not identify block type."));
                return;
            }


            // Run the main logic in a system.run to avoid issues with modifying state/UI directly in event handler
            system.run(() => {
                if (!player.isValid()) return; // Re-check player validity

                const selectionData = getPlayerSelection(player);
                const isSneaking = player.isSneaking;
                const blockIdSimple = targetBlockId.replace('minecraft:', '');

                if (isPalettePicker) {
                    if (isSneaking) { // --- Remove from Active Palette ---
                        if (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0) {
                            player.sendMessage("§cActive Palette is empty. Cannot remove blocks.");
                            return;
                        }

                        const activePalette = selectionData.activePaletteData;
                        const blockIndex = activePalette.findIndex(entry => entry.id === targetBlockId);
                        if (blockIndex !== -1) {
                            activePalette.splice(blockIndex, 1);
                            selectionData.activePaletteData = activePalette;
                            selectionData.activePaletteIsDirty = true;
                            setPlayerSelection(player, selectionData);
                            player.sendMessage(`§eRemoved ${blockIdSimple} from Active Palette.`);
                        } else {
                            player.sendMessage(`§c${blockIdSimple} not found in Active Palette.`);
                        }
                    } else { // --- Add to Active Palette ---
                        selectionData.activePaletteData = selectionData.activePaletteData ?? [];
                        const activePalette = selectionData.activePaletteData;

                        const alreadyExists = activePalette.some(entry => entry.id === targetBlockId);
                        if (alreadyExists) {
                            player.sendMessage(`§e${blockIdSimple} is already in the Active Palette.`);
                        } else {
                            activePalette.push({ id: targetBlockId, weight: 1 }); // Add with default weight 1
                            selectionData.activePaletteData = activePalette;
                            selectionData.activePaletteIsDirty = true;
                            setPlayerSelection(player, selectionData);
                            player.sendMessage(`§aAdded ${blockIdSimple} to Active Palette.`);
                        }
                    }
                }
                else if (isMaskPicker) {
                    const currentMaskInput = selectionData.paintBrushMaskInput ?? "";
                    const parsedCurrent = parseWeightedBlockInput(currentMaskInput); // This returns {blocks, errors}
                    const currentMaskIds = new Set(parsedCurrent.blocks.map(b => b.id)); // Get IDs from parsed blocks

                    if (isSneaking) { // Remove from mask
                        if (currentMaskIds.has(targetBlockId)) {
                            currentMaskIds.delete(targetBlockId);
                            // Reconstruct the string from the Set for simplicity, or manage the parsed array
                            const newMaskInput = Array.from(currentMaskIds).join(", ");
                            selectionData.paintBrushMaskInput = newMaskInput;
                            setPlayerSelection(player, selectionData);
                            player.sendMessage(`§eRemoved ${blockIdSimple} from paint brush mask.`);
                        } else {
                            player.sendMessage(`§c${blockIdSimple} not found in paint brush mask.`);
                        }
                    } else { // Add to mask
                        if (currentMaskIds.has(targetBlockId)) {
                            player.sendMessage(`§e${blockIdSimple} is already in the paint brush mask.`);
                        } else {
                            currentMaskIds.add(targetBlockId);
                            const newMaskInput = Array.from(currentMaskIds).join(", ");
                            selectionData.paintBrushMaskInput = newMaskInput;
                            setPlayerSelection(player, selectionData);
                            player.sendMessage(`§aAdded ${blockIdSimple} to paint brush mask.`);
                        }
                    }
                }
            }); // End of system.run
        } // End of if (isPalettePicker || isMaskPicker)
    }); // End of world.beforeEvents.itemUseOn.subscribe
}
