// scripts/events/player_tick_movement.js

import { world, system, Player, GameMode } from "@minecraft/server";
import {
    PLAYER_SPEED_INTERVAL,
    DEFAULT_PLAYER_SPEED_MULTIPLIER,
    DEFAULT_PLAYER_VERTICAL_SPEED_MULTIPLIER,
    DEFAULT_AUTO_SPECTATOR_DETECTION_DISTANCE // This constant will be used directly
} from "../constants/index.js";
import {
    getPlayerSelection,
    setPlayerSelection
} from "../state/index.js";

// Cooldown duration in ticks. If PLAYER_SPEED_INTERVAL is 3, 10 ticks = 0.5 seconds.
const AUTO_SPECTATOR_COOLDOWN_DURATION_TICKS = 5;

/**
 * Sets up the periodic interval timer for player movement adjustments and auto-spectator logic.
 */
export function setupPlayerTickMovementListener() {
    system.runInterval(() => {
        try {
            for (const player of world.getAllPlayers()) {
                if (!player || !player.isValid()) continue;

                const selectionData = getPlayerSelection(player);
                // Use 'let' for properties that might be modified and saved back.
                let {
                    playerSpeedMultiplier,
                    playerVerticalSpeedMultiplier,
                    autoSpectatorEnabled,
                    wasAutoSwitchedToSpectator,
                    // autoSpectatorDetectionDistance, // <<< REMOVED from player state destructuring
                    autoSpectatorCooldownTicks
                } = selectionData;

                // --- Player Horizontal Speed Multiplier Logic ---
                if (player.getGameMode() === GameMode.creative && player.isSprinting) {
                    try {
                        const speedMultiplierSetting = playerSpeedMultiplier ?? DEFAULT_PLAYER_SPEED_MULTIPLIER;
                        const additionalForceFactor = speedMultiplierSetting - DEFAULT_PLAYER_SPEED_MULTIPLIER;

                        if (additionalForceFactor > 0) {
                            const baseForceMagnitude = 0.15;
                            const forceMagnitude = baseForceMagnitude * additionalForceFactor;
                            const viewDir = player.getViewDirection();
                            player.applyKnockback(viewDir.x, viewDir.z, forceMagnitude, 0.0);
                        }
                    } catch (e) {
                        // console.warn(`[PlayerSpeed] Error applying horizontal speed: ${e}`);
                    }
                }

                // --- Player Vertical Speed Multiplier Logic ---
                if (player.getGameMode() === GameMode.creative && player.isFlying) {
                    try {
                        const verticalSpeedMultiplierSetting = playerVerticalSpeedMultiplier ?? DEFAULT_PLAYER_VERTICAL_SPEED_MULTIPLIER;
                        const additionalVerticalForceFactor = verticalSpeedMultiplierSetting - DEFAULT_PLAYER_VERTICAL_SPEED_MULTIPLIER;

                        if (additionalVerticalForceFactor > 0) {
                            const baseVerticalForceMagnitude = 0.075;
                            const boostImpulse = baseVerticalForceMagnitude * additionalVerticalForceFactor;

                            const currentVelocity = player.getVelocity();
                            const verticalVelocityIntentThreshold = 0.05;

                            if (currentVelocity.y > verticalVelocityIntentThreshold) {
                                player.applyKnockback(0, 0, 0, boostImpulse);
                            } else if (currentVelocity.y < -verticalVelocityIntentThreshold) {
                                player.applyKnockback(0, 0, 0, -boostImpulse);
                            }
                        }
                    } catch (e) {
                        // console.warn(`[PlayerSpeed] Error applying vertical speed: ${e}`);
                    }
                }

                // --- Auto Spectator Cooldown Management ---
                let cooldownDecremented = false;
                if (autoSpectatorCooldownTicks > 0) {
                    selectionData.autoSpectatorCooldownTicks = autoSpectatorCooldownTicks - 1;
                    cooldownDecremented = true;
                    autoSpectatorCooldownTicks = selectionData.autoSpectatorCooldownTicks; // Update local variable
                }


                // --- Auto Spectator Logic ---
                let stateChangedThisTick = cooldownDecremented;

                if (autoSpectatorEnabled) {
                    // <<< Use the fixed constant for detection distance >>>
                    const detectionDistance = DEFAULT_AUTO_SPECTATOR_DETECTION_DISTANCE;
                    let isWallInFront = false;

                    if (player.getGameMode() === GameMode.creative && player.isFlying) {
                        try {
                            const headLocation = player.getHeadLocation();
                            const viewDirection = player.getViewDirection();
                            const raycastStartLocation = {
                                x: headLocation.x + viewDirection.x * 0.01,
                                y: headLocation.y + viewDirection.y * 0.01,
                                z: headLocation.z + viewDirection.z * 0.01,
                            };
                            const blockHit = player.dimension.getBlockFromRay(raycastStartLocation, viewDirection, {
                                maxDistance: detectionDistance
                            });
                            if (blockHit?.block) {
                                isWallInFront = true;
                            }
                        } catch (e) {
                            isWallInFront = false;
                        }
                    }

                    // Condition to switch TO spectator
                    if (player.getGameMode() === GameMode.creative && player.isFlying && isWallInFront) {
                        player.setGameMode(GameMode.spectator);
                        selectionData.wasAutoSwitchedToSpectator = true;
                        selectionData.autoSpectatorCooldownTicks = AUTO_SPECTATOR_COOLDOWN_DURATION_TICKS;
                        stateChangedThisTick = true;
                        // player.sendMessage("§7Auto-switched to Spectator mode (wall detected)."); 
                    }
                    // Condition to switch BACK to creative
                    else if (player.getGameMode() === GameMode.spectator && wasAutoSwitchedToSpectator && autoSpectatorCooldownTicks === 0) {
                        let isPlayerLocationInAir = false;
                        try {
                            const blockAtFeet = player.dimension.getBlock(player.location);
                            const blockAtHead = player.dimension.getBlock(player.getHeadLocation());
                            if (blockAtFeet?.isAir && blockAtHead?.isAir) {
                                isPlayerLocationInAir = true;
                            }
                        } catch (e) {
                            isPlayerLocationInAir = true;
                        }

                        if (isPlayerLocationInAir) {
                            player.setGameMode(GameMode.creative);
                            selectionData.wasAutoSwitchedToSpectator = false;
                            stateChangedThisTick = true;
                            // player.sendMessage("§7Auto-switched back to Creative mode (location clear)."); 
                        }
                    }
                } else if (wasAutoSwitchedToSpectator && player.getGameMode() === GameMode.spectator) {
                    player.setGameMode(GameMode.creative);
                    selectionData.wasAutoSwitchedToSpectator = false;
                    selectionData.autoSpectatorCooldownTicks = 0;
                    stateChangedThisTick = true;
                    // player.sendMessage("§7Auto-Spectator disabled; switched back to Creative mode."); 
                }

                if (stateChangedThisTick) {
                    setPlayerSelection(player, selectionData);
                }

            } // End player loop
        } catch (intervalError) {
            console.error(`[WorldEdit Player Tick Movement] Critical Error: ${intervalError}\n${intervalError.stack}`);
        }
    }, PLAYER_SPEED_INTERVAL);
}
