// scripts/libs/simplex_noise.js
// Based on the public domain implementation by Stefan Gustavson and Peter Eastman
// Adapted for Minecraft Bedrock scripting environment
// Added FBM, Domain Warp, and Worley (Cellular) noise calculation
// Fixed stack overflow in constructor

/**
 * A deterministic simplex noise generator for 2D noise.
 * Also includes FBM, Domain Warp, and Worley (Cellular) noise methods.
 */
export class SimplexNoise {
    /**
     * Initializes a new instance of the SimplexNoise class.
     * @param {number} [seed=0] - A seed value for the noise generator. Defaults to 0.
     * Uses a simple pseudo-random number generator based on the seed.
     */
    constructor(seed = 0) {
        // Simple seeded PRNG (Linear Congruential Generator - basic)
        this.seed = seed; // Store the original seed if needed later
        let currentSeed = seed; // Use a mutable variable for the PRNG state
        const random = () => {
            // Basic LCG parameters
            const a = 1664525;
            const c = 1013904223;
            const m = 2 ** 32; // 2^32
            currentSeed = (a * currentSeed + c) % m; // Update the mutable seed
            return currentSeed / m;
        };

        // Initialize permutation table using the seeded random function
        this.p = new Uint8Array(256);
        for (let i = 0; i < 256; i++) {
            this.p[i] = i;
        }
        for (let i = 255; i > 0; i--) {
            const r = Math.floor(random() * (i + 1));
            [this.p[i], this.p[r]] = [this.p[r], this.p[i]]; // Swap
        }

        // Duplicate the permutation table to avoid buffer overflow (standard practice)
        this.perm = new Uint8Array(512);
        this.permMod12 = new Uint8Array(512);
        for (let i = 0; i < 512; i++) {
            this.perm[i] = this.p[i & 255];
            this.permMod12[i] = this.perm[i] % 12;
        }

        // Skewing and unskewing factors for 2D
        this.F2 = 0.5 * (Math.sqrt(3.0) - 1.0);
        this.G2 = (3.0 - Math.sqrt(3.0)) / 6.0;

        // Gradient vectors for 2D (only 8 needed for 2D)
        this.grad3 = [
            [1, 1, 0], [-1, 1, 0], [1, -1, 0], [-1, -1, 0],
            [1, 0, 1], [-1, 0, 1], [1, 0, -1], [-1, 0, -1],
            [0, 1, 1], [0, -1, 1], [0, 1, -1], [0, -1, -1]
        ];
    }


    /**
     * Computes the 2D Simplex noise value for the given coordinates.
     * @param {number} xin The x-coordinate.
     * @param {number} yin The y-coordinate.
     * @returns {number} The noise value, typically ranging from -1.0 to 1.0.
     */
    noise2D(xin, yin) {
        const perm = this.perm;
        const permMod12 = this.permMod12;
        const grad3 = this.grad3;
        const F2 = this.F2;
        const G2 = this.G2;
        let n0, n1, n2; // Noise contributions from the three corners

        // Skew the input space to determine which simplex cell we're in
        const s = (xin + yin) * F2; // Hairy factor for 2D
        const i = Math.floor(xin + s);
        const j = Math.floor(yin + s);
        const t = (i + j) * G2;
        const X0 = i - t; // Unskew the cell origin back to (x,y) space
        const Y0 = j - t;
        const x0 = xin - X0; // The x,y distances from the cell origin
        const y0 = yin - Y0;

        // For the 2D case, the simplex shape is an equilateral triangle.
        // Determine which simplex we are in.
        let i1, j1; // Offsets for second corner of simplex in (i,j) coords
        if (x0 > y0) { i1 = 1; j1 = 0; } // lower triangle, XY order: (0,0)->(1,0)->(1,1)
        else { i1 = 0; j1 = 1; }      // upper triangle, YX order: (0,0)->(0,1)->(1,1)

        // A step of (1,0) in (i,j) means a step of (1-c,-c) in (x,y), and
        // a step of (0,1) in (i,j) means a step of (-c,1-c) in (x,y), where
        // c = (3-sqrt(3))/6
        const x1 = x0 - i1 + G2; // Offsets for second corner in (x,y) coords
        const y1 = y0 - j1 + G2;
        const x2 = x0 - 1.0 + 2.0 * G2; // Offsets for third corner in (x,y) coords
        const y2 = y0 - 1.0 + 2.0 * G2;

        // Work out the hashed gradient indices of the three simplex corners
        const ii = i & 255;
        const jj = j & 255;
        // Use perm directly, permMod12 was for grad3 lookup which is now direct index
        const gi0 = this.permMod12[ii + this.perm[jj]];
        const gi1 = this.permMod12[ii + i1 + this.perm[jj + j1]];
        const gi2 = this.permMod12[ii + 1 + this.perm[jj + 1]];

        // Calculate the contribution from the three corners
        let t0 = 0.5 - x0 * x0 - y0 * y0;
        if (t0 < 0) n0 = 0.0;
        else {
            t0 *= t0;
            const g0 = grad3[gi0];
            n0 = t0 * t0 * (g0[0] * x0 + g0[1] * y0); // (x,y) of grad3 used for 2D gradient
        }

        let t1 = 0.5 - x1 * x1 - y1 * y1;
        if (t1 < 0) n1 = 0.0;
        else {
            t1 *= t1;
            const g1 = grad3[gi1];
            n1 = t1 * t1 * (g1[0] * x1 + g1[1] * y1);
        }

        let t2 = 0.5 - x2 * x2 - y2 * y2;
        if (t2 < 0) n2 = 0.0;
        else {
            t2 *= t2;
            const g2 = grad3[gi2];
            n2 = t2 * t2 * (g2[0] * x2 + g2[1] * y2);
        }

        // Add contributions from each corner to get the final noise value.
        // The result is scaled to stay just inside [-1,1]
        return 70.0 * (n0 + n1 + n2);
    }

    /**
     * Calculates Fractal Brownian Motion (FBM) noise by summing multiple octaves of Simplex noise.
     * @param {number} x The x-coordinate.
     * @param {number} y The y-coordinate.
     * @param {number} octaves The number of noise layers to combine (e.g., 4-8).
     * @param {number} lacunarity Frequency multiplier between octaves (typically 2.0).
     * @param {number} persistence (Gain) Amplitude multiplier between octaves (typically 0.5).
     * @returns {number} The FBM noise value, normalized to approx [-1, 1].
     */
    fbm2D(x, y, octaves, lacunarity, persistence) {
        let total = 0.0;
        let frequency = 1.0;
        let amplitude = 1.0;
        let maxValue = 0.0; // Used for normalizing result to [-1, 1]

        for (let i = 0; i < octaves; i++) {
            total += this.noise2D(x * frequency, y * frequency) * amplitude;

            maxValue += amplitude; // Accumulate max possible amplitude

            amplitude *= persistence;
            frequency *= lacunarity;
        }

        // Normalize the result to be consistently within [-1, 1] range
        if (maxValue === 0) return 0; // Avoid division by zero
        return total / maxValue;
    }

    /**
     * Applies Domain Warping to input coordinates before sampling a noise function.
     * Uses the main noise instance (`this.noise2D`) for displacement, adding offsets to coordinates.
     * @param {number} x The original x-coordinate.
     * @param {number} y The original y-coordinate.
     * @param {number} warpAmplitude The maximum displacement distance.
     * @param {number} warpFrequency The frequency/scale of the warp noise.
     * @param {(x: number, y: number) => number} noiseFunction The primary noise function (e.g., this.noise2D or bound fbm2D) to call with warped coordinates.
     * @returns {number} The noise value sampled at the warped coordinates.
     */
    domainWarp2D(x, y, warpAmplitude, warpFrequency, noiseFunction) {
        // 1. Get displacement offsets using the *same* noise instance but with coordinate offsets
        // Add large offsets to sample different parts of the noise field for warp X and Y
        const offsetX = this.noise2D((x + 123.45) * warpFrequency, (y + 543.21) * warpFrequency) * warpAmplitude;
        const offsetY = this.noise2D((x - 987.65) * warpFrequency, (y - 654.32) * warpFrequency) * warpAmplitude;

        // 2. Calculate warped coordinates
        const warpedX = x + offsetX;
        const warpedY = y + offsetY;

        // 3. Sample the primary noise function at the warped coordinates
        return noiseFunction(warpedX, warpedY);
    }

    /**
     * Calculates 2D Worley (Cellular) noise (F1 Euclidean distance).
     * Finds the distance to the closest feature point in a grid.
     * @param {number} xin The x-coordinate.
     * @param {number} yin The y-coordinate.
     * @param {number} [jitter=0.8] Controls the randomness of feature points within cells (0 to 1).
     * @returns {number} The distance to the nearest feature point. Range is >= 0.
     * Lower values are closer to feature points.
     */
    worleyF1_2D(xin, yin, jitter = 0.8) {
        const perm = this.perm; // Use the permutation table from constructor

        // Determine the integer cell coordinates
        const ix = Math.floor(xin);
        const iy = Math.floor(yin);

        let minDistSq = Infinity; // Use squared distance for comparison efficiency

        // Check the 3x3 neighborhood of cells
        for (let cy = iy - 1; cy <= iy + 1; cy++) {
            for (let cx = ix - 1; cx <= ix + 1; cx++) {
                // Calculate pseudo-random feature point within the cell (cx, cy)
                // Use permutation table for hashing cell coordinates
                const hashIndex1 = perm[(cx & 255) + perm[cy & 255]] & 255;
                const hashIndex2 = perm[(cx & 255) + 1 + perm[(cy & 255) + 1]] & 255; // Offset indices slightly

                // Generate offsets within the cell using the hash values
                // Scale by jitter to control randomness (0 = center, 1 = full cell range)
                const offsetX = (hashIndex1 / 255.0 - 0.5) * 2.0 * jitter; // Offset range [-jitter, jitter]
                const offsetY = (hashIndex2 / 255.0 - 0.5) * 2.0 * jitter;

                // Feature point location (center of cell + jittered offset)
                const featureX = cx + 0.5 + offsetX;
                const featureY = cy + 0.5 + offsetY;

                // Calculate squared Euclidean distance
                const dx = xin - featureX;
                const dy = yin - featureY;
                const distSq = dx * dx + dy * dy;

                // Update minimum distance squared
                if (distSq < minDistSq) {
                    minDistSq = distSq;
                }
            }
        }

        // Return the actual distance (sqrt)
        return Math.sqrt(minDistSq);
    }

    // --- Potential future additions: worleyF2_2D, worleyF2F1_2D ---
}
