// scripts/state/dynamic_props.js

import { PALETTES_PROPERTY_ID } from "../constants/settings.js";
import { Player } from "@minecraft/server";
// Import necessary functions from player_state
import {
    getActiveSavedPaletteName, // Renamed
    setActiveSavedPaletteName, // Renamed
    discardActivePaletteChanges, // Renamed
} from "./player_state.js";

/** @typedef {import("./player_state.js").WeightedBlock} WeightedBlock */

/**
 * Reads saved palettes from player's dynamic properties.
 * @param {Player} player The player entity.
 * @returns {Object.<string, WeightedBlock[]>}
 */
export function getPlayerPalettes(player) {
    try {
        const palettesJson = player.getDynamicProperty(PALETTES_PROPERTY_ID);
        if (typeof palettesJson === 'string') {
            return JSON.parse(palettesJson);
        }
    } catch (e) {
        console.warn(`[getPlayerPalettes] Error reading or parsing palettes for ${player.name ?? player.id}: ${e}`);
    }
    return {};
}

/**
 * Saves the palettes object to the player's dynamic properties.
 * @param {Player} player The player entity.
 * @param {Object.<string, WeightedBlock[]>} palettesObject
 * @returns {boolean}
 */
export function savePlayerPalettes(player, palettesObject) {
    try {
        const palettesJson = JSON.stringify(palettesObject);
        // Basic size check (dynamic properties have limits)
        if (palettesJson.length >= 32000) {
            console.error(`[savePlayerPalettes] Palettes JSON string too large for ${player.name ?? player.id}. Save failed.`);
            player.sendMessage("§cError: Palette data too large to save.");
            return false;
        }
        player.setDynamicProperty(PALETTES_PROPERTY_ID, palettesJson);
        return true;
    } catch (e) {
        console.error(`[savePlayerPalettes] Error stringifying or saving palettes for ${player.name ?? player.id}: ${e}`);
        player.sendMessage("§cAn internal error occurred while saving palettes.");
        return false;
    }
}

/**
 * Adds or updates a palette in the player's saved palettes.
 * @param {Player} player The player entity.
 * @param {string} name The name of the palette.
 * @param {WeightedBlock[]} blocksWithWeights The array of weighted blocks for the palette.
 * @returns {boolean} True if saving was successful, false otherwise.
 */
export function addPlayerPalette(player, name, blocksWithWeights) {
    if (!name || !blocksWithWeights) {
        console.warn(`[addPlayerPalette] Invalid input provided for ${player.name ?? player.id}. Name: ${name}`);
        return false;
    }
    if (blocksWithWeights.length === 0) {
        console.warn(`[addPlayerPalette] Attempted to save empty palette "${name}" for ${player.name ?? player.id}.`);
        player.sendMessage(`§cCannot save empty palette "${name}".`);
        return false;
    }

    const palettes = getPlayerPalettes(player);
    palettes[name] = blocksWithWeights;
    return savePlayerPalettes(player, palettes);
}

/**
 * Deletes a palette from the player's saved palettes.
 * Also handles clearing active/working state if the deleted palette was active.
 * @param {Player} player The player entity.
 * @param {string} name The name of the palette to delete.
 * @returns {boolean} True if deletion and saving were successful, false otherwise.
 */
export function deletePlayerPalette(player, name) {
    if (!name) {
        console.warn(`[deletePlayerPalette] No name provided for deletion for player ${player.name ?? player.id}.`);
        return false;
    }

    const palettes = getPlayerPalettes(player);

    if (palettes[name]) {
        const currentActiveSaved = getActiveSavedPaletteName(player); // Use renamed getter
        delete palettes[name];

        if (currentActiveSaved === name) {
            setActiveSavedPaletteName(player, undefined); // Use renamed setter
            // Also discard Active Palette changes as they likely relate to the deleted active saved palette
            discardActivePaletteChanges(player, false); // Use renamed function, no message needed here
            console.log(`Cleared active states for ${player.name} because deleted palette "${name}" was the active saved one.`);
        }

        return savePlayerPalettes(player, palettes);
    } else {
        console.warn(`[deletePlayerPalette] Palette "${name}" not found for player ${player.name ?? player.id}.`);
        return false;
    }
}
