// scripts/state/global_state.js

import { world } from "@minecraft/server";
import {
    GLOBAL_SETTINGS_PROPERTY_ID,
    OUTLINE_PARTICLE,
    MAX_UNDO_HISTORY,
    DEFAULT_BRUSH_SIZE
} from "../constants/settings.js";

/**
 * @typedef {object} GlobalSettings
 * @property {boolean} showActionBarInfo - Whether to display tool info in the action bar.
 * @property {string} outlineParticleId - The particle ID used for previews.
 * @property {number} maxUndoHistory - The maximum number of undo steps stored per player.
 * @property {number} defaultBrushSize - The default brush size used if player hasn't set one.
 * // Add future global settings properties here
 */

/**
 * Default values for global settings. Used if the dynamic property is missing or invalid.
 * @type {GlobalSettings}
 */
export const DEFAULT_GLOBAL_SETTINGS = {
    showActionBarInfo: true,
    outlineParticleId: OUTLINE_PARTICLE, // Use default from constants
    maxUndoHistory: MAX_UNDO_HISTORY,   // Use default from constants
    defaultBrushSize: DEFAULT_BRUSH_SIZE, // Use default from constants
};

/**
 * Retrieves the global settings object from world dynamic properties.
 * Merges saved settings with defaults to ensure all keys are present.
 * @returns {GlobalSettings} The global settings object.
 */
export function getGlobalSettings() {
    let savedSettings = {};
    try {
        const settingsJson = world.getDynamicProperty(GLOBAL_SETTINGS_PROPERTY_ID);
        if (typeof settingsJson === 'string' && settingsJson.length > 0) {
            savedSettings = JSON.parse(settingsJson);
            // Basic type validation (could be more thorough)
            if (typeof savedSettings !== 'object' || savedSettings === null) {
                console.warn("[BWE Global Settings] Invalid JSON type found in dynamic property. Resetting to defaults.");
                savedSettings = {}; // Reset if not an object
            }
        }
    } catch (e) {
        console.warn(`[BWE Global Settings] Error reading or parsing dynamic property '${GLOBAL_SETTINGS_PROPERTY_ID}': ${e}. Using defaults.`);
        savedSettings = {}; // Reset on error
    }

    // Merge saved settings with defaults to ensure all keys exist and have correct types (basic check)
    const finalSettings = { ...DEFAULT_GLOBAL_SETTINGS };
    for (const key in DEFAULT_GLOBAL_SETTINGS) {
        if (Object.hasOwnProperty.call(DEFAULT_GLOBAL_SETTINGS, key)) {
            // Check if saved setting exists and has the same type as the default
            if (savedSettings.hasOwnProperty(key) && typeof savedSettings[key] === typeof DEFAULT_GLOBAL_SETTINGS[key]) {
                finalSettings[key] = savedSettings[key];
            }
            // If saved setting is missing or wrong type, the default value remains
        }
    }

    // Specific validation/clamping for numeric values
    finalSettings.maxUndoHistory = Math.max(0, Math.min(1000, Math.floor(finalSettings.maxUndoHistory))); // Example clamp
    finalSettings.defaultBrushSize = Math.max(1, Math.min(64, Math.floor(finalSettings.defaultBrushSize))); // Example clamp

    return finalSettings;
}

/**
 * Saves the global settings object to world dynamic properties.
 * @param {GlobalSettings} settings The settings object to save.
 * @returns {boolean} True if saving was successful, false otherwise.
 */
export function saveGlobalSettings(settings) {
    try {
        // Ensure we are saving a valid object
        if (typeof settings !== 'object' || settings === null) {
            throw new Error("Attempted to save invalid settings object.");
        }
        const settingsJson = JSON.stringify(settings);
        // Basic size check (dynamic properties have limits, though world props are larger)
        if (settingsJson.length >= 32767) { // Check against a reasonable limit
            console.error("[BWE Global Settings] Settings JSON string potentially too large. Save failed.");
            return false;
        }
        world.setDynamicProperty(GLOBAL_SETTINGS_PROPERTY_ID, settingsJson);
        return true;
    } catch (e) {
        console.error(`[BWE Global Settings] Error stringifying or saving dynamic property '${GLOBAL_SETTINGS_PROPERTY_ID}': ${e}`);
        return false;
    }
}
