// scripts/state/player_state.js

import {
    // Import defaults
    DEFAULT_FOLIAGE_RADIUS,
    DEFAULT_FOLIAGE_DENSITY,
    DEFAULT_FOLIAGE_SOURCE,
    DEFAULT_FOLIAGE_PRESET,
    DEFAULT_FOLIAGE_CUSTOM,
    DEFAULT_BRUSH_MATERIAL,
    DEFAULT_PAINT_MASK_BLOCK,
    DEFAULT_HEIGHT_BRUSH_AMOUNT,
    DEFAULT_PATH_MATERIAL,
    DEFAULT_PATH_CURVE_TYPE,
    DEFAULT_PATH_SMOOTHNESS,
    DEFAULT_PATH_WIDTH,
    // Import terrain defaults
    DEFAULT_TERRAIN_BASE_MATERIAL,
    DEFAULT_TERRAIN_SURFACE_MATERIAL,
    DEFAULT_TERRAIN_SURFACE_DEPTH,
    DEFAULT_TERRAIN_NOISE_TYPE, 
    DEFAULT_TERRAIN_FREQUENCY,
    DEFAULT_TERRAIN_AMPLITUDE,
    DEFAULT_TERRAIN_SEED_TYPE,
    DEFAULT_TERRAIN_OCTAVES,
    DEFAULT_TERRAIN_LACUNARITY,
    DEFAULT_TERRAIN_PERSISTENCE,
    DEFAULT_TERRAIN_WARP_ENABLED,
    DEFAULT_TERRAIN_WARP_AMPLITUDE,
    DEFAULT_TERRAIN_WARP_FREQUENCY,
    DEFAULT_TERRAIN_JITTER,
    // Import player settings defaults
    DEFAULT_PLAYER_SPEED_MULTIPLIER,
    DEFAULT_PLAYER_VERTICAL_SPEED_MULTIPLIER,
    DEFAULT_AUTO_SPECTATOR_ENABLED,
    DEFAULT_EXTENDED_REACH_ENABLED,
    DEFAULT_SHOW_ACTION_FEEDBACK_MESSAGES // Import new default
} from '../constants/settings.js';
import { Player } from "@minecraft/server";
import { getGlobalSettings } from './global_state.js';

// --- Maps ---
const playerSelectionMap = new Map();
const playerClipboardMap = new Map();
const playerUndoMap = new Map();
const playerRedoMap = new Map();

// --- Internal Helper Functions ---
/**
 * Gets data associated with a player ID from a Map.
 * @param {Map<string, any>} map The map to retrieve data from.
 * @param {Player | string} playerOrId The player object or their ID string.
 * @param {any} [defaultValue=undefined] The value to return if the player ID is not found.
 * @returns {any} The data associated with the player ID or the default value.
 */
function getPlayerMapData(map, playerOrId, defaultValue = undefined) {
    const playerId = typeof playerOrId === 'string' ? playerOrId : playerOrId.id;
    return map.get(playerId) ?? defaultValue;
}

/**
 * Sets data associated with a player ID in a Map.
 * @param {Map<string, any>} map The map to store data in.
 * @param {Player | string} playerOrId The player object or their ID string.
 * @param {any} data The data to store.
 */
function setPlayerMapData(map, playerOrId, data) {
    const playerId = typeof playerOrId === 'string' ? playerOrId : playerOrId.id;
    map.set(playerId, data);
}

/**
 * Deletes data associated with a player ID from a Map.
 * @param {Map<string, any>} map The map to delete data from.
 * @param {Player | string} playerOrId The player object or their ID string.
 */
export function deletePlayerMapData(map, playerOrId) {
    const playerId = typeof playerOrId === 'string' ? playerOrId : playerOrId.id;
    map.delete(playerId);
}

// --- Typedefs ---
/**
 * Options controlling structure placement.
 * @typedef {object} PasteOptions
 * @property {boolean} pasteOrigin - Paste relative to original copy location.
 * @property {boolean} placeEntities - Place entities saved in the structure.
 * @property {number} integrity - Chance (0.0-1.0) of blocks being placed.
 * @property {string} seed - Seed for integrity placement (numeric string or empty).
 * @property {("None (0°)"|"90°"|"180°"|"270°")} rotation - Structure rotation override.
 * @property {("None"|"X"|"Z"|"XZ")} mirror - Structure mirroring.
 */

/**
 * Represents the player's current selection state, options, palettes, and tool settings.
 * @typedef {object} PlayerSelection
 * @property {{x:number, y:number, z:number} | undefined} posA - First selection point.
 * @property {{x:number, y:number, z:number} | undefined} posB - Second selection point.
 * @property {number | undefined} selectionState - 0: Inactive, 1: PosA Set, 2: Complete.
 * @property {string | undefined} activeSavedPaletteName - Name of the currently active *saved* palette (used to track which saved palette is loaded into the Active Palette).
 * @property {WeightedBlock[] | undefined} activePaletteData - Array of blocks in the temporary "Active Palette" that the player is currently working with.
 * @property {boolean | undefined} activePaletteIsDirty - Flag indicating unsaved changes in the Active Palette.
 * @property {PasteOptions | undefined} pasteOptions - Settings for the paste tool.
 * @property {boolean | undefined} copyIncludesEntities - Setting for the copy/cut tool.
 * @property {('sphere' | 'cube' | 'cylinder' | 'pyramid' | 'octahedron' | 'dome' | 'bowl' | 'cone_up' | 'cone_down' | 'disk' | 'plate' | 'torus') | undefined} brushShape - Shape for brush tools.
 * @property {number | undefined} brushSize - Size (radius/half-width) for brush tools.
 * @property {boolean | undefined} brushHollow - Hollow setting for the regular brush.
 * @property {('active' | 'manual' | string) | undefined} brushMaterialSource - Source for brush/paint material ('active', 'manual', or saved palette name). Note: 'active' refers to activePaletteData.
 * @property {string | undefined} brushManualMaterial - Manual block input string for brush/paint.
 * @property {boolean | undefined} paintBrushMaskEnabled - Masking enabled state for paint brush.
 * @property {string | undefined} paintBrushMaskInput - Mask block input string for paint brush.
 * @property {number | undefined} heightBrushAmount - Amount for height brush.
 * @property {("Raise"|"Lower") | undefined} heightBrushMode - Mode for height brush.
 * // Foliage Brush
 * @property {number | undefined} foliageBrushRadius - Radius for foliage brush.
 * @property {number | undefined} foliageBrushDensity - Density percentage for foliage brush.
 * @property {('preset' | 'custom' | 'active' | string) | undefined} foliageBrushPlantSource - Source for foliage plants ('preset', 'custom', 'active', or saved palette name).
 * @property {string | undefined} foliageBrushPresetType - Selected preset type for foliage brush.
 * @property {string | undefined} foliageBrushCustomInput - Custom plant input string for foliage brush.
 * // Line Tool
 * @property {{x:number, y:number, z:number} | undefined} linePosA - First point for line tool.
 * @property {number | undefined} lineState - 0: Ready for PosA, 1: PosA Set.
 * @property {('active' | 'manual' | string) | undefined} lineMaterialSource - Source for line material ('active', 'manual', or saved palette name).
 * @property {string | undefined} lineManualMaterial - Manual block input string for line tool.
 * // Path Tool
 * @property {Array<{x: number, y: number, z: number}> | undefined} pathPoints - Array of points set for the path tool.
 * @property {('active' | 'manual' | string) | undefined} pathMaterialSource - Source for path material ('active', 'manual', or saved palette name).
 * @property {string | undefined} pathManualMaterial - Manual block input string for path tool.
 * @property {('Linear' | 'Smooth' | 'Bezier' | 'Circular Arc') | undefined} pathCurveType - Curve type for path tool.
 * @property {number | undefined} pathSmoothness - Smoothness factor for curved paths.
 * @property {number | undefined} pathWidth - Width of the generated path.
 * @property {boolean | undefined} pathClosed - Whether the path should form a closed loop.
 * // Terrain Generation
 * @property {('active' | 'manual' | string) | undefined} terrainBaseMaterialSource - Source for terrain base material.
 * @property {string | undefined} terrainManualBaseMaterial - Manual input for terrain base material.
 * @property {('active' | 'manual' | string) | undefined} terrainSurfaceMaterialSource - Source for terrain surface material.
 * @property {string | undefined} terrainManualSurfaceMaterial - Manual input for terrain surface material.
 * @property {number | undefined} terrainSurfaceDepth - Thickness of the surface layer.
 * @property {string | undefined} terrainNoiseType - Type of noise ('Simplex', 'FBM', 'Worley').
 * @property {number | undefined} terrainFrequency - Base frequency/scale of the noise.
 * @property {number | undefined} terrainAmplitude - Height multiplier for the noise.
 * @property {("Random"|"World Seed"|"Custom") | undefined} terrainSeedType - Source for the noise seed.
 * @property {number | undefined} terrainCustomSeed - Specific seed value if terrainSeedType is 'Custom'.
 * @property {number | undefined} terrainOctaves - Number of octaves for FBM noise.
 * @property {number | undefined} terrainLacunarity - Frequency multiplier between octaves for FBM.
 * @property {number | undefined} terrainPersistence - Amplitude multiplier (gain) between octaves for FBM.
 * @property {boolean | undefined} terrainDomainWarpEnabled - Whether Domain Warp is enabled.
 * @property {number | undefined} terrainWarpAmplitude - Amplitude of the warp displacement.
 * @property {number | undefined} terrainWarpFrequency - Frequency/scale of the warp noise.
 * @property {number | undefined} terrainJitter - Jitter amount for Worley noise (0.0 to 1.0).
 * // Player Specific Settings
 * @property {number | undefined} playerSpeedMultiplier - Multiplier for player horizontal speed in creative.
 * @property {number | undefined} playerVerticalSpeedMultiplier - Multiplier for player vertical speed in creative (flying). 
 * @property {boolean | undefined} autoSpectatorEnabled - Whether the auto-spectator feature is enabled for the player. 
 * @property {boolean | undefined} wasAutoSwitchedToSpectator - Internal flag to track if the addon switched the player to spectator. 
 * @property {number | undefined} autoSpectatorCooldownTicks - Ticks remaining before allowing switch back to creative. 
 * @property {boolean | undefined} extendedReachEnabled;
 * @property {boolean | undefined} showActionFeedbackMessages; // New setting for feedback messages
 */

/**
 * Represents the state of the player's clipboard (copied selection).
 * @typedef {object} PlayerClipboard
 * @property {string} structureId The player-specific identifier for the saved structure.
 * @property {{x:number, y:number, z:number}} size The dimensions of the copied structure.
 * @property {boolean} includesEntities Whether entities were included when this structure was saved.
 * @property {{x:number, y:number, z:number} | undefined} origin The minimum corner where copied (optional).
 * @property {number} timestamp The time the copy was made.
*/

/**
 * Represents a state in the undo/redo history.
 * @typedef {object} HistoryState
 * @property {{min: {x:number, y:number, z:number}, max: {x:number, y:number, z:number}}} affectedArea Bounding box of the change.
 * @property {Array<{location: {x:number, y:number, z:number}, permutation: import('@minecraft/server').BlockPermutation}>} blocks Array of block states *before* the change.
 * @property {number} timestamp Time of the original change.
 * @property {{x:number, y:number, z:number} | undefined} tickingAreaMin - Optional: Min corner of the ticking area used for the original operation.
 * @property {{x:number, y:number, z:number} | undefined} tickingAreaMax - Optional: Max corner of the ticking area used for the original operation.
 */

/**
 * Represents a block type with an associated weight for weighted selection.
 * @typedef {{id: string, weight: number}} WeightedBlock Block ID and its selection weight.
*/


// --- Exported State Getters/Setters ---

/**
 * Gets or initializes the PlayerSelection object for a player, ensuring all properties have default values.
 * @param {Player} player The player entity.
 * @returns {PlayerSelection} The player's selection state object.
 */
export function getPlayerSelection(player) {
    const playerId = player.id;
    let data = playerSelectionMap.get(playerId);
    if (!data) { data = {}; playerSelectionMap.set(playerId, data); }

    const globalSettings = getGlobalSettings();
    const globalDefaultBrushSize = globalSettings.defaultBrushSize;

    // --- Ensure Defaults ---
    data.posA = data.posA ?? undefined;
    data.posB = data.posB ?? undefined;
    data.selectionState = data.selectionState ?? 0;
    data.activePaletteData = data.activePaletteData ?? undefined;
    data.activePaletteIsDirty = data.activePaletteIsDirty ?? false;
    data.activeSavedPaletteName = data.activeSavedPaletteName ?? undefined;
    data.pasteOptions = data.pasteOptions ?? {
        pasteOrigin: false, placeEntities: true, integrity: 1.0, // Changed default to true
        seed: "", rotation: "None (0°)", mirror: "None"
    };
    data.copyIncludesEntities = data.copyIncludesEntities ?? true; // Default for copy can be true
    // Shared Brush Settings
    data.brushShape = data.brushShape ?? 'sphere';
    data.brushSize = data.brushSize ?? globalDefaultBrushSize;
    data.brushHollow = data.brushHollow ?? false;
    data.brushMaterialSource = data.brushMaterialSource ?? 'manual';
    data.brushManualMaterial = data.brushManualMaterial ?? DEFAULT_BRUSH_MATERIAL;
    // Paint Brush Specific
    data.paintBrushMaskEnabled = data.paintBrushMaskEnabled ?? false;
    data.paintBrushMaskInput = data.paintBrushMaskInput ?? DEFAULT_PAINT_MASK_BLOCK;
    // Height Brush Specific
    data.heightBrushAmount = data.heightBrushAmount ?? DEFAULT_HEIGHT_BRUSH_AMOUNT;
    data.heightBrushMode = data.heightBrushMode ?? "Raise";
    // Foliage Brush Specific
    data.foliageBrushRadius = data.foliageBrushRadius ?? DEFAULT_FOLIAGE_RADIUS;
    data.foliageBrushDensity = data.foliageBrushDensity ?? DEFAULT_FOLIAGE_DENSITY;
    data.foliageBrushPlantSource = data.foliageBrushPlantSource ?? DEFAULT_FOLIAGE_SOURCE;
    data.foliageBrushPresetType = data.foliageBrushPresetType ?? DEFAULT_FOLIAGE_PRESET;
    data.foliageBrushCustomInput = data.foliageBrushCustomInput ?? DEFAULT_FOLIAGE_CUSTOM;
    // Line Tool Specific
    data.linePosA = data.linePosA ?? undefined;
    data.lineState = data.lineState ?? 0;
    data.lineMaterialSource = data.lineMaterialSource ?? 'manual';
    data.lineManualMaterial = data.lineManualMaterial ?? DEFAULT_BRUSH_MATERIAL;
    // Path Tool Defaults
    data.pathPoints = data.pathPoints ?? [];
    data.pathMaterialSource = data.pathMaterialSource ?? 'manual';
    data.pathManualMaterial = data.pathManualMaterial ?? DEFAULT_PATH_MATERIAL;
    data.pathCurveType = data.pathCurveType ?? DEFAULT_PATH_CURVE_TYPE;
    data.pathSmoothness = data.pathSmoothness ?? DEFAULT_PATH_SMOOTHNESS;
    data.pathWidth = data.pathWidth ?? DEFAULT_PATH_WIDTH;
    data.pathClosed = data.pathClosed ?? false;
    // Terrain Generation Defaults
    data.terrainBaseMaterialSource = data.terrainBaseMaterialSource ?? 'manual';
    data.terrainManualBaseMaterial = data.terrainManualBaseMaterial ?? DEFAULT_TERRAIN_BASE_MATERIAL;
    data.terrainSurfaceMaterialSource = data.terrainSurfaceMaterialSource ?? 'manual';
    data.terrainManualSurfaceMaterial = data.terrainManualSurfaceMaterial ?? DEFAULT_TERRAIN_SURFACE_MATERIAL;
    data.terrainSurfaceDepth = data.terrainSurfaceDepth ?? DEFAULT_TERRAIN_SURFACE_DEPTH;
    data.terrainNoiseType = data.terrainNoiseType ?? DEFAULT_TERRAIN_NOISE_TYPE;
    data.terrainFrequency = data.terrainFrequency ?? DEFAULT_TERRAIN_FREQUENCY;
    data.terrainAmplitude = data.terrainAmplitude ?? DEFAULT_TERRAIN_AMPLITUDE;
    data.terrainSeedType = data.terrainSeedType ?? DEFAULT_TERRAIN_SEED_TYPE;
    data.terrainCustomSeed = data.terrainCustomSeed ?? undefined;
    data.terrainOctaves = data.terrainOctaves ?? DEFAULT_TERRAIN_OCTAVES;
    data.terrainLacunarity = data.terrainLacunarity ?? DEFAULT_TERRAIN_LACUNARITY;
    data.terrainPersistence = data.terrainPersistence ?? DEFAULT_TERRAIN_PERSISTENCE;
    data.terrainDomainWarpEnabled = data.terrainDomainWarpEnabled ?? DEFAULT_TERRAIN_WARP_ENABLED;
    data.terrainWarpAmplitude = data.terrainWarpAmplitude ?? DEFAULT_TERRAIN_WARP_AMPLITUDE;
    data.terrainWarpFrequency = data.terrainWarpFrequency ?? DEFAULT_TERRAIN_WARP_FREQUENCY;
    data.terrainJitter = data.terrainJitter ?? DEFAULT_TERRAIN_JITTER;
    // Player Specific Settings Defaults
    data.playerSpeedMultiplier = data.playerSpeedMultiplier ?? DEFAULT_PLAYER_SPEED_MULTIPLIER;
    data.playerVerticalSpeedMultiplier = data.playerVerticalSpeedMultiplier ?? DEFAULT_PLAYER_VERTICAL_SPEED_MULTIPLIER;
    data.autoSpectatorEnabled = data.autoSpectatorEnabled ?? DEFAULT_AUTO_SPECTATOR_ENABLED;
    data.wasAutoSwitchedToSpectator = data.wasAutoSwitchedToSpectator ?? false;
    data.autoSpectatorCooldownTicks = data.autoSpectatorCooldownTicks ?? 0;
    data.extendedReachEnabled = data.extendedReachEnabled ?? DEFAULT_EXTENDED_REACH_ENABLED;
    data.showActionFeedbackMessages = data.showActionFeedbackMessages ?? DEFAULT_SHOW_ACTION_FEEDBACK_MESSAGES; // Initialize new setting

    return data;
}

/**
 * Sets the PlayerSelection object for a player.
 * @param {Player} player The player entity.
 * @param {PlayerSelection} selectionData The selection data object to set.
 */
export function setPlayerSelection(player, selectionData) {
    setPlayerMapData(playerSelectionMap, player, selectionData);
}

/**
 * Gets the name of the saved palette currently considered "active" (i.e., loaded into the Active Palette).
 * @param {Player} player The player entity.
 * @returns {string | undefined} The name of the active saved palette, or undefined if none is active.
 */
export function getActiveSavedPaletteName(player) {
    const selectionData = getPlayerSelection(player);
    return selectionData.activeSavedPaletteName;
}

/**
 * Sets the active saved palette name for the player. This implies that this saved palette's data
 * would typically be loaded into the activePaletteData.
 * @param {Player} player The player entity.
 * @param {string | undefined} paletteName The name of the saved palette to set as active, or undefined to clear.
 */
export function setActiveSavedPaletteName(player, paletteName) {
    const selectionData = getPlayerSelection(player);
    selectionData.activeSavedPaletteName = paletteName;
    setPlayerSelection(player, selectionData);
}

/**
 * Clears the Active Palette data and its dirty flag.
 * @param {Player} player The player entity.
 * @param {boolean} [showMessage=true] Whether to send a confirmation message to the player.
 */
export function discardActivePaletteChanges(player, showMessage = true) {
    const selectionData = getPlayerSelection(player);
    if (!selectionData.activePaletteData && !selectionData.activePaletteIsDirty) {
        if (showMessage) player.sendMessage("§eNo Active Palette changes or data to discard.");
        return;
    }
    selectionData.activePaletteData = undefined;
    selectionData.activePaletteIsDirty = false;
    setPlayerSelection(player, selectionData);
    if (showMessage) player.sendMessage("§aActive Palette changes discarded.");
}

/**
 * Gets the PlayerClipboard object for a player.
 * @param {Player | string} playerOrId The player object or their ID string.
 * @returns {PlayerClipboard | undefined} The player's clipboard data, or undefined if none exists.
 */
export function getPlayerClipboard(playerOrId) {
    return getPlayerMapData(playerClipboardMap, playerOrId);
}

/**
 * Sets the PlayerClipboard object for a player.
 * @param {Player} player The player entity.
 * @param {PlayerClipboard | undefined} clipboardData The clipboard data object to set, or undefined to clear.
 */
export function setPlayerClipboard(player, clipboardData) {
    setPlayerMapData(playerClipboardMap, player, clipboardData);
}

/**
 * Gets the Undo stack (HistoryState[]) for a player.
 * @param {Player} player The player entity.
 * @returns {HistoryState[]} The player's undo stack (array of history states).
 */
export function getPlayerUndoStack(player) {
    return getPlayerMapData(playerUndoMap, player, []);
}

/**
 * Sets the Undo stack (HistoryState[]) for a player.
 * @param {Player} player The player entity.
 * @param {HistoryState[]} undoStack The undo stack array to set.
 */
export function setPlayerUndoStack(player, undoStack) {
    setPlayerMapData(playerUndoMap, player, undoStack);
}

/**
 * Gets the Redo stack (HistoryState[]) for a player.
 * @param {Player} player The player entity.
 * @returns {HistoryState[]} The player's redo stack (array of history states).
 */
export function getPlayerRedoStack(player) {
    return getPlayerMapData(playerRedoMap, player, []);
}

/**
 * Sets the Redo stack (HistoryState[]) for a player.
 * @param {Player} player The player entity.
 * @param {HistoryState[]} redoStack The redo stack array to set.
 */
export function setPlayerRedoStack(player, redoStack) {
    setPlayerMapData(playerRedoMap, player, redoStack);
}

// Export the maps explicitly grouped only for the player_leave cleanup listener.
export const PlayerStateMaps = {
    selection: playerSelectionMap,
    clipboard: playerClipboardMap,
    undo: playerUndoMap,
    redo: playerRedoMap
};
