// scripts/ui/brush_menu.js

import { Player } from "@minecraft/server";
import { ModalFormData } from "@minecraft/server-ui";
import {
    getPlayerSelection,
    setPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName // Renamed
} from "../state/index.js";
import {
    // General Brush Limits/Defaults
    MAX_BRUSH_SIZE,
    DEFAULT_BRUSH_MATERIAL,
    DEFAULT_PAINT_MASK_BLOCK,
    // Height Brush Limits/Defaults
    MAX_HEIGHT_BRUSH_AMOUNT,
    DEFAULT_HEIGHT_BRUSH_AMOUNT,
    // Foliage Brush Limits/Defaults & Data
    MAX_FOLIAGE_RADIUS,
    DEFAULT_FOLIAGE_RADIUS,
    DEFAULT_FOLIAGE_DENSITY,
    DEFAULT_FOLIAGE_SOURCE,
    DEFAULT_FOLIAGE_PRESET,
    DEFAULT_FOLIAGE_CUSTOM,
    FOLIAGE_PRESETS
} from "../constants/index.js";
// Import parsing/validation utils needed by forms
import { parseWeightedBlockInput, validateBlockIds } from "../utils/index.js";

// --- Helper Function for Consistent Shape Options ---
function getBrushShapeOptions() {
    const shapeMap = { "Sphere": 'sphere', "Cube": 'cube', "Cylinder": 'cylinder', "Pyramid": 'pyramid', "Octahedron": 'octahedron', "Dome": 'dome', "Bowl": 'bowl', "Cone (Point Up)": 'cone_up', "Cone (Point Down)": 'cone_down', "Disk": 'disk', "Plate": 'plate', "Torus": 'torus' };
    const reverseShapeMap = Object.fromEntries(Object.entries(shapeMap).map(([key, value]) => [value, key]));
    const shapeOptions = Object.keys(shapeMap);
    return { shapeOptions, shapeMap, reverseShapeMap };
}

// --- Helper Function for Dynamic Source Dropdown ---
/**
 * Builds the options, mappings, and default index for the Material/Plant Source dropdown.
 * @param {Player} player The player object.
 * @param {string | undefined} currentSourceValue The current source value from state ('manual', 'active', 'preset', 'custom', or a palette name).
 * @param {boolean} includePresets Should the "Preset" option be included (for Foliage Brush)?
 * @param {boolean} includeCustom Should the "Custom Input" option be included (for Foliage Brush)? Default true.
 * @returns {{sourceOptions: string[], sourceValueMap: object, reverseSourceMap: object, currentSourceIndex: number}}
 */
export function buildSourceDropdownOptions(player, currentSourceValue, includePresets = false, includeCustom = true) {
    const selectionData = getPlayerSelection(player);
    const palettes = getPlayerPalettes(player);
    const savedPaletteNames = Object.keys(palettes).sort();

    const sourceOptions = [];
    const sourceValueMap = {};
    const reverseSourceMap = {};

    if (includeCustom) {
        sourceOptions.push("Manual Input");
        sourceValueMap["Manual Input"] = 'manual';
        reverseSourceMap['manual'] = "Manual Input";
    }
    if (includePresets) {
        sourceOptions.push("Preset");
        sourceValueMap["Preset"] = 'preset';
        reverseSourceMap['preset'] = "Preset";
    }

    // Add Active Palette option
    const activePaletteData = selectionData.activePaletteData; // Renamed
    const activeStatus = activePaletteData ? ` (${activePaletteData.length} items${selectionData.activePaletteIsDirty ? ' §c*' : ''})` : ' (Empty)'; // Renamed
    const activeOptionLabel = `Active Palette${activeStatus}`; // Renamed
    sourceOptions.push(activeOptionLabel);
    sourceValueMap[activeOptionLabel] = 'active'; // Maps to 'active' state value // Renamed
    reverseSourceMap['active'] = activeOptionLabel; // Renamed

    savedPaletteNames.forEach(name => {
        const label = `Saved: ${name}`;
        sourceOptions.push(label);
        sourceValueMap[label] = name;
        reverseSourceMap[name] = label;
    });

    let defaultStateValue = includeCustom ? 'manual' : (includePresets ? 'preset' : 'active'); // Renamed default
    let currentSourceIndex = sourceOptions.indexOf(reverseSourceMap[defaultStateValue]);
    if (currentSourceIndex < 0) currentSourceIndex = 0;

    const currentSourceLabel = reverseSourceMap[currentSourceValue];
    if (currentSourceLabel && sourceOptions.includes(currentSourceLabel)) {
        currentSourceIndex = sourceOptions.indexOf(currentSourceLabel);
    } else if (currentSourceValue && !['manual', 'active', 'preset', 'custom'].includes(currentSourceValue)) { // Renamed
        console.warn(`Source '${currentSourceValue}' not found in options, defaulting.`);
    }

    if (currentSourceValue === 'active') { // Renamed
        currentSourceIndex = sourceOptions.findIndex(opt => opt.startsWith('Active Palette')); // Renamed
        if (currentSourceIndex < 0) currentSourceIndex = 0;
    }

    return { sourceOptions, sourceValueMap, currentSourceIndex };
}


// --- Regular Brush Settings Form ---
/**
 * Shows the form to configure REGULAR brush settings.
 * @param {Player} player
 */
export async function showRegularBrushSettingsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentShape = selectionData.brushShape ?? 'sphere';
    const currentSize = selectionData.brushSize;
    const currentHollow = selectionData.brushHollow;
    const currentSourceValue = selectionData.brushMaterialSource ?? 'manual';
    const currentManualMaterial = selectionData.brushManualMaterial;

    const { shapeOptions, shapeMap, reverseShapeMap } = getBrushShapeOptions();
    const currentShapeIndex = shapeOptions.indexOf(reverseShapeMap[currentShape] ?? shapeOptions[0]);

    const { sourceOptions, sourceValueMap, currentSourceIndex } = buildSourceDropdownOptions(player, currentSourceValue);

    const form = new ModalFormData().title("Regular Brush Settings")
        .dropdown("Shape", shapeOptions, currentShapeIndex)
        .slider("Size (Radius/Half-Width)", 1, MAX_BRUSH_SIZE, 1, currentSize)
        .toggle("Hollow Shape", currentHollow)
        .dropdown("Material Source", sourceOptions, currentSourceIndex)
        .textField("Manual Blocks (id:wt,...)", "Used if Source is 'Manual Input'", currentManualMaterial);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eRegular Brush settings unchanged.");
            return;
        }

        let formIndex = 0;
        const formVals = response.formValues;
        const newShapeName = shapeOptions[formVals[formIndex++]];
        const newSize = formVals[formIndex++];
        const newHollow = formVals[formIndex++];
        const selectedOptionLabel = sourceOptions[formVals[formIndex++]];
        const newManualMaterialInput = formVals[formIndex++].trim();

        const newShape = shapeMap[newShapeName] ?? 'sphere';
        const newSourceValue = sourceValueMap[selectedOptionLabel] ?? 'manual';

        let manualInputIsValid = true;
        let finalManualMaterialString = currentManualMaterial;
        if (newSourceValue === 'manual') {
            if (!newManualMaterialInput) { manualInputIsValid = false; player.sendMessage("§cManual Blocks cannot be empty..."); }
            else {
                const parsedResult = parseWeightedBlockInput(newManualMaterialInput);
                if (parsedResult.errors.length > 0 || parsedResult.blocks.length === 0) { manualInputIsValid = false; player.sendMessage("§cError parsing manual blocks..."); }
                else { const { invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id)); if (invalidIds.length > 0) { manualInputIsValid = false; player.sendMessage(`§cInvalid IDs: ${invalidIds.join(',')}`); } else { finalManualMaterialString = newManualMaterialInput; } }
            }
            if (!manualInputIsValid) { player.sendMessage("§eRegular Brush settings NOT updated due to invalid manual input."); return; }
        }

        if (newSourceValue === 'active' && (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0)) { // Renamed
            player.sendMessage("§cActive Palette is empty. Select 'Manual Input' or load/create a palette first. Settings not updated."); return; // Renamed
        }

        if (!['manual', 'active'].includes(newSourceValue)) { // Renamed
            const palettes = getPlayerPalettes(player);
            if (!palettes[newSourceValue]) {
                player.sendMessage(`§cSelected saved palette "${newSourceValue}" no longer exists. Settings not updated.`); return;
            }
        }

        selectionData.brushShape = newShape;
        selectionData.brushSize = newSize;
        selectionData.brushHollow = newHollow;
        selectionData.brushMaterialSource = newSourceValue;
        if (newSourceValue === 'manual') {
            selectionData.brushManualMaterial = finalManualMaterialString;
        }
        setPlayerSelection(player, selectionData);

        const sourceDesc = selectedOptionLabel;
        const hollowDesc = newHollow ? '(Hollow)' : '';
        player.sendMessage(`§aRegular Brush updated: ${newShapeName}${hollowDesc}, Size ${newSize}, Src ${sourceDesc}`);

    } catch (e) {
        console.error(`[Regular Brush Settings Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing regular brush settings form.");
    }
}


// --- Paint Brush Settings Form ---
/**
 * Shows the form to configure PAINT brush settings.
 * @param {Player} player
 */
export async function showPaintBrushSettingsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentShape = selectionData.brushShape ?? 'sphere';
    const currentSize = selectionData.brushSize;
    const currentSourceValue = selectionData.brushMaterialSource ?? 'manual';
    const currentManualMaterial = selectionData.brushManualMaterial;
    const currentMaskEnabled = selectionData.paintBrushMaskEnabled;
    const currentMaskInput = selectionData.paintBrushMaskInput;

    const { shapeOptions, shapeMap, reverseShapeMap } = getBrushShapeOptions();
    const currentShapeIndex = shapeOptions.indexOf(reverseShapeMap[currentShape] ?? shapeOptions[0]);

    const { sourceOptions, sourceValueMap, currentSourceIndex } = buildSourceDropdownOptions(player, currentSourceValue);

    const form = new ModalFormData().title("Paint Brush Settings")
        .dropdown("Shape", shapeOptions, currentShapeIndex)
        .slider("Size (Radius/Half-Width)", 1, MAX_BRUSH_SIZE, 1, currentSize)
        .dropdown("Paint Material Source", sourceOptions, currentSourceIndex)
        .textField("Manual Paint Blocks (id:wt,...)", "Used if Source is 'Manual Input'", currentManualMaterial)
        .toggle("Enable Masking", currentMaskEnabled)
        .textField("Mask Blocks (IDs, comma-sep)", "e.g. stone,dirt,grass", currentMaskInput);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§ePaint Brush settings unchanged.");
            return;
        }

        let formIndex = 0;
        const formVals = response.formValues;
        const newShapeName = shapeOptions[formVals[formIndex++]];
        const newSize = formVals[formIndex++];
        const selectedOptionLabel = sourceOptions[formVals[formIndex++]];
        const newManualMaterialInput = formVals[formIndex++].trim();
        const newMaskEnabled = formVals[formIndex++];
        const newMaskInput = formVals[formIndex++].trim();

        const newShape = shapeMap[newShapeName] ?? 'sphere';
        const newSourceValue = sourceValueMap[selectedOptionLabel] ?? 'manual';

        let manualInputIsValid = true;
        let finalManualMaterialString = currentManualMaterial;
        if (newSourceValue === 'manual') {
            if (!newManualMaterialInput) { manualInputIsValid = false; player.sendMessage("§cManual Paint Blocks cannot be empty..."); }
            else {
                const parsedResult = parseWeightedBlockInput(newManualMaterialInput);
                if (parsedResult.errors.length > 0 || parsedResult.blocks.length === 0) { manualInputIsValid = false; player.sendMessage("§cError parsing manual blocks..."); }
                else { const { invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id)); if (invalidIds.length > 0) { manualInputIsValid = false; player.sendMessage(`§cInvalid IDs: ${invalidIds.join(',')}`); } else { finalManualMaterialString = newManualMaterialInput; } }
            }
            if (!manualInputIsValid) { player.sendMessage("§ePaint Brush settings NOT updated due to invalid manual input."); return; }
        }

        if (newSourceValue === 'active' && (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0)) { // Renamed
            player.sendMessage("§cActive Palette is empty. Select 'Manual Input' or load/create a palette first. Settings not updated."); return; // Renamed
        }

        if (!['manual', 'active'].includes(newSourceValue)) { // Renamed
            const palettes = getPlayerPalettes(player);
            if (!palettes[newSourceValue]) {
                player.sendMessage(`§cSelected saved palette "${newSourceValue}" no longer exists. Settings not updated.`); return;
            }
        }

        let maskInputIsValid = true;
        let finalMaskInputString = currentMaskInput;
        if (newMaskEnabled) {
            if (!newMaskInput) { maskInputIsValid = false; player.sendMessage("§cMask Blocks cannot be empty..."); }
            else {
                const parsedMask = parseWeightedBlockInput(newMaskInput);
                if (parsedMask.blocks.length === 0) { maskInputIsValid = false; player.sendMessage("§cNo valid blocks found in Mask Blocks..."); }
                else { const { invalidIds } = validateBlockIds(parsedMask.blocks.map(b => b.id)); if (invalidIds.length > 0) { maskInputIsValid = false; player.sendMessage(`§cInvalid Mask IDs: ${invalidIds.join(', ')}...`); } else { finalMaskInputString = newMaskInput; } }
            }
            if (!maskInputIsValid) { player.sendMessage("§ePaint Brush settings NOT updated due to invalid mask input."); return; }
        }

        selectionData.brushShape = newShape;
        selectionData.brushSize = newSize;
        selectionData.brushMaterialSource = newSourceValue;
        if (newSourceValue === 'manual') {
            selectionData.brushManualMaterial = finalManualMaterialString;
        }
        selectionData.paintBrushMaskEnabled = newMaskEnabled;
        if (newMaskEnabled) {
            selectionData.paintBrushMaskInput = finalMaskInputString;
        }
        setPlayerSelection(player, selectionData);

        const paintSourceDesc = selectedOptionLabel;
        const maskDesc = newMaskEnabled ? `ON` : 'OFF';
        player.sendMessage(`§aPaint Brush updated: ${newShapeName}, Size ${newSize}, Paint ${paintSourceDesc}, Mask ${maskDesc}`);

    } catch (e) {
        console.error(`[Paint Brush Settings Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing paint brush settings form.");
    }
}


// --- Height Brush Settings Form ---
export async function showHeightBrushSettingsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentShape = selectionData.brushShape ?? 'sphere';
    const currentSize = selectionData.brushSize;
    const currentAmount = selectionData.heightBrushAmount ?? DEFAULT_HEIGHT_BRUSH_AMOUNT;
    const currentMode = selectionData.heightBrushMode ?? "Raise";

    const { shapeOptions, shapeMap, reverseShapeMap } = getBrushShapeOptions();
    const currentShapeIndex = shapeOptions.indexOf(reverseShapeMap[currentShape] ?? shapeOptions[0]);
    const modeOptions = ["Raise", "Lower"];
    const currentModeIndex = modeOptions.indexOf(currentMode);

    const form = new ModalFormData().title("Height Brush Settings")
        .dropdown("Shape", shapeOptions, currentShapeIndex)
        .slider("Size (Radius/Half-Width)", 1, MAX_BRUSH_SIZE, 1, currentSize)
        .slider("Amount", 1, MAX_HEIGHT_BRUSH_AMOUNT, 1, currentAmount)
        .dropdown("Mode", modeOptions, Math.max(0, currentModeIndex));

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) { player.sendMessage("§eHeight Brush settings unchanged."); return; }

        let formIndex = 0;
        const formVals = response.formValues;
        const newShapeName = shapeOptions[formVals[formIndex++]];
        const newSize = formVals[formIndex++];
        const newAmount = formVals[formIndex++];
        const newModeName = modeOptions[formVals[formIndex++]];
        const newShape = shapeMap[newShapeName] ?? 'sphere';

        selectionData.brushShape = newShape;
        selectionData.brushSize = newSize;
        selectionData.heightBrushAmount = newAmount;
        selectionData.heightBrushMode = newModeName;
        setPlayerSelection(player, selectionData);

        player.sendMessage(`§aHeight Brush updated: ${newShapeName}, Size ${newSize}, Amt ${newAmount}, Mode ${newModeName}`);
    } catch (e) { console.error(`[Height Brush Settings Form] Error: ${e}\n${e.stack}`); player.sendMessage("§cError processing height brush settings form."); }
}


// --- Foliage Brush Settings Form ---
/**
 * Shows the form to configure Foliage brush settings.
 * @param {Player} player
 */
export async function showFoliageBrushSettingsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentRadius = selectionData.foliageBrushRadius ?? DEFAULT_FOLIAGE_RADIUS;
    const currentDensity = selectionData.foliageBrushDensity ?? DEFAULT_FOLIAGE_DENSITY;
    const currentSourceValue = selectionData.foliageBrushPlantSource ?? DEFAULT_FOLIAGE_SOURCE;
    const currentPreset = selectionData.foliageBrushPresetType ?? DEFAULT_FOLIAGE_PRESET;
    const currentCustom = selectionData.foliageBrushCustomInput ?? DEFAULT_FOLIAGE_CUSTOM;

    const { sourceOptions, sourceValueMap, currentSourceIndex } = buildSourceDropdownOptions(player, currentSourceValue, true, true);

    const presetOptions = Object.keys(FOLIAGE_PRESETS);
    let currentPresetIndex = presetOptions.indexOf(currentPreset);
    if (currentPresetIndex === -1 || presetOptions.length === 0) { currentPresetIndex = 0; }

    const form = new ModalFormData().title("Foliage Brush Settings")
        .slider("Scatter Radius", 1, MAX_FOLIAGE_RADIUS, 1, currentRadius)
        .slider("Density (%)", 0, 100, 1, currentDensity)
        .dropdown("Plant Source", sourceOptions, currentSourceIndex)
        .dropdown("Preset Type", presetOptions.length > 0 ? presetOptions : ["No Presets Defined"], currentPresetIndex)
        .textField("Custom Plants (id:wt,...)", "Used if Source is 'Custom Input'", currentCustom);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eFoliage Brush settings unchanged.");
            return;
        }

        const formVals = response.formValues;
        const newRadius = formVals[0];
        const newDensity = formVals[1];
        const selectedOptionLabel = sourceOptions[formVals[2]];
        const newPresetName = presetOptions.length > 0 ? presetOptions[formVals[3]] : DEFAULT_FOLIAGE_PRESET;
        const newCustomInput = formVals[4].trim();

        const newSourceValue = sourceValueMap[selectedOptionLabel] ?? DEFAULT_FOLIAGE_SOURCE;

        let customInputIsValid = true;
        let finalCustomInputString = currentCustom;
        if (newSourceValue === 'manual') { // Changed 'custom' to 'manual'
            if (!newCustomInput) { customInputIsValid = false; player.sendMessage("§cCustom Plants cannot be empty..."); }
            else {
                const parsedResult = parseWeightedBlockInput(newCustomInput);
                if (parsedResult.errors.length > 0 || parsedResult.blocks.length === 0) { customInputIsValid = false; player.sendMessage("§cError parsing custom blocks..."); }
                else { const { invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id)); if (invalidIds.length > 0) { customInputIsValid = false; player.sendMessage(`§cInvalid IDs: ${invalidIds.join(',')}`); } else { finalCustomInputString = newCustomInput; } }
            }
            if (!customInputIsValid) { player.sendMessage("§cFoliage Brush settings NOT updated due to invalid manual input."); return; }
        }

        if (newSourceValue === 'active' && (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0)) { // Renamed
            player.sendMessage("§cActive Palette is empty. Select another source or load/create a palette first. Settings not updated."); return; // Renamed
        }

        if (!['preset', 'manual', 'active'].includes(newSourceValue)) { // Changed 'custom' to 'manual'
            const palettes = getPlayerPalettes(player);
            if (!palettes[newSourceValue]) {
                player.sendMessage(`§cSelected saved palette "${newSourceValue}" no longer exists. Settings not updated.`); return;
            }
        }

        if (newSourceValue === 'preset' && presetOptions.length === 0) {
            player.sendMessage("§cCannot select 'Preset' source as no presets are defined. Settings not updated."); return;
        }

        selectionData.foliageBrushRadius = newRadius;
        selectionData.foliageBrushDensity = newDensity;
        selectionData.foliageBrushPlantSource = newSourceValue;
        selectionData.foliageBrushPresetType = newPresetName;
        if (newSourceValue === 'manual') { // Changed 'custom' to 'manual'
            selectionData.foliageBrushCustomInput = finalCustomInputString;
        }
        setPlayerSelection(player, selectionData);

        const sourceDesc = selectedOptionLabel;
        player.sendMessage(`§aFoliage Brush updated: Radius=${newRadius}, Density=${newDensity}%, Source=${sourceDesc}`);

    } catch (e) {
        console.error(`[Foliage Brush Settings Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing foliage brush settings form.");
    }
}
