// scripts/ui/fill_menu.js

import { Player, system } from "@minecraft/server";
import { ActionFormData, ModalFormData } from "@minecraft/server-ui";
import { fillSelection } from "../core/index.js";
import { showEditMenu } from "./main_menu.js";
import {
    getPlayerSelection,
    getPlayerPalettes,
    addPlayerPalette // For saving manual input
} from "../state/index.js";
import {
    parseWeightedBlockInput,
    validateBlockIds
} from "../utils/index.js";
import { buildSourceDropdownOptions } from "./brush_menu.js"; // Re-use for source dropdown
import { DEFAULT_BRUSH_MATERIAL } from "../constants/index.js"; // For default material
import { promptSavePalette } from "./options_menu.js"; // For palette saving prompt

/**
 * Shows the submenu for selecting the type of fill operation.
 * This function now also handles getting the material input before calling the core fillSelection.
 * @param {Player} player The player initiating the action.
 */
export async function showFillOptionsMenu(player) {
    const fillTypeForm = new ActionFormData()
        .title("Fill Options")
        .body("Select the type of fill to perform on the selection.")
        .button("Fill (Solid)", "textures/ui/icon_recipe_construction")
        .button("Fill (Hollow)", "textures/ui/icon_recipe_item")
        .button("Fill (Frame/Edges)", "textures/ui/wireframe")
        .button("Fill (Walls)", "textures/ui/brick")
        .button("Back", "textures/ui/cancel");

    try {
        const fillTypeResponse = await fillTypeForm.show(player);

        if (fillTypeResponse.canceled || fillTypeResponse.selection === undefined || fillTypeResponse.selection === 4) {
            system.run(async () => { if (player.isValid()) await showEditMenu(player); });
            return;
        }

        const fillTypes = ["Solid", "Hollow", "Frame", "Walls"];
        const selectedFillType = fillTypes[fillTypeResponse.selection];

        // Now, get material input
        const selectionData = getPlayerSelection(player);
        // Determine preferred source: active, then active saved, then manual
        let preferredSourceValue = 'manual';
        if (selectionData.activePaletteData && selectionData.activePaletteData.length > 0) {
            preferredSourceValue = 'active';
        } else if (selectionData.activeSavedPaletteName) {
            preferredSourceValue = selectionData.activeSavedPaletteName;
        }

        const { sourceOptions, sourceValueMap, currentSourceIndex } = buildSourceDropdownOptions(
            player, preferredSourceValue, false, true // No presets, include manual/active/saved
        );

        let initialBlockText = DEFAULT_BRUSH_MATERIAL; // Fallback
        if (preferredSourceValue === 'active' && selectionData.activePaletteData) {
            initialBlockText = selectionData.activePaletteData.map(b => `${b.id.replace('minecraft:', '')}${b.weight > 1 ? ':' + b.weight : ''}`).join(', ');
        } else if (preferredSourceValue !== 'manual' && preferredSourceValue !== 'active') {
            const palettes = getPlayerPalettes(player);
            const paletteData = palettes[preferredSourceValue];
            if (paletteData) {
                initialBlockText = paletteData.map(b => `${b.id.replace('minecraft:', '')}${b.weight > 1 ? ':' + b.weight : ''}`).join(', ');
            }
        }

        const materialForm = new ModalFormData().title(`Fill Selection (${selectedFillType})`);
        materialForm.dropdown("Block Source", sourceOptions, currentSourceIndex); // 0
        materialForm.textField("Block IDs / Palette Data", "e.g., stone / stone:5, dirt:3", initialBlockText); // 1
        materialForm.toggle("Save Manual Input as New Palette?", false); // 2

        const materialResponse = await materialForm.show(player);

        if (materialResponse.canceled || !materialResponse.formValues) {
            player.sendMessage("§eFill cancelled.");
            system.run(async () => { if (player.isValid()) await showEditMenu(player); });
            return;
        }

        const formVals = materialResponse.formValues;
        const selectedOptionLabel = sourceOptions[formVals[0]];
        const manualInputText = formVals[1]?.toString().trim() ?? "";
        const savePaletteRequested = formVals[2] ?? false;
        const finalSourceValue = sourceValueMap[selectedOptionLabel] ?? 'manual';

        let blocksToUseWeighted = [];
        let rawBlockInputStringsForSave = []; // For saving palette from manual input
        let sourceDesc = selectedOptionLabel; // User-friendly description
        let finalSourceWasManual = false;

        switch (finalSourceValue) {
            case "active": // Changed from 'working'
                if (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0) { // Changed
                    player.sendMessage("§cActive Palette is currently empty. Cannot fill."); // Changed
                    system.run(async () => { if (player.isValid()) await showEditMenu(player); });
                    return;
                }
                blocksToUseWeighted = selectionData.activePaletteData; // Changed
                sourceDesc = "Active Palette"; // Changed
                break;
            case "manual":
                sourceDesc = "manual input";
                finalSourceWasManual = true;
                if (!manualInputText) {
                    player.sendMessage("§cBlock input cannot be empty for manual source.");
                    system.run(async () => { if (player.isValid()) await showEditMenu(player); });
                    return;
                }
                const parsedResult = parseWeightedBlockInput(manualInputText);
                if (parsedResult.errors.length > 0) {
                    player.sendMessage(`§cInput errors:\n${parsedResult.errors.join('\n')}`);
                    system.run(async () => { if (player.isValid()) await showEditMenu(player); });
                    return;
                }
                if (parsedResult.blocks.length === 0) {
                    player.sendMessage("§cNo valid blocks parsed from manual input. Cannot fill.");
                    system.run(async () => { if (player.isValid()) await showEditMenu(player); });
                    return;
                }
                blocksToUseWeighted = parsedResult.blocks;
                rawBlockInputStringsForSave = blocksToUseWeighted.map(b => b.id); // Store original IDs for saving
                break;
            default: // Saved palette name
                const palettes = getPlayerPalettes(player);
                const paletteData = palettes[finalSourceValue];
                if (!paletteData || paletteData.length === 0) {
                    player.sendMessage(`§cSaved palette "${finalSourceValue}" is empty or invalid. Cannot fill.`);
                    system.run(async () => { if (player.isValid()) await showEditMenu(player); });
                    return;
                }
                blocksToUseWeighted = paletteData;
                sourceDesc = `palette "${finalSourceValue}"`;
                break;
        }

        // Call core logic
        await fillSelection(player, selectedFillType, blocksToUseWeighted, sourceDesc);

        if (finalSourceWasManual && savePaletteRequested && rawBlockInputStringsForSave.length > 0) {
            // Use system.run to ensure promptSavePalette (which shows a form) is called safely
            system.run(async () => {
                if (player.isValid()) {
                    // Pass the original block strings for saving, not the weighted objects
                    await promptSavePalette(player, rawBlockInputStringsForSave);
                }
            });
        }

        // Return to main edit menu after fill attempt
        system.run(async () => { if (player.isValid()) await showEditMenu(player); });

    } catch (e) {
        console.error(`[Fill Options Menu] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError displaying fill options menu.");
        system.run(async () => { if (player.isValid()) await showEditMenu(player); });
    }
}
