// scripts/ui/global_settings_menu.js

import { Player, system, GameMode, EquipmentSlot } from "@minecraft/server";
import { ModalFormData, ActionFormData } from "@minecraft/server-ui";
import {
    getGlobalSettings,
    saveGlobalSettings,
} from "../state/global_state.js";
import {
    getPlayerSelection,
    setPlayerSelection
} from "../state/player_state.js";
import {
    MAX_UNDO_HISTORY,
    MAX_BRUSH_SIZE,
    DEFAULT_PLAYER_SPEED_MULTIPLIER,
    MIN_PLAYER_SPEED_MULTIPLIER,
    MAX_PLAYER_SPEED_MULTIPLIER,
    DEFAULT_PLAYER_VERTICAL_SPEED_MULTIPLIER,
    MIN_PLAYER_VERTICAL_SPEED_MULTIPLIER,
    MAX_PLAYER_VERTICAL_SPEED_MULTIPLIER,
    DEFAULT_AUTO_SPECTATOR_ENABLED,
    DEFAULT_EXTENDED_REACH_ENABLED,
    DEFAULT_SHOW_ACTION_FEEDBACK_MESSAGES // Import new default
} from "../constants/settings.js";

/**
 * Shows the main hub for settings, allowing navigation to World or Player settings.
 * @param {Player} player The player opening the menu.
 */
export async function showMainSettingsHubForm(player) {
    const form = new ActionFormData()
        .title("Addon Settings")
        .body("Choose the type of settings you want to configure.")
        .button("World Settings", "textures/ui/world_glyph_color_2x")
        .button("Player Settings", "textures/ui/icon_steve")
        .button("Cancel", "textures/ui/cancel");

    try {
        const response = await form.show(player);
        if (response.canceled || response.selection === undefined) {
            return;
        }

        switch (response.selection) {
            case 0: // World Settings
                await showWorldSettingsForm(player);
                break;
            case 1: // Player Settings
                await showPlayerSettingsForm(player);
                break;
            case 2: // Cancel
                return;
        }
    } catch (e) {
        console.error(`[Main Settings Hub Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError opening settings hub.");
    }
}


/**
 * Shows the ModalForm to configure global addon (World) settings.
 * @param {Player} player The player opening the menu.
 */
async function showWorldSettingsForm(player) {
    const currentSettings = getGlobalSettings();

    const form = new ModalFormData()
        .title("World Settings (Global)")
        .toggle("Show Action Bar Info", currentSettings.showActionBarInfo) // 0
        .textField("Outline Particle ID", "e.g., minecraft:basic_flame_particle", currentSettings.outlineParticleId) // 1
        .slider("Max Undo Steps (per player)", 0, 100, 1, currentSettings.maxUndoHistory) // 2
        .slider("Default Brush Size", 1, MAX_BRUSH_SIZE, 1, currentSettings.defaultBrushSize); // 3

    try {
        const response = await form.show(player);

        if (response.canceled || !response.formValues) {
            player.sendMessage("§eWorld settings unchanged.");
            system.run(() => showMainSettingsHubForm(player));
            return;
        }

        const formVals = response.formValues;
        const newSettings = { ...currentSettings };
        let validationErrors = [];

        newSettings.showActionBarInfo = formVals[0];

        const newParticleId = formVals[1].trim();
        if (!newParticleId) {
            validationErrors.push("Outline Particle ID cannot be empty.");
        } else {
            newSettings.outlineParticleId = newParticleId.includes(':') ? newParticleId : `minecraft:${newParticleId}`;
        }

        const newMaxUndo = formVals[2];
        if (typeof newMaxUndo === 'number' && Number.isInteger(newMaxUndo)) {
            newSettings.maxUndoHistory = Math.max(0, Math.min(1000, newMaxUndo));
        } else {
            validationErrors.push("Invalid value for Max Undo Steps.");
        }

        const newDefaultBrushSize = formVals[3];
        if (typeof newDefaultBrushSize === 'number' && Number.isInteger(newDefaultBrushSize)) {
            newSettings.defaultBrushSize = Math.max(1, Math.min(MAX_BRUSH_SIZE, newDefaultBrushSize));
        } else {
            validationErrors.push("Invalid value for Default Brush Size.");
        }

        if (validationErrors.length > 0) {
            player.sendMessage("§cInvalid input:\n" + validationErrors.join("\n") + "\n§eWorld settings not saved.");
        } else {
            if (saveGlobalSettings(newSettings)) {
                player.sendMessage("§aGlobal World settings saved successfully.");
            } else {
                player.sendMessage("§cFailed to save global World settings. Check server logs.");
            }
        }
        system.run(() => showMainSettingsHubForm(player));

    } catch (e) {
        console.error(`[World Settings Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError displaying or processing World settings form.");
        system.run(() => showMainSettingsHubForm(player));
    }
}

/**
 * Shows the ModalForm to configure player-specific settings.
 * @param {Player} player The player opening the menu.
 */
async function showPlayerSettingsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentPlayerSpeedMultiplier = selectionData.playerSpeedMultiplier ?? DEFAULT_PLAYER_SPEED_MULTIPLIER;
    const currentPlayerVerticalSpeedMultiplier = selectionData.playerVerticalSpeedMultiplier ?? DEFAULT_PLAYER_VERTICAL_SPEED_MULTIPLIER;
    const currentAutoSpectatorEnabled = selectionData.autoSpectatorEnabled ?? DEFAULT_AUTO_SPECTATOR_ENABLED;
    const currentExtendedReachEnabled = selectionData.extendedReachEnabled ?? DEFAULT_EXTENDED_REACH_ENABLED;
    const currentShowActionFeedbackMessages = selectionData.showActionFeedbackMessages ?? DEFAULT_SHOW_ACTION_FEEDBACK_MESSAGES; // Get current value

    const form = new ModalFormData()
        .title("Player Settings")
        .slider(
            "Creative Horizontal Speed Multiplier",
            MIN_PLAYER_SPEED_MULTIPLIER,
            MAX_PLAYER_SPEED_MULTIPLIER,
            0.1, // Step
            currentPlayerSpeedMultiplier
        ) // 0
        .slider(
            "Creative Vertical Speed Multiplier (Flying)",
            MIN_PLAYER_VERTICAL_SPEED_MULTIPLIER,
            MAX_PLAYER_VERTICAL_SPEED_MULTIPLIER,
            0.1, // Step
            currentPlayerVerticalSpeedMultiplier
        ) // 1
        .toggle(
            "Auto Spectator Near Walls (Creative Flying)",
            currentAutoSpectatorEnabled
        ) // 2
        .toggle(
            "Enable Extended Reach Placement",
            currentExtendedReachEnabled
        ) // 3
        .toggle( // New toggle for feedback messages
            "Show Action Feedback Messages",
            currentShowActionFeedbackMessages
        ); // 4

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§ePlayer settings unchanged.");
            system.run(() => showMainSettingsHubForm(player));
            return;
        }

        const formVals = response.formValues;
        const newSpeedMultiplier = formVals[0];
        const newVerticalSpeedMultiplier = formVals[1];
        const newAutoSpectatorEnabled = formVals[2];
        const newExtendedReachEnabled = formVals[3];
        const newShowActionFeedbackMessages = formVals[4]; // Get new value

        let settingsChanged = false;
        let messages = [];

        // Horizontal Speed
        if (typeof newSpeedMultiplier === 'number' &&
            newSpeedMultiplier >= MIN_PLAYER_SPEED_MULTIPLIER &&
            newSpeedMultiplier <= MAX_PLAYER_SPEED_MULTIPLIER) {
            if (selectionData.playerSpeedMultiplier !== parseFloat(newSpeedMultiplier.toFixed(2))) {
                selectionData.playerSpeedMultiplier = parseFloat(newSpeedMultiplier.toFixed(2));
                settingsChanged = true;
            }
        } else {
            messages.push(`§cInvalid horizontal speed multiplier. Must be between ${MIN_PLAYER_SPEED_MULTIPLIER} and ${MAX_PLAYER_SPEED_MULTIPLIER}. Horizontal speed not changed.`);
        }

        // Vertical Speed
        if (typeof newVerticalSpeedMultiplier === 'number' &&
            newVerticalSpeedMultiplier >= MIN_PLAYER_VERTICAL_SPEED_MULTIPLIER &&
            newVerticalSpeedMultiplier <= MAX_PLAYER_VERTICAL_SPEED_MULTIPLIER) {
            if (selectionData.playerVerticalSpeedMultiplier !== parseFloat(newVerticalSpeedMultiplier.toFixed(2))) {
                selectionData.playerVerticalSpeedMultiplier = parseFloat(newVerticalSpeedMultiplier.toFixed(2));
                settingsChanged = true;
            }
        } else {
            messages.push(`§cInvalid vertical speed multiplier. Must be between ${MIN_PLAYER_VERTICAL_SPEED_MULTIPLIER} and ${MAX_PLAYER_VERTICAL_SPEED_MULTIPLIER}. Vertical speed not changed.`);
        }

        // Auto Spectator Toggle
        if (typeof newAutoSpectatorEnabled === 'boolean') {
            if (selectionData.autoSpectatorEnabled !== newAutoSpectatorEnabled) {
                selectionData.autoSpectatorEnabled = newAutoSpectatorEnabled;
                settingsChanged = true;
            }
        } else {
            messages.push("§cInvalid value for Auto Spectator toggle. Setting not changed.");
        }

        // Extended Reach Toggle
        if (typeof newExtendedReachEnabled === 'boolean') {
            if (selectionData.extendedReachEnabled !== newExtendedReachEnabled) {
                selectionData.extendedReachEnabled = newExtendedReachEnabled;
                settingsChanged = true;
            }
        } else {
            messages.push("§cInvalid value for Extended Reach toggle. Setting not changed.");
        }

        // Show Action Feedback Messages Toggle
        if (typeof newShowActionFeedbackMessages === 'boolean') {
            if (selectionData.showActionFeedbackMessages !== newShowActionFeedbackMessages) {
                selectionData.showActionFeedbackMessages = newShowActionFeedbackMessages;
                settingsChanged = true;
            }
        } else {
            messages.push("§cInvalid value for Show Action Feedback Messages toggle. Setting not changed.");
        }


        if (settingsChanged) {
            setPlayerSelection(player, selectionData);
            messages.unshift(`§aPlayer settings updated: HSpeed=${selectionData.playerSpeedMultiplier}x, VSpeed=${selectionData.playerVerticalSpeedMultiplier}x, AutoSpec=${selectionData.autoSpectatorEnabled ? 'ON' : 'OFF'}, ExtReach=${selectionData.extendedReachEnabled ? 'ON' : 'OFF'}, FeedbackMsgs=${selectionData.showActionFeedbackMessages ? 'ON' : 'OFF'}`);
        } else if (!response.canceled && messages.length === 0) {
            messages.push("§eNo changes made to player settings.");
        }

        if (messages.length > 0) {
            player.sendMessage(messages.join('\n'));
        }

        system.run(() => showMainSettingsHubForm(player));

    } catch (e) {
        console.error(`[Player Settings Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError displaying or processing Player settings form.");
        system.run(() => showMainSettingsHubForm(player));
    }
}
