// scripts/ui/line_menu.js

import { Player } from "@minecraft/server";
import { ModalFormData } from "@minecraft/server-ui";
import {
    getPlayerSelection,
    setPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName // Renamed
} from "../state/index.js";
import { DEFAULT_BRUSH_MATERIAL } from "../constants/index.js";
import { parseWeightedBlockInput, validateBlockIds } from "../utils/index.js";
import { buildSourceDropdownOptions } from "./brush_menu.js"; // Import the helper

/**
 * Shows the form to configure Line Tool settings (Material Source, Manual Material).
 * Triggered by sneak-using the line tool (bn:line_tool).
 * @param {Player} player The player opening the settings form.
 */
export async function showLineToolSettingsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentSourceValue = selectionData.lineMaterialSource ?? 'manual';
    const currentManualMaterial = selectionData.lineManualMaterial ?? DEFAULT_BRUSH_MATERIAL;

    // Line tool uses 'manual', 'active', or 'saved palette'
    const { sourceOptions, sourceValueMap, currentSourceIndex } = buildSourceDropdownOptions(player, currentSourceValue, false, true);

    const form = new ModalFormData().title("Line Tool Settings")
        .dropdown("Material Source", sourceOptions, currentSourceIndex)
        .textField("Manual Material (id:wt,...)", "Used if Source is 'Manual Input'", currentManualMaterial);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eLine Tool settings unchanged.");
            return;
        }

        const formVals = response.formValues;
        const selectedOptionLabel = sourceOptions[formVals[0]];
        const newManualMaterialInput = formVals[1].trim();

        const newSourceValue = sourceValueMap[selectedOptionLabel] ?? 'manual';

        let manualInputIsValid = true;
        let finalManualMaterialString = currentManualMaterial;
        if (newSourceValue === 'manual') {
            if (!newManualMaterialInput) {
                player.sendMessage("§cManual Material cannot be empty when source is 'Manual'. Keeping previous setting.");
                manualInputIsValid = false;
            } else {
                const parsedResult = parseWeightedBlockInput(newManualMaterialInput);
                if (parsedResult.errors.length > 0) {
                    player.sendMessage(`§cManual Material input errors:\n${parsedResult.errors.join('\n')}\nKeeping previous setting.`);
                    manualInputIsValid = false;
                } else if (parsedResult.blocks.length === 0) {
                    player.sendMessage("§cNo blocks found in Manual Material input. Keeping previous setting.");
                    manualInputIsValid = false;
                } else {
                    const { invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id));
                    if (invalidIds.length > 0) {
                        player.sendMessage(`§cInvalid Block ID(s) in Manual Material: ${invalidIds.join(', ')}. Keeping previous setting.`);
                        manualInputIsValid = false;
                    } else {
                        finalManualMaterialString = newManualMaterialInput;
                    }
                }
            }
            if (!manualInputIsValid) {
                player.sendMessage("§eLine Tool settings NOT updated due to invalid manual input.");
                return;
            }
        }

        if (newSourceValue === 'active' && (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0)) { // Renamed
            player.sendMessage("§cActive Palette is empty. Select 'Manual Input' or load/create a palette first. Settings not updated."); return; // Renamed
        }

        if (!['manual', 'active'].includes(newSourceValue)) { // Renamed
            const palettes = getPlayerPalettes(player);
            if (!palettes[newSourceValue]) {
                player.sendMessage(`§cSelected saved palette "${newSourceValue}" no longer exists. Settings not updated.`); return;
            }
        }

        selectionData.lineMaterialSource = newSourceValue;
        if (newSourceValue === 'manual') {
            selectionData.lineManualMaterial = finalManualMaterialString;
        }

        setPlayerSelection(player, selectionData);

        const sourceDesc = selectedOptionLabel;
        player.sendMessage(`§aLine Tool settings updated: Material Source=${sourceDesc}`);

    } catch (e) {
        console.error(`[Line Tool Settings Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing line tool settings form.");
    }
}
