// scripts/ui/options_menu.js

import { Player, system } from "@minecraft/server";
import { ActionFormData, ModalFormData, MessageFormData } from "@minecraft/server-ui";
import {
    // State functions needed
    getPlayerSelection,
    setPlayerSelection,
    getPlayerPalettes,
    addPlayerPalette
} from "../state/index.js";
import {
    // Constants needed for sliders/limits
    MAX_EXPAND_CONTRACT
} from "../constants/index.js";
import {
    // Utils needed for selection info/expand
    getMinCorner,
    getMaxCorner,
    calculateVolume
} from "../utils/index.js";

// --- Selection Options ---
export async function showSelectionOptionsMenu(player) {
    const selection = getPlayerSelection(player);
    const selectionState = selection.selectionState ?? 0;
    const hasSelection = selectionState > 0;
    const canModify = selectionState === 2;

    let statusText = "§cInactive";
    if (selectionState === 1) statusText = "§6Pos A Set";
    else if (selectionState === 2) statusText = "§aComplete";

    const form = new ActionFormData()
        .title("Selection Options")
        .body(`Current selection status: ${statusText}§r`)
        .button("Expand / Contract") // Index 0
        .button("Selection Info")   // Index 1
        .button("Clear Selection")   // Index 2
        .button("Cancel");           // Index 3

    try {
        const response = await form.show(player);
        if (response.canceled || response.selection === undefined || response.selection === 3) {
            return;
        }

        let refreshMenu = false;

        switch (response.selection) {
            case 0: // Expand / Contract
                if (canModify) {
                    await showExpandContractForm(player);
                    refreshMenu = true;
                } else {
                    player.sendMessage("§cPlease complete the selection first (Set Pos A & Pos B).");
                    refreshMenu = true;
                }
                break;
            case 1: // Selection Info
                if (canModify) {
                    await showSelectionInfo(player);
                } else {
                    player.sendMessage("§cPlease complete the selection first (Set Pos A & Pos B).");
                    refreshMenu = true;
                }
                break;
            case 2: // Clear Selection
                if (hasSelection) {
                    selection.posA = undefined;
                    selection.posB = undefined;
                    selection.selectionState = 0;
                    setPlayerSelection(player, selection);
                    player.sendMessage("§aSelection cleared.");
                } else {
                    player.sendMessage("§eNo selection to clear.");
                }
                refreshMenu = true;
                break;
        }

        if (refreshMenu) {
            system.run(async () => {
                if (player.isValid()) {
                    await showSelectionOptionsMenu(player);
                }
            });
        }

    } catch (e) {
        console.error(`[Selection Options Menu] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError showing selection options menu.");
    }
}

async function showExpandContractForm(player) {
    const selection = getPlayerSelection(player);
    if (selection.selectionState !== 2 || !selection.posA || !selection.posB) {
        player.sendMessage("§cCannot expand/contract: Selection is incomplete.");
        return;
    }

    const form = new ModalFormData().title("Expand/Contract Selection")
        .slider("Amount East (+X)", -MAX_EXPAND_CONTRACT, MAX_EXPAND_CONTRACT, 1, 0)
        .slider("Amount West (-X)", -MAX_EXPAND_CONTRACT, MAX_EXPAND_CONTRACT, 1, 0)
        .slider("Amount Up (+Y)", -MAX_EXPAND_CONTRACT, MAX_EXPAND_CONTRACT, 1, 0)
        .slider("Amount Down (-Y)", -MAX_EXPAND_CONTRACT, MAX_EXPAND_CONTRACT, 1, 0)
        .slider("Amount South (+Z)", -MAX_EXPAND_CONTRACT, MAX_EXPAND_CONTRACT, 1, 0)
        .slider("Amount North (-Z)", -MAX_EXPAND_CONTRACT, MAX_EXPAND_CONTRACT, 1, 0);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eExpansion/Contraction cancelled.");
            return;
        }

        const [expandEast, expandWest, expandUp, expandDown, expandSouth, expandNorth] = response.formValues;
        const { posA, posB } = selection;

        const min = getMinCorner(posA, posB);

        let newPosA = { ...posA };
        let newPosB = { ...posB };

        if (posA.x === min.x) { newPosA.x -= expandWest; newPosB.x += expandEast; }
        else { newPosB.x -= expandWest; newPosA.x += expandEast; }

        if (posA.y === min.y) { newPosA.y -= expandDown; newPosB.y += expandUp; }
        else { newPosB.y -= expandDown; newPosA.y += expandUp; }

        if (posA.z === min.z) { newPosA.z -= expandNorth; newPosB.z += expandSouth; }
        else { newPosB.z -= expandNorth; newPosA.z += expandSouth; }

        const newMin = getMinCorner(newPosA, newPosB);
        const newMax = getMaxCorner(newPosA, newPosB);
        if (newMin.x > newMax.x || newMin.y > newMax.y || newMin.z > newMax.z) {
            player.sendMessage("§cInvalid expansion/contraction: Resulting selection size would be zero or negative.");
            return;
        }

        selection.posA = newPosA;
        selection.posB = newPosB;
        setPlayerSelection(player, selection);

        const newVolume = calculateVolume(newPosA, newPosB);
        player.sendMessage(`§aSelection modified. New volume: ${newVolume}`);
        player.sendMessage(`§aNew Pos A: §f${newPosA.x}, ${newPosA.y}, ${newPosA.z}`);
        player.sendMessage(`§bNew Pos B: §f${newPosB.x}, ${newPosB.y}, ${newPosB.z}`);

    } catch (e) {
        console.error(`[Expand/Contract Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing expand/contract values.");
    }
}

async function showSelectionInfo(player) {
    const selection = getPlayerSelection(player);
    const { posA, posB, selectionState } = selection;

    if (selectionState !== 2 || !posA || !posB) {
        player.sendMessage("§cCannot show info: Selection is incomplete.");
        return;
    }

    const min = getMinCorner(posA, posB);
    const max = getMaxCorner(posA, posB);
    const size = { x: max.x - min.x + 1, y: max.y - min.y + 1, z: max.z - min.z + 1 };
    const volume = size.x * size.y * size.z;

    const body = `§fPos A: §a${posA.x}, ${posA.y}, ${posA.z}\n` +
        `§fPos B: §b${posB.x}, ${posB.y}, ${posB.z}\n` +
        `§fMin Corner: §e${min.x}, ${min.y}, ${min.z}\n` +
        `§fMax Corner: §e${max.x}, ${max.y}, ${max.z}\n` +
        `§fSize (X, Y, Z): §d${size.x}, ${size.y}, ${size.z}\n` +
        `§fVolume: §d${volume} blocks`;

    const form = new MessageFormData()
        .title("Selection Info")
        .body(body)
        .button1("OK");

    try {
        await form.show(player);
    } catch (e) {
        console.error(`[Selection Info Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError displaying selection info.");
    }
}

// --- Paste Options ---
export async function showPasteOptionsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentOptions = selectionData.pasteOptions ?? { pasteOrigin: false, placeEntities: true, integrity: 1.0, seed: "", rotation: "None (0°)", mirror: "None" };

    const rotationOptions = ["None (0°)", "90°", "180°", "270°"];
    const mirrorOptions = ["None", "X", "Z", "XZ"];

    const currentRotationIndex = Math.max(0, rotationOptions.indexOf(currentOptions.rotation));
    const currentMirrorIndex = Math.max(0, mirrorOptions.indexOf(currentOptions.mirror));

    const form = new ModalFormData().title("Advanced Paste Options")
        .toggle("Paste at Original Location", currentOptions.pasteOrigin) // 0
        .toggle("Place Entities (if copied)", currentOptions.placeEntities) // 1
        .slider("Integrity (%)", 0, 100, 1, Math.round(currentOptions.integrity * 100)) // 2
        .textField("Integrity Seed (number, optional)", "Leave blank for random", currentOptions.seed) // 3
        .dropdown("Rotation", rotationOptions, currentRotationIndex) // 4
        .dropdown("Mirror", mirrorOptions, currentMirrorIndex); // 5

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) { player.sendMessage("§ePaste options unchanged."); return; }

        const formVals = response.formValues;
        let newOptions = {
            pasteOrigin: formVals[0],
            placeEntities: formVals[1],
            integrity: formVals[2] / 100.0,
            seed: "",
            rotation: rotationOptions[formVals[4]],
            mirror: mirrorOptions[formVals[5]],
        };
        let errorMessages = [];

        const seedStr = formVals[3].trim();
        if (seedStr !== "") {
            if (isNaN(parseInt(seedStr, 10))) {
                errorMessages.push("Integrity Seed must be a whole number or blank.");
            }
            newOptions.seed = seedStr;
        } else {
            newOptions.seed = "";
        }

        if (errorMessages.length > 0) { player.sendMessage("§cInvalid input:\n" + errorMessages.join("\n") + "\n§ePaste options not updated."); return; }

        selectionData.pasteOptions = newOptions;
        setPlayerSelection(player, selectionData);
        player.sendMessage("§aAdvanced paste options updated.");

    } catch (e) { console.error(`[Paste Options Form] Error: ${e}\n${e.stack}`); player.sendMessage("§cError setting paste options."); }
}

// --- Copy Options ---
export async function showCopyOptionsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentIncludeEntities = selectionData.copyIncludesEntities ?? true;

    const form = new ModalFormData().title("Copy Options")
        .toggle("Include Entities", currentIncludeEntities); // Index 0

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eCopy options unchanged.");
            return;
        }
        const newIncludeEntities = response.formValues[0]; // This is a boolean

        // Log the value being set
        console.log(`[WorldEdit CopyOptionsForm] Player: ${player.name}, Setting 'copyIncludesEntities' to: ${newIncludeEntities}, Type: ${typeof newIncludeEntities}`);

        selectionData.copyIncludesEntities = newIncludeEntities;
        setPlayerSelection(player, selectionData);

        // Log after attempting to set, to confirm it's in selectionData
        const updatedSelectionData = getPlayerSelection(player);
        console.log(`[WorldEdit CopyOptionsForm] Player: ${player.name}, 'copyIncludesEntities' in state after update: ${updatedSelectionData.copyIncludesEntities}, Type: ${typeof updatedSelectionData.copyIncludesEntities}`);

        player.sendMessage(`§aCopy options updated: Include Entities = ${newIncludeEntities ? 'ON' : 'OFF'}`);
    } catch (e) {
        console.error(`[Copy Options Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError setting copy options.");
    }
}


// --- Palette Saving Prompt ---
export async function promptSavePalette(player, blockIds) {
    if (!blockIds || blockIds.length === 0) return;

    const form = new ModalFormData()
        .title("Save New Palette from Manual Input")
        .textField("Palette Name", "Enter a unique name for this set of blocks", "");

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) { player.sendMessage("§ePalette save cancelled."); return; }
        const paletteName = response.formValues[0]?.trim();
        if (!paletteName) { player.sendMessage("§cPalette name cannot be empty. Not saved."); return; }

        const existingPalettes = getPlayerPalettes(player);
        if (existingPalettes[paletteName]) { player.sendMessage(`§cPalette named "${paletteName}" already exists. Not saved.`); return; }

        const blocksWithWeights = blockIds.map(id => ({ id: id, weight: 1 }));
        if (addPlayerPalette(player, paletteName, blocksWithWeights)) {
            player.sendMessage(`§aPalette "${paletteName}" saved successfully with default weights.`);
        }
    } catch (e) { console.error(`[WorldEdit Save Palette Prompt] Error: ${e}\n${e.stack}`); player.sendMessage("§cError trying to save palette."); }
}
