// scripts/ui/palette_menu.js

import { Player, system } from "@minecraft/server";
import { ActionFormData, ModalFormData } from "@minecraft/server-ui";
import {
    getPlayerSelection,
    getPlayerPalettes,
    savePlayerPalettes,
    addPlayerPalette,
    deletePlayerPalette,
    setActiveSavedPaletteName, // Renamed
    getActiveSavedPaletteName, // Renamed
    setPlayerSelection,
    discardActivePaletteChanges // Renamed
} from "../state/index.js";
import { parseWeightedBlockInput, validateBlockIds } from "../utils/index.js";
import { showEditMenu } from "./main_menu.js"; // For back button
import { analyzeSelectionForPalette } from "../core/analyze_selection.js";

// --- Main Palette Menu ---
/** Shows the main menu for managing palettes */
export async function showPaletteMenu(player) {
    const selectionData = getPlayerSelection(player);
    const activeSaved = getActiveSavedPaletteName(player); // Renamed getter
    const activePaletteData = selectionData.activePaletteData; // Renamed field
    const isDirty = !!selectionData.activePaletteIsDirty; // Renamed field
    const activeItemsCount = activePaletteData?.length ?? 0;
    const activeStatus = activePaletteData ? ` (${activeItemsCount} items${isDirty ? ' §c*' : ''})` : ' (Empty/Inactive)'; // Updated text
    const hasCompleteSelection = selectionData.selectionState === 2;

    const form = new ActionFormData()
        .title("Manage Palettes")
        .body(`§eActive Palette:§r ${activeStatus}\n§eLoaded Saved Palette:§r ${activeSaved ? `§b${activeSaved}` : "§7None"}`) // Updated text
        .button("Load Saved Palette", "textures/ui/book_glyph_color")
        .button("Edit Saved Palette", "textures/ui/book_edit_default")
        .button("Save Options...", "textures/ui/icon_save")
        .button(`Create from Selection... ${hasCompleteSelection ? '' : '§7(Needs Sel)'}`, "textures/ui/icon_import")
        .button("Duplicate Loaded Saved Palette", "textures/ui/copy") // Clarified button text
        .button("Delete Saved Palette", "textures/ui/trash")
        .button("Back", "textures/ui/cancel");

    try {
        const response = await form.show(player);
        if (response.canceled || response.selection === undefined) return;

        let refreshMenu = true;

        switch (response.selection) {
            case 0: await showLoadPaletteList(player); break;
            case 1: await showEditPaletteList(player); break;
            case 2: await showSaveOptionsMenu(player); refreshMenu = false; break;
            case 3:
                if (hasCompleteSelection) {
                    const uniqueBlockIds = await analyzeSelectionForPalette(player);
                    if (uniqueBlockIds) {
                        await showCreatePaletteFromSelectionForm(player, uniqueBlockIds);
                        refreshMenu = false;
                    } else {
                        refreshMenu = true;
                    }
                } else {
                    player.sendMessage("§cComplete selection needed to create palette.");
                    refreshMenu = true;
                }
                break;
            case 4: await duplicateActiveSavedPalette(player); break; // Renamed function call
            case 5: await showDeletePaletteMenu(player); break;
            case 6:
                refreshMenu = false;
                await system.run(async () => { if (player.isValid()) await showEditMenu(player); });
                return;
        }

        if (refreshMenu) {
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
        }
    } catch (e) {
        console.error(`[WorldEdit Palette Menu] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError displaying palette menu.");
    }
}

// --- Sub-Menu Functions (Called by showPaletteMenu or each other) ---

async function showLoadPaletteList(player) {
    const palettes = getPlayerPalettes(player);
    const paletteNames = Object.keys(palettes);
    const currentSelection = getPlayerSelection(player);
    const isDirty = !!currentSelection.activePaletteIsDirty; // Renamed

    const form = new ActionFormData().title("Load Saved Palette");
    let body = "";
    if (isDirty) {
        body += "§cWarning: Loading will discard unsaved Active Palette changes!§r\n"; // Renamed
    }
    body += "Select saved palette to load into Active Palette:"; // Renamed
    form.body(body);

    if (paletteNames.length === 0) {
        form.body(body + "\n§eNo saved palettes available.§r");
    } else {
        paletteNames.forEach(name => form.button(name));
    }
    form.button("§c[Cancel]");

    try {
        const response = await form.show(player);
        const cancelIndex = paletteNames.length;

        if (response.canceled || response.selection === undefined || response.selection === cancelIndex) {
            return;
        }

        if (response.selection < cancelIndex) {
            const selectedName = paletteNames[response.selection];

            if (isDirty) {
                const confirmForm = new ModalFormData()
                    .title("Discard Unsaved Changes?")
                    .toggle(`Loading "${selectedName}" will discard unsaved changes in the Active Palette. Continue?`, false); // Renamed
                const confirmResponse = await confirmForm.show(player);
                if (confirmResponse.canceled || !confirmResponse.formValues || !confirmResponse.formValues[0]) {
                    player.sendMessage("§eLoad cancelled.");
                    return;
                }
            }

            const paletteDataToLoad = palettes[selectedName];
            if (!paletteDataToLoad) {
                player.sendMessage(`§cError: Palette "${selectedName}" data not found.`);
                return;
            }

            currentSelection.activePaletteData = JSON.parse(JSON.stringify(paletteDataToLoad)); // Renamed field
            currentSelection.activeSavedPaletteName = selectedName; // Renamed field
            currentSelection.activePaletteIsDirty = false; // Renamed field
            setPlayerSelection(player, currentSelection);
            player.sendMessage(`§aLoaded "${selectedName}" into Active Palette and set as loaded saved palette.`); // Renamed
        }
    } catch (e) {
        console.error(`[WorldEdit Load Palette List] Error: ${e}\n${e.stack}`);
        player.sendMessage(`§cError displaying load palette list.`);
    }
}

async function showEditPaletteList(player) {
    const palettes = getPlayerPalettes(player);
    const paletteNames = Object.keys(palettes);

    const form = new ActionFormData().title("Edit Saved Palette");
    form.body("Select saved palette to edit directly:");

    if (paletteNames.length === 0) {
        form.body("§eNo saved palettes available to edit.§r");
    } else {
        paletteNames.forEach(name => form.button(name));
    }
    form.button("§c[Cancel]");

    try {
        const response = await form.show(player);
        const cancelIndex = paletteNames.length;

        if (response.canceled || response.selection === undefined || response.selection === cancelIndex) {
            return;
        }

        if (response.selection < cancelIndex) {
            const paletteNameToEdit = paletteNames[response.selection];
            system.run(async () => {
                if (player.isValid()) {
                    await editTextPalette(player, paletteNameToEdit);
                    // await showPaletteMenu(player); // Refresh is handled by editTextPalette or its caller
                }
            });
        }
    } catch (e) {
        console.error(`[WorldEdit Edit Palette List] Error: ${e}\n${e.stack}`);
        player.sendMessage(`§cError selecting palette to edit.`);
    }
}

async function editTextPalette(player, paletteName) {
    const palettes = getPlayerPalettes(player);
    const currentPaletteData = palettes[paletteName];

    if (!currentPaletteData) {
        player.sendMessage(`§cPalette "${paletteName}" not found (maybe deleted?).`);
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); }); // Go back
        return;
    }

    const currentBlocksString = currentPaletteData.map(b => `${b.id.replace('minecraft:', '')}:${b.weight}`).join(', ');

    const form = new ModalFormData()
        .title(`Edit Saved Palette: ${paletteName}`)
        .textField("Palette Name", "Edit name (must be unique)", paletteName)
        .textField("Blocks (format: id:weight, ...)", "Edit blocks and weights", currentBlocksString);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eEdit cancelled.");
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); }); // Go back
            return;
        }

        const newPaletteName = response.formValues[0]?.trim();
        const newBlockInput = response.formValues[1]?.trim();

        if (!newPaletteName) { player.sendMessage("§cPalette name cannot be empty. Not saved."); /* Fall through to refresh */ }
        else if (!newBlockInput) { player.sendMessage("§cBlock input cannot be empty. Not saved."); /* Fall through to refresh */ }
        else {
            const { blocks: newBlocksWithWeights, errors } = parseWeightedBlockInput(newBlockInput);
            if (errors.length > 0) { player.sendMessage(`§cInput errors:\n${errors.join('\n')}`); }
            else if (newBlocksWithWeights.length === 0) { player.sendMessage("§cNo valid blocks parsed. Palette not saved."); }
            else {
                const { invalidIds } = validateBlockIds(newBlocksWithWeights.map(b => b.id));
                if (invalidIds.length > 0) { player.sendMessage(`§cInvalid block ID(s) resolved: ${invalidIds.join(', ')}. Palette not saved.`); }
                else {
                    const updatedPalettes = getPlayerPalettes(player);
                    if (newPaletteName !== paletteName && updatedPalettes[newPaletteName]) {
                        player.sendMessage(`§cAnother palette named "${newPaletteName}" already exists. Cannot rename/save.`);
                    } else {
                        if (newPaletteName !== paletteName) {
                            delete updatedPalettes[paletteName];
                        }
                        updatedPalettes[newPaletteName] = newBlocksWithWeights;

                        if (savePlayerPalettes(player, updatedPalettes)) {
                            player.sendMessage(`§aSaved palette "${newPaletteName}" updated successfully.`);
                            const selectionData = getPlayerSelection(player);
                            if (selectionData.activeSavedPaletteName === paletteName && newPaletteName !== paletteName) { // Renamed
                                setActiveSavedPaletteName(player, newPaletteName); // Renamed
                                player.sendMessage(`§eLoaded saved palette updated to "${newPaletteName}".`);
                            }
                        }
                    }
                }
            }
        }
    } catch (e) {
        console.error(`[WorldEdit Edit Palette Text] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError editing palette.");
    }
    // Always refresh palette menu after edit attempt
    await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
}

async function showSaveOptionsMenu(player) {
    const selectionData = getPlayerSelection(player);
    const activeSavedName = selectionData.activeSavedPaletteName; // Renamed
    const activePaletteData = selectionData.activePaletteData; // Renamed
    const isDirty = !!selectionData.activePaletteIsDirty; // Renamed
    const hasActiveData = !!activePaletteData && activePaletteData.length > 0;

    const form = new ActionFormData().title("Save Options");
    let body = `§eActive Palette Status:§r ${hasActiveData ? `${activePaletteData.length} items` : 'Empty'}${isDirty ? ' §c(Unsaved Changes)' : ' (No Changes)'}\n`; // Renamed
    body += `§eLoaded Saved Palette:§r ${activeSavedName ? `§b${activeSavedName}` : "§7None"}`; // Renamed
    form.body(body);

    let buttonIndexMap = {};
    let currentButtonIndex = 0;

    if (hasActiveData) {
        form.button("Save Active Palette as New..."); // Renamed
        buttonIndexMap[currentButtonIndex++] = "save_as";
    } else {
        form.button("Save Active Palette as New... §7(Empty)"); // Renamed
        buttonIndexMap[currentButtonIndex++] = "save_as_disabled";
    }

    if (activeSavedName && isDirty && hasActiveData) {
        form.button(`Update Saved '${activeSavedName}'`);
        buttonIndexMap[currentButtonIndex++] = "update_active";
    } else {
        form.button(`Update Loaded Saved Palette §7(N/A)`); // Clarified
        buttonIndexMap[currentButtonIndex++] = "update_disabled";
    }

    if (hasActiveData || isDirty) {
        form.button("Discard Active Palette Changes"); // Renamed
        buttonIndexMap[currentButtonIndex++] = "discard";
    } else {
        form.button("Discard Active Palette Changes §7(N/A)"); // Renamed
        buttonIndexMap[currentButtonIndex++] = "discard_disabled";
    }

    form.button("Back");
    buttonIndexMap[currentButtonIndex++] = "back";

    try {
        const response = await form.show(player);
        if (response.canceled || response.selection === undefined) {
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }

        const action = buttonIndexMap[response.selection];
        let returnToMainMenu = true;

        switch (action) {
            case "save_as": await saveActivePaletteAs(player); break; // Renamed
            case "save_as_disabled":
                player.sendMessage("§cActive Palette is empty, cannot 'Save As'."); // Renamed
                returnToMainMenu = false;
                await system.run(async () => { if (player.isValid()) await showSaveOptionsMenu(player); });
                break;
            case "update_active": await confirmOverwriteActiveSavedPalette(player); break; // Renamed
            case "update_disabled":
                player.sendMessage("§cCannot update: No loaded saved palette, or no unsaved changes, or Active Palette is empty."); // Renamed
                returnToMainMenu = false;
                await system.run(async () => { if (player.isValid()) await showSaveOptionsMenu(player); });
                break;
            case "discard": await discardActivePaletteChanges(player); break; // Renamed
            case "discard_disabled":
                player.sendMessage("§eNo Active Palette changes or data to discard."); // Renamed
                returnToMainMenu = false;
                await system.run(async () => { if (player.isValid()) await showSaveOptionsMenu(player); });
                break;
            case "back": break;
            default: console.warn(`Invalid action selected in Save Options Menu: ${response.selection}`); break;
        }

        if (returnToMainMenu) {
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
        }

    } catch (e) {
        console.error(`[WorldEdit Save Options Menu] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError in save options menu.");
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
    }
}

async function saveActivePaletteAs(player) { // Renamed
    const selectionData = getPlayerSelection(player);
    const activePaletteData = selectionData.activePaletteData; // Renamed

    if (!activePaletteData || activePaletteData.length === 0) { // Renamed
        player.sendMessage("§cActive Palette is empty. Add blocks using Picker first."); // Renamed
        return;
    }

    const form = new ModalFormData()
        .title("Save Active Palette As...") // Renamed
        .textField("New Palette Name", "Enter a unique name", "");

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eSave cancelled.");
            return;
        }
        const newPaletteName = response.formValues[0]?.trim();
        if (!newPaletteName) {
            player.sendMessage("§cPalette name cannot be empty.");
            return;
        }

        const existingPalettes = getPlayerPalettes(player);
        if (existingPalettes[newPaletteName]) {
            player.sendMessage(`§cPalette named "${newPaletteName}" already exists.`);
            return;
        }

        if (addPlayerPalette(player, newPaletteName, activePaletteData)) { // Renamed
            player.sendMessage(`§aActive Palette saved as "${newPaletteName}".`); // Renamed
            selectionData.activeSavedPaletteName = newPaletteName; // Renamed
            selectionData.activePaletteIsDirty = false; // Renamed
            setPlayerSelection(player, selectionData);
        }
    } catch (e) {
        console.error(`[WorldEdit Save Active As] Error: ${e}\n${e.stack}`); // Renamed
        player.sendMessage("§cError saving palette.");
    }
}

async function confirmOverwriteActiveSavedPalette(player) { // Renamed
    const selectionData = getPlayerSelection(player);
    const activePaletteData = selectionData.activePaletteData; // Renamed
    const activeSavedName = selectionData.activeSavedPaletteName; // Renamed
    const isDirty = !!selectionData.activePaletteIsDirty; // Renamed

    if (!activeSavedName) { player.sendMessage("§cNo loaded saved palette set to overwrite."); return; } // Renamed
    if (!activePaletteData || activePaletteData.length === 0) { player.sendMessage("§cActive Palette is empty. Cannot overwrite."); return; } // Renamed
    if (!isDirty) { player.sendMessage("§eNo unsaved changes in Active Palette to overwrite."); return; } // Renamed

    const form = new ModalFormData()
        .title("Overwrite Saved Palette?") // Renamed
        .toggle(`Overwrite saved palette "${activeSavedName}" with current Active Palette changes? This cannot be undone.`, false); // Renamed

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues || !response.formValues[0]) {
            player.sendMessage("§eOverwrite cancelled.");
            return;
        }

        if (addPlayerPalette(player, activeSavedName, activePaletteData)) { // Renamed
            player.sendMessage(`§aSaved palette "${activeSavedName}" updated with Active Palette changes.`); // Renamed
            selectionData.activePaletteIsDirty = false; // Renamed
            setPlayerSelection(player, selectionData);
        }
    } catch (e) {
        console.error(`[WorldEdit Overwrite Active Saved] Error: ${e}\n${e.stack}`); // Renamed
        player.sendMessage("§cError overwriting palette.");
    }
}

async function duplicateActiveSavedPalette(player) { // Renamed
    const activeSavedName = getActiveSavedPaletteName(player); // Renamed getter

    if (!activeSavedName) { // Renamed
        player.sendMessage("§cNo loaded saved palette selected to duplicate."); // Renamed
        return;
    }

    const palettes = getPlayerPalettes(player);
    const paletteToCopy = palettes[activeSavedName]; // Renamed

    if (!paletteToCopy) {
        player.sendMessage(`§cCould not find data for loaded saved palette "${activeSavedName}".`); // Renamed
        return;
    }

    const form = new ModalFormData()
        .title(`Duplicate Saved Palette: ${activeSavedName}`) // Renamed
        .textField("New Palette Name", `Enter a unique name for the copy`, `${activeSavedName} Copy`);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§eDuplicate cancelled.");
            return;
        }

        const newPaletteName = response.formValues[0]?.trim();
        if (!newPaletteName) {
            player.sendMessage("§cNew palette name cannot be empty.");
            return;
        }
        if (palettes[newPaletteName]) {
            player.sendMessage(`§cPalette named "${newPaletteName}" already exists.`);
            return;
        }
        if (newPaletteName === activeSavedName) { // Renamed
            player.sendMessage(`§cNew name must be different from the original.`);
            return;
        }

        const copiedData = JSON.parse(JSON.stringify(paletteToCopy));

        if (addPlayerPalette(player, newPaletteName, copiedData)) {
            player.sendMessage(`§aPalette "${activeSavedName}" duplicated as "${newPaletteName}".`); // Renamed
        }
    } catch (e) {
        console.error(`[WorldEdit Duplicate Palette] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError duplicating palette.");
    }
}

async function showDeletePaletteMenu(player) {
    const palettes = getPlayerPalettes(player);
    const paletteNames = Object.keys(palettes);

    const form = new ActionFormData().title("Delete Saved Palette");

    if (paletteNames.length === 0) {
        form.body("No saved palettes to delete.");
    } else {
        form.body("Select a saved palette to delete:");
        paletteNames.forEach(name => form.button(name));
    }
    form.button("§c[Back]");

    try {
        const response = await form.show(player);
        const backIndex = paletteNames.length;

        if (response.canceled || response.selection === undefined || response.selection === backIndex) {
            return; // Handled by caller to refresh main palette menu
        }

        if (response.selection < backIndex) {
            const selectedName = paletteNames[response.selection];

            const confirmForm = new ModalFormData()
                .title("Confirm Deletion")
                .toggle(`Delete your saved palette "${selectedName}"? This cannot be undone.`, false);
            const confirmResponse = await confirmForm.show(player);

            if (!confirmResponse.canceled && confirmResponse.formValues?.[0] === true) {
                if (deletePlayerPalette(player, selectedName)) {
                    player.sendMessage(`§aSaved palette "${selectedName}" deleted.`);
                } else {
                    // deletePlayerPalette already sends a message on failure if palette not found
                }
            } else {
                player.sendMessage("§eDeletion cancelled.");
            }
        }
    } catch (e) {
        console.error(`[WorldEdit Delete Palette] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError showing/processing delete palette menu.");
    }
}


export async function showCreatePaletteFromSelectionForm(player, uniqueBlockIds) {
    if (!uniqueBlockIds || uniqueBlockIds.length === 0) {
        player.sendMessage("§cCannot create palette: No blocks provided for form.");
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
        return;
    }

    const form = new ModalFormData().title("Create Palette from Selection");
    form.textField("New Palette Name", "Enter a unique name", "");
    const formIndexToBlockId = {}; let currentIndex = 1;
    for (const blockId of uniqueBlockIds) { const simpleName = blockId.replace('minecraft:', ''); form.toggle(`Include ${simpleName}?`, true); formIndexToBlockId[currentIndex] = { type: 'toggle', blockId: blockId }; currentIndex++; form.slider(`Weight (${simpleName})`, 1, 100, 1, 1); formIndexToBlockId[currentIndex] = { type: 'slider', blockId: blockId }; currentIndex++; }

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§ePalette creation cancelled.");
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        const formValues = response.formValues; const newPaletteName = formValues[0]?.trim();
        if (!newPaletteName) {
            player.sendMessage("§cPalette name cannot be empty. Not saved.");
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        const existingPalettes = getPlayerPalettes(player);
        if (existingPalettes[newPaletteName]) {
            player.sendMessage(`§cPalette named "${newPaletteName}" already exists. Not saved.`);
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        const newPaletteData = [];
        for (let i = 1; i < formValues.length; i++) {
            const mapping = formIndexToBlockId[i];
            if (mapping?.type === 'toggle') {
                const includeBlock = formValues[i];
                const sliderIndex = i + 1;
                const weightMapping = formIndexToBlockId[sliderIndex];
                if (includeBlock && weightMapping?.type === 'slider' && weightMapping?.blockId === mapping.blockId) {
                    const weight = formValues[sliderIndex];
                    newPaletteData.push({ id: mapping.blockId, weight: weight });
                }
                i++;
            }
        }
        if (newPaletteData.length === 0) {
            player.sendMessage("§cNo blocks were included in the palette. Not saved.");
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        if (addPlayerPalette(player, newPaletteName, newPaletteData)) {
            player.sendMessage(`§aPalette "${newPaletteName}" created and saved successfully.`);
        }
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
    } catch (e) {
        console.error(`[Create Palette from Selection Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing create palette form.");
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
    }
}
