// scripts/ui/path_menu.js

import { Player, system } from "@minecraft/server";
import { ActionFormData, ModalFormData } from "@minecraft/server-ui";
import {
    getPlayerSelection,
    setPlayerSelection,
    getPlayerPalettes,
    getActiveSavedPaletteName, // Renamed
    addPlayerPalette
} from "../state/index.js";
import {
    DEFAULT_PATH_MATERIAL,
    DEFAULT_PATH_CURVE_TYPE,
    DEFAULT_PATH_SMOOTHNESS,
    DEFAULT_PATH_WIDTH,
    MAX_PATH_WIDTH
} from "../constants/index.js";
import { drawPath } from "../core/path.js";
import { parseWeightedBlockInput, validateBlockIds } from "../utils/index.js";
import { showEditMenu } from "./main_menu.js";
import { buildSourceDropdownOptions } from "./brush_menu.js";
import { analyzeSelectionForPalette } from "../core/analyze_selection.js";


/**
 * Shows the main options menu for the Path Tool.
 * Triggered by sneak-using the path tool.
 * @param {Player} player The player opening the menu.
 */
export async function showPathToolOptionsMenu(player) {
    const selectionData = getPlayerSelection(player);
    const points = selectionData.pathPoints ?? [];
    const pointCount = points.length;
    const curveType = selectionData.pathCurveType ?? DEFAULT_PATH_CURVE_TYPE;
    const width = selectionData.pathWidth ?? DEFAULT_PATH_WIDTH;
    const isClosed = selectionData.pathClosed ?? false;

    const form = new ActionFormData()
        .title("Path Tool Options")
        .body(`Points Set: ${pointCount}\nCurve Type: ${curveType}\nWidth: ${width}\nClosed Path: ${isClosed ? 'Yes' : 'No'}`)
        .button(`Finish Path (${curveType}${isClosed ? ' Loop' : ''}) ${pointCount < (isClosed ? 3 : 2) ? `§7(Need >=${isClosed ? 3 : 2} pts)` : ''}`)
        .button("Set Material...")
        .button("Set Path Shape...")
        .button(`Undo Last Point ${pointCount === 0 ? '§7(No pts)' : ''}`)
        .button(`Clear All Points ${pointCount === 0 ? '§7(No pts)' : ''}`)
        .button("Cancel");

    try {
        const response = await form.show(player);
        if (response.canceled || response.selection === undefined || response.selection === 5) {
            return;
        }

        let refreshMenu = false;

        switch (response.selection) {
            case 0:
                const minPoints = isClosed ? 3 : 2;
                if (pointCount >= minPoints) {
                    system.run(async () => { if (player.isValid()) await drawPath(player); });
                    // Don't refresh menu after drawing, path points are cleared by drawPath
                } else {
                    player.sendMessage(`§cCannot finish path, need at least ${minPoints} points${isClosed ? ' for a closed path' : ''}.`);
                    refreshMenu = true;
                }
                break;
            case 1:
                await showPathMaterialForm(player);
                refreshMenu = true;
                break;
            case 2:
                await showPathShapeOptionsForm(player);
                refreshMenu = true;
                break;
            case 3:
                if (pointCount > 0) {
                    selectionData.pathPoints.pop();
                    setPlayerSelection(player, selectionData);
                    player.sendMessage(`§aLast path point removed (${selectionData.pathPoints.length} remaining).`);
                } else {
                    player.sendMessage("§cNo points to remove.");
                }
                refreshMenu = true;
                break;
            case 4:
                if (pointCount > 0) {
                    selectionData.pathPoints = [];
                    setPlayerSelection(player, selectionData);
                    player.sendMessage("§aAll path points cleared.");
                } else {
                    player.sendMessage("§cNo points to clear.");
                }
                refreshMenu = true;
                break;
        }

        if (refreshMenu) {
            system.run(async () => { if (player.isValid()) await showPathToolOptionsMenu(player); });
        }

    } catch (e) {
        console.error(`[Path Tool Options Menu] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError showing path tool options.");
    }
}

/** Shows the form to configure Path Tool material settings. */
async function showPathMaterialForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentSourceValue = selectionData.pathMaterialSource ?? 'manual';
    const currentManualMaterial = selectionData.pathManualMaterial ?? DEFAULT_PATH_MATERIAL;
    // Path tool uses 'manual', 'active', or 'saved palette'
    const { sourceOptions, sourceValueMap, currentSourceIndex } = buildSourceDropdownOptions(player, currentSourceValue, false, true);

    const form = new ModalFormData().title("Path Material Settings")
        .dropdown("Material Source", sourceOptions, currentSourceIndex)
        .textField("Manual Material (id:wt,...)", "Used if Source is 'Manual Input'", currentManualMaterial);
    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) { player.sendMessage("§ePath material settings unchanged."); return; }

        const formVals = response.formValues;
        const selectedOptionLabel = sourceOptions[formVals[0]];
        const newManualMaterialInput = formVals[1].trim();
        const newSourceValue = sourceValueMap[selectedOptionLabel] ?? 'manual';

        let manualInputIsValid = true;
        let finalManualMaterialString = currentManualMaterial;
        if (newSourceValue === 'manual') {
            if (!newManualMaterialInput) { manualInputIsValid = false; player.sendMessage("§cManual Material cannot be empty when source is 'Manual'. Keeping previous setting."); }
            else {
                const parsedResult = parseWeightedBlockInput(newManualMaterialInput);
                if (parsedResult.errors.length > 0) { manualInputIsValid = false; player.sendMessage(`§cManual Material input errors:\n${parsedResult.errors.join('\n')}\nKeeping previous setting.`); }
                else if (parsedResult.blocks.length === 0) { manualInputIsValid = false; player.sendMessage("§cNo blocks found in Manual Material input. Keeping previous setting."); }
                else {
                    const { invalidIds } = validateBlockIds(parsedResult.blocks.map(b => b.id));
                    if (invalidIds.length > 0) { manualInputIsValid = false; player.sendMessage(`§cInvalid Block ID(s) in Manual Material: ${invalidIds.join(', ')}. Keeping previous setting.`); }
                    else { finalManualMaterialString = newManualMaterialInput; }
                }
            }
            if (!manualInputIsValid) { player.sendMessage("§ePath material settings NOT updated due to invalid manual input."); return; }
        }

        if (newSourceValue === 'active' && (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0)) {
            player.sendMessage("§cActive Palette is empty. Select 'Manual Input' or load/create a palette first. Settings not updated."); return;
        }

        if (!['manual', 'active'].includes(newSourceValue)) {
            const palettes = getPlayerPalettes(player);
            if (!palettes[newSourceValue]) {
                player.sendMessage(`§cSelected saved palette "${newSourceValue}" no longer exists. Settings not updated.`); return;
            }
        }

        selectionData.pathMaterialSource = newSourceValue;
        if (newSourceValue === 'manual') {
            selectionData.pathManualMaterial = finalManualMaterialString;
        }
        setPlayerSelection(player, selectionData);
        player.sendMessage(`§aPath Material updated: Source=${selectedOptionLabel}`);
    } catch (e) {
        console.error(`[Path Material Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing path material form.");
    }
}

/**
 * Shows the form to configure Path Tool shape settings (Curve Type, Smoothness, Width, Closed).
 * @param {Player} player
 */
async function showPathShapeOptionsForm(player) {
    const selectionData = getPlayerSelection(player);
    const currentCurveType = selectionData.pathCurveType ?? DEFAULT_PATH_CURVE_TYPE;
    const currentSmoothness = selectionData.pathSmoothness ?? DEFAULT_PATH_SMOOTHNESS;
    const currentWidth = selectionData.pathWidth ?? DEFAULT_PATH_WIDTH;
    const currentClosed = selectionData.pathClosed ?? false;

    const curveOptions = ["Linear", "Smooth", "Bezier", "Circular Arc"];
    const currentCurveIndex = curveOptions.indexOf(currentCurveType);

    const widthOptions = [];
    const widthMap = {};
    for (let w = 1; w <= MAX_PATH_WIDTH; w += 2) {
        widthOptions.push(`Width: ${w}`);
        widthMap[widthOptions.length - 1] = w;
    }
    let currentWidthIndex = widthOptions.findIndex((_v, index) => widthMap[index] === currentWidth);
    if (currentWidthIndex < 0) currentWidthIndex = 0;


    const form = new ModalFormData().title("Path Shape Settings")
        .dropdown("Curve Type", curveOptions, Math.max(0, currentCurveIndex))
        .slider("Smoothness (for Smooth/Bezier/Arc)", 2, 20, 1, currentSmoothness)
        .dropdown("Path Width", widthOptions, currentWidthIndex)
        .toggle("Closed Path (Loop)", currentClosed);

    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§ePath shape settings unchanged.");
            return;
        }

        const newCurveType = curveOptions[response.formValues[0]];
        const newSmoothness = response.formValues[1];
        const newWidth = widthMap[response.formValues[2]];
        const newClosed = response.formValues[3];

        if (newWidth === undefined || newWidth < 1) {
            player.sendMessage("§cInvalid path width selected.");
            return;
        }
        if (newClosed && (selectionData.pathPoints?.length ?? 0) < 3) {
            player.sendMessage("§cCannot enable 'Closed Path' with fewer than 3 points set.");
            // Note: The setting for 'closed' will still be saved, but drawing might fail or result in a line.
        }

        selectionData.pathCurveType = newCurveType;
        selectionData.pathSmoothness = newSmoothness;
        selectionData.pathWidth = newWidth;
        selectionData.pathClosed = newClosed;
        setPlayerSelection(player, selectionData);

        player.sendMessage(`§aPath Shape updated: Type=${newCurveType}${newCurveType === 'Smooth' || newCurveType === 'Bezier' || newCurveType === 'Circular Arc' ? `, Smoothness=${newSmoothness}` : ''}, Width=${newWidth}, Closed=${newClosed ? 'Yes' : 'No'}`);

    } catch (e) {
        console.error(`[Path Shape Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing path shape form.");
    }
}

// (showCreatePaletteFromSelectionForm is part of palette_menu.js and remains unchanged here)
/** Shows the form to configure and save a new palette based on analyzed selection blocks. */
async function showCreatePaletteFromSelectionForm(player, uniqueBlockIds) {
    if (!uniqueBlockIds || uniqueBlockIds.length === 0) {
        player.sendMessage("§cCannot create palette: No blocks provided for form.");
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
        return;
    }
    const form = new ModalFormData().title("Create Palette from Selection");
    form.textField("New Palette Name", "Enter a unique name", ""); // Index 0
    const formIndexToBlockId = {}; let currentIndex = 1;
    for (const blockId of uniqueBlockIds) {
        const simpleName = blockId.replace('minecraft:', '');
        form.toggle(`Include ${simpleName}?`, true);
        formIndexToBlockId[currentIndex] = { type: 'toggle', blockId: blockId };
        currentIndex++;
        form.slider(`Weight (${simpleName})`, 1, 100, 1, 1);
        formIndexToBlockId[currentIndex] = { type: 'slider', blockId: blockId };
        currentIndex++;
    }
    try {
        const response = await form.show(player);
        if (response.canceled || !response.formValues) {
            player.sendMessage("§ePalette creation cancelled.");
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        const formValues = response.formValues;
        const newPaletteName = formValues[0]?.trim();
        if (!newPaletteName) {
            player.sendMessage("§cPalette name cannot be empty. Not saved.");
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        const existingPalettes = getPlayerPalettes(player);
        if (existingPalettes[newPaletteName]) {
            player.sendMessage(`§cPalette named "${newPaletteName}" already exists. Not saved.`);
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        const newPaletteData = [];
        for (let i = 1; i < formValues.length; i++) {
            const mapping = formIndexToBlockId[i];
            if (mapping?.type === 'toggle') {
                const includeBlock = formValues[i];
                const sliderIndex = i + 1;
                const weightMapping = formIndexToBlockId[sliderIndex];
                if (includeBlock && weightMapping?.type === 'slider' && weightMapping?.blockId === mapping.blockId) {
                    const weight = formValues[sliderIndex];
                    newPaletteData.push({ id: mapping.blockId, weight: weight });
                }
                i++;
            }
        }
        if (newPaletteData.length === 0) {
            player.sendMessage("§cNo blocks were included in the palette. Not saved.");
            await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
            return;
        }
        if (addPlayerPalette(player, newPaletteName, newPaletteData)) {
            player.sendMessage(`§aPalette "${newPaletteName}" created and saved successfully.`);
        }
        // No explicit else needed as addPlayerPalette handles its own error messages
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
    } catch (e) {
        console.error(`[Create Palette from Selection Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError processing create palette form.");
        await system.run(async () => { if (player.isValid()) await showPaletteMenu(player); });
    }
}
