// scripts/ui/presets_menu.js

import { Player, system } from "@minecraft/server";
import { ActionFormData } from "@minecraft/server-ui";
import { presetRemoveBlocks, presetFlattenTerrain } from "../core/index.js";
import {
    TREE_BLOCK_ID_SET,
    LIQUID_BLOCK_ID_SET,
    FOLIAGE_BLOCK_ID_SET
} from "../constants/index.js";
import { showEditMenu } from "./main_menu.js"; // For back button

/**
 * Shows the menu for preset actions (Remove Trees, Flatten, etc.).
 * @param {Player} player The player accessing the menu.
 */
export async function showPresetsMenu(player) {
    const form = new ActionFormData()
        .title("Preset Actions")
        .body("Select a preset action to perform on the current selection.")
        .button("Remove Trees", "textures/items/sapling_oak")             // Index 0
        .button("Remove Liquids", "textures/items/bucket_water")          // Index 1
        .button("Remove Foliage", "textures/items/wheat_seeds")           // Index 2
        .button("Flatten Terrain (Top Layer)", "textures/items/iron_shovel") // Index 3
        .button("Back", "textures/ui/cancel");                             // Index 4

    try {
        const response = await form.show(player);
        if (response.canceled || response.selection === undefined || response.selection === 4) {
            // Go back to edit menu if cancelled or 'Back' is pressed
            await system.run(async () => { if (player.isValid()) await showEditMenu(player); });
            return;
        }

        // Selection completeness check is done within the core preset functions

        let refreshMenu = true; // Refresh this menu after action by default

        switch (response.selection) {
            case 0: // Remove Trees
                await presetRemoveBlocks(player, TREE_BLOCK_ID_SET, "Tree");
                break;
            case 1: // Remove Liquids
                await presetRemoveBlocks(player, LIQUID_BLOCK_ID_SET, "Liquid");
                break;
            case 2: // Remove Foliage
                await presetRemoveBlocks(player, FOLIAGE_BLOCK_ID_SET, "Foliage");
                break;
            case 3: // Flatten Terrain
                await presetFlattenTerrain(player); // This function includes its own UI
                break;
            // Case 4 (Back) handled above
        }

        // After executing a preset, refresh this menu to allow another action
        // Use system.run to avoid potential UI stack issues
        if (refreshMenu) {
            await system.run(async () => { if (player.isValid()) await showPresetsMenu(player); });
        }

    } catch (e) {
        console.error(`[WorldEdit Presets Menu] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError displaying or processing presets menu.");
        // Attempt to go back to edit menu on error
        await system.run(async () => { if (player.isValid()) await showEditMenu(player); });
    }
}