// scripts/ui/terrain_menu.js

import { Player, system, world } from "@minecraft/server";
import { ModalFormData } from "@minecraft/server-ui";
import {
    getPlayerSelection,
    setPlayerSelection,
    getPlayerPalettes
} from "../state/index.js";
import {
    DEFAULT_TERRAIN_BASE_MATERIAL,
    DEFAULT_TERRAIN_SURFACE_MATERIAL,
    DEFAULT_TERRAIN_SURFACE_DEPTH,
    DEFAULT_TERRAIN_NOISE_TYPE,
    DEFAULT_TERRAIN_FREQUENCY,
    DEFAULT_TERRAIN_AMPLITUDE,
    DEFAULT_TERRAIN_SEED_TYPE,
    DEFAULT_TERRAIN_OCTAVES,
    DEFAULT_TERRAIN_LACUNARITY,
    DEFAULT_TERRAIN_PERSISTENCE,
    DEFAULT_TERRAIN_WARP_ENABLED,
    DEFAULT_TERRAIN_WARP_AMPLITUDE,
    DEFAULT_TERRAIN_WARP_FREQUENCY,
    DEFAULT_TERRAIN_JITTER
} from "../constants/index.js";
import { generateTerrain } from "../core/index.js";
import { getSingleBlockPermutation, validateBlockIds } from "../utils/validation.js";
import { parseWeightedBlockInput } from "../utils/parsing.js";
import { buildSourceDropdownOptions } from "./brush_menu.js";

/**
 * Shows the ModalForm to get parameters for Terrain Generation.
 * @param {Player} player The player initiating the action.
 */
export async function showTerrainGeneratorForm(player) {
    const selectionData = getPlayerSelection(player);

    const currentBaseMatSource = selectionData.terrainBaseMaterialSource ?? 'manual';
    const currentManualBaseMat = selectionData.terrainManualBaseMaterial ?? DEFAULT_TERRAIN_BASE_MATERIAL;
    const currentSurfaceMatSource = selectionData.terrainSurfaceMaterialSource ?? 'manual';
    const currentManualSurfaceMat = selectionData.terrainManualSurfaceMaterial ?? DEFAULT_TERRAIN_SURFACE_MATERIAL;

    const currentSurfaceDepth = selectionData.terrainSurfaceDepth ?? DEFAULT_TERRAIN_SURFACE_DEPTH;
    const currentNoiseType = selectionData.terrainNoiseType ?? DEFAULT_TERRAIN_NOISE_TYPE;
    const currentFrequency = selectionData.terrainFrequency ?? DEFAULT_TERRAIN_FREQUENCY;
    const currentAmplitude = selectionData.terrainAmplitude ?? DEFAULT_TERRAIN_AMPLITUDE;
    const currentSeedType = selectionData.terrainSeedType ?? DEFAULT_TERRAIN_SEED_TYPE;
    const currentCustomSeed = selectionData.terrainCustomSeed ?? "";
    const currentOctaves = selectionData.terrainOctaves ?? DEFAULT_TERRAIN_OCTAVES;
    const currentLacunarity = selectionData.terrainLacunarity ?? DEFAULT_TERRAIN_LACUNARITY;
    const currentPersistence = selectionData.terrainPersistence ?? DEFAULT_TERRAIN_PERSISTENCE;
    const currentWarpEnabled = selectionData.terrainDomainWarpEnabled ?? DEFAULT_TERRAIN_WARP_ENABLED;
    const currentWarpAmplitude = selectionData.terrainWarpAmplitude ?? DEFAULT_TERRAIN_WARP_AMPLITUDE;
    const currentWarpFrequency = selectionData.terrainWarpFrequency ?? DEFAULT_TERRAIN_WARP_FREQUENCY;
    const currentJitter = selectionData.terrainJitter ?? DEFAULT_TERRAIN_JITTER;

    const seedOptions = ["Random", "World Seed", "Custom"];
    const currentSeedIndex = seedOptions.indexOf(currentSeedType);
    const noiseTypeOptions = ["Simplex", "FBM", "Worley"];
    const currentNoiseTypeIndex = noiseTypeOptions.indexOf(currentNoiseType);

    // Terrain generator uses 'manual', 'active', or 'saved palette'
    const {
        sourceOptions: baseSourceOptions,
        sourceValueMap: baseSourceValueMap,
        currentSourceIndex: currentBaseSourceIndex
    } = buildSourceDropdownOptions(player, currentBaseMatSource, false, true);

    const {
        sourceOptions: surfaceSourceOptions,
        sourceValueMap: surfaceSourceValueMap,
        currentSourceIndex: currentSurfaceSourceIndex
    } = buildSourceDropdownOptions(player, currentSurfaceMatSource, false, true);

    const form = new ModalFormData()
        .title("Generate Terrain")
        .dropdown("Base Material Source", baseSourceOptions, currentBaseSourceIndex)
        .textField("Manual Base Material (id:wt,...)", "e.g. stone, deepslate:3", currentManualBaseMat)
        .dropdown("Surface Material Source", surfaceSourceOptions, currentSurfaceSourceIndex)
        .textField("Manual Surface Material (id:wt,...)", "e.g. grass_block, dirt:2", currentManualSurfaceMat)
        .slider("Surface Layer Depth", 1, 10, 1, currentSurfaceDepth)
        .dropdown("Noise Type", noiseTypeOptions, Math.max(0, currentNoiseTypeIndex))
        .slider("Amplitude (Height)", 1, 64, 1, currentAmplitude)
        .textField("Frequency (Scale)", "Smaller = larger features (e.g., 0.02)", String(currentFrequency))
        .dropdown("Seed Source", seedOptions, Math.max(0, currentSeedIndex))
        .textField("Custom Seed (if selected)", "Enter number for custom seed", String(currentCustomSeed))
        .slider("Octaves (for FBM)", 1, 8, 1, currentOctaves)
        .textField("Lacunarity (FBM Freq Mult)", "Usually 2.0", String(currentLacunarity))
        .textField("Persistence (FBM Amp Mult)", "Usually 0.5 (Gain)", String(currentPersistence))
        .toggle("Enable Domain Warp (Not for Worley)", currentWarpEnabled)
        .textField("Warp Amplitude", "Strength of coordinate distortion", String(currentWarpAmplitude))
        .textField("Warp Frequency", "Scale of warp distortion", String(currentWarpFrequency))
        .slider("Worley Jitter (Cell Randomness)", 0.0, 1.0, 0.1, currentJitter);

    try {
        const response = await form.show(player);

        if (response.canceled || !response.formValues) {
            player.sendMessage("§eTerrain Generation cancelled.");
            return;
        }

        const formVals = response.formValues;
        let index = 0;
        const newBaseMatSourceLabel = baseSourceOptions[formVals[index++]];
        const newManualBaseMatInput = formVals[index++].trim();
        const newSurfaceMatSourceLabel = surfaceSourceOptions[formVals[index++]];
        const newManualSurfaceMatInput = formVals[index++].trim();
        const newSurfaceDepth = formVals[index++];
        const newNoiseType = noiseTypeOptions[formVals[index++]];
        const newAmplitude = formVals[index++];
        const newFrequencyStr = formVals[index++].trim();
        const newSeedType = seedOptions[formVals[index++]];
        const newCustomSeedStr = formVals[index++].trim();
        const newOctaves = formVals[index++];
        const newLacunarityStr = formVals[index++].trim();
        const newPersistenceStr = formVals[index++].trim();
        const newWarpEnabled = formVals[index++];
        const newWarpAmplitudeStr = formVals[index++].trim();
        const newWarpFrequencyStr = formVals[index++].trim();
        const newJitter = formVals[index++];

        const finalBaseMatSource = baseSourceValueMap[newBaseMatSourceLabel] ?? 'manual';
        const finalSurfaceMatSource = surfaceSourceValueMap[newSurfaceMatSourceLabel] ?? 'manual';

        let validationErrors = [];
        let finalFrequency = parseFloat(newFrequencyStr);
        let finalCustomSeed = parseInt(newCustomSeedStr, 10);
        let finalLacunarity = parseFloat(newLacunarityStr);
        let finalPersistence = parseFloat(newPersistenceStr);
        let finalWarpAmplitude = parseFloat(newWarpAmplitudeStr);
        let finalWarpFrequency = parseFloat(newWarpFrequencyStr);
        let finalJitter = typeof newJitter === 'number' ? newJitter : DEFAULT_TERRAIN_JITTER;

        let baseMaterialParam;
        if (finalBaseMatSource === 'manual') {
            if (!newManualBaseMatInput) validationErrors.push("Manual Base Material cannot be empty.");
            else {
                const parsed = parseWeightedBlockInput(newManualBaseMatInput);
                if (parsed.errors.length > 0) validationErrors.push(`Base Material: ${parsed.errors.join(', ')}`);
                else if (parsed.blocks.length === 0) validationErrors.push("No valid blocks in Manual Base Material.");
                else baseMaterialParam = parsed.blocks;
            }
        } else if (finalBaseMatSource === 'active') { // Changed from 'working'
            if (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0) validationErrors.push("Active Palette (Base) is empty."); // Changed
            else baseMaterialParam = selectionData.activePaletteData; // Changed
        } else {
            const palettes = getPlayerPalettes(player);
            if (!palettes[finalBaseMatSource] || palettes[finalBaseMatSource].length === 0) validationErrors.push(`Saved Palette "${finalBaseMatSource}" (Base) is empty or not found.`);
            else baseMaterialParam = palettes[finalBaseMatSource];
        }

        let surfaceMaterialParam;
        if (finalSurfaceMatSource === 'manual') {
            if (!newManualSurfaceMatInput) validationErrors.push("Manual Surface Material cannot be empty.");
            else {
                const parsed = parseWeightedBlockInput(newManualSurfaceMatInput);
                if (parsed.errors.length > 0) validationErrors.push(`Surface Material: ${parsed.errors.join(', ')}`);
                else if (parsed.blocks.length === 0) validationErrors.push("No valid blocks in Manual Surface Material.");
                else surfaceMaterialParam = parsed.blocks;
            }
        } else if (finalSurfaceMatSource === 'active') { // Changed from 'working'
            if (!selectionData.activePaletteData || selectionData.activePaletteData.length === 0) validationErrors.push("Active Palette (Surface) is empty."); // Changed
            else surfaceMaterialParam = selectionData.activePaletteData; // Changed
        } else {
            const palettes = getPlayerPalettes(player);
            if (!palettes[finalSurfaceMatSource] || palettes[finalSurfaceMatSource].length === 0) validationErrors.push(`Saved Palette "${finalSurfaceMatSource}" (Surface) is empty or not found.`);
            else surfaceMaterialParam = palettes[finalSurfaceMatSource];
        }

        if (isNaN(finalFrequency) || finalFrequency <= 0) validationErrors.push("Frequency must be a positive number.");
        if (newSeedType === "Custom" && isNaN(finalCustomSeed)) validationErrors.push("Custom Seed must be a valid number when 'Custom' source is selected.");
        if (isNaN(finalLacunarity) || finalLacunarity < 1.1) validationErrors.push("Lacunarity must be >= 1.1.");
        if (isNaN(finalPersistence) || finalPersistence <= 0 || finalPersistence > 1.0) validationErrors.push("Persistence must be between 0.0 (exclusive) and 1.0 (inclusive).");
        if (isNaN(finalWarpAmplitude) || finalWarpAmplitude < 0) validationErrors.push("Warp Amplitude must be a non-negative number.");
        if (isNaN(finalWarpFrequency) || finalWarpFrequency <= 0) validationErrors.push("Warp Frequency must be a positive number.");

        if (validationErrors.length > 0) {
            player.sendMessage("§cInvalid input:\n" + validationErrors.join("\n") + "\n§eTerrain generation cancelled.");
            return;
        }

        let finalSeed;
        if (newSeedType === "World Seed") {
            finalSeed = Math.floor(world.getAbsoluteTime());
        } else if (newSeedType === "Custom") {
            finalSeed = finalCustomSeed;
        } else {
            finalSeed = Date.now() + Math.floor(Math.random() * 10000);
        }
        // Message about seed type is sent by core function now.

        selectionData.terrainBaseMaterialSource = finalBaseMatSource;
        selectionData.terrainManualBaseMaterial = (finalBaseMatSource === 'manual') ? newManualBaseMatInput : DEFAULT_TERRAIN_BASE_MATERIAL;
        selectionData.terrainSurfaceMaterialSource = finalSurfaceMatSource;
        selectionData.terrainManualSurfaceMaterial = (finalSurfaceMatSource === 'manual') ? newManualSurfaceMatInput : DEFAULT_TERRAIN_SURFACE_MATERIAL;

        selectionData.terrainSurfaceDepth = newSurfaceDepth;
        selectionData.terrainNoiseType = newNoiseType;
        selectionData.terrainFrequency = finalFrequency;
        selectionData.terrainAmplitude = newAmplitude;
        selectionData.terrainSeedType = newSeedType;
        selectionData.terrainCustomSeed = (newSeedType === "Custom") ? finalCustomSeed : undefined;
        selectionData.terrainOctaves = newOctaves;
        selectionData.terrainLacunarity = finalLacunarity;
        selectionData.terrainPersistence = finalPersistence;
        selectionData.terrainDomainWarpEnabled = newWarpEnabled;
        selectionData.terrainWarpAmplitude = finalWarpAmplitude;
        selectionData.terrainWarpFrequency = finalWarpFrequency;
        selectionData.terrainJitter = finalJitter;
        setPlayerSelection(player, selectionData);

        const parameters = {
            baseMaterialParam: baseMaterialParam,
            surfaceMaterialParam: surfaceMaterialParam,
            surfaceDepth: newSurfaceDepth,
            noiseType: newNoiseType,
            frequency: finalFrequency,
            amplitude: newAmplitude,
            seed: finalSeed,
            octaves: newOctaves,
            lacunarity: finalLacunarity,
            persistence: finalPersistence,
            warpEnabled: newWarpEnabled,
            warpAmplitude: finalWarpAmplitude,
            warpFrequency: finalWarpFrequency,
            jitter: finalJitter
        };

        system.run(async () => {
            if (player.isValid()) {
                await generateTerrain(player, parameters);
            }
        });

    } catch (e) {
        console.error(`[Terrain Generator Form] Error: ${e}\n${e.stack}`);
        player.sendMessage("§cError displaying or processing terrain generator form.");
    }
}
