// scripts/utils/algorithms.js

/**
 * Calculates all integer block coordinates along a 3D line between two points
 * using a 3D adaptation of Bresenham's line algorithm.
 * @param {{x: number, y: number, z: number}} start The starting coordinate object.
 * @param {{x: number, y: number, z: number}} end The ending coordinate object.
 * @returns {Array<{x: number, y: number, z: number}>} An array of coordinate objects representing the blocks on the line.
 */
export function getLinePoints3D(start, end) {
    const points = [];
    let x1 = Math.floor(start.x);
    let y1 = Math.floor(start.y);
    let z1 = Math.floor(start.z);
    const x2 = Math.floor(end.x);
    const y2 = Math.floor(end.y);
    const z2 = Math.floor(end.z);

    const dx = Math.abs(x2 - x1);
    const dy = Math.abs(y2 - y1);
    const dz = Math.abs(z2 - z1);
    const sx = x1 < x2 ? 1 : -1;
    const sy = y1 < y2 ? 1 : -1;
    const sz = z1 < z2 ? 1 : -1;

    points.push({ x: x1, y: y1, z: z1 }); // Add start point

    // Driving axis is axis with largest delta
    if (dx >= dy && dx >= dz) { // X is dominant
        let err1 = 2 * dy - dx;
        let err2 = 2 * dz - dx;
        while (x1 !== x2) {
            x1 += sx;
            if (err1 >= 0) { y1 += sy; err1 -= 2 * dx; }
            if (err2 >= 0) { z1 += sz; err2 -= 2 * dx; }
            err1 += 2 * dy;
            err2 += 2 * dz;
            points.push({ x: x1, y: y1, z: z1 });
        }
    } else if (dy >= dx && dy >= dz) { // Y is dominant
        let err1 = 2 * dx - dy;
        let err2 = 2 * dz - dy;
        while (y1 !== y2) {
            y1 += sy;
            if (err1 >= 0) { x1 += sx; err1 -= 2 * dy; }
            if (err2 >= 0) { z1 += sz; err2 -= 2 * dy; }
            err1 += 2 * dx;
            err2 += 2 * dz;
            points.push({ x: x1, y: y1, z: z1 });
        }
    } else { // Z is dominant
        let err1 = 2 * dy - dz;
        let err2 = 2 * dx - dz;
        while (z1 !== z2) {
            z1 += sz;
            if (err1 >= 0) { y1 += sy; err1 -= 2 * dz; }
            if (err2 >= 0) { x1 += sx; err2 -= 2 * dz; }
            err1 += 2 * dy;
            err2 += 2 * dx;
            points.push({ x: x1, y: y1, z: z1 });
        }
    }
    return points;
}


/**
 * Calculates points along a Catmull-Rom spline segment using the standard formula.
 * Generates 'smoothness' number of points *between* p1 and p2.
 * @param {{x: number, y: number, z: number}} p0 Control point before the segment start.
 * @param {{x: number, y: number, z: number}} p1 Start point of the segment.
 * @param {{x: number, y: number, z: number}} p2 End point of the segment.
 * @param {{x: number, y: number, z: number}} p3 Control point after the segment end.
 * @param {number} smoothness Number of intermediate points to calculate between p1 and p2. Higher is smoother. Must be >= 1.
 * @returns {Array<{x: number, y: number, z: number}>} An array of points along the spline segment, including p1 and p2.
 */
export function getCatmullRomSplinePoints(p0, p1, p2, p3, smoothness) {
    const points = [];
    const numSteps = Math.max(1, Math.floor(smoothness)); // Ensure at least 1 step

    for (let i = 0; i <= numSteps; i++) { // Include t=0 (p1) and t=1 (p2)
        const t = i / numSteps; // Parameter t ranges from 0 to 1.0
        const t2 = t * t;
        const t3 = t2 * t;

        // Catmull-Rom spline formula for each coordinate
        const x = 0.5 * ((2 * p1.x) +
            (-p0.x + p2.x) * t +
            (2 * p0.x - 5 * p1.x + 4 * p2.x - p3.x) * t2 +
            (-p0.x + 3 * p1.x - 3 * p2.x + p3.x) * t3);
        const y = 0.5 * ((2 * p1.y) +
            (-p0.y + p2.y) * t +
            (2 * p0.y - 5 * p1.y + 4 * p2.y - p3.y) * t2 +
            (-p0.y + 3 * p1.y - 3 * p2.y + p3.y) * t3);
        const z = 0.5 * ((2 * p1.z) +
            (-p0.z + p2.z) * t +
            (2 * p0.z - 5 * p1.z + 4 * p2.z - p3.z) * t2 +
            (-p0.z + 3 * p1.z - 3 * p2.z + p3.z) * t3);

        points.push({ x, y, z });
    }
    return points;
}


// --- NEW Bezier Curve Implementation ---
/**
 * Calculates points along a cubic Bezier curve segment defined by 4 points.
 * @param {{x: number, y: number, z: number}} p0 Start point.
 * @param {{x: number, y: number, z: number}} p1 Control point 1.
 * @param {{x: number, y: number, z: number}} p2 Control point 2.
 * @param {{x: number, y: number, z: number}} p3 End point.
 * @param {number} smoothness Number of intermediate points to calculate. Higher is smoother. Must be >= 1.
 * @returns {Array<{x: number, y: number, z: number}>} An array of points along the Bezier curve segment, including start and end points.
 */
export function getBezierCurvePoints(p0, p1, p2, p3, smoothness) {
    const points = [];
    const numSteps = Math.max(1, Math.floor(smoothness)); // Ensure at least 1 step

    for (let i = 0; i <= numSteps; i++) {
        const t = i / numSteps; // Parameter t ranges from 0 to 1.0
        const u = 1 - t;
        const tt = t * t;
        const uu = u * u;
        const uuu = uu * u;
        const ttt = tt * t;

        // Cubic Bezier formula: B(t) = (1-t)^3*P0 + 3(1-t)^2*t*P1 + 3(1-t)*t^2*P2 + t^3*P3
        const p = { x: 0, y: 0, z: 0 };
        p.x = uuu * p0.x + 3 * uu * t * p1.x + 3 * u * tt * p2.x + ttt * p3.x;
        p.y = uuu * p0.y + 3 * uu * t * p1.y + 3 * u * tt * p2.y + ttt * p3.y;
        p.z = uuu * p0.z + 3 * uu * t * p1.z + 3 * u * tt * p2.z + ttt * p3.z;

        points.push(p);
    }
    return points;
}


// --- NEW Circular Arc Implementation (3-Point Method) ---
/**
 * Calculates points along a circular arc segment defined by three points (start, mid, end).
 * Handles collinear points by returning a straight line.
 * @param {{x: number, y: number, z: number}} p1 Start point of the arc.
 * @param {{x: number, y: number, z: number}} p2 A point the arc passes through.
 * @param {{x: number, y: number, z: number}} p3 End point of the arc.
 * @param {number} smoothness Number of line segments to approximate the arc. Higher is smoother. Must be >= 1.
 * @returns {Array<{x: number, y: number, z: number}>} An array of points approximating the circular arc, including start and end points.
 */
export function getCircularArcPoints(p1, p2, p3, smoothness) {
    const numSegments = Math.max(1, Math.floor(smoothness));

    // --- Vector Math Helpers ---
    const vecSub = (a, b) => ({ x: a.x - b.x, y: a.y - b.y, z: a.z - b.z });
    const vecAdd = (a, b) => ({ x: a.x + b.x, y: a.y + b.y, z: a.z + b.z });
    const vecScale = (v, s) => ({ x: v.x * s, y: v.y * s, z: v.z * s });
    const vecDot = (a, b) => a.x * b.x + a.y * b.y + a.z * b.z;
    const vecCross = (a, b) => ({
        x: a.y * b.z - a.z * b.y,
        y: a.z * b.x - a.x * b.z,
        z: a.x * b.y - a.y * b.x
    });
    const vecMagSq = (v) => v.x * v.x + v.y * v.y + v.z * v.z;
    const vecMag = (v) => Math.sqrt(vecMagSq(v));
    const vecNormalize = (v) => { const m = vecMag(v); return m === 0 ? { x: 0, y: 0, z: 0 } : vecScale(v, 1 / m); };

    // --- Calculate Circle Center and Radius ---
    const v12 = vecSub(p2, p1);
    const v13 = vecSub(p3, p1);

    const crossProd = vecCross(v12, v13);
    const crossMagSq = vecMagSq(crossProd);
    const epsilon = 1e-8; // Small tolerance for floating point errors

    // Check for collinearity (cross product magnitude is near zero)
    if (crossMagSq < epsilon) {
        console.warn("Circular Arc points are collinear. Drawing straight line segment.");
        // Return a straight line from p1 to p3 (or maybe p1 to p2?) Let's do p1 to p3.
        return getLinePoints3D(p1, p3);
    }

    // Plane normal
    const planeNormal = vecNormalize(crossProd);

    // Use formula for circumcenter C of triangle P1,P2,P3 (derived from vector projections)
    // C = P1 + alpha * v12 + beta * v13
    // Solving leads to:
    const v12MagSq = vecMagSq(v12);
    const v13MagSq = vecMagSq(v13);
    const dot12_13 = vecDot(v12, v13);

    const denom = 2 * (v12MagSq * v13MagSq - dot12_13 * dot12_13); // 2 * |v12 x v13|^2
    if (Math.abs(denom) < epsilon) { // Should be caught by crossMagSq check, but good safety
        console.warn("Circular Arc denominator near zero. Drawing straight line segment.");
        return getLinePoints3D(p1, p3);
    }

    const alpha = (v13MagSq * (v12MagSq - dot12_13)) / denom;
    const beta = (v12MagSq * (v13MagSq - dot12_13)) / denom;

    const center = vecAdd(p1, vecAdd(vecScale(v12, alpha), vecScale(v13, beta)));
    const radius = vecMag(vecSub(p1, center));

    // --- Calculate Angles ---
    // Project points onto the plane defined by the center and two orthogonal vectors within the plane
    const basisX = vecNormalize(vecSub(p1, center)); // Vector from center to p1
    const basisY = vecNormalize(vecCross(planeNormal, basisX)); // Vector orthogonal to basisX and normal

    // Find angles of p1, p2, p3 relative to the center using the basis vectors
    const angle1 = 0; // p1 is on basisX by definition
    const vC2 = vecSub(p2, center);
    const angle2 = Math.atan2(vecDot(vC2, basisY), vecDot(vC2, basisX));
    const vC3 = vecSub(p3, center);
    const angle3 = Math.atan2(vecDot(vC3, basisY), vecDot(vC3, basisX));

    // Determine the sweep angle and direction
    let startAngle = angle1;
    let endAngle = angle3;
    let totalAngle = endAngle - startAngle;

    // Adjust angles to ensure the arc passes through angle2
    // Normalize angles to [0, 2PI)
    const twoPi = 2 * Math.PI;
    let normAngle2 = (angle2 % twoPi + twoPi) % twoPi;
    let normAngle3 = (angle3 % twoPi + twoPi) % twoPi;

    // If p2 is "behind" p3 relative to p1, sweep the long way
    if (normAngle2 > normAngle3) {
        totalAngle = twoPi - (normAngle2 - normAngle3); // Should this be angle3 + 2pi - angle1?
        // Let's recalculate totalAngle based on normalized values
        if (normAngle3 < normAngle2) { // If angle3 is smaller than angle2 (e.g., angle1=0, angle2=3pi/2, angle3=pi/2)
            endAngle += twoPi; // Make angle3 larger than angle1
        }
        totalAngle = endAngle - startAngle;
    } else {
        totalAngle = normAngle3; // Since startAngle is 0
    }
    // Ensure totalAngle has the correct sign based on cross product direction?
    // Let's assume shorter arc for now unless angle2 forces the longer one.

    // --- Generate Points along the Arc ---
    const points = [];
    for (let i = 0; i <= numSegments; i++) {
        const t = i / numSegments;
        const currentAngle = startAngle + t * totalAngle;

        // Calculate 3D point on the circle using center, radius, and basis vectors
        const pt_3D = vecAdd(center, vecAdd(vecScale(basisX, radius * Math.cos(currentAngle)), vecScale(basisY, radius * Math.sin(currentAngle))));
        points.push(pt_3D);
    }

    return points;
}
