// scripts/utils/history_helpers.js

import { Dimension, Player, BlockPermutation } from "@minecraft/server"; // Import types for JSDoc/clarity
// import { MAX_UNDO_HISTORY } from "../constants/settings.js"; // <<< REMOVE: Get from global settings
import {
    getPlayerUndoStack,
    setPlayerUndoStack,
    getPlayerRedoStack,
    setPlayerRedoStack,
    getGlobalSettings // <<< ADD: Import global settings getter
} from "../state/index.js"; // Import state functions

/**
 * Gets block data (location and permutation) for all blocks within a bounding box.
 * Handles potential unloaded chunk errors gracefully.
 * @param {Dimension} dimension The dimension to get blocks from.
 * @param {{x:number, y:number, z:number}} min The minimum corner of the bounding box.
 * @param {{x:number, y:number, z:number}} max The maximum corner of the bounding box.
 * @returns {Array<{location: {x:number, y:number, z:number}, permutation: BlockPermutation}>} An array of block state objects.
 */
export function getBlocksInArea(dimension, min, max) {
    const blocks = [];
    // Basic validation of bounds
    if (!min || !max || min.x > max.x || min.y > max.y || min.z > max.z) {
        console.warn("[getBlocksInArea] Invalid bounds provided:", min, max);
        return blocks;
    }

    try {
        for (let x = min.x; x <= max.x; x++) {
            for (let y = min.y; y <= max.y; y++) {
                for (let z = min.z; z <= max.z; z++) {
                    const location = { x, y, z };
                    try {
                        // Use getBlock for potential unloaded chunk safety - it returns undefined if unloaded
                        const block = dimension.getBlock(location);
                        if (block) {
                            // Store the location and permutation
                            blocks.push({ location: { ...location }, permutation: block.permutation });
                        }
                        // If block is undefined (unloaded), it's simply not added to the array
                    } catch (e) {
                        // Catch any unexpected errors from getBlock, though it usually returns undefined
                        console.warn(`[getBlocksInArea] Unexpected error getting block at ${x},${y},${z}: ${e}`);
                    }
                }
            }
        }
    } catch (e) {
        // Catch potential errors during the iteration itself (less likely)
        console.error(`[WorldEdit Utils] Error during getBlocksInArea iteration: ${e}`);
    }
    return blocks;
}

/**
 * Restores blocks in the world based on a saved HistoryState object.
 * @param {Player} player The player performing the action (used for dimension access).
 * @param {import("../state/player_state.js").HistoryState} historyState The state object containing blocks to restore.
 */
export function restoreBlocksFromState(player, historyState) {
    if (!historyState || !historyState.blocks) {
        console.warn("[restoreBlocksFromState] Invalid or missing history state/blocks.");
        return;
    }
    const dimension = player.dimension;
    try {
        for (const blockState of historyState.blocks) {
            try {
                // Check if block exists before trying to set permutation
                const block = dimension.getBlock(blockState.location);
                if (block) {
                    block.setPermutation(blockState.permutation);
                } else {
                    // Optional: Log if a block couldn't be found (e.g., area partially unloaded during restore?)
                    // console.warn(`[restoreBlocksFromState] Block not found at ${blockState.location.x},${blockState.location.y},${blockState.location.z}`);
                }
            } catch (e) {
                // Ignore single block errors (e.g., invalid permutation somehow?)
                // console.warn(`[restoreBlocksFromState] Error setting block: ${e}`);
            }
        }
    } catch (e) {
        // Catch errors iterating through the block states
        console.error(`[WorldEdit Utils] Error in restoreBlocksFromState loop: ${e}`);
        player.sendMessage("§cAn error occurred while restoring blocks.");
    }
}

/**
 * Pushes a new HistoryState onto the player's undo stack, manages stack size, and clears the redo stack.
 * @param {Player} player The player performing the action.
 * @param {import("../state/player_state.js").HistoryState} historyState The state object to add to the undo history.
 */
export function pushUndoState(player, historyState) {
    const undoStack = getPlayerUndoStack(player);

    // Validate the state object before pushing
    if (!historyState || !historyState.affectedArea || !historyState.blocks) {
        console.warn("[pushUndoState] Attempted to push invalid history state object.");
        return;
    }
    // Allow pushing states with empty blocks array (needed for redo capture after deletion)

    undoStack.push(historyState);

    // <<< Get max history from global settings >>>
    const globalSettings = getGlobalSettings();
    const maxHistory = globalSettings.maxUndoHistory;

    // Limit history size by removing the oldest entry if necessary
    while (undoStack.length > maxHistory && maxHistory > 0) { // Only trim if maxHistory > 0
        undoStack.shift();
    }
    // If maxHistory is 0, don't limit the stack (effectively infinite undo, though memory is a concern)
    // Or alternatively, if maxHistory is 0, don't push anything? Let's allow infinite for now.

    // Save the modified undo stack
    setPlayerUndoStack(player, undoStack);

    // Clear the redo stack whenever a new action is performed and added to undo
    setPlayerRedoStack(player, []);
}
