// scripts/utils/location.js

/**
 * Calculates the volume (number of blocks) of a cuboid defined by two points.
 * @param {{x:number, y:number, z:number}} posA First corner.
 * @param {{x:number, y:number, z:number}} posB Second corner.
 * @returns {number} The volume, or 0 if input is invalid.
 */
export function calculateVolume(posA, posB) {
    if (!posA || !posB) return 0;
    const min = getMinCorner(posA, posB);
    const max = getMaxCorner(posA, posB);
    // Add 1 to each dimension difference because the coordinates are inclusive
    return (max.x - min.x + 1) * (max.y - min.y + 1) * (max.z - min.z + 1);
}

/**
 * Gets the minimum corner (lowest coordinates on all axes) of a cuboid defined by two points.
 * @param {{x:number, y:number, z:number}} posA First corner.
 * @param {{x:number, y:number, z:number}} posB Second corner.
 * @returns {{x:number, y:number, z:number}} The minimum corner coordinates.
 */
export function getMinCorner(posA, posB) {
    return {
        x: Math.min(posA.x, posB.x),
        y: Math.min(posA.y, posB.y),
        z: Math.min(posA.z, posB.z)
    };
}

/**
 * Gets the maximum corner (highest coordinates on all axes) of a cuboid defined by two points.
 * @param {{x:number, y:number, z:number}} posA First corner.
 * @param {{x:number, y:number, z:number}} posB Second corner.
 * @returns {{x:number, y:number, z:number}} The maximum corner coordinates.
 */
export function getMaxCorner(posA, posB) {
    return {
        x: Math.max(posA.x, posB.x),
        y: Math.max(posA.y, posB.y),
        z: Math.max(posA.z, posB.z)
    };
}

/**
 * Gets the cardinal direction (North, South, East, West) the player is primarily facing based on their view direction vector.
 * @param {{x: number, y: number, z: number}} viewDirection Player's view direction vector.
 * @returns {"North" | "South" | "East" | "West"} The cardinal direction string.
 */
export function getPlayerDirection(viewDirection) {
    // Compare the absolute values of X and Z components
    if (Math.abs(viewDirection.x) > Math.abs(viewDirection.z)) {
        // Primarily facing East or West
        return viewDirection.x > 0 ? "East" : "West";
    } else {
        // Primarily facing North or South (includes cases where |x| == |z|)
        return viewDirection.z > 0 ? "South" : "North";
    }
}

/**
 * Rotates a relative position vector within a cuboid of a given size around the Y axis.
 * Assumes rotation center is the geometric center of the cuboid base (X, Z).
 * @param {{x:number, y:number, z:number}} relativePos The position relative to the cuboid's minimum corner.
 * @param {{x:number, y:number, z:number}} size The size of the cuboid (used to find center).
 * @param {0 | 90 | 180 | 270} rotationDegrees The rotation angle in degrees.
 * @returns {{x:number, y:number, z:number}} The new relative position after rotation, rounded to nearest integer coordinates.
 */
export function rotateRelativePos(relativePos, size, rotationDegrees) {
    // No rotation needed
    if (rotationDegrees === 0) return { ...relativePos };

    // Calculate the center of the cuboid's base (XZ plane)
    // Subtract 1 because size is count, indices are 0 to size-1
    const centerX = (size.x - 1) / 2;
    const centerZ = (size.z - 1) / 2;

    // Calculate the position relative to the center
    const dx = relativePos.x - centerX;
    const dz = relativePos.z - centerZ;

    let rotatedX = relativePos.x;
    let rotatedZ = relativePos.z;

    // Apply rotation formula based on degrees (counter-clockwise)
    switch (rotationDegrees) {
        case 90: // Rotate West view -> North view (90 deg counter-clockwise)
            rotatedX = -dz + centerX; // New X is -dZ relative to center
            rotatedZ = dx + centerZ;  // New Z is dX relative to center
            break;
        case 180: // Rotate South view -> North view (180 deg)
            rotatedX = -dx + centerX; // New X is -dX
            rotatedZ = -dz + centerZ; // New Z is -dZ
            break;
        case 270: // Rotate East view -> North view (270 deg counter-clockwise / -90 deg)
            rotatedX = dz + centerX;  // New X is dZ
            rotatedZ = -dx + centerZ; // New Z is -dX
            break;
        // case 0 handled above
    }

    // Round results to get integer relative coordinates, as block positions are integers
    return { x: Math.round(rotatedX), y: relativePos.y, z: Math.round(rotatedZ) };
}