// scripts/utils/messaging.js

import { Player } from "@minecraft/server";
import { getPlayerSelection } from "../state/player_state.js";

/**
 * Sends a message to the player if their 'showActionFeedbackMessages' setting is true,
 * or if the message is forced.
 * @param {Player} player The player to send the message to.
 * @param {string} message The message string to send.
 * @param {boolean} [forceShow=false] If true, the message will be shown regardless of player settings. Useful for errors or critical instructions.
 */
export function sendPlayerMessage(player, message, forceShow = false) {
    if (!player || !player.isValid()) {
        // console.warn("[sendPlayerMessage] Attempted to send message to invalid player.");
        return;
    }

    try {
        const selectionData = getPlayerSelection(player);
        const showFeedback = selectionData.showActionFeedbackMessages ?? false; // Default to false if undefined

        if (showFeedback || forceShow) {
            player.sendMessage(message);
        }
    } catch (e) {
        // Fallback or log error if sending message fails for some reason
        // (though player.sendMessage itself is usually quite robust if player is valid)
        console.warn(`[sendPlayerMessage] Error sending message to ${player.name}: ${e}`);
        // As a last resort, if forceShow was true, try to log it so info isn't lost.
        if (forceShow) {
            console.log(`[FORCED MESSAGE for ${player.name}]: ${message}`);
        }
    }
}

/**
 * Sends an error message to the player. This message will always be shown.
 * @param {Player} player The player to send the error message to.
 * @param {string} errorMessage The error message string.
 */
export function sendErrorMessage(player, errorMessage) {
    sendPlayerMessage(player, `§c${errorMessage}`, true); // Prepend red color code and force show
}

/**
 * Sends a success message to the player. Obeys player's feedback setting.
 * @param {Player} player The player to send the success message to.
 * @param {string} successMessage The success message string.
 */
export function sendSuccessMessage(player, successMessage) {
    sendPlayerMessage(player, `§a${successMessage}`); // Prepend green color code
}

/**
 * Sends an informational/warning message to the player. Obeys player's feedback setting.
 * @param {Player} player The player to send the info message to.
 * @param {string} infoMessage The info message string.
 */
export function sendInfoMessage(player, infoMessage) {
    sendPlayerMessage(player, `§e${infoMessage}`); // Prepend yellow color code
}
