// scripts/utils/parsing.js

import { BlockPermutation } from "@minecraft/server";

/**
 * Parses a comma-separated string of block IDs with optional weights.
 * Handles missing namespaces (defaults to 'minecraft:') and weight validation.
 * Format examples: "stone", "stone:5, dirt:3", "minecraft:glass", "my_addon:custom_block:10"
 * @param {string} textInput The input string.
 * @returns {{blocks: Array<{id: string, weight: number}>, errors: string[]}} An object containing an array of parsed WeightedBlock objects and an array of error messages encountered during parsing.
 */
export function parseWeightedBlockInput(textInput) {
    const b = []; // Array to hold successfully parsed blocks {id: string, weight: number}
    const e = []; // Array to hold error messages

    // Basic input validation
    if (textInput === null || textInput === undefined || typeof textInput !== 'string') {
        e.push("Invalid input type (expected string).");
        return { blocks: b, errors: e };
    }
    if (textInput.trim() === "") {
        // Allow empty input string? Let's treat it as no blocks, no error.
        // e.push("Input string cannot be empty.");
        return { blocks: b, errors: e };
    }

    const entries = textInput.split(','); // Split the input string by commas

    for (const entry of entries) {
        const t = entry.trim(); // Remove leading/trailing whitespace from each entry
        if (!t) continue; // Skip empty entries resulting from multiple commas, etc.

        // Find the last colon, which separates ID from weight (if weight exists)
        const lastColonIndex = t.lastIndexOf(':');
        let idPart = t;
        let weightPart = "1"; // Default weight
        let w = 1;

        // Check if a colon exists and if the part after it is potentially a weight
        if (lastColonIndex > 0) { // Ensure colon is not the first character
            const potentialWeightStr = t.substring(lastColonIndex + 1).trim();
            const potentialWeight = parseInt(potentialWeightStr, 10);

            // Check if it's a valid positive integer weight AND the part before colon is not empty
            if (!isNaN(potentialWeight) &&
                potentialWeight > 0 &&
                Number.isInteger(potentialWeight) &&
                potentialWeightStr === String(potentialWeight) && // Avoid partial parses like "stone:5a" -> 5
                t.substring(0, lastColonIndex).trim() !== "") {
                // Looks like a valid weight suffix
                idPart = t.substring(0, lastColonIndex).trim();
                weightPart = potentialWeightStr; // Keep original string for now
                w = potentialWeight;
            } else {
                // Treat the whole string as the ID if the part after last colon isn't a valid weight
                idPart = t;
                // Weight remains default 1
            }
        }
        // If no colon, idPart is already 't', weight is default 1

        let id = idPart; // Assign the determined ID part

        // Validate the determined ID part
        if (!id) {
            e.push(`Empty ID processed from: '${t}'. Skip.`);
            continue;
        }

        // Add default "minecraft:" namespace if missing
        if (!id.includes(':')) {
            id = `minecraft:${id}`;
        }

        // Final check: Try resolving the block ID to catch truly invalid ones
        try {
            // BlockPermutation.resolve will throw an error if the ID (including namespace) is invalid
            BlockPermutation.resolve(id);
            // Add the valid block and its weight to the results
            b.push({ id: id, weight: w });
        } catch (err) {
            // Add an error message if the ID is invalid
            e.push(`Invalid Block ID: '${id}' (from entry '${t}'). Skip.`);
        }
    }

    // Return the parsed blocks and any errors encountered
    return { blocks: b, errors: e };
}


/**
 * Selects a block ID from an array of weighted blocks based on their weights.
 * @param {Array<{id: string, weight: number}>} blocksWithWeights Array of WeightedBlock objects.
 * @returns {string | undefined} The selected block ID, or undefined if the input array is empty or invalid.
 */
export function selectWeightedBlock(blocksWithWeights) {
    if (!blocksWithWeights || blocksWithWeights.length === 0) return undefined;

    // Ensure weights are valid positive integers, default to 1 otherwise
    const validWeightedBlocks = blocksWithWeights.map(b => ({
        id: b.id,
        weight: (Number.isInteger(b.weight) && b.weight > 0) ? b.weight : 1
    }));

    // Calculate the total weight
    const totalWeight = validWeightedBlocks.reduce((sum, block) => sum + block.weight, 0);

    // If total weight is zero (e.g., all weights were invalid/zero), return the first ID as a fallback
    if (totalWeight <= 0) {
        return validWeightedBlocks[0]?.id; // Use optional chaining for safety
    }

    // Generate a random number between 0 (inclusive) and totalWeight (exclusive)
    let randomWeight = Math.random() * totalWeight;

    // Iterate through the blocks, subtracting their weight until the random number is exceeded
    for (const block of validWeightedBlocks) {
        if (randomWeight < block.weight) {
            return block.id; // This block is selected
        }
        randomWeight -= block.weight; // Subtract current block's weight and check next
    }

    // Fallback: Should theoretically not be reached if totalWeight > 0, but return last block ID just in case of floating point issues
    return validWeightedBlocks[validWeightedBlocks.length - 1]?.id;
}