// scripts/utils/particle.js

import { Dimension, Player } from "@minecraft/server"; // Import types
import {
    PARTICLE_EDGE_OFFSET,
    PREVIEW_PARTICLE_DENSITY,
    MIN_PREVIEW_PARTICLES_PER_EDGE,
} from "../constants/settings.js";
import { getLinePoints3D } from "./algorithms.js"; // Import line algorithm
import { drawBoxOutline as utilDrawBoxOutline } from "./particle.js";
import { getGlobalSettings } from "../state/global_state.js";

/**
 * Spawns a particle at a location, offset slightly towards the outside edge
 * based on the location's position relative to min/max bounds.
 * Ignores errors silently (e.g., particle out of loaded area).
 * @param {Dimension} dimension - The dimension to spawn the particle in.
 * @param {string} particleId - The ID of the particle to spawn.
 * @param {{x:number, y:number, z:number}} location - The integer block coordinate where the particle conceptually lies.
 * @param {{x:number, y:number, z:number}} min - The minimum corner of the bounding box for offset calculation.
 * @param {{x:number, y:number, z:number}} max - The maximum corner of the bounding box for offset calculation.
 */
export function silentSpawnParticle(dimension, particleId, location, min, max) {
    const edgeOffset = PARTICLE_EDGE_OFFSET;
    const cornerOffset = 0.1;

    // Handle single block case - all 8 corners
    if (min.x === max.x && min.y === max.y && min.z === max.z) {
        try {
            // Bottom corners
            dimension.spawnParticle(particleId, { x: location.x - cornerOffset, y: location.y - cornerOffset, z: location.z - cornerOffset });
            dimension.spawnParticle(particleId, { x: location.x - cornerOffset, y: location.y - cornerOffset, z: location.z + 1 + cornerOffset });
            dimension.spawnParticle(particleId, { x: location.x + 1 + cornerOffset, y: location.y - cornerOffset, z: location.z - cornerOffset });
            dimension.spawnParticle(particleId, { x: location.x + 1 + cornerOffset, y: location.y - cornerOffset, z: location.z + 1 + cornerOffset });

            // Top corners
            dimension.spawnParticle(particleId, { x: location.x - cornerOffset, y: location.y + 1 + cornerOffset, z: location.z - cornerOffset });
            dimension.spawnParticle(particleId, { x: location.x - cornerOffset, y: location.y + 1 + cornerOffset, z: location.z + 1 + cornerOffset });
            dimension.spawnParticle(particleId, { x: location.x + 1 + cornerOffset, y: location.y + 1 + cornerOffset, z: location.z - cornerOffset });
            dimension.spawnParticle(particleId, { x: location.x + 1 + cornerOffset, y: location.y + 1 + cornerOffset, z: location.z + 1 + cornerOffset });
        } catch (e) { }
        return;
    }

    // Calculate thin dimensions
    const isThinX = min.x === max.x;
    const isThinY = min.y === max.y;
    const isThinZ = min.z === max.z;

    // Generate all possible offset combinations for thin dimensions
    const xOffsets = isThinX ?
        [location.x - edgeOffset, location.x + 1 + edgeOffset] :
        [Math.abs(location.x - min.x) < 0.1 ? location.x - edgeOffset : location.x + 1 + edgeOffset];

    const yOffsets = isThinY ?
        [location.y - edgeOffset, location.y + 1 + edgeOffset] :
        [Math.abs(location.y - min.y) < 0.1 ? location.y - edgeOffset : location.y + 1 + edgeOffset];

    const zOffsets = isThinZ ?
        [location.z - edgeOffset, location.z + 1 + edgeOffset] :
        [Math.abs(location.z - min.z) < 0.1 ? location.z - edgeOffset : location.z + 1 + edgeOffset];

    // Spawn particles for all combinations of offsets
    try {
        xOffsets.forEach(x => {
            yOffsets.forEach(y => {
                zOffsets.forEach(z => {
                    dimension.spawnParticle(particleId, { x, y, z });
                });
            });
        });
    } catch (e) {
        // Silently ignore errors
    }
}

/**
 * Draws the wireframe outline of a box defined by min and max corners using particles.
 * Uses silentSpawnParticle for offsetting.
 * @param {Player} player The player context (used for dimension).
 * @param {{x:number, y:number, z:number}} min The minimum corner of the box.
 * @param {{x:number, y:number, z:number}} max The maximum corner of the box.
 */
export function drawBoxOutline(player, min, max) {
    const globalSettings = getGlobalSettings(); // <<< Get settings
    const outlineParticleId = globalSettings.outlineParticleId;

    const dimension = player.dimension;
    const size = { x: max.x - min.x + 1, y: max.y - min.y + 1, z: max.z - min.z + 1 };

    if (size.x <= 0 || size.y <= 0 || size.z <= 0) return;

    const density = PREVIEW_PARTICLE_DENSITY;
    const minParticles = MIN_PREVIEW_PARTICLES_PER_EDGE;
    const steps = {
        x: Math.max(minParticles, Math.ceil(size.x * density)),
        y: Math.max(minParticles, Math.ceil(size.y * density)),
        z: Math.max(minParticles, Math.ceil(size.z * density))
    };
    const edgeLengthX = Math.max(0, size.x - 1);
    const edgeLengthY = Math.max(0, size.y - 1);
    const edgeLengthZ = Math.max(0, size.z - 1);

    const particleLocations = new Set();
    const addParticleLocation = (loc) => {
        particleLocations.add(`${Math.round(loc.x)},${Math.round(loc.y)},${Math.round(loc.z)}`);
    };

    // Edges parallel to X axis
    for (let i = 0; i <= steps.x; i++) {
        const t = steps.x > 0 ? i / steps.x : 0;
        const x = min.x + t * edgeLengthX;
        addParticleLocation({ x: x, y: min.y, z: min.z });
        addParticleLocation({ x: x, y: max.y, z: min.z });
        addParticleLocation({ x: x, y: min.y, z: max.z });
        addParticleLocation({ x: x, y: max.y, z: max.z });
    }
    // Edges parallel to Y axis
    for (let i = 0; i <= steps.y; i++) {
        const t = steps.y > 0 ? i / steps.y : 0;
        const y = min.y + t * edgeLengthY;
        addParticleLocation({ x: min.x, y: y, z: min.z });
        addParticleLocation({ x: max.x, y: y, z: min.z });
        addParticleLocation({ x: min.x, y: y, z: max.z });
        addParticleLocation({ x: max.x, y: y, z: max.z });
    }
    // Edges parallel to Z axis
    for (let i = 0; i <= steps.z; i++) {
        const t = steps.z > 0 ? i / steps.z : 0;
        const z = min.z + t * edgeLengthZ;
        addParticleLocation({ x: min.x, y: min.y, z: z });
        addParticleLocation({ x: max.x, y: min.y, z: z });
        addParticleLocation({ x: min.x, y: max.y, z: z });
        addParticleLocation({ x: max.x, y: max.y, z: z });
    }

    particleLocations.forEach(locString => {
        const [x, y, z] = locString.split(',').map(Number);
        silentSpawnParticle(dimension, outlineParticleId, { x, y, z }, min, max);
    });
}


/**
 * Draws an approximate particle outline for various brush shapes.
 * Note: This visually represents the SOLID shape outline, regardless of brush hollow settings.
 * @param {Player} player The player requesting the outline.
 * @param {{x:number, y:number, z:number}} center The center block location.
 * @param {string} shape The shape type internal name (e.g., 'sphere', 'octahedron').
 * @param {number} size The radius or half-width of the shape.
 */
export function drawShapeOutline(player, center, shape, size) {
    const globalSettings = getGlobalSettings(); // <<< Get settings
    const outlineParticleId = globalSettings.outlineParticleId;

    const dimension = player.dimension;
    const radius = Math.max(0, Math.floor(size)); // Ensure non-negative integer radius/half-width

    // Calculate overall bounding box for particle offsetting
    // Use a simplified bounding box for preview purposes (center +/- radius)
    const minBound = { x: center.x - radius, y: center.y - radius, z: center.z - radius };
    const maxBound = { x: center.x + radius, y: center.y + radius, z: center.z + radius };

    // Handle simple cube shape by drawing its bounding box directly
    if (shape === 'cube') {
        // Use the correct Y bounds for cube preview based on radius
        const cubeMin = { x: center.x - radius, y: center.y - radius, z: center.z - radius };
        const cubeMax = { x: center.x + radius, y: center.y + radius, z: center.z + radius };
        utilDrawBoxOutline(player, cubeMin, cubeMax); // Call the box outline function
        return; // Exit early for cube
    }

    const density = PREVIEW_PARTICLE_DENSITY;
    const minParticles = MIN_PREVIEW_PARTICLES_PER_EDGE;
    const particleLocations = new Set(); // Use Set for unique integer locations
    const addParticleLocation = (loc) => {
        particleLocations.add(`${Math.round(loc.x)},${Math.round(loc.y)},${Math.round(loc.z)}`);
    };
    // Helper to add all points from getLinePoints3D
    const addLinePoints = (start, end) => {
        const points = getLinePoints3D(start, end);
        points.forEach(p => addParticleLocation(p));
    };

    // --- Draw shape-specific outlines ---
    const steps = Math.max(minParticles * 2, Math.ceil(2 * Math.PI * radius * density)); // Steps for circles

    switch (shape) {
        case 'sphere': {
            // Draw 3 orthogonal great circles
            if (steps <= 0 && radius === 0) { addParticleLocation(center); break; }
            if (steps <= 0) break;
            for (let i = 0; i < steps; i++) {
                const angle = (i / steps) * 2 * Math.PI;
                const cosA = Math.cos(angle); const sinA = Math.sin(angle);
                addParticleLocation({ x: center.x + radius * cosA, y: center.y + radius * sinA, z: center.z }); // XY
                addParticleLocation({ x: center.x + radius * cosA, y: center.y, z: center.z + radius * sinA }); // XZ
                addParticleLocation({ x: center.x, y: center.y + radius * cosA, z: center.z + radius * sinA }); // YZ
            }
            break;
        }
        case 'cylinder': {
            // Draw top and bottom circles and vertical lines
            const cylHeight = (radius * 2); // Approximate height based on radius for preview
            const cylStartY = center.y - Math.floor(cylHeight / 2); // Example centered Y
            const cylEndY = center.y + Math.ceil(cylHeight / 2);
            const heightSteps = Math.max(minParticles, Math.ceil(cylHeight * density));

            // Top and Bottom circles
            if (steps > 0) {
                for (let i = 0; i < steps; i++) {
                    const angle = (i / steps) * 2 * Math.PI;
                    const xOffset = radius * Math.cos(angle); const zOffset = radius * Math.sin(angle);
                    addParticleLocation({ x: center.x + xOffset, y: cylStartY, z: center.z + zOffset }); // Bottom
                    addParticleLocation({ x: center.x + xOffset, y: cylEndY, z: center.z + zOffset });   // Top
                }
            } else { // Handle radius 0 case (single vertical line)
                if (heightSteps > 0) {
                    for (let i = 0; i <= heightSteps; i++) {
                        const y = cylStartY + (cylHeight > 0 ? (i / heightSteps) * cylHeight : 0);
                        addParticleLocation({ x: center.x, y: y, z: center.z });
                    }
                } else {
                    addParticleLocation(center); // Single point if radius and height are 0
                }
            }
            // Vertical lines at cardinal directions if radius > 0 and height > 0
            if (radius > 0 && heightSteps > 0) {
                for (let i = 0; i <= heightSteps; i++) {
                    const y = cylStartY + (cylHeight > 0 ? (i / heightSteps) * cylHeight : 0);
                    addParticleLocation({ x: center.x + radius, y: y, z: center.z }); // +X side
                    addParticleLocation({ x: center.x - radius, y: y, z: center.z }); // -X side
                    addParticleLocation({ x: center.x, y: y, z: center.z + radius }); // +Z side
                    addParticleLocation({ x: center.x, y: y, z: center.z - radius }); // -Z side
                }
            }
            break;
        }
        case 'pyramid': {
            // Draw base square outline and edges to apex
            const apex = { x: center.x, y: center.y + radius, z: center.z }; // Assume height = radius for preview
            const baseY = center.y;
            const baseCorners = [
                { x: center.x + radius, y: baseY, z: center.z + radius }, { x: center.x + radius, y: baseY, z: center.z - radius },
                { x: center.x - radius, y: baseY, z: center.z - radius }, { x: center.x - radius, y: baseY, z: center.z + radius }
            ];
            // Base outline
            addLinePoints(baseCorners[0], baseCorners[1]); addLinePoints(baseCorners[1], baseCorners[2]);
            addLinePoints(baseCorners[2], baseCorners[3]); addLinePoints(baseCorners[3], baseCorners[0]);
            // Edges to apex
            baseCorners.forEach(corner => addLinePoints(corner, apex));
            break;
        }
        // --- NEW SHAPE PREVIEWS ---
        case 'octahedron': {
            // Connect the 6 axis endpoints
            const vertices = [
                { x: center.x + radius, y: center.y, z: center.z }, { x: center.x - radius, y: center.y, z: center.z },
                { x: center.x, y: center.y + radius, z: center.z }, { x: center.x, y: center.y - radius, z: center.z },
                { x: center.x, y: center.y, z: center.z + radius }, { x: center.x, y: center.y, z: center.z - radius }
            ];
            // Top pyramid edges
            addLinePoints(vertices[0], vertices[2]); addLinePoints(vertices[1], vertices[2]);
            addLinePoints(vertices[4], vertices[2]); addLinePoints(vertices[5], vertices[2]);
            // Bottom pyramid edges
            addLinePoints(vertices[0], vertices[3]); addLinePoints(vertices[1], vertices[3]);
            addLinePoints(vertices[4], vertices[3]); addLinePoints(vertices[5], vertices[3]);
            // Middle square edges
            addLinePoints(vertices[0], vertices[4]); addLinePoints(vertices[4], vertices[1]);
            addLinePoints(vertices[1], vertices[5]); addLinePoints(vertices[5], vertices[0]);
            break;
        }
        case 'dome': { // Hemisphere Up
            if (steps <= 0 && radius === 0) { addParticleLocation(center); break; }
            if (steps <= 0) break;
            for (let i = 0; i < steps; i++) {
                const angle = (i / steps) * Math.PI; // Only 180 degrees for vertical semi-circles
                const fullAngle = (i / steps) * 2 * Math.PI; // Full circle for base
                const cosA = Math.cos(angle); const sinA = Math.sin(angle);
                const cosFull = Math.cos(fullAngle); const sinFull = Math.sin(fullAngle);
                // XY Semi-Circle (Top Half)
                addParticleLocation({ x: center.x + radius * cosA, y: center.y + radius * Math.abs(sinA), z: center.z });
                // YZ Semi-Circle (Top Half)
                addParticleLocation({ x: center.x, y: center.y + radius * Math.abs(cosA), z: center.z + radius * sinA });
                // XZ Base Circle
                addParticleLocation({ x: center.x + radius * cosFull, y: center.y, z: center.z + radius * sinFull });
            }
            break;
        }
        case 'bowl': { // Hemisphere Down
            if (steps <= 0 && radius === 0) { addParticleLocation(center); break; }
            if (steps <= 0) break;
            for (let i = 0; i < steps; i++) {
                const angle = (i / steps) * Math.PI;
                const fullAngle = (i / steps) * 2 * Math.PI;
                const cosA = Math.cos(angle); const sinA = Math.sin(angle);
                const cosFull = Math.cos(fullAngle); const sinFull = Math.sin(fullAngle);
                // XY Semi-Circle (Bottom Half)
                addParticleLocation({ x: center.x + radius * cosA, y: center.y - radius * Math.abs(sinA), z: center.z });
                // YZ Semi-Circle (Bottom Half)
                addParticleLocation({ x: center.x, y: center.y - radius * Math.abs(cosA), z: center.z + radius * sinA });
                // XZ Top Circle
                addParticleLocation({ x: center.x + radius * cosFull, y: center.y, z: center.z + radius * sinFull });
            }
            break;
        }
        case 'cone_up': { // Point Up
            const apexY = center.y + radius; // Assume height = radius for preview
            const baseY = center.y;
            // Base circle
            if (steps > 0) {
                for (let i = 0; i < steps; i++) {
                    const angle = (i / steps) * 2 * Math.PI;
                    addParticleLocation({ x: center.x + radius * Math.cos(angle), y: baseY, z: center.z + radius * Math.sin(angle) });
                }
            } else { addParticleLocation({ x: center.x, y: baseY, z: center.z }); } // Base point if radius 0
            // Lines to apex from cardinal points on base
            if (radius > 0) {
                addLinePoints({ x: center.x + radius, y: baseY, z: center.z }, { x: center.x, y: apexY, z: center.z });
                addLinePoints({ x: center.x - radius, y: baseY, z: center.z }, { x: center.x, y: apexY, z: center.z });
                addLinePoints({ x: center.x, y: baseY, z: center.z + radius }, { x: center.x, y: apexY, z: center.z });
                addLinePoints({ x: center.x, y: baseY, z: center.z - radius }, { x: center.x, y: apexY, z: center.z });
            }
            break;
        }
        case 'cone_down': { // Point Down
            const apexY = center.y - radius; // Assume height = radius for preview
            const baseY = center.y;
            // Base circle
            if (steps > 0) {
                for (let i = 0; i < steps; i++) {
                    const angle = (i / steps) * 2 * Math.PI;
                    addParticleLocation({ x: center.x + radius * Math.cos(angle), y: baseY, z: center.z + radius * Math.sin(angle) });
                }
            } else { addParticleLocation({ x: center.x, y: baseY, z: center.z }); }
            // Lines to apex
            if (radius > 0) {
                addLinePoints({ x: center.x + radius, y: baseY, z: center.z }, { x: center.x, y: apexY, z: center.z });
                addLinePoints({ x: center.x - radius, y: baseY, z: center.z }, { x: center.x, y: apexY, z: center.z });
                addLinePoints({ x: center.x, y: baseY, z: center.z + radius }, { x: center.x, y: apexY, z: center.z });
                addLinePoints({ x: center.x, y: baseY, z: center.z - radius }, { x: center.x, y: apexY, z: center.z });
            }
            break;
        }
        case 'disk':
        case 'torus': { // Use same circular preview for Torus
            if (steps <= 0 && radius === 0) { addParticleLocation(center); break; }
            if (steps <= 0) break;
            const yLevel = center.y;
            for (let i = 0; i < steps; i++) {
                const angle = (i / steps) * 2 * Math.PI;
                addParticleLocation({ x: center.x + radius * Math.cos(angle), y: yLevel, z: center.z + radius * Math.sin(angle) });
            }
            // Optionally draw inner circle for torus? Keep it simple for now.
            break;
        }
        case 'plate': {
            const yLevel = center.y;
            const baseMin = { x: center.x - radius, y: yLevel, z: center.z - radius };
            const baseMax = { x: center.x + radius, y: yLevel, z: center.z + radius };
            const sideLength = (radius * 2);
            const sideSteps = Math.max(minParticles, Math.ceil(sideLength * density));

            if (sideSteps > 0) {
                for (let i = 0; i <= sideSteps; i++) {
                    const t = i / sideSteps;
                    const currentSideLength = sideLength > 0 ? sideLength : 0;
                    addParticleLocation({ x: baseMax.x, y: yLevel, z: baseMin.z + t * currentSideLength }); // +X edge
                    addParticleLocation({ x: baseMin.x, y: yLevel, z: baseMin.z + t * currentSideLength }); // -X edge
                    addParticleLocation({ x: baseMin.x + t * currentSideLength, y: yLevel, z: baseMax.z }); // +Z edge
                    addParticleLocation({ x: baseMin.x + t * currentSideLength, y: yLevel, z: baseMin.z }); // -Z edge
                }
            } else { addParticleLocation({ x: center.x, y: yLevel, z: center.z }); } // Center point if radius 0
            break;
        }
        // No blob preview yet
    } // End switch

    // --- Spawn particles from the unique calculated locations ---
    particleLocations.forEach(locString => {
        const [x, y, z] = locString.split(',').map(Number);
        // Pass the overall shape min/max bounds for correct offsetting by silentSpawnParticle
        silentSpawnParticle(dimension, outlineParticleId, { x, y, z }, minBound, maxBound);
    });
}

/**
 * Draws particles along a line between two points using a 3D line algorithm.
 * Intended for the Line Tool preview.
 * @param {Player} player The player context (used for dimension).
 * @param {{x:number, y:number, z:number}} posA Start point of the line.
 * @param {{x:number, y:number, z:number}} posB End point of the line.
 */
export function drawParticleLine(player, posA, posB) {
    const globalSettings = getGlobalSettings(); // <<< Get settings
    const outlineParticleId = globalSettings.outlineParticleId;

    if (!posA || !posB) return;
    const dimension = player.dimension;
    const linePoints = getLinePoints3D(posA, posB);
    const pointCount = linePoints.length;
    if (pointCount === 0) return;
    const density = PREVIEW_PARTICLE_DENSITY * 1.5;
    const minParticles = MIN_PREVIEW_PARTICLES_PER_EDGE;
    const numParticles = Math.max(minParticles, Math.ceil(pointCount * density));
    const step = Math.max(1, Math.floor(pointCount / numParticles));
    for (let i = 0; i < pointCount; i += step) {
        const point = linePoints[i];
        try {
            dimension.spawnParticle(outlineParticleId, { x: point.x + 0.5, y: point.y + 0.5, z: point.z + 0.5 });
        } catch (e) { /* ignore spawn errors */ }
    }
    if (pointCount > 0 && (pointCount - 1) % step !== 0) {
        const lastPoint = linePoints[pointCount - 1];
        try {
            dimension.spawnParticle(outlineParticleId, { x: lastPoint.x + 0.5, y: lastPoint.y + 0.5, z: lastPoint.z + 0.5 });
        } catch (e) { /* ignore spawn errors */ }
    }
}