// scripts/utils/raycasting.js

import { Player } from "@minecraft/server"; // Import types
import { MAX_SELECTION_DISTANCE } from "../constants/settings.js";

/**
 * Determines the block location the player is looking at within a specified maximum distance.
 * Prioritizes hitting solid blocks, then any block (including air), then calculates endpoint if no block is hit.
 * @param {Player} player The player whose view direction is used.
 * @param {number} [maxDistance=MAX_SELECTION_DISTANCE] The maximum distance to raycast. Defaults to constant.
 * @returns {{x:number, y:number, z:number} | undefined} The target block location, or undefined if an error occurs.
 */
export function determineLookTargetPos(player, maxDistance = MAX_SELECTION_DISTANCE) {
    try {
        if (!player?.isValid()) return undefined; // Check if player object is valid

        // Option 1: Raycast for solid blocks only
        const solidHit = player.getBlockFromViewDirection({
            maxDistance: maxDistance,
            excludeBlockTypes: ['minecraft:air'] // Explicitly exclude air
        });
        if (solidHit?.block) {
            return solidHit.block.location; // Return location of the first solid block hit
        }

        // Option 2: If no solid block hit, raycast for *any* block (including air)
        const anyHit = player.getBlockFromViewDirection({
            maxDistance: maxDistance
        });
        if (anyHit?.block) {
            return anyHit.block.location; // Return location of the first block hit (could be air)
        }

        // Option 3: If no block was hit at all within maxDistance, calculate the endpoint
        const headLoc = player.getHeadLocation();
        const viewDir = player.getViewDirection();
        const targetPos = {
            x: Math.floor(headLoc.x + viewDir.x * maxDistance),
            y: Math.floor(headLoc.y + viewDir.y * maxDistance),
            z: Math.floor(headLoc.z + viewDir.z * maxDistance)
        };
        return targetPos; // Return the calculated position in the air/unloaded chunk

    } catch (e) {
        // Handle potential errors during raycasting (e.g., player becomes invalid during call)
        // Avoid spamming logs for common invalid entity errors when player leaves etc.
        if (!(e?.message?.includes("Player not found") || e?.message?.includes("Invalid entity"))) {
            console.warn(`[WorldEdit Raycast] Error determining look target for ${player.name ?? player.id}: ${e}`);
        }
        return undefined; // Indicate failure
    }
}