// scripts/utils/tickingarea_manager.js

import { Dimension, system } from "@minecraft/server";

/**
 * Attempts to add a temporary, preloaded ticking area.
 * The name is generated to be unique for WorldEdit operations.
 *
 * @param {Dimension} dimension The dimension in which to add the ticking area.
 * @param {{x: number, y: number, z: number}} min The minimum block coordinates of the area.
 * @param {{x: number, y: number, z: number}} max The maximum block coordinates of the area.
 * @param {string} operationId A string to identify the operation (e.g., "paste", "fill", "terrain").
 * @returns {Promise<string | null>} The name of the created ticking area if successful, otherwise null.
 */
export async function addTemporaryTickingArea(dimension, min, max, operationId) {
    // Generate a unique name for the ticking area
    // Prefix with "we_" for WorldEdit, "temp_" for temporary, operationId, and timestamp for uniqueness.
    const areaName = `we_temp_${operationId}_${Date.now()}`;

    // The /tickingarea add command uses block coordinates for the <from> and <to> positions.
    // The y-coordinate doesn't define the height of the ticking area but is part of the position.
    // It will create a column that includes all chunks touched by the defined box.
    // The 'true' at the end is for the [preload: Boolean] parameter.
    const command = `/tickingarea add ${min.x} ${min.y} ${min.z} ${max.x} ${max.y} ${max.z} ${areaName} true`;

    try {
        // Execute the command to add the ticking area
        await dimension.runCommandAsync(command);
        console.log(`[WorldEdit] Added temporary ticking area: ${areaName}`);
        return areaName; // Return the name on success
    } catch (error) {
        // Log errors if the command fails (e.g., max ticking areas reached, invalid syntax)
        console.warn(`[WorldEdit] Failed to add temporary ticking area '${areaName}': ${error}`);
        // Check for specific error messages if needed, e.g., "No more room for new ticking areas"
        if (error.message && error.message.toLowerCase().includes("no more room")) {
            // Potentially try to remove an old "we_temp_" area if at limit - complex, for future
            console.warn(`[WorldEdit] Maximum number of ticking areas (10) might have been reached.`);
        }
        return null; // Return null on failure
    }
}

/**
 * Attempts to remove a temporary ticking area by its name.
 *
 * @param {Dimension} dimension The dimension from which to remove the ticking area.
 * @param {string} name The name of the ticking area to remove.
 * @returns {Promise<void>}
 */
export async function removeTemporaryTickingArea(dimension, name) {
    if (!name) {
        console.warn("[WorldEdit] Attempted to remove ticking area with no name.");
        return;
    }

    const command = `/tickingarea remove "${name}"`; // Enclose name in quotes in case it contains spaces, though ours don't

    try {
        // Execute the command to remove the ticking area
        await dimension.runCommandAsync(command);
        console.log(`[WorldEdit] Removed temporary ticking area: ${name}`);
    } catch (error) {
        // Log errors if the command fails (e.g., area not found)
        // It's common for this to error if the area was already removed or never existed, so a warning is usually fine.
        if (error.message && !error.message.toLowerCase().includes("did not match any ticking area")) {
            console.warn(`[WorldEdit] Error removing temporary ticking area '${name}': ${error}`);
        } else {
            // console.log(`[WorldEdit] Ticking area '${name}' likely already removed or never existed.`);
        }
    }
}
