// scripts/utils/validation.js

import { BlockPermutation } from "@minecraft/server";

/**
 * Validates an array of block IDs using BlockPermutation.resolve().
 * Automatically handles adding 'minecraft:' namespace if missing.
 * @param {string[]} blockIds Array of block ID strings to validate.
 * @returns {{validPermutations: BlockPermutation[], invalidIds: string[]}} An object containing an array of valid BlockPermutation objects and an array of invalid ID strings.
 */
export function validateBlockIds(blockIds) {
    const validPermutations = [];
    const invalidIds = [];

    if (!Array.isArray(blockIds)) {
        console.warn("[validateBlockIds] Input was not an array.");
        return { validPermutations, invalidIds: ['Invalid input type'] };
    }

    for (const id of blockIds) {
        // Ensure working with a string and trim whitespace
        const trimmedId = typeof id === 'string' ? id.trim() : '';
        if (!trimmedId) continue; // Skip empty strings

        try {
            // BlockPermutation.resolve handles adding 'minecraft:' if needed
            const permutation = BlockPermutation.resolve(trimmedId);
            // If resolve succeeds without throwing, add it to the valid list
            validPermutations.push(permutation);
        } catch (e) {
            // If resolve throws an error, the ID is invalid
            invalidIds.push(trimmedId); // Add the problematic ID string to the invalid list
        }
    }
    return { validPermutations, invalidIds };
}

/**
 * Gets a single BlockPermutation object for a given block ID string.
 * Validates the ID and automatically adds the 'minecraft:' namespace if missing.
 * @param {string} blockId The block ID string (e.g., "stone", "minecraft:wool").
 * @returns {BlockPermutation | undefined} The resolved BlockPermutation, or undefined if the ID is invalid.
 */
export function getSingleBlockPermutation(blockId) {
    if (!blockId || typeof blockId !== 'string') {
        // console.warn("[getSingleBlockPermutation] Invalid input:", blockId);
        return undefined;
    }

    let id = blockId.trim();
    if (!id) {
        // console.warn("[getSingleBlockPermutation] Input was empty string.");
        return undefined;
    }

    // Add default namespace if missing
    if (!id.includes(':')) {
        id = `minecraft:${id}`;
    }

    try {
        // Attempt to resolve the permutation
        return BlockPermutation.resolve(id);
    } catch (e) {
        // If resolve fails, return undefined
        // console.warn(`[getSingleBlockPermutation] Failed to resolve ID '${id}': ${e}`);
        return undefined;
    }
}